//___________________________________________________________________
//
//                          Class  CheckBox
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/checkbox.h>
#include <wx/frame.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "LogoObject.h"
#include "Int.h"
#include "UInt8.h"
#include "Color.h"
#include "Point.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "CheckBox.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(CheckBox, wxCheckBox)
	EVT_CHECKBOX	(-1, CheckBox::OnClick	)
END_EVENT_TABLE()

CheckBox::CheckBox(CheckBoxNodeP checkBoxNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=checkBoxNode_;
	wxWindow *parent=(wxWindow*)console;
	if (checkBoxNode_->parent)
		parent=checkBoxNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			_("CheckBox"));
}

CheckBox::CheckBox(CheckBoxNodeP checkBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxCheckBox(parent, id, label, 
			pos, size, style, 
			validator, name)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=checkBoxNode_;
	checkBoxNode_->window=WindowP(this);
	checkBoxNode_->setonClick(onClick_);
}

CheckBox::~CheckBox()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
}

void CheckBox::OnClick(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventCheckBox(this);
	if throwing return;
	*th->eventFiFo << evt;
}

wxString CheckBox::wxstring()
{
	wxString str;
	str << _("CheckBox") << " [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << (*d(onClick))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " 
		<< GetValue();
	return str;
}

DataInputStream &CheckBox::read(DataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style, state;
	
	istr >> label >> name >> (int&)*d(onClick)
		>> pos.x >> pos.y >> size.x >> size.y >> style >> state;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetValue(state!=0);
	return istr;
}
wxDataOutputStream &CheckBox::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(*d(onClick))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle()) << int(GetValue());
	return ostr;
}

PairP CheckBoxNode::onClick()
{
	if (checkBox)
		return PairP(*CheckBoxP(window)->d(CheckBox::onClick));
	return 0;
}

void CheckBoxNode::setonClick(NodeP nd) 
{
	if (window)
	{	ThreadP th=Thread::This();
		*checkBox->d(CheckBox:: onClick) = nd;
		*checkBox->d(CheckBox::uonClick) = th->lexufun;
		*checkBox->d(CheckBox::tonClick) = th->thisobj;
		check_oldyoung(nd);
	}
}

CheckBoxNode::CheckBoxNode(): 
	WindowNode(CHECKBOX),
	checkBox((CheckBoxP&)window)
{}

CheckBoxNode::CheckBoxNode(WindowNodeP parent_): 
	WindowNode(CHECKBOX, parent_),
	checkBox((CheckBoxP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void CheckBoxNode::free() 
{	
	if (window) 
	{	ThreadP th=Thread::This();
		*CheckBoxP(window)->d(CheckBox::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		window->Destroy(); 
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tCheckBox:
				if (WindowP(((EventCheckBox*)ev)->checkBox)==window)
					((EventCheckBox*)ev)->checkBox=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		window=0;
	}
}
void CheckBoxNode::inc_gc() {fifo << onClick();}
int CheckBoxNode::youngs() {return gmark(onClick());}
int CheckBoxNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onClick(), ptrToInt);
}
void CheckBoxNode::check_oldyoungs() {check_oldyoung(onClick());}

DataInputStream &CheckBoxNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode;
		window=WindowP(new CheckBox(CheckBoxNodeP(this)));
		if (!window)
		{	error(OUT_OF_MEM); return istr;
		}
		window->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &CheckBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (window)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent);
		window->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void CheckBoxNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=WindowNodeP((WindowNode*)intToPtr.find(int(parent)));
	if (window)
	{	setonClick(PairP((Pair*)intToPtr.find(int(CheckBoxP(window)->onClick))));
	}
}

void CheckBoxNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	CheckBoxP c=CheckBoxP(window), c2=0;
	if (!window) return;
	c2=new CheckBox(CheckBoxNodeP(this),
		p, -1, c->GetLabel(),
		PairP(*c->d(CheckBox::onClick)), c->GetPosition(), c->GetSize(), c->GetWindowStyle(), 
		wxDefaultValidator, c->GetName());
	if (!c2)
	{	error(OUT_OF_MEM); return;
	}
	c2->SetValue(c->GetValue());
	window=WindowP(c2);
	setonClick(*c->d(CheckBox::onClick));
	delete c;
	window=WindowP(c2);
}

StringP CheckBoxNode::string()
{
	if (window)
		return new String(window->wxstring());
	return _Null_Word->name();
}

wxString CheckBoxNode::wxstring()
{
	if (window)
		return window->wxstring();
	return wxEmptyString;
}

void CheckBoxNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lCheckBox(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		l=new String(l);
		label=l->ptr();
	}
	PairP onClick=0;
	if (exist(3))
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxWindowID id=-1;
	if (exist(7))
	{	int i=arg[7]->a_int(7);
		if throwing return Unbound;
		id=i;	
	}
	wxString name=_("CheckBox");
	if (exist(8))
	{	StringP n=arg[8]->a_string(8);
		if throwing return Unbound;
		name=n->ptr();
	}
//	GUI_ENTER
	CheckBoxNodeP checkBoxNode=new CheckBoxNode(parentNode);
	if throwing return Unbound;
	CheckBoxP cb=new CheckBox(checkBoxNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	if (!cb)
	{	error(OUT_OF_MEM); return Unbound;
	}
//	GUI_LEAVE
	return checkBoxNode;
}

NodeP lCheckBoxOnClick(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	checkBox->setonClick(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lCheckBoxDestroy(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=checkBox->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(checkBox));
			p->check_oldyoung(p->children);
		}
	}
	checkBox->window->Destroy();
//	GUI_LEAVE
	return Unbound;
}

NodeP lCheckBoxSet(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	CheckBoxP(checkBox->window)->SetValue(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lCheckBoxValue(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
//	GUI_ENTER
	bool state=CheckBoxP(checkBox->window)->GetValue();
//	GUI_LEAVE
	return Node::boolnode(state);
}

NodeP lCheckBoxEnable(PairPP arg)
{
	CheckBoxNodeP checkBox=arg[1]->a_checkBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	checkBox->window->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}
