//___________________________________________________________________
//
//                         ComboBox Header
//___________________________________________________________________

#ifndef WINDOWNODE
	#include "Window.h"
#endif

extern Function 
	lComboBox, lComboBoxDestroy,
	lComboBoxSelection, lComboBoxSetSelection,
	lComboBoxSetChoices, 
	lComboBoxAppend, 
	lComboBoxSetItem, lComboBoxRemoveItem, 
	lComboBoxCount,
	lComboBoxValue, lComboBoxSetValue,
	lComboBoxSetBackgroundColor, lComboBoxSetColor,
	lComboBoxSetFontSize, lComboBoxSetFontName,
	lComboBoxSetFontStyle, lComboBoxSetFontWeight,
	lComboBoxOnChar, lComboBoxOnKeyDown, lComboBoxOnKeyUp,
	lComboBoxOnSelect, lComboBoxOnChange, lComboBoxOnEnter,
	lComboBoxEnable;

class wxStaticText;
class MyComboBox;

class ComboBox: public wxControl
{
	public:
		wxStaticText *label;
		MyComboBox *comboBox;
		wxSize bestSize;
		wxString value;
		wxArrayString choices;
		enum Data
		{	WINDOW_DATA,
			onChange, uonChange, tonChange, 
			onEnter, uonEnter, tonEnter, 
			onSelect, uonSelect, tonSelect, 
			dataCount
		};
		NodeP data[dataCount];

		static int vmt;

		virtual NodePP d(int i) {return &data[i];}

		ComboBox(wxWindow *parent);
		ComboBox(ComboBoxNodeP comboBoxNode_, int style);
		ComboBox(ComboBoxNodeP comboBoxNode_,
			wxWindow* parent, wxWindowID id, 
			const wxString &value_,
			const wxArrayString &choices_, 
			PairP onSelect_=PairP(0),
			PairP onChange_=PairP(0),
			PairP onEnter_=PairP(0),
			const wxPoint &pos=wxDefaultPosition, 
			const wxSize &size=wxDefaultSize, 
			long style = 0, 
			const wxValidator &validator = wxDefaultValidator, 
			const wxString &name = _("ComboBox"));

		bool create(wxWindow* parent, wxWindowID id, 
			const wxString &value_,
			const wxArrayString &choices_, 
			const wxPoint &pos, 
			const wxSize &size, 
			long style, 
			const wxValidator &validator, 
			const wxString &name);

		~ComboBox();

		virtual void DoSetSize(int x, int y, int width, int height,
						   int sizeFlags = wxSIZE_AUTO);

		virtual wxSize DoGetBestSize();

		virtual void Set(wxArrayString &choices);
		virtual void SetValue(const wxString &value);
		virtual wxString GetValue() const;

		void 
			OnClose	(wxCloseEvent &event),
			OnSelect	(wxCommandEvent &event),
			OnChange	(wxCommandEvent &event),
			OnEnter	(wxCommandEvent &event),
			OnChar	(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp	(wxKeyEvent &event);

		wxString wxstring();
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		virtual bool SetBackgroundColour(const wxColour& colour);
		virtual bool SetForegroundColour(const wxColour& colour);
		virtual bool SetFont(const wxFont& font);

		DECLARE_EVENT_TABLE()
};
typedef ComboBox *ComboBoxP;

class ComboBoxNode: public WindowNode
{
	public:
		ComboBoxP &comboBox;
		static const ComboBoxNode aComboBoxNode;

		operator_new;
		PairP onSelect();
		PairP onChange();
		PairP onEnter();
		PairP onChar();
		PairP onKeyDown();
		PairP onKeyUp();
		void setonSelect	(NodeP function);
		void setonChange	(NodeP function);
		void setonEnter	(NodeP function);
		void setonChar		(NodeP function);
		void setonKeyDown	(NodeP function);
		void setonKeyUp	(NodeP function);

		ComboBoxNode();
		ComboBoxNode(WindowNodeP parent_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr(PtrToInt_ &ptrToInt);
		void check_oldyoungs();
		
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
		void recreate();

		StringP string();
		wxString wxstring();
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		StringP _typeof();
};
