#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "StringToPtr.h"
#include "PtrToString.h"
#include "String.h"
#include "Pair.h"
#include "Object.h"
#include "Line.h"
#include "PtrTable.h"
#include "Compiled.h"
#include "Colon.h"
#include "Tree.h"
#include "Error.h"
#include "Init.h"
#include "Thread.h"
#include "Color.h"
#include "Console.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

Compiled::Compiled(ObjectP object_, int size_):
	Node(COMPILED), _object(object_), size(size_), length(0),
	_assembly(0), code(0), links(0), refs(0), tint(0)
{
	code=FunctionP(GC_::malloc(size));
	memset(code, 0x90, size);
	links=new PtrTable(7);
	refs=new PtrTable(7);
}

void Compiled::free()
{
	GC_::free(code); code=0;
}

void Compiled::setassembly(PairP nd) {_assembly=nd;  check_oldyoung(nd);}
void Compiled::setlinks(PtrTableP nd) {links=nd;  check_oldyoung(nd);}
void Compiled::setrefs(PtrTableP nd) {refs=nd;  check_oldyoung(nd);}

void Compiled::inc_gc()
{
	fifo << _object << _assembly << links << refs;
}

int Compiled::youngs() 
{
	return gmark(_object)
			+gmark(_assembly)
			+gmark(links)
			+gmark(refs);
}

int Compiled::youngsPtr(PtrToInt_ &ptrToInt) 
{
	return gmarkPtr(_object, ptrToInt)
			+gmarkPtr(_assembly, ptrToInt)
			+gmarkPtr(links, ptrToInt)
			+gmarkPtr(refs, ptrToInt);
}

void Compiled::check_oldyoungs() 
{
	check_oldyoung(_object); 
	check_oldyoung(_assembly); 
	check_oldyoung(links); 
	check_oldyoung(refs);
}

void Compiled::setPtrs(IntToPtr_ &intToPtr)
{	_object=(ObjectP((Object*)intToPtr.find(int(_object))));
	setassembly(PairP((Pair*)intToPtr.find(int(_assembly))));
	setlinks   (PtrTableP((PtrTable*)intToPtr.find(int(links))));
	setrefs    (PtrTableP((PtrTable*)intToPtr.find(int(refs))));
}
DataInputStream &Compiled::read(DataInputStream &istr) 
{	return istr 
	>> (int&)_object >> (int&)_assembly >> (int&)links >> (int&)refs >> wxString(charP(code)) >> size;
}
wxDataOutputStream &Compiled::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(_object) 
		<< GC_::ptrToInt->find(_assembly)
		<< GC_::ptrToInt->find(links)
		<< GC_::ptrToInt->find(refs)
		<< (charP&)code
		<< size;
}

NodeP Compiled::copy(PtrToInt_ &ptrToInt)
{	CompiledP cmp=new Compiled();
	if throwing return Unbound;
	cmp->_object=ObjectP(ptrToInt.find(_object));
	cmp->_assembly=PairP(ptrToInt.find(_assembly));
	cmp->links=PtrTableP(ptrToInt.find(links));
	cmp->refs=PtrTableP(ptrToInt.find(refs));
	memcpy(cmp->code, code, size);
	cmp->size=size;
	return cmp;
}

void Compiled::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	if (_assembly)
		_assembly->print(strm, width, depth, noBrackets, noSpacers);
}

void Compiled::printBin(FILEP strm)
{
	if (code)
		for (int i=0; i<size; i++)
			::printBin(strm, charP(code)[i]);
}

NodeP Compiled::contentsTree(	wxArrayString &namearr, 
										wxArrayString &valarr, 
										wxArrayPtrVoid &objarr, 
										ObjectType want, 
										ObjectType dont, 
										ObjectType must, 
										int &count, int level)
{
	PairP nd=_assembly;
	objarr[count]=nd;
	NodeP contentsList=0;
	namearr[count]=wxString(' ',level-1);
	if (nd)
	{	if (nd->expandable())
			if (nd->expanded())
				namearr[count]+='-';
			else
				namearr[count]+='+';
		if (nd->expanded())
		{	valarr[count]=nd->_typeof()->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
			contentsList=nd->contentsTree(
				namearr, valarr, objarr, want, dont, must, count, level+1);
		}else
		{	valarr[count]=_typeof()->ptr();
			incContentsArrays(namearr, valarr, objarr, count);
		}
	}
	if (!contentsList || contentsList==Unbound)
		return 0;
	return this;
}

int Compiled::compare(NodeP nd)
{	return _assembly->compare(nd);
}
int Compiled::count()
{	return _assembly->count();
}
int Compiled::_sizeof()
{	return _assembly->_sizeof();
}
bool Compiled::equalP(NodeP b)
{	return _assembly->equalP(b);
}
NodeP Compiled::first_()
{	return _assembly->first_();
}
NodeP Compiled::butFirst()
{	return _assembly->butFirst();
}
NodeP Compiled::last()
{	return _assembly->last();
}
NodeP Compiled::butLast()
{	return _assembly->butLast();
}
PairP Compiled::list()
{	return _assembly;
}

NodeP Compiled::call(PairPP arg)
{
	console->flush();
	return (*FunctionP(code))(arg);
}

Compiled &Compiled::operator<<(const CodeMacro &m) 
{
	c_out << m << endL;
	if (length+m.length > size)
		if (resize(size*2)!=_True)
			return *this;

	memcpy((void*)((unsigned int)code+length), m.code, m.length);
	length+=m.length;
	setassembly(new Pair(new String(m.name), assembly()));
	return *this;
}

NodeP Compiled::resize(int newsize)
{
	FunctionP newcode=FunctionP(GC_::realloc(code, newsize));
	if (newcode && newcode!=code) 
	{	code=newcode; size=newsize; return _True;
	}
	return _False;
}

void Compiled::erase_backto(int oldlength) {length=oldlength;}
