#ifndef Compiled_included
#ifndef Node_included
	#include "Node.h"
#endif

enum Reg {AX, CX, DX, BX, SP, BP, SI, DI,  SIB=SP, DISP32=BP}; // The Reg field
			// in R/M: SP indicates a SIB (Shifted Index Byte) following the ModR/M byte
			// in R/M with Mod=0: BP indicates displacement32

enum Mod {INDIRECT, INDIRECT8, INDIRECT32, REG};	// The Mod field

enum JmpCond {O,NO, B,NB, Z,NZ, BE,NBE, S,NS, P,NP, L,NL, LE, NLE};

enum OpCode 
{	add=0x01, addr=add|2, addc=0x81, 
	call=0xE8, 
	cmp=0x39, cmpr=cmp|2, cmpc=0x81,
	decr=0x48, decm=0xFF, 
	enter=0xC8, 
	incr=0x40, incm=0xFF, 
	jcc=0x0F, jcc2=0x80, 
	jmp=0xE9,
	leave=0xC9, 
	mov=0x89, movr=mov|2, movc=0xB8, movmc=0xC7,
	pushr=0x50, push_=0xFF, pushc=0x68, popr=0x58, pop=0x8F, 
	ret=0xC3, 
	sub=0x29, subr=sub|2, subc=0x81,
};

class ModRM 
{	public: 
		char b;
		ModRM(Mod mod, Reg reg, Reg rm): b(char((mod<<6)+(reg<<3)+rm)) {}
		operator char() {return b;}
};

class CodeMacro
{	public:
		int length;
		char code[0x100];
		wxString name;

		CodeMacro(wxString name_, OpCode o);
		CodeMacro(wxString name_, OpCode o, char i);
		CodeMacro(wxString name_, OpCode o, short i);
		CodeMacro(wxString name_, OpCode o, short i, char i2);
		CodeMacro(wxString name_, OpCode o, int i);
		CodeMacro(wxString name_, OpCode o, Reg r);
		CodeMacro(wxString name_, OpCode o, Reg r, int i);
		CodeMacro(wxString name_, OpCode o, ModRM m);
		CodeMacro(wxString name_, OpCode o, ModRM m, void *mem);
		CodeMacro(wxString name_, OpCode o, ModRM m, void *mem, int i);
		CodeMacro(wxString name_, OpCode o, ModRM m, char c);
		CodeMacro(wxString name_, OpCode o, ModRM m, int c);
		CodeMacro(wxString name_, OpCode o, OpCode o2, JmpCond jc, int i);
		
		friend File &operator<<(File &f, const CodeMacro &m) {return f << m.name;}
};

class Compiled: public Node
{
	public:
		ObjectP _object;
		PairP _assembly;
		PtrTableP links, refs;
		FunctionP code;
		int length, size;
		union {ObjectType flag; ObjectBits bit; int tint;};

		static const Compiled aCompiled;

		operator_new;
		Compiled() { }
		Compiled(ObjectP object_, int size_=0x10000);
		void free();

		ObjectP object() {return _object;}
		PairP assembly() {return _assembly;}

		void setassembly(PairP nd);
		void setlinks(PtrTableP nd);
		void setrefs(PtrTableP nd);

		void inc_gc();
		int youngs();
		int youngsPtr(PtrToInt_ &ptrToInt);
		void check_oldyoungs();

		void setPtrs(IntToPtr_ &intToPtr);
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		NodeP copy(PtrToInt_ &ptrToInt);
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		void printBin(FILEP strm);

		NodeP contentsTree(	wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont, 
									ObjectType must, 
									int &count, int level);
		bool expanded() {return bit.table_expand;}
		bool expandable() {return true;}
		void setExpanded(bool state) {bit.table_expand=state;}

		int compare(NodeP nd);
		int count();
		int _sizeof();
		StringP _typeof();
		bool equalP(NodeP b);
		NodeP first_();
		NodeP butFirst();
		NodeP last();
		NodeP butLast();
		PairP list();

		NodeP call(PairPP arg);
		Compiled &operator<<(const CodeMacro &m);
		NodeP resize(int newsize);
		NodeP link();
		NodeP relocate(charP newcode);
		void erase_backto(int oldlength);
};

#define Compiled_included
#endif