
#include "stdafx.h"

#ifndef WIN32
	#include <termios.h>
#endif

#include "Main.h"
#include "Node.h"
#include "File.h"
#include "Pair.h"
#include "Int.h"
#include "Color.h"
#include "Point.h"
#include "History.h"
#include "Graph.h"
#include "LogoApp.h"
#include "Error.h"
#include "Event.h"
#include "Thread.h"
#include "Consolew.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

Consolew *consolew;

BEGIN_EVENT_TABLE(Consolew, wxFrame)
	EVT_CHAR				(Consolew::OnChar					)
	EVT_KEY_DOWN		(Consolew::OnKeyDown				)
	EVT_KEY_UP			(Consolew::OnKeyUp				)
END_EVENT_TABLE()

Consolew::Consolew():
wxFrame(0, -1, _T("aUCBLogo Consolew"), 
	wxDefaultPosition, wxDefaultSize, 
	wxDEFAULT_FRAME_STYLE, _T("Consolew"))
{
#ifdef WIN32
	HMODULE kernel32 = LoadLibrary("kernel32.dll");
	GetConsoleWindow = (TGetConsoleWindow)
		GetProcAddress(kernel32, "GetConsoleWindow");
	if (GetConsoleWindow)
	{	hwnd=GetConsoleWindow();
		consoleWP.length=sizeof(WINDOWPLACEMENT);
		GetWindowPlacement(hwnd, &consoleWP);
		hConsoleInput=GetStdHandle(STD_INPUT_HANDLE);
		hConsoleOutput=GetStdHandle(STD_OUTPUT_HANDLE);
		SetConsoleMode(hConsoleInput,
			ENABLE_ECHO_INPUT | 
//			ENABLE_INSERT_MODE | 
			0x20 |
			ENABLE_MOUSE_INPUT |
//			ENABLE_PROCESSED_INPUT |
			ENABLE_WINDOW_INPUT);
		SetConsoleMode(hConsoleOutput,
			ENABLE_PROCESSED_OUTPUT);
		SetConsoleCP(1200);
	}else
		hwnd=0;
#else
//	TODO: GetConsoleWindow() in Linux???
	hwnd=0;
#endif

	lineAvailable=false;
	charAvailable=false;
	readIndex=0;
	pos=0;
	coord.X=0;
	coord.Y=0;
	coord0=coord;
}

Consolew::~Consolew()
{
#ifdef WIN32
	show(consoleWP.showCmd);
#else
//TODO
#endif
}

void Consolew::show(int state)
{
	if (hwnd)
#ifdef WIN32
		ShowWindow(hwnd, state);
#else
//TODO
		;
#endif
}

void Consolew::SetFocus()
{
#ifdef WIN32
	::SetFocus(hwnd);
#else
//TODO
#endif
}

wxPoint Consolew::GetPosition()
{
#ifdef WIN32
	RECT rect;
	GetWindowRect(hwnd, &rect);
	return wxPoint(rect.left, rect.top);
#else
//TODO
	return wxPoint(0,0);
#endif
}

wxSize Consolew::GetSize()
{
#ifdef WIN32
	RECT rect;
	GetWindowRect(hwnd, &rect);
	return wxSize(rect.right-rect.left, rect.bottom-rect.top);
#else
//TODO
	return wxSize(600,400);
#endif
}

NodeP Consolew::_Cursor()
{
#ifdef WIN32
	long x=1,y=1;
	CONSOLE_SCREEN_BUFFER_INFO ConsoleScreenBufferInfo;
	BOOL result=GetConsoleScreenBufferInfo(
		hConsoleOutput, &ConsoleScreenBufferInfo);
	if (result)
		return new Pair
		(	new Int(ConsoleScreenBufferInfo.dwCursorPosition.X),
			new Pair(new Int(ConsoleScreenBufferInfo.dwCursorPosition.X),0) 
		);
	else
		return Unbound;
#else
//TODO
	return Unbound;
#endif
}

NodeP Consolew::_setCursor(PairP arg)
{
	PairP vec = arg->a_pos_int_vector(1);
	if (!throwing) 
	{	int x=IntP(vec-> first())->int_,
			 y=IntP(vec->second())->int_;

		if ((x < 0 || y < 0) && !throwing) 
		{	error(BAD_DATA, vec); return Unbound;
		}
#ifdef WIN32
		COORD dwCursorPosition={x,y};
		BOOL result=SetConsoleCursorPosition(
			consolew->hConsoleOutput, dwCursorPosition);	
#else
//TODO
#endif
	}
	return Unbound;
}

NodeP Consolew::_setTextColor(PairPP arg)
{
	int c=arg[1]->a_color(1);
#ifdef WIN32
	BOOL result=SetConsoleTextAttribute(
		hConsoleOutput, c);	
#else
//TODO
#endif
	return Unbound;
}

void Consolew::SetInsertionCoord(int x, int y)
{
	pos=x;
	coord0.X=x;
	coord0.Y=y;
}

void Consolew::WriteText(wxString str)
{
	readLine+=str;
#ifdef WIN32
	DWORD cNumWritten;
	WriteConsoleOutputCharacter(hConsoleOutput, 
		(readLine+' ').c_str(), readLine.Length()+1, coord0, &cNumWritten);
	SetConsoleCursorPosition(hConsoleOutput, coord);
#else
//TODO
//	::puts(str);
#endif
}

void Consolew::finishLine(wxString line)
{
	readLine=line;
	theApp->history->append(line);
	lineAvailable=true;
	readIndex=0;
	pos=0;
	coord.X=0;
	coord.Y++;
}

void Consolew::ClientToScreen(int *x, int *y)
{
	wxPoint p=GetPosition();
	wxSize s=GetSize();
#ifdef WIN32
	BOOL result=FALSE;
//	CONSOLE_FONT_INFO ConsoleCurrentFont;
//	BOOL result=GetCurrentConsoleFont(
//		hConsoleOutput, FALSE, &ConsoleCurrentFont);
//	COORD fs=GetConsoleFontSize(
//		hConsoleOutput, ConsoleCurrentFont.nFont);
	COORD fs={8,10};
	CONSOLE_SCREEN_BUFFER_INFO ConsoleScreenBufferInfo;
	result=GetConsoleScreenBufferInfo(
		hConsoleOutput, &ConsoleScreenBufferInfo);
	SMALL_RECT sw=ConsoleScreenBufferInfo.srWindow;
	TITLEBARINFO pti;
	pti.cbSize=sizeof(TITLEBARINFO);
	GetTitleBarInfo(hwnd, &pti);
	*x=p.x;
	if (fs.Y)
		*y=pti.rcTitleBar.bottom+(coord.Y-sw.Top)*fs.Y;
#else
//TODO
	*x=0;
	*y=0;
#endif
}

void Consolew::charModeOn()
{
	charMode=true; 
}

void Consolew::charModeOff() 
{
	if (!EventChar::function)
		charMode=false;
}

bool Consolew::kbhit()
{
#ifdef WIN32
	return charAvailable;
#else
	struct termios term, oterm; 
	int fd = 0; 
	int c = 0; 
	tcgetattr(fd, &oterm); 
	memcpy(&term, &oterm, sizeof(term)); 
	term.c_lflag = term.c_lflag & (!ICANON); 
	term.c_cc[VMIN] = 0; 
	term.c_cc[VTIME] = 1; 
	tcsetattr(fd, TCSANOW, &term); 
	c = getchar(); 
	tcsetattr(fd, TCSANOW, &oterm); 
	if (c != -1) 
		ungetc(c, stdin); 
	return (c != -1); 
#endif
}

char Consolew::getch()
{
#ifdef WIN32
	charAvailable=false;
#else
	ch=::getchar();
#endif
	return ch;
}

int Consolew::getchExt()
{
	charAvailable=false;
	return chExt;
}

char Consolew::getc()
{
	char c;
#ifdef WIN32
	BOOL result;
	DWORD cNumRead, cNumWritten;
	ThreadP th=Thread::This();
	if (!lineAvailable) 
	{	if (th->graph)
			th->graph->updateGraph();
		CONSOLE_SCREEN_BUFFER_INFO screenInfo;
		GetConsoleScreenBufferInfo(hConsoleOutput, &screenInfo);
		coord0=coord=screenInfo.dwCursorPosition;
		readLine=_T("");
		pos=0;

		while (!throwing)
		{	do
			{	if (th->GuiTryEnter())
				{	//if (th->graph)
					//	th->graph->mutexLeaveOrEnter();

					MSG msg;
					if (::PeekMessage(&msg, hwnd, WM_KEYFIRST, WM_KEYLAST, PM_NOREMOVE))
						break;
//					theApp->dispatch();
//					theApp->processIdle();
					if (::PeekMessage(&msg, 0, 0, 0, PM_NOREMOVE))
					{	BOOL rc=::GetMessage(&msg, (HWND) NULL, 0, 0);
						if (!rc)
						{	theApp->action=EXIT_LOGO;
						}
						::TranslateMessage(&msg);
						::DispatchMessage(&msg);
					}
					ACTION;
					switch (theApp->action)
					{
					case EXIT_LOGO:
					case EXIT_DELAYED:
					case RECREATE_GUI:
					case HIDE_MAIN:
						return EOF;
					}
					th->GuiLeave();
					Sleep(50);
				}
				result=GetNumberOfConsoleInputEvents(
					hConsoleInput, &cNumRead);
				if (lineAvailable)
					break;
			} while (!throwing && cNumRead==0);
			
			INPUT_RECORD irInBuf[128];
			ReadConsoleInput( 
					hConsoleInput,      // input buffer handle 
					irInBuf,     // buffer to read into 
					128,         // size of read buffer 
					&cNumRead);	 // number of records read 
			// Dispatch the events to the appropriate handler. 
	 
			for (DWORD i = 0; i < cNumRead; i++) 
			{
				switch(irInBuf[i].EventType) 
				{ 
				case KEY_EVENT: // keyboard input 
				{	KEY_EVENT_RECORD kev=irInBuf[i].Event.KeyEvent;
					if (kev.bKeyDown)
					{	c=kev.uChar.AsciiChar; 
						switch (kev.wVirtualKeyCode)
						{
						case VK_SHIFT:
						case VK_CONTROL:
						case VK_MENU:
						case VK_CAPITAL:
							break;

						case VK_BACK:
							if (pos > 0)
							{	readLine.Remove(pos-1, 1);
								pos--;
								coord.X--;
							}
							break;

						case VK_DELETE:
							if (pos < readLine.Length())
							{	readLine.Remove(pos, 1);
							}
							break;

						case VK_LEFT:
							if (pos > 0)
							{	pos--;
								coord.X--;
							}
							break;

						case VK_RIGHT:
							if (pos < readLine.Length())
							{	pos++;
								coord.X++;
							}
							break;

						case VK_HOME:
							pos=0;
							coord.X=0;
							break;

						case VK_END:
							pos=readLine.Length();
							coord.X=pos;
							break;

						case VK_UP:
						case VK_DOWN:
							theApp->history->show(coord.Y+GetPosition().y, readLine);
							while (theApp->history->win->IsShown())
								th->dispatch();
							break;

						default:
							if (c=='\r')
							{	c='\n';
								readLine+=c;
							}else
								readLine=readLine.Left(pos)+c+readLine.Mid(pos);
							pos++;
							coord.X++;
							if (c=='\n')
							{	finishLine(readLine);
								SetConsoleCursorPosition(hConsoleOutput, coord);
								goto ready;
							}
						}
						WriteConsoleOutputCharacter(hConsoleOutput, 
							(readLine+' ').c_str(), readLine.Length()+1, coord0, &cNumWritten);
						SetConsoleCursorPosition(hConsoleOutput, coord);
					}
					break; 
				}
				case MOUSE_EVENT: // mouse input 
				case WINDOW_BUFFER_SIZE_EVENT: // scrn buf. resizing 
				case FOCUS_EVENT:  // disregard focus events 
				case MENU_EVENT:   // disregard menu events 
				break; 
				} 
			}
			if (lineAvailable)
			{	if (readLine.Last()!='\n')
					readLine+='\n';
				break;
			}
		}
	}
ready:
	if (readIndex <= readLine.length())
	{	c=readLine[readIndex];
		readIndex++;
	}
	if (c=='q'-'a'+1 || c=='p'-'a'+1 || c=='\n') // control-q, control-p
	{	lineAvailable=false;
		readIndex=0;
//		readLine.resize(0);
		charAvailable=false;
		if (readLine=="reset\n")
			readLine=oldReadLine;
	}
	return c;
#else
//TODO
	return ::getchar();
#endif
}

void Consolew::OnChar(wxKeyEvent &ev)
{
}

void Consolew::OnKeyDown(wxKeyEvent &event)
{
	ThreadP th=Thread::This();
	int c=event.GetKeyCode(), o=0;
	EventP evt=new EventChar(event);
	if throwing return;
	*th->eventFiFo << evt;
	if (charMode) 
	{	if (theApp->action!=NOP)
			return;
		switch (c)
		{
		case WXK_CANCEL:
			WriteText('\n');
			theApp->action=CONTROL_PAUSE;
			return;
		
		case WXK_PAUSE:
			WriteText('\n');
			if (event.ControlDown())
				theApp->action=CONTROL_PAUSE;
			else
				theApp->action=PAUSE;
			return;
		}
		if (c > 255)
		{	ch=char(255);
			chExt=c;
		}else
			ch=char(c);
		charAvailable=true;
		// we must handle Events ourselfes, 
		// because no foreign code is allowed on the stack for the persistence
		return;
	}
	switch(c)
	{
	case WXK_CANCEL:	WriteText('\n');	
							theApp->action=CONTROL_PAUSE;	break;
	case WXK_PAUSE:	WriteText('\n');
							if (event.ControlDown())
								theApp->action=CONTROL_PAUSE;
							else
								theApp->action=PAUSE;
							return;
	default:				
		if (event.ControlDown())
			switch (c)
			{
			case WXK_UP:
			case WXK_DOWN: 
					theApp->history->show(coord.Y+GetPosition().y, readLine);
					return;
			}
	}

	bool capslock=wxGetKeyState(WXK_CAPITAL),
		numlock=wxGetKeyState(WXK_NUMLOCK),
		scrolllock=wxGetKeyState(WXK_SCROLL),
		enhanced=c > 0xff;

	if (c > 0xff)
		switch (c)
		{
		case WXK_NUMPAD0: o='0'; break;
		case WXK_NUMPAD1: o='1'; break;
		case WXK_NUMPAD2: o='2'; break;
		case WXK_NUMPAD3: o='3'; break;
		case WXK_NUMPAD4: o='4'; break;
		case WXK_NUMPAD5: o='5'; break;
		case WXK_NUMPAD6: o='6'; break;
		case WXK_NUMPAD7: o='7'; break;
		case WXK_NUMPAD8: o='8'; break;
		case WXK_NUMPAD9: o='9'; break;
		case WXK_NUMPAD_DECIMAL: o='.'; break;
		case WXK_NUMPAD_MULTIPLY:	o='*'; break;
		case WXK_NUMPAD_DIVIDE:		o='/'; break;
		case WXK_NUMPAD_SUBTRACT:	o='-'; break;
		case WXK_NUMPAD_ADD:			o='+'; break;
		default:
			o=c;
		}
	else
	{	if (event.ShiftDown()^capslock)
		{	switch (c)
			{
			case '\\': o='�'; break;
			case '1': o='!'; break;
			case '2': o='"'; break;
			case '3': o='�'; break;
			case '4': o='$'; break;
			case '5': o='%'; break;
			case '6': o='&'; break;
			case '7': o='/'; break;
			case '8': o='('; break;
			case '9': o=')'; break;
			case '0': o='='; break;
			case '[': o='?'; break;
			case '�': o='`'; break;
			case '<': o='>'; break;
			case ',': o=';'; break;
			case '.': o=':'; break;
			case '-': o='_'; break;
			case '/': o='\'';break;
			case '+': o='*'; break;
			case '~': o='�'; break;
			case '\'':o='�'; break;
			case ';': o='�'; break;
			default:
				o=event.GetKeyCode();
			}
		}else if (event.AltDown())
		{	switch (c)
			{
			case '2': o='�'; break;
			case '3': o='�'; break;
			case '7': o='{'; break;
			case '8': o='['; break;
			case '9': o=']'; break;
			case '0': o='}'; break;
			case '�': o='\\'; break;
			case '<': o='|'; break;
			case 'Q': o='@'; break;
			case '+': o='~'; break;
			default:
				o=event.GetKeyCode();
			}
		}else
			o=wxTolower((unsigned char)(event.GetKeyCode()));
	}
#ifdef WIN32
	INPUT_RECORD ir;
	ir.EventType=KEY_EVENT;
	ir.Event.KeyEvent.wRepeatCount=1;
	ir.Event.KeyEvent.uChar.AsciiChar=o;
	ir.Event.KeyEvent.wVirtualKeyCode=event.GetRawKeyCode();
	ir.Event.KeyEvent.bKeyDown=TRUE;

	ir.Event.KeyEvent.dwControlKeyState=
		event.ShiftDown()*SHIFT_PRESSED+
		event.ControlDown()*LEFT_CTRL_PRESSED+
		event.AltDown()*LEFT_ALT_PRESSED+
		event.MetaDown()*RIGHT_ALT_PRESSED+
		capslock*CAPSLOCK_ON+
		numlock*NUMLOCK_ON+
		scrolllock*SCROLLLOCK_ON+
		enhanced*ENHANCED_KEY;

	DWORD numWritten;
	BOOL result=WriteConsoleInput(
		hConsoleInput, &ir, 1, &numWritten);
#else
//TODO
#endif
}

void Consolew::OnKeyUp(wxKeyEvent &ev)
{
#ifdef WIN32
	INPUT_RECORD ir;
	ir.EventType=KEY_EVENT;
	ir.Event.KeyEvent.wRepeatCount=1;
	if (ev.ShiftDown()^wxGetKeyState(WXK_CAPITAL))
		ir.Event.KeyEvent.uChar.AsciiChar=ev.GetRawKeyCode();
	else
		ir.Event.KeyEvent.uChar.AsciiChar=wxTolower(unsigned char(ev.GetRawKeyCode()));
	ir.Event.KeyEvent.wVirtualKeyCode=ev.GetRawKeyCode();
	ir.Event.KeyEvent.bKeyDown=FALSE;
	ir.Event.KeyEvent.dwControlKeyState=ev.ShiftDown()*SHIFT_PRESSED;
	DWORD numWritten;
	BOOL result=WriteConsoleInput(
		hConsoleInput, &ir, 1, &numWritten);
#else
//TODO
#endif
}
