//___________________________________________________________________
//
//                      aUCBLogo Editor module
//                         based on stctest
//            written by Otto Wyss and Andreas Micheler
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/filedlg.h>  // file dialog support
#include <wx/filename.h> // filename support
#include <wx/notebook.h> // notebook support
#include <wx/settings.h> // system settings
#include <wx/string.h>   // strings support
#include <wx/image.h>    // images support
#include <wx/html/helpctrl.h> // html help
#include <wx/listctrl.h>
#include <wx/numdlg.h>
#include <wx/aui/aui.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>

#include <fstream>

#include "Main.h"
#include "String.h"
#include "File.h"
#include "Pair.h"
#include "Line.h"
#include "List.h"
#include "Quote.h"
#include "Tree.h"
#include "Error.h"
#include "Eval.h"
#include "Object.h"
#include "Table.h"
#include "UInt8.h"
#include "Point.h"
#include "Procedure.h"
#include "LogoObject.h"

#include "Stack.h"
#include "NodePStack.h"
#include "Workspace.h"
#include "Files.h"
#include "Init.h"
#include "Color.h"

#include "EdDefsExt.h"     // Additional definitions
#include "Edit.h"        // Edit module
#include "EdPrefs.h"       // Prefs
#include "EdFrame.h"
#include "Console.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "AboutDlg.h"
#include "Thread.h"
#include "Vars.h"
#include "Calls.h"
#include "DebugState.h"
#include "Aui.h"

// the application icon (under Windows and OS/2 it is in resources)
#if defined(__WXGTK__) || defined(__WXMOTIF__) || defined(__WXMAC__) || defined(__WXMGL__) || defined(__WXX11__)
//	#include "mondrian.xpm"
#endif

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________
//
//                           declarations
//___________________________________________________________________

#define APP_NAME _T("aUCBLogo")
#define APP_MAINT _T("Andreas Micheler")
#define APP_LICENCE _T("GNU GPL")

#define APP_VERSION _T("4.7")
#define APP_BUILD __DATE__

#define APP_WEBSITE _T("http://www.Physik.Uni-Augsburg.de/~micheler")
#define APP_MAIL _T("mailto:Andreas.Micheler@Physik.Uni-Augsburg.de")

wxString NONAME=_("noname.lg");

wxPrintData *g_printData=0;
wxPageSetupData *g_pageSetupData=0;

#ifdef WIN32
bool FileDropTarget::OnDropFiles(wxCoord WXUNUSED(x), wxCoord WXUNUSED(y),
											const wxArrayString& filenames)
{
	for (size_t i=0; i<filenames.Count(); i++)
	{	theApp->editorShow();
		if (theApp->edFrame)
			theApp->edFrame->FileOpen(filenames[i]);
	}
	return true;
}
#endif

//___________________________________________________________________
//
//                            EdFrame
//___________________________________________________________________

BEGIN_EVENT_TABLE (EdFrame, wxPanel)
	// common
	EVT_CLOSE (                      EdFrame::OnClose)
	// file
	EVT_MENU (wxID_NEW,              EdFrame::OnFileNew)
	EVT_MENU (wxID_OPEN,             EdFrame::OnFileOpen)
	EVT_MENU (wxID_SAVE,             EdFrame::OnFileSave)
	EVT_MENU (wxID_SAVEAS,           EdFrame::OnFileSaveAs)
	EVT_MENU (myID_EXPORT_HTML,      EdFrame::OnEdit)
	EVT_MENU (wxID_CLOSE,            EdFrame::OnFileClose)
	// properties
	EVT_MENU (myID_PROPERTIES,       EdFrame::OnProperties)
	// print and exit
	EVT_MENU (wxID_PRINT_SETUP,      EdFrame::OnPrintSetup)
	EVT_MENU (wxID_PREVIEW,          EdFrame::OnPrintPreview)
	EVT_MENU (wxID_PRINT,            EdFrame::OnPrint)
	EVT_MENU (wxID_EXIT,             EdFrame::OnExit)
	// edit
	EVT_MENU (wxID_CLEAR,            EdFrame::OnEdit)
	EVT_MENU (wxID_CUT,              EdFrame::OnEdit)
	EVT_MENU (wxID_COPY,             EdFrame::OnEdit)
	EVT_MENU (wxID_PASTE,            EdFrame::OnEdit)
	EVT_MENU (myID_INDENTINC,        EdFrame::OnEdit)
	EVT_MENU (myID_INDENTRED,        EdFrame::OnEdit)
	EVT_MENU (wxID_SELECTALL,        EdFrame::OnEdit)
	EVT_MENU (myID_SELECTLINE,       EdFrame::OnEdit)
	EVT_MENU (wxID_REDO,             EdFrame::OnEdit)
	EVT_MENU (wxID_UNDO,             EdFrame::OnEdit)
	// find
	EVT_MENU (wxID_FIND,             EdFrame::OnEdit)
	EVT_MENU (myID_FINDNEXT,         EdFrame::OnEdit)
	EVT_MENU (myID_FINDPREVIOUS,     EdFrame::OnEdit)
	EVT_MENU (myID_REPLACE,          EdFrame::OnEdit)
	EVT_MENU (myID_REPLACENEXT,      EdFrame::OnEdit)
	EVT_MENU (myID_REPLACEALL,       EdFrame::OnEdit)
	EVT_MENU (myID_FINDWHOLEWORD,    EdFrame::OnEdit)
	EVT_MENU (myID_SKIP_COMMENTS,    EdFrame::OnEdit)
	EVT_MENU (myID_BRACEMATCH,       EdFrame::OnEdit)
	EVT_MENU (myID_GOTO,             EdFrame::OnEdit)
	EVT_MENU (myID_GOTODEF,          EdFrame::OnEdit)
	EVT_MENU (myID_GOTOPREV,         EdFrame::OnEdit)
	// view
	EVT_MENU_RANGE (myID_HILIGHTFIRST, myID_HILIGHTLAST,
								 EdFrame::OnEdit)
	EVT_MENU (myID_DISPLAYEOL,       EdFrame::OnEdit)
	EVT_MENU (myID_INDENTGUIDE,      EdFrame::OnEdit)
	EVT_MENU (myID_LINENUMBER,       EdFrame::OnEdit)
	EVT_MENU (myID_LONGLINEON,       EdFrame::OnEdit)
	EVT_MENU (myID_WHITESPACE,       EdFrame::OnEdit)
	EVT_MENU (myID_FOLDTOGGLE,       EdFrame::OnEdit)
	EVT_MENU (myID_FOLDTOGGLEALL,    EdFrame::OnEdit)
	EVT_MENU (myID_OVERTYPE,         EdFrame::OnEdit)
	EVT_MENU (myID_READONLY,         EdFrame::OnEdit)
	EVT_MENU (myID_WRAPMODEON,       EdFrame::OnEdit)
	EVT_MENU (myID_FONT,             EdFrame::OnSetFont)
	// run
	EVT_MENU (myID_RUN,              EdFrame::callOnRun)
	EVT_MENU (myID_STEP,             EdFrame::callOnStep)
	EVT_MENU (myID_REVERSE_STEP,     EdFrame::callOnReverseStep)
	EVT_MENU (myID_STEP_OVER,        EdFrame::callOnStepOver)
	EVT_MENU (myID_REVERSE_STEP_OVER,EdFrame::callOnReverseStepOver)
	EVT_MENU (myID_BREAKPOINT,       EdFrame::OnBreakpoint)
	EVT_MENU (myID_LOAD,             EdFrame::OnLoad)
	EVT_MENU (myID_AUTOLOAD,         EdFrame::OnAutoLoad)
	EVT_MENU (myID_SETRINGBUFFER,    EdFrame::OnSetRingBuffer)
	EVT_MENU (myID_SETRUNDIR,        EdFrame::OnSetRunDir)

	// extra
	EVT_MENU (myID_BACKUP,           EdFrame::OnMakeBackupFiles)
	EVT_MENU (myID_CHANGEUPLOW,      EdFrame::OnEdit)
	EVT_MENU (myID_CHANGELOWER,      EdFrame::OnEdit)
	EVT_MENU (myID_CHANGEUPPER,      EdFrame::OnEdit)
	EVT_MENU (myID_CONVERTCR,        EdFrame::OnEdit)
	EVT_MENU (myID_CONVERTCRLF,      EdFrame::OnEdit)
	EVT_MENU (myID_CONVERTLF,        EdFrame::OnEdit)
	EVT_MENU (myID_MAKETOASSIGN,     EdFrame::OnEdit)
	EVT_MENU (myID_CHARSETANSI,      EdFrame::OnEdit)
	EVT_MENU (myID_CHARSETMAC,       EdFrame::OnEdit)
	// window
	EVT_MENU (myID_PAGENEXT,       EdFrame::OnPageNext)
	EVT_MENU (myID_PAGEPREV,       EdFrame::OnPagePrev)
	EVT_MENU (myID_CONSOLE_SET_FOCUS,EdFrame::OnConsoleSetFocus)

	// help
	EVT_MENU (wxID_HELP,             EdFrame::OnEdit)
	EVT_MENU (myID_SETHELPFILENAME,  EdFrame::OnSetHelpFileName)
	EVT_MENU (wxID_ABOUT,            EdFrame::OnAbout)
	EVT_NOTEBOOK_PAGE_CHANGED(myID_NOTEBOOKSEL, EdFrame::OnSelChange)
	EVT_SET_FOCUS		(EdFrame::OnSetFocus)
END_EVENT_TABLE ()

EdFrame::EdFrame (wxWindow *parent, const wxString &title, wxPoint pos, wxSize size,
						  int argc, wxChar** argv)
:wxPanel(parent, wxID_ANY, pos, size, 
			wxNO_BORDER // | wxFRAME_FLOAT_ON_PARENT 
//			wxFRAME_FLOAT_ON_PARENT | wxCAPTION
//			(wxDEFAULT_FRAME_STYLE | wxNO_FULL_REPAINT_ON_RESIZE)
			)
{
	inReverseStep=false;
	fontSize=theApp->editFontSize;
	// intitialize important variables
	g_printData = new wxPrintData;
	g_pageSetupData = new wxPageSetupDialogData;

	// set icon and background
//	SetTitle(aUCBLogo+_T(" Editor"));
#ifdef WIN32
//	SetIcons(wxIconBundle(wxIcon(
//		theApp->runDirectory+"/logo.ico", wxBITMAP_TYPE_ICO)));
//	SetIcons(wxIconBundle(wxIcon("ICON_LOGO", wxBITMAP_TYPE_ICO_RESOURCE)));
#endif
	SetBackgroundColour(_T("WHITE"));

	// create menu
	m_menuBar = new wxFlatMenuBar(this);
	CreateMenu();

	notebook = new wxNotebook(this, myID_NOTEBOOKSEL);

	wxBoxSizer *sizer=new wxBoxSizer(wxVERTICAL);
	sizer->Add(m_menuBar, 0, wxEXPAND);
	sizer->Add(notebook, 1, wxEXPAND);
	SetSizer(sizer);

	Show();
	if (argc <= 1)
		;//OnFileNew(wxCommandEvent());
	else 
	if (wxGetUserName().Upper()!="SYSTEM" 
	&&  wxGetUserName().Upper()!=".DEFAULT")
		for (int i=1; i<argc; i++)
		{	wxString s(argv[i]);
			if (s.StartsWith("-run"))
				break;
			if (wxFileExists(s))
				FileOpen(s);
		}

#ifdef WIN32
	fileDropTarget=new FileDropTarget();
	SetDropTarget(fileDropTarget);
#endif
	theApp->editorShown=true;
	show();
	if (theApp->mainFrame && theApp->mainFrame->viewMenuEditor)
		theApp->mainFrame->viewMenuEditor->Check(theApp->editorShown);
}

EdFrame::~EdFrame()
{
	if (g_printData) delete g_printData; g_printData=0;
	if (g_pageSetupData) delete g_pageSetupData; g_pageSetupData=0;
	if (theApp)
	{	theApp->edFrame=0;
		theApp->editorShown=false;
		if (theApp->mainFrame && theApp->mainFrame->viewMenuEditor)
			theApp->mainFrame->viewMenuEditor->Check(theApp->editorShown);
	}
}

void EdFrame::show(bool shown)
{
	wxAuiPaneInfo &pane=theApp->aui->GetPane("Editor");
	if (pane.IsOk())
	{	pane.Show();
	}else
		theApp->aui->AddPane(this, 
			wxAuiPaneInfo()
			.Name("Editor").Caption(_("Editor"))
			.Layer(2).Right()
			.BestSize(600,1000)
			.MinimizeButton().MaximizeButton().PinButton()
			.DestroyOnClose());
	theApp->aui->Update();
}

// common event handlers
void EdFrame::OnClose (wxCloseEvent &event)
{
	theApp->saveEditorConfig();
	wxCommandEvent evt;
	OnFileClose (evt);
	for (size_t i=1; i < notebook->GetPageCount(); i++)
	{	Edit *edit = (Edit*)notebook->GetPage(i);
		if (edit && edit->Modified())
		{	if (event.CanVeto()) 
				event.Veto (true);
			if (event.GetVeto())
				return;
		}
	}
	theApp->aui->DetachPane(this);
	delete this;
}

void EdFrame::OnAbout (wxCommandEvent &WXUNUSED(event))
{
	AboutDlg aboutDlg;
	aboutDlg.ShowModal();
}

void EdFrame::OnExit (wxCommandEvent &WXUNUSED(event))
{
	theApp->action=EXIT_LOGO;
//	Close (true);
}

// file event handlers
void EdFrame::OnFileNew(wxCommandEvent &WXUNUSED(event))
{
	wxString fname=NONAME;
	Edit *edit = new Edit(fname, notebook, wxID_ANY, 
		wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER|wxVSCROLL, fontSize);
	if (!edit)
	{	error(OUT_OF_MEM); return;
	}
	notebook->AddPage(edit, NONAME, true);
//	SetTitle(NONAME+_T(" - ")+aUCBLogo+_T(" Editor"));
	edit->InitializePrefs(edit->DeterminePrefs(fname));
	edit->SetFocus();
}

void EdFrame::OnFileOpen (wxCommandEvent &WXUNUSED(event))
{
	wxString fname;
	wxFileDialog dlg (this, _("Open file"), _T(""), _T(""), 
		_("Logo Files (*.lg)|*.lg|All Files (*.*)|*.*"),
		wxOPEN | wxFILE_MUST_EXIST | wxCHANGE_DIR);
	if (dlg.ShowModal() != wxID_OK) return;
	fname = dlg.GetPath ();
	FileOpen (fname);
}

void EdFrame::OnFileSave (wxCommandEvent &WXUNUSED(event))
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (!edit->Modified())
	 {   wxMessageBox (_("There is nothing to save!"), _("Save file"),
					  wxOK | wxICON_EXCLAMATION);
		return;
	}
	if (edit->GetFilename()==NONAME)
	{	wxCommandEvent cev;		
		OnFileSaveAs(cev);
		return;
	}
	edit->SaveFile ();
}

void EdFrame::OnFileSaveAs (wxCommandEvent &WXUNUSED(event))
{
	wxString filename = wxEmptyString;
	wxFileDialog dlg (this, _("Save file"), _T(""), _T(""), 
		_("Logo Files (*.lg)|*.lg|All Files (*.*)|*.*"),
		wxSAVE|wxOVERWRITE_PROMPT);
	if (dlg.ShowModal() != wxID_OK) return;
	filename = dlg.GetPath();
	int page=notebook->GetSelection();
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->SaveFile (filename);
	edit->SetFilename(filename);
	wxFileName fn=filename;
	notebook->SetPageText(page, fn.GetFullName());
//	SetTitle(edit->GetFilename()+_T(" - ")+aUCBLogo+_T(" Editor"));
}

void EdFrame::OnFileClose (wxCommandEvent &WXUNUSED(event))
{
	int page=notebook->GetSelection();
	if (page < 0)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (!edit)
		return;
	if (edit->Modified())
	{	if (wxMessageBox (_("Text is not saved, save before closing?"), _("Close"),
						  wxYES_NO | wxICON_QUESTION) == wxYES)
		{	edit->SaveFile();
		 if (edit->Modified())
			{	wxMessageBox (_("Text could not be saved!"), _("Close abort"),
							  wxOK | wxICON_EXCLAMATION);
				return;
			}
		}
	}
	notebook->DeletePage(page);
//	if (notebook->GetPageCount()==0)
//		SetTitle(aUCBLogo+_T(" Editor"));
}

void EdFrame::OnSetFont(wxCommandEvent &WXUNUSED(event))
{
	fontSize=wxGetNumberFromUser(_("Input the new font size:"), "(0..100):", _("FontSize"), 10, 4, 100);
	int page=notebook->GetSelection();
	if (page < 0)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (!edit)
		return;
	edit->fontSize=fontSize;
	edit->setFonts();
}

void EdFrame::callOnRun(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_RUN;
}

void EdFrame::OnRun(wxCommandEvent &WXUNUSED(event))
{
	ThreadP th=Thread::This();
	debugStates.switchToLast();
	root->clearflag( PROCS, PROC_STEPPED );
	if (th->thisobj)
		th->thisobj->clearflag( PROCS, PROC_STEPPED );
	run();
}

void EdFrame::OnStep(wxCommandEvent &WXUNUSED(event))
{
	ThreadP th=Thread::This();
	if (th->stepArmed)
		return;
	th->stepArmed=true;
	th->stepNr=1;
	while (th->stepArmed && debugStates.switchToNext() && updateVarsOnStep)
	{	lupdateVars(0);
		lupdateCalls(0);
		if (th->stepNr < 10)
			wxMilliSleep(100-10*th->stepNr);
		th->stepNr++;
	}
	if (!th->stepArmed)
		return;
	root->setflag( PROCS, PROC_STEPPED );
	if (th->thisobj)
		th->thisobj->setflag( PROCS, PROC_STEPPED );
	th->stepArmed=true;
	th->stepNr=1;
	run();
}

void EdFrame::callOnStep(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_STEP;
}

void EdFrame::callOnReverseStep(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_REVERSE_STEP;
}

void EdFrame::OnReverseStep(wxCommandEvent &WXUNUSED(event))
{
	ThreadP th=Thread::This();
	if (th->stepArmed)
		return;
	if (inReverseStep)
		return;
	inReverseStep=true;
	debugStates.switchToPrev();
	if (updateVarsOnStep)
	{	lupdateVars(0);
		lupdateCalls(0);
	}
	inReverseStep=false;
}

void EdFrame::callOnStepOver(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_STEP_OVER;
}

void EdFrame::OnStepOver(wxCommandEvent &WXUNUSED(event))
{
	ThreadP th=Thread::This();
	if (th->stepArmed)
		return;
	if (debugStates.switchToNext() && updateVarsOnStep)
	{	lupdateVars(0);
		lupdateCalls(0);
		return;
	}
	root->clearflag( PROCS, PROC_STEPPED );
	if (th->thisobj)
		th->thisobj->clearflag( PROCS, PROC_STEPPED );
	if (th->ufun)
		th->ufun->bit.proc_stepped=true;
	if (th->last_ufun)
		th->last_ufun->bit.proc_stepped=true;
	th->stepArmed=true;
	run();
}

void EdFrame::callOnReverseStepOver(wxCommandEvent &WXUNUSED(event))
{
	theApp->action=ON_REVERSE_STEP_OVER;
}

void EdFrame::OnReverseStepOver(wxCommandEvent &WXUNUSED(event))
{
	ThreadP th=Thread::This();
	if (th->stepArmed)
		return;
	debugStates.switchToPrev();
	if (updateVarsOnStep)
	{	lupdateVars(0);
		lupdateCalls(0);
	}
}

void EdFrame::run()
{
	ThreadP th=Thread::This();
	console->idle=false;
	debugStates.copyRoot();
	if (th->inpause)
	{	th->stoppingFlag=THROWING;
		th->throw_node=_pause;
		console->readLine='\n';
		console->lineAvailable=true;
		return;
	}
	if (th->instep)
	{	console->readLine='\n';
		console->lineAvailable=true;
		return;
	}
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (edit->Modified())
	{	if (edit->GetFilename()==NONAME)
		{	wxCommandEvent cev;
			OnFileSaveAs(cev);
		}
		else
			edit->SaveFile();
	}
	debugStates.first=debugStates.last;
	StringP myfn=0;
	{	wxString fn=edit->GetFilename();
		NodeP pn=new String(fn);
		if throwing return;
		if (theApp->erallOnRun)
			leraseAll(0);

		NodePP asp=th->argStack->index;
		Push(th->argStack, pn, run);
		lload(PairPP(asp));
		th->argStack->pop_till(PairPP(asp));

		if (!th->startup || th->startup==Unbound)
		{	int pos=fn.Find('.', true);
			fn.Remove(pos);
			pos=fn.Find(wxFileName::GetPathSeparator(), true);
			if (pos>=0)
			{	wxString path=fn.Left(pos);
				if (!wxSetWorkingDirectory(path))
					error(FILE_ERROR, new List(
						new String(_("couldn't change dir to '")),
						new String(path),	_ListEnd));
				fn.Remove(0,pos+1);
			}
			myfn=new String(fn);
			if throwing return;
		}
	}
	unmark(0);
	if (!throwing)
	{	if (theApp->setFocusOnRun)
			console->SetFocus();
		if (!th->startup || th->startup==Unbound)
		{	th->currentLine=myfn;
			PairP execList = th->currentLine->parser(true);
			if (execList) 
			{	th->nodeList=execList;
				th->val_status = no_value_allowed;
				*th->what=__begin_line;
				th->Entry();
			}
		}
		if throwing
		{	     if (th->throw_node==_system  ) return;
			else if (th->throw_node==_error   ) error_print();
			else if (th->throw_node!=_toplevel)
			{	error(NO_CATCH_TAG, th->throw_node);
				error_print();
			}
			th->stoppingFlag=RUN;
		}
		if (stopping || outputting)
		{	c_out << _("; You must be in a procedure to use OUTPUT or STOP.") << endL;
			th->stoppingFlag=RUN;
		}
//		if (th->updateVarsOnStep)
//		{	lupdateVars(0);
//		}
		th->argStack->clean();
	}
	if (th->graph)
		th->graph->updateGraph();
	File::charModeOff();

	console->readLine='\n';
	console->lineAvailable=true;
}

void EdFrame::OnLoad(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (edit->Modified())
	{	if (edit->GetFilename()==NONAME)
		{	wxCommandEvent cev;
			OnFileSaveAs(cev);
		}else
			edit->SaveFile();
	}
	wxString fn=edit->GetFilename();
	NodeP pn=new String(fn);
	if throwing return;
	NodePP asp=th->argStack->index;
	Push2(th->argStack, pn, _Editing, load);
	ControlType sf=th->stoppingFlag;
	NodeP tn=th->throw_node, on=th->output_node;
	LineP nen=_Not_Enough_node, nen2;
	nen2=_Not_Enough_node=new Line(0, 0);
	if throwing return;
	bool tdk=Tree::dont_know_how_to;
	th->autoThisLine=0;
	lloadLocal(PairPP(asp));	// try it first
	if (_Editing->table() != Unbound)
		_Editing->table()->makeTree(false);
	if (th->stoppingFlag==RUN && _Not_Enough_node==nen2 && !Tree::dont_know_how_to)
	{	th->stoppingFlag=sf;
		lload(PairPP(asp));
		if (updateVarsOnStep)
		{	lupdateVars(0);
			lupdateCalls(0);
		}
	}else
	{	if (!th->err_msg && _Not_Enough_node!=nen2)
			error(NOT_ENOUGH, 0);
		error_print();
		th->stoppingFlag=sf;
		th->throw_node=tn;
		th->output_node=on;
		_Not_Enough_node=nen;
		Tree::dont_know_how_to=tdk;
	}
	th->argStack->pop_till(PairPP(asp));
}

void EdFrame::OnAutoLoad(wxCommandEvent &WXUNUSED(event))
{
	theApp->autoloadEdit = !theApp->autoloadEdit;
}

void EdFrame::togglePause(Edit *edit, int col, int row, StringP fn, PairP parent)
{
	PairP list=parent->first(), last=0;
//		cout << parent << endL;
	while (list)
	{	if (list->first()  &&  list->first()->is.list)
			togglePause(edit, col, row, fn, list);
		else
		{	if (LineP(list)->row == row+1  &&  list->first())
			{	edit->togglemark(col, row);

				if (list->first() == NodeP(_pause))
				{	if (list->tail())
					{	list->setfirst(list->tail()->first());
						list->settail(list->tail()->tail());
					}
					break;
				}
				Thread::This()->row=row+1;
				Thread::This()->col=col;
				Thread::This()->setfilename(fn);
				list->settype_LINE();
				LineP line=new Line(_pause, list);
				if throwing return;
				if (last)
					last->settail(line);
				else
					parent->setfirst(line);
				break;
			}
		}
		last=list;
		next(list);
	}
}

void EdFrame::togglePauseTree(Edit *edit, int col, int row, StringP fn, PairP parent, bool clear)
{
	PairP list=0, last=0;
	if (parent->type==TREE)
		list=TreeP(parent)->tree();
	else if (parent->type==QUOTE)
		list=QuoteP(parent)->quote();
	else
		list=parent;
	while (list)
	{	if (LineP(list)->row == row+1)// && list->first()->type!=QUOTE)
		{	
			if (clear)
			{	if (list->first() == NodeP(_pause))
				{	if (list->tail())
					{	list->setfirst(list->tail()->first());
						list->settail(list->tail()->tail());
					}else
						TreeP(parent)->settree(parent->tail());
					break;
				}
			}else
			{	Thread::This()->row=row+1;
				Thread::This()->col=col;
				Thread::This()->setfilename(fn);
				LineP line=new Line(new Line(_pause, 0), list);
				if throwing return;
				if (last)
					last->settail(line);
				else
				{	TreeP(parent)->settree(line);
				}
				break;
			}
		}
		int n=0;
		togglePauseTree2(edit, col, row, fn, list, n, clear);
		last=list;
		next(list);
	}
}

void EdFrame::togglePauseTree2(Edit *edit, int col, int row, StringP fn, PairP parent, int &n, bool clear)
{
	PairP list=parent, last=0;
	if (list->type==QUOTE && list->first() && list->first()->type==TREE)
		list=TreeP(QuoteP(list)->quote())->tree();
	else
		list=list->first();
	while (list && (list->is.list || list->type==QUOTE || list->type==TREE))
	{	if (list->is.list || list->type==QUOTE || list->type==TREE)
		{	togglePauseTree2(edit, col, row, fn, list, n, clear);
		}
		if (n==0 && LineP(list)->row == row+1)
		{	if (clear)
			{	if (list->first() == NodeP(_pause))
				{	if (list->tail())
					{	if (list->tail()->type==TREE)
						{	list->settype_TREE();
							TreeP(list)->settree(TreeP(list->tail())->tree());
						}
						list->setfirst(list->tail()->first());
						list->settail(list->tail()->tail());
					}else if (parent->tail())
					{	parent->setfirst(parent->tail()->first());
						parent->settail(parent->tail()->tail());
					}
					n--;
					break;
				}
			}else
			{	ThreadP th=Thread::This();
				th->col=col;
				th->row=row+1;
				th->setfilename(fn);
				LineP line=0;
				line=new Line(_pause, 0);
				if throwing return;
				if (last)
					last->settail(line);
				else if (parent->type==QUOTE)
					TreeP(QuoteP(parent)->quote())->settree(line);
				else
				{	if (LineP(list)->col >= col)
					{	PairP l=new Line(list, parent->tail());
						if throwing return;
						parent->setfirst(line);
						parent->settail(l);
					}else
					{	PairP l=new Line(line, parent->tail());
						if throwing return;
						parent->setfirst(list);
						parent->settail(l);
					}
				}
				n++;
				break;
			}
		}
		last=list;
		next(list);
	}
}

void EdFrame::OnBreakpoint(wxCommandEvent &WXUNUSED(event))
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	int pos=edit->GetCurrentPos(),
		row=edit->GetCurrentLine(),
		pos0=edit->PositionFromLine(row),
		col=pos-pos0;

	if (edit->GetTextRange(pos, pos+6)=="pause ")
	{	int oldpos=edit->GetCurrentPos();
		edit->SetSelection(pos, pos+6);
		edit->DeleteBack();
		if (oldpos-pos < 6)
			edit->GotoPos(pos);
		else
			edit->GotoPos(oldpos-6);
	}else
		edit->InsertText(pos, "pause ");
	bool set=edit->togglemark(col, row);
	StringP filename=new String(edit->GetFilename());
	if throwing return;
	StringP fn=0;
	PairP fnlist=0;
	for (fnlist=Thread::This()->filenames;  fnlist;  next(fnlist))
		if (fnlist->first() && fnlist->first()->equal(filename))
			fn=StringP(fnlist->first());

	if (!fn)
	{/*wxMessageBox(_("Setting breakpoint not possible: "
			"File not found in the Logo root table"), 
			_("aUCBLogo Editor Error"), wxOK, this);
	*/	return;
	}
	ProcedureP proc=root->findProc(fn, row+1); // in Logo the lines start from 1!!!

	if (proc)
	{
//		c_out << TreeP(proc->bodylist()) << endL;
//		togglePause(edit, col, row, fn, proc->bodylist());
		togglePauseTree(edit, col, row, fn, proc->bodylist(), true);
		if (set)
			togglePauseTree(edit, col, row, fn, proc->bodylist(), false);
//		c_out << TreeP(proc->bodylist()) << endL;
	}	
}

void EdFrame::OnSetHelpFileName(wxCommandEvent &WXUNUSED(event))
{
#ifdef WIN32
	wxFileDialog dlg(this, _("Set help file name"), _T(""),
		theApp->helpFileName, _("Compiled HTML Module (*.chm)|*.chm"),
					  wxOPEN | wxFILE_MUST_EXIST);
#else
	wxFileDialog dlg(this, _("Set help file name"), _T(""),
		theApp->helpFileName, _("HTML Help Package (*.hhp)|*.hhp"),
					  wxOPEN | wxFILE_MUST_EXIST);
#endif
	if (dlg.ShowModal() != wxID_OK) return;
	theApp->helpFileName = dlg.GetPath ();
}

void EdFrame::OnSetRingBuffer(wxCommandEvent &WXUNUSED(event))
{
	int size = wxGetNumberFromUser( 
		_("Set debug ringbuffer size."
		"Higher values need more memory and give lower speed."), 
		_("Enter new size (0..1024):"), 
		_("Set debug ringbuffer size"), 
		debugStates.size, 0, 1024, console, wxDefaultPosition);

	if (size != debugStates.size)
	{	debugStates.free();
		debugStates.init(size);
	}
}

void EdFrame::OnSetRunDir(wxCommandEvent &WXUNUSED(event))
{
	wxDirDialog dlg(this, _("Set run dir"), theApp->runDirectory);
	if (dlg.ShowModal() != wxID_OK) return;
	theApp->runDirectory = dlg.GetPath();
}

// properties event handlers
void EdFrame::OnProperties (wxCommandEvent &WXUNUSED(event))
{
	Edit *edit=(Edit*)notebook->GetPage(notebook->GetSelection());
	if (!edit) return;

	EditProperties dlg(edit, 0);
}

// print event handlers
void EdFrame::OnPrintSetup (wxCommandEvent &WXUNUSED(event))
{
	(*g_pageSetupData) = *g_printData;
	wxPageSetupDialog pageSetupDialog(this, g_pageSetupData);
	pageSetupDialog.ShowModal();
	(*g_printData) = pageSetupDialog.GetPageSetupData().GetPrintData();
	(*g_pageSetupData) = pageSetupDialog.GetPageSetupData();
}

void EdFrame::OnPrintPreview (wxCommandEvent &WXUNUSED(event))
{
	Edit *edit=(Edit*)notebook->GetPage(notebook->GetSelection());
	wxPrintDialogData printDialogData( *g_printData);
	wxPrintPreview *preview =
		new wxPrintPreview (new EditPrint (edit),
							new EditPrint (edit),
							&printDialogData);
	if (!preview)
	{	error(OUT_OF_MEM); return;
	}
	if (!preview->Ok())
	 {   delete preview;
		wxMessageBox (_("There was a problem with previewing.\n\
						 Perhaps your current printer is not correctly?"),
					  _("Previewing"), wxOK);
		return;
	}
	wxRect rect = DeterminePrintSize();
	wxPreviewFrame *frame = new wxPreviewFrame (preview, this, _("Print Preview"));
	if (!frame)
	{	error(OUT_OF_MEM); return;
	}
	frame->SetSize (rect);
	frame->Centre(wxBOTH);
	frame->Initialize();
	frame->Show(true);
}

void EdFrame::OnPrint (wxCommandEvent &WXUNUSED(event))
{
	Edit *edit=(Edit*)notebook->GetPage(notebook->GetSelection());
	wxPrintDialogData printDialogData(*g_printData);
	wxPrinter printer (&printDialogData);
	EditPrint printout (edit);
	if (!printer.Print (this, &printout, true))
	 {   if (wxPrinter::GetLastError() == wxPRINTER_ERROR)
		  {
				wxMessageBox (_("There was a problem with printing.\n\
						 Perhaps your current printer is not correctly?"),
					  _("Previewing"), wxOK);
			return;
		}
	}
	(*g_printData) = printer.GetPrintDialogData().GetPrintData();
}

// edit events
void EdFrame::OnEdit (wxCommandEvent &event)
{
	int page=notebook->GetSelection();
	if (page < 0)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	if (edit) edit->ProcessEvent (event);
}

void EdFrame::OnSelChange(wxNotebookEvent &event)
{
	size_t page=event.GetSelection();
	if (page < notebook->GetPageCount())
	{	Edit *edit=(Edit*)notebook->GetPage(page);
	//	SetTitle(edit->GetFilename()+_T(" - ")+aUCBLogo+_T(" Editor"));
		edit->SetFocus();
	}else
		event.Skip();
}

void EdFrame::OnSetFocus(wxFocusEvent &event)
{
	int page=notebook->GetSelection();
	if (page >= 0)
	{	Edit *edit=(Edit*)notebook->GetPage(page);
		if (edit)
			edit->SetFocus();
	}
	event.Skip();
}

void EdFrame::OnPageNext(wxCommandEvent &WXUNUSED(event))
{
	notebook->AdvanceSelection();
}

void EdFrame::OnPagePrev(wxCommandEvent &WXUNUSED(event))
{
	notebook->AdvanceSelection(false);
}

void EdFrame::OnConsoleSetFocus(wxCommandEvent &WXUNUSED(event))
{
	if (theApp->consoleFrame)
		theApp->consoleFrame->SetFocus();
	if (console)
		console->SetFocus();
}

void EdFrame::OnMakeBackupFiles(wxCommandEvent &WXUNUSED(event))
{
	theApp->backup=!theApp->backup;
}

// private functions
void EdFrame::CreateMenu ()
{
	// File menu
	wxFlatMenu *menuFile = new wxFlatMenu;
	if (!menuFile)
		return;
	
	menuFile->Append (wxID_NEW, _("&New ..\tCtrl+N"));
	menuFile->Append (wxID_OPEN, _("&Open ..\tCtrl+O"));
	menuFile->Append (wxID_SAVE, _("&Save\tCtrl+S"));
	menuFile->Append (wxID_SAVEAS, _("Save &as ..\tCtrl+Shift+S"));
	menuFile->Append (myID_EXPORT_HTML, _("Export as &HTML"));
	menuFile->Append (wxID_CLOSE, 
#ifdef WIN32
		_("&Close\tCtrl+F4"));
#else
		_("&Close\tCtrl+W"));
#endif
	menuFile->AppendSeparator();
	menuFile->Append (myID_PROPERTIES, _("Proper&ties ..\tCtrl+I"));
	menuFile->AppendSeparator();
	menuFile->Append (wxID_PRINT_SETUP, _("Print Set&up .."));
	menuFile->Append (wxID_PREVIEW, _("Print Pre&view\tCtrl+Shift+P"));
	menuFile->Append (wxID_PRINT, _("&Print ..\tCtrl+P"));
	menuFile->AppendSeparator();
	menuFile->Append (wxID_EXIT, _("E&xit\tAlt+X"));



	// Edit menu
	wxFlatMenu *menuEdit = new wxFlatMenu;
	if (!menuEdit)
		return;
	menuEdit->Append (wxID_UNDO, _("&Undo\tCtrl+Z"));
	menuEdit->Append (wxID_REDO, _("&Redo\tCtrl+Shift+Z"));
	menuEdit->AppendSeparator();
	menuEdit->Append (wxID_CUT, _("Cu&t\tCtrl+X"));
	menuEdit->Append (wxID_COPY, _("&Copy\tCtrl+C"));
	menuEdit->Append (wxID_PASTE, _("&Paste\tCtrl+V"));
	menuEdit->Append (wxID_CLEAR, _("&Delete\tDel"));
	menuEdit->AppendSeparator();
	menuEdit->Append (wxID_FIND, _("&Find\tCtrl+F"));
	menuEdit->Append (myID_FINDNEXT, _("Find &next\tF3"));
	menuEdit->Append (myID_FINDPREVIOUS, _("Find pr&evious\tShift+F3"));
	menuEdit->Append (myID_REPLACE, _("&Replace\tCtrl+H"));
	menuEdit->Append (myID_REPLACENEXT, _("Replace &again\tCtrl+F3"));
	menuEdit->Append (myID_REPLACEALL, _("Rep&lace all"));
	wxFlatMenuItem *menuEditFindWholeWords=menuEdit->AppendCheckItem(
		myID_FINDWHOLEWORD, _("Find &whole words"));
	menuEditFindWholeWords->Check(true);
	wxFlatMenuItem *menuEditSkipComments=menuEdit->AppendCheckItem (
		myID_SKIP_COMMENTS, _("&Skip comments"));
	menuEditSkipComments->Check(false);
	menuEdit->AppendSeparator();
	menuEdit->Append (myID_BRACEMATCH, _("&Match brace\tCtrl+M"));
	menuEdit->Append (myID_GOTO, _("&Goto\tCtrl+G"));
	menuEdit->Append (myID_GOTODEF, _("Goto definition\tF4"));
	menuEdit->Append (myID_GOTOPREV, _("Goto previous position\tShift+F4"));
	menuEdit->AppendSeparator();
	menuEdit->Append (myID_INDENTINC, _("&Indent increase\tTab"));
	menuEdit->Append (myID_INDENTRED, _("I&ndent reduce\tShift+Tab"));
	menuEdit->AppendSeparator();
	menuEdit->Append (wxID_SELECTALL, _("Select &all\tCtrl+A"));
//	menuEdit->Append (myID_SELECTLINE, _("Select &line\tCtrl+L"));

	// hilight submenu
	wxFlatMenu *menuHilight = new wxFlatMenu;
	if (!menuHilight)
		return;
	int Nr;
	for (Nr = 0; Nr < g_LanguagePrefsSize; Nr++)
	 {   menuHilight->Append (myID_HILIGHTFIRST + Nr,
							 g_LanguagePrefs [Nr].name);
	}

	// charset submenu
	wxFlatMenu *menuCharset = new wxFlatMenu;
	if (!menuCharset)
		return;
	menuCharset->Append (myID_CHARSETANSI, _("&ANSI (Windows)"));
	menuCharset->Append (myID_CHARSETMAC, _("&MAC (Macintosh)"));

	// View menu
	wxFlatMenu *menuView = new wxFlatMenu;
	if (!menuView)
		return;
	menuView->Append (myID_HILIGHTLANG, _("&Hilight language .."), menuHilight);
	menuView->AppendSeparator();
	menuView->AppendCheckItem (myID_FOLDTOGGLE, _("&Toggle current fold\tCtrl+T"));
	menuView->AppendCheckItem (myID_FOLDTOGGLEALL, _("&Toggle all folds\tCtrl+Alt+T"));
	menuView->AppendCheckItem (myID_OVERTYPE, _("&Overwrite mode\tIns"));
	menuView->AppendCheckItem (myID_WRAPMODEON, _("&Wrap mode\tCtrl+U"));
	menuView->AppendSeparator();
	menuView->AppendCheckItem (myID_DISPLAYEOL, _("Show line &endings"));
	menuView->AppendCheckItem (myID_INDENTGUIDE, _("Show &indent guides"));
	menuView->AppendCheckItem (myID_LINENUMBER, _("Show line &numbers"));
	menuView->AppendCheckItem (myID_LONGLINEON, _("Show &long line marker"));
	menuView->AppendCheckItem (myID_WHITESPACE, _("Show white&space"));
	menuView->AppendSeparator();
	menuView->Append (myID_USECHARSET, _("Use &code page of .."), menuCharset);
	menuView->Append (myID_FONT, _("Set &Font"));

	// change case submenu
	wxFlatMenu *menuChangeCase = new wxFlatMenu;
	if (!menuChangeCase)
		return;
	menuChangeCase->Append (myID_CHANGEUPLOW, _("&Correct case"));
	menuChangeCase->Append (myID_CHANGEUPPER, _("&Upper case"));
	menuChangeCase->Append (myID_CHANGELOWER, _("&Lower case"));

	// convert EOL submenu
	wxFlatMenu *menuConvertEOL = new wxFlatMenu;
	if (!menuConvertEOL)
		return;
	menuConvertEOL->Append (myID_CONVERTCR, _("CR (&Linux)"));
	menuConvertEOL->Append (myID_CONVERTCRLF, _("CR+LF (&Windows)"));
	menuConvertEOL->Append (myID_CONVERTLF, _("LF (&Macintosh)"));

	// Run menu
	wxFlatMenu *menuRun = new wxFlatMenu;
	if (!menuRun)
		return;
	menuRun->Append (myID_RUN, _("&Run\tF9"));
	menuRun->Append (myID_STEP, _("&Step\t F 7"));
	menuRun->Append (myID_REVERSE_STEP, _("R&everse Step\tShift+F7"));
	menuRun->Append (myID_STEP_OVER, _("Step &Over\tF 8"));
	menuRun->Append (myID_REVERSE_STEP_OVER, _("Re&verse Step Over\tShift+F8"));
	menuRun->Append (myID_BREAKPOINT, _("Set &Breakpoint\tCtrl+F8"));
	menuEdit->AppendSeparator();
	menuRun->Append (myID_LOAD, _("&Load file to Workspace\tCtrl+L"));
	menuRun->AppendCheckItem (myID_AUTOLOAD, _("&AutoLoad file to Workspace"));
	menuEdit->AppendSeparator();
	menuRun->Append (myID_SETRINGBUFFER, _("Se&t ringbuffer size"));
	menuRun->Append (myID_SETRUNDIR, _("Set run &Directory"));

	// Extra menu
	wxFlatMenu *menuExtra = new wxFlatMenu;
	if (!menuExtra)
		return;
	wxFlatMenuItem *menuExtraBackup=menuExtra->AppendCheckItem (
		myID_BACKUP, _("Make &backup files"));
	menuExtraBackup->Check(theApp->backup);
	menuExtra->AppendCheckItem (myID_READONLY, _("&Readonly mode"));
	menuExtra->AppendSeparator();
	menuExtra->Append (myID_CHANGECASE, _("Change &case to .."), menuChangeCase);
	menuExtra->AppendSeparator();
	menuExtra->Append (myID_CONVERTEOL, _("Convert line &endings to .."), menuConvertEOL);
	menuExtra->Append (myID_MAKETOASSIGN, _("Convert &make to ="));

	// Window menu
	wxFlatMenu *menuWindow = new wxFlatMenu;
	if (!menuWindow)
		return;
	menuWindow->Append (myID_PAGENEXT, _("&Next\tF6"));
	menuWindow->Append (myID_PAGEPREV, _("&Previous\tShift+F6"));
	menuWindow->Append (myID_CONSOLE_SET_FOCUS, _("&Set Focus to Console\tCtrl+Tab"));

	// Help menu
	wxFlatMenu *menuHelp = new wxFlatMenu;
	if (!menuHelp)
		return;
	menuHelp->Append (wxID_HELP, _("&Keyword lookup\tF1"));
	menuHelp->Append (myID_SETHELPFILENAME, _("&Set help file name"));
	menuHelp->Append (wxID_ABOUT, _("&About ..\tShift+F1"));

	// construct menu
	m_menuBar->Append (menuFile  , _("&File"));
	m_menuBar->Append (menuEdit  , _("&Edit"));
	m_menuBar->Append (menuView  , _("&View"));
	m_menuBar->Append (menuRun   , _("&Run"));
	m_menuBar->Append (menuExtra , _("Ex&tra"));
	m_menuBar->Append (menuWindow, _("&Window"));
	m_menuBar->Append (menuHelp  , _("&Help"));
//	SetMenuBar (m_menuBar);
}

void EdFrame::FileOpen (wxString fname)
{
	wxFileName w(fname); w.Normalize(); fname = w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page!=-1)
	{	notebook->SetSelection(page);
		((Edit*)notebook->GetPage(page))->LoadFile(fname);
		notebook->GetPage(page)->SetFocus();
		return;
	}
	Edit *edit = new Edit(fname, notebook, wxID_ANY, 
		wxDefaultPosition, wxDefaultSize,  
		wxSUNKEN_BORDER|wxVSCROLL, fontSize);
	if (!edit)
	{	error(OUT_OF_MEM); return;
	}
	wxString s=w.GetFullName();
	notebook->AddPage(edit, s, true);
	edit->LoadFile(fname);
//	SetTitle(edit->GetFilename()+_T(" - ")+aUCBLogo+_T(" Editor"));
	edit->SetFocus();
}

void EdFrame::FileClose(charP filename)
{
	wxFileName w(filename); w.Normalize();
	wxString fname=w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page==-1)
		return;

	notebook->SetSelection(page);
	wxCommandEvent cev;
	OnFileClose(cev);
}

void EdFrame::FileSave(charP filename)
{
	wxFileName w(filename); w.Normalize();
	wxString fname=w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page==-1)
		return;

	notebook->SetSelection(page);
	((Edit*)notebook->GetPage(page))->SaveFile ();
}

void EdFrame::mark(charP filename, int col, int row, int markernr)
{
	if (theApp->autoloadEdit)
		return;

	wxFileName w(filename); w.Normalize();
	wxString fname=w.GetFullPath();

	int page=-1;
	for (unsigned int i=0; i<notebook->GetPageCount(); i++)
		if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
			page=i;

	if (page==-1)
	{	if (wxFileExists(fname))
		{	FileOpen(fname);
			for (unsigned int i=0; i<notebook->GetPageCount(); i++)
				if (((Edit*)notebook->GetPage(i))->GetFilename()==fname)
					page=i;
		}
	}
	if (page==-1)
		return;

	notebook->SetSelection(page);
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->mark(col-1, row-1, markernr);
//	console->SetFocus();
//	edit->SetFocus();
}

void EdFrame::unmark(int markernr)
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->unmark(markernr);
}

void EdFrame::gotoLine(int line)
{
	int page=notebook->GetSelection();
	if (page==-1)
		return;
	Edit *edit=(Edit*)notebook->GetPage(page);
	edit->gotoLine(line);
}

wxRect EdFrame::DeterminePrintSize ()
{
	wxSize scr = wxGetDisplaySize();

	// determine position and size (shifting 16 left and down)
	wxRect rect = GetRect();
	rect.x += 16;
	rect.y += 16;
	rect.width = wxMin (rect.width, (scr.x - rect.x));
	rect.height = wxMin (rect.height, (scr.x - rect.y));

	return rect;
}
