//////////////////////////////////////////////////////////////////////////////
// File:        edit.h
// Purpose:     STC test module
// Maintainer:  Wyo
// Created:     2003-09-01
// RCS-ID:      $Id: edit.h,v 1.3 2004/09/28 15:13:24 ABX Exp $
// Copyright:   (c) wxGuide
// Licence:     wxWindows licence
//////////////////////////////////////////////////////////////////////////////

#ifndef _EDIT_H_
#define _EDIT_H_

//----------------------------------------------------------------------------
// informations
//----------------------------------------------------------------------------


//----------------------------------------------------------------------------
// headers
//----------------------------------------------------------------------------

//! wxWidgets headers

//! wxWidgets/contrib headers
#include <wx/stc/stc.h>  // styled text control


//============================================================================
// declarations
//============================================================================

class EditPrint;
class EditProperties;

#ifdef WIN32
class FileDropTarget: public wxFileDropTarget
{
public:
	bool OnDropFiles(wxCoord x, wxCoord y, const wxArrayString& filenames);
};
#endif

struct LanguageInfo;

WX_DECLARE_OBJARRAY(int, intArray);

//----------------------------------------------------------------------------
//! Edit

class Edit: public wxStyledTextCtrl 
{
		friend class EditProperties;
		friend class EditPrint;

	public:
		int fontSize;

		//! constructor
		Edit (wxString filename, 
			wxWindow *parent, wxWindowID id = wxID_ANY,
			const wxPoint &pos = wxDefaultPosition,
			const wxSize &size = wxDefaultSize,
			long style = wxSUNKEN_BORDER|wxVSCROLL,
			int fontSize_=10);

		//! destructor
		~Edit ();

		void GotoPosVisible(int pos), selectWord();

		// event handlers
		// common
		void OnSetFocus(wxFocusEvent &event);
		void OnEnterWindow(wxMouseEvent &event);
		void OnLeaveWindow(wxMouseEvent &event);
		void OnChar(wxKeyEvent &event);
		void OnKeyUp(wxKeyEvent &event);
		void OnKeyDown(wxKeyEvent &event);
		void OnSize( wxSizeEvent &event );
		// file
		void OnExportHTML(wxCommandEvent &event);

		// edit
		void OnEditRedo  (wxCommandEvent &event);
		void OnEditUndo  (wxCommandEvent &event);
		void OnEditClear (wxCommandEvent &event);
		void OnEditCut   (wxCommandEvent &event);
		void OnEditCopy  (wxCommandEvent &event);
		void OnEditPaste (wxCommandEvent &event);
		// find
		void OnFind          (wxCommandEvent &event);
		void OnFindNext      (wxCommandEvent &event);
		void OnFindPrevious  (wxCommandEvent &event);
		void OnReplace       (wxCommandEvent &event);
		void OnReplaceNext   (wxCommandEvent &event);
		void OnReplaceAll    (wxCommandEvent &event);
		void OnFindWholeWord (wxCommandEvent &event);
		void OnSkipComments  (wxCommandEvent &event);
		void OnBraceMatch    (wxCommandEvent &event);
		void OnGoto          (wxCommandEvent &event);
		void gotoLine(int line);
		void OnGotoDef       (wxCommandEvent &event);
		void OnGotoPrev      (wxCommandEvent &event);
		void OnEditIndentInc (wxCommandEvent &event);
		void OnEditIndentRed (wxCommandEvent &event);
		void OnEditSelectAll (wxCommandEvent &event);
		void OnEditSelectLine(wxCommandEvent &event);
		//! view
		void OnHilightLang (wxCommandEvent &event);
		void OnDisplayEOL  (wxCommandEvent &event);
		void OnIndentGuide (wxCommandEvent &event);
		void OnLineNumber  (wxCommandEvent &event);
		void OnLongLineOn  (wxCommandEvent &event);
		void OnWhiteSpace  (wxCommandEvent &event);
		void OnFoldToggle  (wxCommandEvent &event);
		void OnFoldToggleAll(wxCommandEvent &event);
		void OnSetOverType (wxCommandEvent &event);
		void OnSetReadOnly (wxCommandEvent &event);
		void OnWrapmodeOn  (wxCommandEvent &event);
		void OnUseCharset  (wxCommandEvent &event);
		//! extra
		void OnCorrectCase (wxCommandEvent &event);
		void OnChangeCase  (wxCommandEvent &event);
		void OnConvertEOL  (wxCommandEvent &event);
		void OnMakeToAssign(wxCommandEvent &event);
		void OnHelp        (wxCommandEvent &event);
		// stc
		void OnMarginClick (wxStyledTextEvent &event);
		void OnCharAdded   (wxStyledTextEvent &event);
		void OnModified    (wxStyledTextEvent &event);

		//! language/lexer
		void setFonts();
		wxString DeterminePrefs (const wxString &filename);
		void prelex();
		bool InitializePrefs (const wxString &filename);
		bool UserSettings (const wxString &filename);
		LanguageInfo const* GetLanguageInfo () {return m_language;};

		//! load/save file
		bool LoadFile ();
		bool LoadFile (const wxString &filename);
		bool SaveFile ();
		bool SaveFile (const wxString &filename);
		bool Modified ();
		wxString GetFilename () {return m_filename;};
		void SetFilename (const wxString &filename) {m_filename = filename;};

		void mark(int col, int row, int markernr);
		void unmark(int markernr);
		bool togglemark(int col, int row);
		bool gotoDef(wxString preposition, wxString word, int pos, int start, int end);
		const wxString RenderAsHTML();

	private:
		int dummy;
		// file
		wxString m_filename, lastFindExpr, lastReplaceExpr;

		// lanugage properties
		LanguageInfo const* m_language;

		// margin variables
		int m_LineNrID;
		int m_LineNrMargin;
		int m_FoldingID;
		int m_FoldingMargin;
		int m_DividerID;
		int m_wholeword;
		bool allFolded, skipComments;

#ifdef WIN32
		FileDropTarget *fileDropTarget;
#endif
		
		intArray posArray, marker;

#define contextMenuNr 1000
		wxMenu *contextMenu[contextMenuNr];

		struct MenuTree
		{	int level;
			charP str;
		};

		static MenuTree command[];

		bool HasWord(wxString word, wxString &wordlist);

		static int __cdecl CompareCaseInsensitive(
			const wxString& first, 
			const wxString& second);

		wxArrayString SortedKeywordArray();
		wxString SortedKeywords();
		static inline bool IsAWordChar(const int ch);

		void
			createMenu(int &used, int &i, int level),
			OnContextMenu(wxContextMenuEvent &event),
			OnInsert(wxCommandEvent &event);

		DECLARE_EVENT_TABLE()
};

//----------------------------------------------------------------------------
//! EditProperties
class EditProperties: public wxDialog 
{
	public:
		//! constructor
		EditProperties (Edit *edit, long style = 0);

	private:
};

//----------------------------------------------------------------------------
//! EditPrint
class EditPrint: public wxPrintout 
{
	public:
		//! constructor
		EditPrint (Edit *edit, wxChar *title = _T(""));

		//! event handlers
		bool OnPrintPage (int page);
		bool OnBeginDocument (int startPage, int endPage);

		//! print functions
		bool HasPage (int page);
		void GetPageInfo (int *minPage, int *maxPage, int *selPageFrom, int *selPageTo);

	private:
		Edit *m_edit;
		int m_printed;
		wxRect m_pageRect;
		wxRect m_printRect;

		bool PrintScaling (wxDC *dc);
};

extern Function lrenderLogoAsHTML, lrenderLogoFileAsHTML;

#endif // _EDIT_H_

