#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/frame.h>
#include <wx/listctrl.h>
#include <wx/treectrl.h>
#include <wx/timer.h>
#include <wx/notebook.h>
#include <wx/dialog.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Tree.h"
#include "Object.h"
#include "Line.h"
#include "List.h"
#include "Int.h"
#include "UInt8.h"

#include "Error.h"
#include "NodePStack.h"
#include "Init.h"

#include "Vars.h"
#include "Calls.h"
#include "Color.h"
#include "MainFrame.h"
#include "LogoApp.h"

#include "EdFrame.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif


void error_print(void) 
{
	ThreadP th=Thread::This();
	ControlType save_flag=th->stoppingFlag;
	
	LineP &err=th->err_msg;
	if (!err) return;
	
	th->stoppingFlag=RUN;
	th->backslashes=true;
	
	th->GuiEnter();
#ifdef WIN32
	if (theApp->mainFrame)
		ShowWindow((HWND)theApp->mainFrame->GetHWND(), SW_SHOW);
#else
// TODO
#endif
	if (err->second())
	{	int width=10000;
		err->second()->print(File::conout, width, 5, 1);
		if (console)
		{	width=10000;
			err->second()->print(stdout, width, 5, 1);
		}
	}
	if (err->tail()->tail())
	{	if (err->third()) 
		{	c_out << _("  in ") << err->third();
			if (console)
				std_err << _("  in ") << err->third();
		}	
		if (err->tail()->tail()->tail())
		{	c_out	<< ' ' << err->fourth();
			if (console)
				std_err << ' ' << err->fourth();
		}
		if (err->tail()->tail()->tail()->tail())
		{	if (err->fifth() && err->sixth())
			{	c_out << '[' << err->fifth() << ' '  << err->sixth() << ']' << endL;
				if (console)
					std_err << '[' << err->fifth() << ' ' << err->sixth() << ']' << endL;
			}
			if (err->fourth() && err->fifth())
			{	charP fn=StringP(err->fourth())->ptr();
				int col=IntP(err->fifth())->int_, row=IntP(err->sixth())->int_;
				if (th->IsMain() && theApp->showGuiOnError)
				{	theApp->mainShow();
					theApp->editorShow();
					if (theApp->edFrame)
						theApp->edFrame->mark(fn, col, row, 0);
				}
			}
		}
	}
	err=0;
	c_out << endL;
	if (console)
		std_err << endL;
	
	th->GuiLeave();

	th->backslashes=false;
	th->stoppingFlag=save_flag;
}

PairP error(ErrorType error_type, NodeP error_desc, int nr)
{
	ThreadP th=Thread::This();
	if throwing return PairP(Unbound);

	bool recoverable=false, warning=false, uplevel=false;
	PairP err_act=0, val = (PairP)Unbound;
	LineP &err=th->err_msg;
	
	switch(error_type) 
	{
	case FATAL:
			c_out << _("Logo: Fatal Internal Error.") << endL;
			exit(1);

	case OUT_OF_MEM_UNREC:
			c_out << _("Logo: Out of Memory.") << endL;
			exit(1);

	case OUT_OF_MEM:
			GC_::reserve_tank_use();
			if (theApp && theApp->heapReserve)
			{	delete theApp->heapReserve;
				theApp->heapReserve=0;
			}
			err = LineP(ErrorMessage_Out_of_Mem->var());
			break;

	case STACK_OVERFLOW:
			err = LineP(ErrorMessage_Stack_Overflow->var()); break;

	case TURTLE_OUT_OF_BOUNDS:
			err = new Line(_("; turtle out of bounds"));	break;

	case BAD_GRAPH_INIT:
			err = new Line(_("; couldn't initialize graphics")); break;
			break;

	case BAD_DATA_UNREC:
			err = new List
				(	new String(_("; ")),
					th->fun,
					new String(_("doesn\'t like")), 
					error_desc,
					new String(_("as input")), 
					_ListEnd
				);
			break;

	case DIDNT_OUTPUT:
			if (th->didnt_output_name) 
				th->last_call=th->didnt_output_name;
			
			if (!error_desc && th->didnt_get_output) 
			{	error_desc = th->didnt_get_output->fun;
				     th->ufun = th->didnt_get_output->ufun;
				th->this_line = th->didnt_get_output->this_line;
			}
			err = new List
				(	new String(_("; ")),
					th->last_call,
					new String(_("didn\'t output to")),
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case NOT_ENOUGH:
			if (!error_desc)
				err = new List
					(	new String(_("; not enough inputs to")),
						th->fun, 
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; not enough inputs to")),
						error_desc, 
						_ListEnd
					);
			break;

	case BAD_DATA:
			if (nr==1)
				err = new List
					(	new String(_("; ")),
						th->fun,
						new String(_("doesn\'t like")), 
						error_desc,
						new String(_("as input")), 
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; ")),
						th->fun,
						new String(_("doesn\'t like")), 
						error_desc,
						new List(
							new String(_("as input ")), 
							new Int(nr), _ListEnd), 
						_ListEnd
					);
			recoverable=true;
			break;

	case APPLY_BAD_DATA:
			err = new List
				(	new String(_("; APPLY doesn\'t like")),
					error_desc,
					new String(_("as input")), 
					_ListEnd
				);
			recoverable=true;
			break;

	case TOO_MUCH:
			err = new Line(_("; too much inside ()\'s"));	break;

	case DK_WHAT_UP:
			uplevel=true;

	case DK_WHAT:
			err = new List
				(	error_desc, 
					new String(_(" ;-)")),	// "You don`t say what to do with x" was too annoying
					_ListEnd
				);
			break;

	case PAREN_MISMATCH:
			if (error_desc && error_desc->type==OBJECT)	
				err = new List
					(	new String(_("; too many (\'s ")),
						ObjectP(error_desc)->longName(),
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; too many (\'s ")),
						error_desc,
						_ListEnd
					);
			recoverable=true;
			break;

	case BRACKET_MISMATCH:
			if (error_desc && error_desc->type==OBJECT)	
				err = new List
					(	new String(_("; too many [\'s ")),
						ObjectP(error_desc)->longName(),
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; too many [\'s ")),
						error_desc,
						_ListEnd
					);
			recoverable=true;
			break;

	case BRACE_MISMATCH:
			if (error_desc && error_desc->type==OBJECT)	
				err = new List
					(	new String(_("; too many {\'s ")),
						ObjectP(error_desc)->longName(),
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; too many {\'s ")),
						error_desc,
						_ListEnd
					);
			recoverable=true;
			break;

	case VBAR_MISMATCH:
			if (error_desc && error_desc->type==OBJECT)	
				err = new List
					(	new String(_("; too many |\'s ")),
						ObjectP(error_desc)->longName(),
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; too many |\'s ")),
						error_desc,
						_ListEnd
					);
			recoverable=true;
			break;

	case END_MISMATCH:
			if (error_desc && error_desc->type==OBJECT)	
				err = new List
					(	new String(_("; end expected ")),
						ObjectP(error_desc)->longName(),
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; end expected ")),
						error_desc,
						_ListEnd
					);
			recoverable=true;
			break;

	case NO_VALUE:
			if (error_desc && error_desc->type==OBJECT)	
				err = new List
					(	new String(_("; ")),
						ObjectP(error_desc)->longName(),
						new String(_("has no value")),
						_ListEnd
					);
			else
				err = new List
					(	new String(_("; ")),
						error_desc,
						new String(_("has no value")),
						_ListEnd
					);
			recoverable=true;
			break;

	case UNEXPECTED_PAREN:
			err = new List
				(	new String(_("; unexpected paren \')\'")),
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case UNEXPECTED_BRACKET:
			err = new List
				(	new String(_("; unexpected bracket \']\'")),
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case UNEXPECTED_BRACE:
			err = new List
				(	new String(_("; unexpected brace \'}\'")),
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case UNEXPECTED_END:
			err = new List
				(	new String(_("; unexpected end")),
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case DK_HOW:
			recoverable=true;

	case DK_HOW_UNREC:
			if (error_desc && error_desc->type==OBJECT)
				err = new List
				(	new String(_("; I don\'t know how  to")),
					ObjectP(error_desc)->longName(), 
					_ListEnd
				);
			else
				err = new List
				(	new String(_("; I don\'t know how  to")),
					error_desc, 
					_ListEnd
				);
			break;

	case NO_CATCH_TAG:
			err = new List
				(	new String(_("; Can't find catch tag for")),
					error_desc, 
					_ListEnd
				);
			break;

	case ALREADY_DEFINED:
			err = new List
				(	new String(_("; ")),
					error_desc,
					new String(_("is already defined")),
					_ListEnd
				);
			break;

	case STOP_ERROR:
			err = new Line(_("; Stopping..."));		break;

	case ALREADY_DRIBBLING:
			err = new Line(_("; Already dribbling"));	break;

	case FILE_ERROR:
			err = new List
				(	new String(_("; File system error:")),
					error_desc, 
					_ListEnd
				);
			break;

	case IF_WARNING:
			err = new Line(_("; Assuming you mean IFELSE, not IF"));
			warning=true;
			break;

	case SHADOW_WARN:
			err = new List
				(	new String(_("; ")),
					error_desc,
					new String(_("shadowed by local in procedure call")),
					_ListEnd
				);
			warning=true;
			break;

	case USER_ERR:
			if (error_desc==Unbound)
				err = new Line(_("; throw \"error"));
			else 
			{	uplevel=true;
				if (error_desc->is.list)
					err = (LineP)error_desc;
				else
					err = new Line(error_desc, 0);
			}
			break;

	case IS_PRIM:
			err = new List
				(	new String(_("; ")),
					error_desc,
					new String(_("is a primitive")),

					_ListEnd
				);
			break;

	case NOT_INSIDE:
			err = new Line(_("; Can't use TO inside a procedure"));	
			break;

	case AT_TOPLEVEL:
			err = new List
				(	new String(_("; Can only use")),
					error_desc,
					new String(_("inside a procedure")),
					_ListEnd
				);
			break;

	case NO_TEST:
			err = new List
				(	new String(_("; ")),
					th->fun, 
					new String(_("without TEST")),
					_ListEnd
				);
			break;

	case ERR_MACRO:
			err = new List
				(	new String(_("; Macro returned")),
					error_desc,
					new String(_("instead of a list")),
					_ListEnd
				);
			break;

	case DEEPEND:
			if (!error_desc) 
				err = new Line(_("; END inside multi-line instruction.\n"));
			else 
			{	err = new List
					(	new String(_("; END inside multi-line instruction in")),
						error_desc,
						_ListEnd
					);
			}
			break;

	case NOT_YET_SUPPORTED:
			err = new List
				(	new String(_("; ")),
					th->fun,
					new String(_("is not yet supported")), 
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case SECURITY_ERROR:
			err = new List
				(	new String(_("; ")),
					th->fun,
					new String(_("is not allowed")), 
					error_desc,
					_ListEnd
				);
			recoverable=true;
			break;

	case SECURITY_MEMSIZE:
			th->stoppingFlag=THROWING;
			err = LineP(ErrorMessage_Security_MemSize->var());
			err->first()->setsecond(th->fun);
			recoverable=true;
			break;

	case SECURITY_NODES:
			th->stoppingFlag=THROWING;
			err = LineP(ErrorMessage_Security_Nodes->var());
			err->first()->setsecond(th->fun);
			recoverable=true;
			break;

	case IS_ALREADY_A_PROC:
			err = new List
				(	new String(_("; ")),
					error_desc,
					new String(_("is already a procedure")),
					_ListEnd
				);
			break;

	default:
			c_out << _("Unknown error condition - internal error.") << endL;
			exit(1);
	}
	switch (error_type)
	{
	case OUT_OF_MEM:
	case SECURITY_MEMSIZE:
	case SECURITY_NODES:
			break;
	default:
			lupdateCalls(0);
			lupdateVars(0);
	}

	th->didnt_output_name=0;

	if (th->ufun  &&  uplevel) 
	{	th->ufun = th->last_ufun;
		th->this_line = th->last_line;
	}
	switch (error_type)
	{
	case OUT_OF_MEM:
	case SECURITY_MEMSIZE:
	case SECURITY_NODES:
			break;

	default:
		if (th->ufun 
		&&  theApp->autoloadEdit 
		&&  th->autoThisLine)
			err = new List
				(	err, 
					th->ufun->longName(),
					new String(LineP(th->autoThisLine)->filename()),
					new Int(LineP(th->autoThisLine)->col),
					new Int(LineP(th->autoThisLine)->row),
					_ListEnd
				);
		else if (th->ufun 
		&&  th->this_line 
		&&  th->this_line->is.line)
			err = new List
				(	err, 
					th->ufun->longName(),
					new String(LineP(th->this_line)->filename()),
					new Int(LineP(th->this_line)->col),
					new Int(LineP(th->this_line)->row),
					_ListEnd
				);
		else if (error_type==NOT_ENOUGH)
			err = new List
				(	err, 
					0, 
					_Not_Enough_node->filename(),
					new Int(_Not_Enough_node->col),
					new Int(_Not_Enough_node->row),
					_ListEnd
				);
		else if (error_type==PAREN_MISMATCH
				|| error_type==BRACKET_MISMATCH
				|| error_type==BRACE_MISMATCH
				|| error_type==END_MISMATCH
				|| error_type==UNEXPECTED_PAREN 
				|| error_type==UNEXPECTED_BRACKET
				|| error_type==UNEXPECTED_BRACE
				|| error_type==UNEXPECTED_END)
			err = new List
				(	err, 
					0, 
					err->filename(),
					new Int(err->col),
					new Int(err->row),
					_ListEnd
				);
		else
			err = new List
				(	err, 
					0, 
					0,
					0,
					0,
					_ListEnd
				);
	}
	err = new Line
		(	new Int(error_type), 
			err
		);
	if (warning) 
	{	error_print();
		return (PairP)Unbound;
	}
	err_act = (PairP)th->erract;

	if (err_act  &&  err_act != Unbound) 
	{	if (error_type != th->erract_errtype) 
		{	ValStatus save_val_status = th->val_status;
			
			th->erract_errtype = error_type;
			th->erract=0;
			th->val_status = value_maybe_ok_in_macro;

			val = PairP(err_act->err_eval_driver());

			th->val_status = save_val_status;
			th->erract= err_act;

			if (recoverable==true  &&  val != Unbound) 
			{	return val;
			}else 
				if (recoverable==false  &&  val != Unbound) 
				{	c_out << _("; You don't say what to do with ") << val << endL;
					val = (PairP)Unbound;
					th->throw_node = _toplevel;
				} else 
				{	return (PairP)Unbound;
				}
		}else 
		{	c_out << _("; _Erract loop") << endL;
			th->throw_node=_toplevel;
		}
	}else 
	{	// no erract
		th->throw_node=_error;
	}
	th->stoppingFlag=THROWING;
	th->output_node = (PairP)Unbound;
	return val;
}

NodeP lError(PairPP) 
{
	ThreadP th=Thread::This();
	NodeP val=th->err_msg;
	th->err_msg=0;
	return val;
}

NodeP lpause(PairPP) 
{
	ThreadP th=Thread::This();
	bool oldinpause=th->inpause;
	th->inpause=true;
	th->stepArmed=false;

	lupdateCalls(0);
	lupdateVars(0);

	NodeP val=Unbound;
	if (th->err_msg) error_print();
	ObjectP uname=th->ufun;

	c_out << _("continue ; Pausing ") << uname;

	LineP &this_line=th->this_line;
	if (this_line
	&&  this_line->is.line)
		c_out << _(" at pos") << " [" << th->this_line->col << " " << th->this_line->row << "]";
	c_out << "..." << endL;
	if (this_line  
	&&  this_line->is.line  
	&&  this_line->filename()
	&&  this_line->filename()->is.string)
	{	theApp->editorShow();
		if (theApp->edFrame)
			theApp->edFrame->mark(this_line->filename()->ptr(), 
			this_line->col, this_line->row, 1);
	}
	jmp_buf sav_iblk;
	memcpy(&sav_iblk, &th->inputBlockingBuffer, sizeof(jmp_buf));

	int saveInputBlocking=th->inputBlocking;
	th->inputBlocking=0;

	ValStatus save_val_status = th->val_status;
	bool save_instep=th->instep;
	th->instep=false;

	while (running) 
	{	StringP str=0;	str=c_in.read("");	// "? ");
		PairP elist=0;	if (!throwing) elist = str->parser(true);

		th->val_status=value_maybe_ok_in_macro;
		if (elist) 
			elist->eval_driver();
		if (throwing) 
		{	if (th->throw_node->equal(_pause)) 
			{	val=th->output_node;
				th->output_node = (PairP)Unbound;
				th->stoppingFlag=RUN;

				memcpy(&th->inputBlockingBuffer, 
					&sav_iblk, sizeof(jmp_buf));

				th->inputBlocking = saveInputBlocking;
				th->val_status = save_val_status;
				if (uname) 
				{	th->ufun = uname;
				}

				theApp->editorShow();
				if (theApp->edFrame)
					theApp->edFrame->unmark(1);
				th->inpause=oldinpause;
				return val;
			}
			else if (th->throw_node->equal(_error)) 
			{	error_print();
				th->stoppingFlag=RUN;
			}
		}
	}
	memcpy(&th->inputBlockingBuffer, 
		&sav_iblk, sizeof(jmp_buf));

	th->instep=save_instep;
	th->inputBlocking = saveInputBlocking;
	File::unblockInput();
	th->val_status = save_val_status;
	if (uname) 
		th->ufun=uname;
	
	th->inpause=oldinpause;
	return val;
}

NodeP lcontinue(PairPP arg) 
{
	ThreadP th=Thread::This();
	th->stoppingFlag=THROWING;
	th->throw_node=_pause;
	if (NodePP(arg)!=th->argStack->index)
		th->output_node = arg[1];
	else
		th->output_node = (PairP)Unbound;
	return Unbound;
}

NodeP lErrorAction(PairPP arg)
{
	ThreadP th=Thread::This();
	return th->erract;
}

NodeP lsetErrorAction(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP erract=arg[1]->a_runnable(1);
	if throwing return Unbound;
	th->erract=erract;
	return Unbound;
}
