//___________________________________________________________________
//
//                       Evaluator Header
//___________________________________________________________________

#ifndef Node_included
	#include "Node.h"
#endif

//#define PROC_EVAL

#define Eval_Labels(x)					\
	x( fetch_command				)		\
	x( begin_line					)		\
	x( begin_sequence				)		\
	x( end_line						)		\
	x( tail_dispatch				)		\
	x( dispatch						)		\
	x( proc_with_args				)		\
	x( arg_loop						)		\
	x( accumulate_arg				)		\
	x( args_done					)		\
	x( apply_dispatch				)		\
	x( compound_apply				)		\
												\
		x( lambda_apply			)		\
			x( begin_formal_loop	)		\
				x( set_args_continue	)	\
			x( set_args_continue2)		\
		x( end_formal_loop		)		\
		x( after_formals			)		\
												\
	x( sequence						)		\
	x( non_tail_eval				)		\
	x( sequence_continue			)		\
	x( compound_apply_continue	)		\
	x( macro_return				)		\
												\
	x( runresult_continuation	)		\
	x( runresult_followup		)		\
	x( repeat_continuation		)		\
	x( repeat_again				)		\
	x( repeat_followup			)		\
	x( for_continuation			)		\
	x( for_again					)		\
	x( for_followup				)		\
	x( ifor_continuation			)		\
	x( ifor_again					)		\
	x( ifor_followup				)		\
	x( while_continuation		)		\
	x( while_cond					)		\
	x( while_again					)		\
	x( while_followup				)		\
	x( while_end					)		\
	x( until_continuation		)		\
	x( until_cond					)		\
	x( until_again					)		\
	x( until_followup				)		\
	x( until_end					)		\
	x( do_while_continuation	)		\
	x( do_while_cond				)		\
	x( do_while_again				)		\
	x( do_while_followup			)		\
	x( do_while_end				)		\
	x( do_until_continuation	)		\
	x( do_until_cond				)		\
	x( do_until_again				)		\
	x( do_until_followup			)		\
	x( do_until_end				)		\
	x( foreach_continuation		)		\
	x( foreach_again				)		\
	x( foreach_followup			)		\
	x( foreach_end					)		\
	x( catch_continuation		)		\
	x( catch_followup				)		\
	x( goto_continuation			)		\
												\
	x( begin_apply					)		\
	x( questionmark_continue	)		\
	x( questionmark_failed		)		\
	x( after_lambda				)		\
	x( end							)		\

#define create_enum(x) x,

enum EvalCommand
{
	Eval_Labels(create_enum)
};

class __single_inheritance Evaluator;

typedef int (Evaluator::*Do)();

struct ToDo
{
#ifdef PROC_EVAL
	#ifdef _DEBUG
		EvalCommand mnemonic;
	#endif
		union 
		{	Do instruction;
			int i;
		};
#else
		EvalCommand mnemonic;
#endif
};

#ifdef _DEBUG
//	#undef new
//	#define new new
#endif


#ifdef PTR_DEBUG
	class ToDoNode;	typedef NPtr<ToDoNode> ToDoNodeP;
#else
	class ToDoNode;	typedef ToDoNode* ToDoNodeP;
#endif


class ToDoNode: public Node		// is a Do Evaluator function and a ToDoStack pointer
{
	NodeP _tail;

	public:
		ToDo toDo;
		static const ToDoNode ToDoNode::aToDoNode;
		operator_new;

		NodeP tail() {return _tail;}
		void settail(NodeP nd) {_tail=nd;  check_oldyoung(nd);}

		ToDoNode(): Node(TODO) {}

		ToDoNode(ToDo toDo_, NodeP tail_): Node(TODO)
			{toDo = toDo_;   settail(tail_);}

		void inc_gc() {fifo << tail();}
		int youngs() {return gmark(tail());}
		int youngsPtr() {return gmarkPtr(tail());}
		void check_oldyoungs() {check_oldyoung(tail());}
		
		void setPtrs();
		wxDataInputStream &read(wxDataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
};

#ifdef _DEBUG
	#undef new
#endif

enum TailCall
{	tailcall_arg=-1,
	in_sequence,  // 0
	tailcall_tail,  // 1
};

enum ValStatus 
{	no_value_allowed,  // body of command = 0
	value_required,    // argument = 1
	output_ok,         // body of function = 2
	dont_care,       // function inside catch = 3
	no_value_in_macro,  // repeat = 4
	value_maybe_ok_in_macro  // catch = 5
};

class NodePStack;
class IntStack;
class FloatStack;
class ToDoStack;
class VarStack;
class VarPStack;
class PairPPStack;
class ErrorStack;

class Evaluator
{
	bool tracing;  // are we tracing the current procedure? 
	int old_ift_iff;
	TailCall oldtailcall;  // in case of reentrant use of Evaluator 

	NodeP		// Registers 
		val,     // the value of the last expression 
		catch_tag;

	PairP
		nodeList,    // flat list of expressions to evaluate
		unev,       // list of unevaluated expressions
		exp;       // the current expression 

	NodeP proc; // the procedure definition

	PairP
		formals,  // list of formal parameters 
		parm,    // the current formal 
		arg;    // the current actual 

	int asp;		// argStack pointer
	int vsp;	  // varStack pointer

	public:
		Evaluator() {}
		Evaluator(PairP nodeList_, ToDo where, ThreadP th);
		NodeP eval(PairP nodeList_, ToDo where, ThreadP th);


#ifdef PROC_EVAL
	#define create_function(x)	_##x(),

		int Eval_Labels(create_function)		dummy;
#endif

		ThreadP th;

		static const char commandName[][end+1];

	friend class Pair;
};

#ifdef PROC_EVAL
	#ifdef _DEBUG
		#define create_ToDo(x)	__##x={x, Evaluator::_##x},
	#else
		#define create_ToDo(x)	__##x={Evaluator::_##x},
	#endif
	const ToDo Eval_Labels(create_ToDo) 
		eval_dummy_label={Evaluator::_fetch_command};
#else
	#define create_ToDo(x)	__##x={x},
	const ToDo Eval_Labels(create_ToDo) 
		eval_dummy_label={fetch_command};
#endif
