#ifndef Node_included
	#include "Node.h"
#endif

#include <wx/wfstream.h>
#include <wx/datstrm.h>
#include <wx/txtstrm.h>
#include <wx/zipstrm.h>

#include "DataInputStream.h"
#include "TextOutputStream.h"
#include "TextInputStream.h"

#define c_in		(*File::CIn)
#define c_out		(*File::COut)
#define std_err	(*File::StdOut)
#define reader	(*Thread::This()->Reader)
#define writer	(*Thread::This()->Writer)
#define loader	(*Thread::This()->Loader)
#define dribbler	(*Thread::This()->Dribbler)
#define stringbuffer	(*(Thread::This()->Stringbuffer))
#define endL	'\n'

class MAPM;

class File: public Node
{
	FileP _tail;
	StringP _name, _mode;
	FILEP _file;
	int _pos;

	public:
		static const File aFile;
		static FileP list;
		static FileP Null;
		static FileP StdIn;
		static FileP StdOut;
		static FileP ConIn;
		static FileP ConOut;
		static FileP CIn;
		static FileP COut;
		static FileP TextCtrl;
		static const FILEP textcontrol;
		static const FILEP conin;
		static const FILEP conout;
		static const int lineMax;

		operator_new;

		   FileP tail() {return _tail;}
		 StringP name() {return _name;}
		 StringP mode() {return _mode;}
		   FILEP file() {return _file;}
		operator bool() {return _file != 0;}

		void settail(FileP   nd) {_tail=nd; check_oldyoung(nd);}
		void setname(StringP nd) {_name=nd; check_oldyoung(NodeP(nd));}
		void setmode(StringP nd) {_mode=nd; check_oldyoung(NodeP(nd));}
		void setfile(FILEP file_) {_file=file_;}

		File(): Node(FILE_), _tail(0), _name(0), _mode(0), 
			_file(0) {}

		File(FILEP file_, StringP name_): 
			Node(FILE_), _tail(0), _mode(0), _file(file_)
		{	setname(name_);
		}
		File(FILEP file_, StringP name_, StringP mode_, FileP tail_): 
			Node(FILE_), _file(file_)
		{	setname(name_); setmode(mode_); settail(tail_);
		}
		void inc_gc();
		int youngs();
		int youngsPtr(PtrToInt_ &ptrToInt);
		void check_oldyoungs();
		
		void setPtrs(IntToPtr_ &intToPtr);
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		NodeP copy(PtrToInt_ &ptrToInt);

		StringP string() {return name();}
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		void printBin(FILEP strm);

		StringP _typeof();
		int compare(NodeP nd);
		FileP findInFilelist(bool remove);

		int read_getc();
		int eof();
		int getc_();
		int ungetc(int c);
		int putc_(int c);
		int pos();
		int seek(int pos);
		int close();
		size_t read(void *buffer, size_t size, size_t count);
		StringP read(charP prompt);

		static bool kbhit();
		static bool exists(const char *filename);
		static int getch();
		static int getchExt();
		static int readConsolec();
		static int read_getc(FILEP strm);
		static void init();
		static void	charModeOn();
		static void charModeOff();
		static void print(FILEP strm, int &width, char ch);
		static void print(FILEP strm, NodeP nd, int noBrackets=0);
		static void printStack(PairPP arg, bool space);
		static void printStackBin(PairPP arg);
		static void printPrompt(charP str);
		static void unblockInput();

		File &operator<<(const   char c);
		File &operator<<(const    int i);
		File &operator<<(const double d);
		File &operator<<(const  NodeP n);
		File &operator<<(const  LineP l);
		File &operator<<(const QuoteP q);
		File &operator<<(const PairPP p);
		File &operator<<(const  char* s);
		File &operator<<(const   VarP v);
};

extern void 
	print(FILEP strm, int &width, char ch), 
	printIndents(FILEP strm, int &width),
	printBin(FILEP strm, char i),
	print(FILEP strm, int &width, unsigned char i),
	printBin(FILEP strm, unsigned char i),
	print(FILEP strm, int &width, short i),
	printBin(FILEP strm, short i),
	print(FILEP strm, int &width, int i),
	printBin(FILEP strm, int i),
	printHex(FILEP strm, int i),
	print(FILEP strm, int &width, double d),
	printBin(FILEP strm, double d),
	print(FILEP strm, int &width, MAPM d),
	printBin(FILEP strm, MAPM d),
	print(FILEP strm, PairPP arg);

extern void __cdecl print(FILEP strm, int &width, const char* fmt, ...);

#define File_included
