#ifndef Node_included
	#include "Node.h"
#endif

#ifndef Float_included

class Float: public Node 
{
	public: 
		double float_;

		static const int digits;
		static const int mantDigits;
		static const double max_;
		static const double pi;
		static const double deg2rad;
		static const double NaN;
		static const double quietNaN;
		static const double epsilon;
		static const Float aFloat;

		struct x87_ControlWord 
		{	unsigned short
				invalid_operation: 1, denormalized_operand: 1, zerodivide: 1, overflow: 1,	
				underflow: 1, precision: 1, dummy: 1, mask_interrupts: 1,
				precision_control: 2, rounding_control: 2,
				infinity: 1;
		};
		enum x87_Precision {_24bit=0, reserved=1, _53bit=2, _64bit=3};
		enum x87_Rounding {nearest=0, down=1, up=2, chop=3};
	
		static x87_ControlWord x87_status;
		static x87_ControlWord x87_overflow_mask;

		operator_new;
		Float():         Node(FLOAT_), float_(0) {}
		Float(double f): Node(FLOAT_), float_(f) {}

		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		NodeP copy(PtrToInt_ &ptrToInt);

		StringP   string();
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		void printBin(FILEP strm);

		int   numberP() {return 1;}
		double getnumber() {return float_;}
		 NodeP    number() {return this;}

		OPERATORS_DECL
		int compare(NodeP nd);
		int _sizeof();
		StringP _typeof();
		double _min(double d);
		double _max(double d);
		NodeP Norm();
		NodeP maxNorm();
		NodeP saturateAbove(double limit);
		NodeP saturateBelow(double limit);
		NodeP item(NodeP index, NodeP result);
		NodeP bititem(NodeP index);
		NodeP memberP(NodeP subobj, bool notp=false, bool substr=false);
		NodeP reverse();
		NodeP readbin(FileP file);
		NodeP upper();
		NodeP lower();
		PairP list();
		void setPixelXY(NodeP yin, NodeP color);
		void setPixelXYZ(NodeP yin, NodeP zin, NodeP color);
		void setXYZ(NodeP ynode, NodeP znode);

		static void init();
		static void reset_FPU();
		static double rnd();
		static double abs(double x);
		static double sqr(double x);
		static double cutError(double x);
		static double Remainder(double x, double d);
		static double mod(double x, double d);
};

extern Function 
	lsetPrintPrecision, lreadFloatBin, lsetFPUStatusFlags;

#define Float_included
#endif
