#ifndef Node_included
	#include "Node.h"
#endif

#ifndef FloatArray_included

struct FloatArrayStruct
{
	int length, origin;
	doubleP data;
};

typedef FloatArrayStruct *FloatArrayStructP;

class FloatArray: public Node // is an float array with a length and an origin.
{
	int _length, _origin;
	doubleP _data;

	public:
		union {ObjectType flag; ObjectBits bit; int tint;};

		static const FloatArray aFloatArray;

		operator_new;
		FloatArray(): Node(FLOATARRAY), _length(0), _origin(0), _data(0) {}
		FloatArray(int length_, int origin_=1);
		FloatArray(FloatArrayP array_);
		void free();

		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		NodeP copy(PtrToInt_ &ptrToInt);

		StringP string();
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		void printBin(FILEP strm);
		NodeP contentsTree(	wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont, 
									ObjectType must, 
									int &count, int level);
		bool expanded() {return bit.table_expand;}
		bool expandable() {return true;}
		void setExpanded(bool state) {bit.table_expand=state;}

		OPERATORS_DECL
		int compare(NodeP nd);
		int count();
		int _sizeof();
		StringP _typeof();
		bool equalP(NodeP b);
		NodeP first_();
		NodeP butFirst();
		NodeP last();
		NodeP butLast();
		NodeP fPut(NodeP nd);
		NodeP lPut(NodeP nd);
		double _min(double d);
		double _max(double d);
		NodeP Norm();
		NodeP Norm3();
		NodeP maxNorm();
		NodeP resize(int newsize);
		NodeP lowPassFilter(double weight);
		NodeP lowPassFilter2(double weight);
		NodeP saturateAbove(double limit);
		NodeP saturateBelow(double limit);
		NodeP item(int index);
		NodeP item(NodeP index, NodeP result);
		NodeP items(NodeP start, NodeP end);
		NodeP setitem(NodeP index, NodeP newitem, bool safe);
		NodeP memberP(NodeP subobj, bool notp=false, bool substr=false);
		NodeP setitems(NodeP start, NodeP newitems);
		NodeP reverse();
		NodeP rotate(int places);
		NodeP shuffle();
		PairP list();
		ArrayP array(int origin=1);
		IntArrayP intarray(int origin=1);
		Int16ArrayP int16array(int origin=1);
		FloatArrayP floatarray(int origin=1);
		void combSortInPlace();
		NodeP combSort();
		GRAPHIC_DECL
		NodeP readbin(FileP file);

		 int length() {return _length;}
		 int origin() {return _origin;}
		doubleP data() {return _data;}
		void setorigin(int origin_) {_origin=origin_;}

	friend NodeP lDynamicLibraryCall(PairPP arg);
};

extern Function lreadFloatArrayBin;

	#define FloatArray_included
#endif
