#include "stdafx.h"
//___________________________________________________________________
//
//                        Class FloatControl
//
//                            Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/spinctrl.h>

#include "spinctld.h"

#include "Main.h"
#include "String.h"
#include "File.h"
#include "Object.h"
#include "Int.h"
#include "Float.h"
#include "Pair.h"
#include "WindowPair.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "Color.h"
#include "FloatControl.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Point.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(FloatControl, wxSpinCtrlDbl)
	EVT_SPINCTRL	(-1, FloatControl::OnChange)
END_EVENT_TABLE()

FloatControl::FloatControl(FloatControlNodeP floatControlNode_, int style)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=floatControlNode_;
	wxWindow *parent=(wxWindow*)console;
	if (floatControlNode_->parent)
		parent=floatControlNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			1,1,1,1,-1,
			_("FloatControl"));
}

FloatControl::FloatControl(FloatControlNodeP floatControlNode_, 
			wxWindow* parent, wxWindowID id, 
			double value, double minValue, double maxValue, double increment,
			int digits, int format_,
			PairP onChange_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
wxSpinCtrlDbl(*parent, id, 
			wxEmptyString,
			pos, size, style, 
			minValue, maxValue, value, increment, digits,
			name)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=floatControlNode_;
	*(PairP*)(d(onChange))=onChange_;
	format=format_;
	SetDigits(digits, formatType(format));
}

FloatControl::~FloatControl()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
}

void FloatControl::OnChange(wxSpinEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventFloatControl(this);
	if throwing return;
	*th->eventFiFo << evt;
	event.Skip();
}

wxString FloatControl::wxstring()
{
	wxString str;
	str << _("FloatControl ") << GetMin() << ' ' << GetValue() << ' ' << GetMax() << ' ' 
		<< GetIncrement() << ' ' << GetDigits() << ' ' << format << ' '
		<< '[' << GetName() << "] ";
	if (onChange)
		str << (*d(onChange))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &FloatControl::read(DataInputStream &istr) 
{	
	double value, min, max, increment;
	int digits;
	wxPoint pos;
	wxSize size;
	wxString name;
	
	istr >> min >> value >> max >> increment >> digits >> format
		>> name >> (int&)*d(onChange)
		>> pos.x >> pos.y >> size.x >> size.y;

	SetRange(min, max);
	SetValue(value);
	SetIncrement(increment, digits);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetDigits(digits, formatType(format));
	return istr;
}
wxDataOutputStream &FloatControl::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetMin() << GetValue() << GetMax() << GetIncrement() << GetDigits()
		<< format << GetName() 
		<< GC_::ptrToInt->find(*d(onChange))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;
	return ostr;
}

PairP FloatControlNode::onChange()
{
	if (floatControl)
		return PairP(*floatControl->d(FloatControl::onChange));
	return 0;
}

void FloatControlNode::setonChange(NodeP nd) 
{
	if (floatControl)
	{	*floatControl->d(FloatControl::onChange) = PairP(nd);  check_oldyoung(nd);
	}
}

FloatControlNode::FloatControlNode(): 
	WindowNode(FLOATCONTROL), 
	floatControl((FloatControlP&)window)
{}

FloatControlNode::FloatControlNode(WindowNodeP parent_): 
	WindowNode(FLOATCONTROL, parent_), 
	floatControl((FloatControlP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void FloatControlNode::free() 
{	
	if (floatControl) 
	{	ThreadP th=Thread::This();
//		GUI_ENTER
		*floatControl->d(FloatControl::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		floatControl->Destroy(); 
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tFloatControl:
				if (((EventFloatControl*)ev)->floatControl==floatControl)
					((EventFloatControl*)ev)->floatControl=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		floatControl=0;
//		GUI_LEAVE
	}
}
void FloatControlNode::inc_gc() {fifo << onChange();}
int FloatControlNode::youngs() {return gmark(onChange());}
int FloatControlNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onChange(), ptrToInt);
}
void FloatControlNode::check_oldyoungs() {check_oldyoung(onChange());}

DataInputStream &FloatControlNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode >> style;
		floatControl=new FloatControl(FloatControlNodeP(this), style);
		if (!floatControl)
		{	error(OUT_OF_MEM); return istr;
		}
		floatControl->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &FloatControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (floatControl)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent) 
			<< style;
		floatControl->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void FloatControlNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (floatControl)
		setonChange((Pair*)intToPtr.find(int(floatControl->onChange)));
}

void FloatControlNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	FloatControlP f=floatControl, f2=0;
	if (!f) return;
	f2=new FloatControl(FloatControlNodeP(this),
			p, -1, 
			f->GetValue(), f->GetMin(), f->GetMax(), f->GetIncrement(),
			f->GetDigits(), f->format,
			PairP(*f->d(FloatControl::onChange)), f->GetPosition(), f->GetSize(), style, 
			wxDefaultValidator, f->GetName());
	if (!f2)
	{	error(OUT_OF_MEM); return;
	}
	floatControl=f2;
	setonChange(*f->d(FloatControl::onChange));
	delete f;
	floatControl=f2;
}

StringP FloatControlNode::string()
{
	if (floatControl)
		return new String(floatControl->wxstring());
	return _Null_Word->name();
}

wxString FloatControlNode::wxstring()
{
	if (floatControl)
		return floatControl->wxstring();
	return wxEmptyString;
}

void FloatControlNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lFloatControl(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_("FloatControl");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	double minValue;
	{	double m=arg[3]->a_float(3);
		if throwing return Unbound;
		minValue=m;	
	}
	double value;
	{	double v=arg[4]->a_float(4);
		if throwing return Unbound;
		value=v;	
	}
	double maxValue;
	{	double m=arg[5]->a_float(5);
		if throwing return Unbound;
		maxValue=m;	
	}
	double increment;
	{	double i=arg[6]->a_float(6);
		if throwing return Unbound;
		increment=i;	
	}
	int digits=3;
	if (exist(7))
	{	int d=arg[7]->a_int(7);
		if throwing return Unbound;
		digits=d;	
	}
	PairP onChange=0;
	if (exist(8))
	{	PairP os=arg[8]->a_runnable(8);
		if throwing return Unbound;
		onChange=os;
	}
	int format=wxSpinCtrlDbl::lg_fmt;
	if (exist(9) && arg[9])
	{	int f=arg[9]->a_int(9);
		if throwing return Unbound;
		format=f;
	}
	long style = wxSP_ARROW_KEYS;
	if (exist(10) && arg[10])
	{	int s=arg[10]->a_int(10);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(11))
	{	PairP p=arg[11]->a_int_vector(11);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(12))
	{	PairP s=arg[12]->a_int_vector(12);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	FloatControlNodeP floatControlNode=new FloatControlNode(parentNode);
	if throwing return Unbound;
	FloatControlP floatControl=new FloatControl(floatControlNode,
			parent, id, 
			value, minValue, maxValue, increment, digits, format, 
			onChange, pos, size, style, wxDefaultValidator, name);
	if (!floatControl)
	{	error(OUT_OF_MEM); return Unbound;
	}
	floatControlNode->floatControl=floatControl;
	floatControlNode->style=style;
//	GUI_LEAVE
	return floatControlNode;
}

NodeP lFloatControlValue(PairPP arg)
{
	ThreadP th=Thread::This();
	if (exist(1))
	{	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
		if throwing return Unbound;
//		GUI_ENTER
		FloatP val=new Float(floatControl->floatControl->GetValue());
//		GUI_LEAVE
		return val;
	}
	else
		return new Float(th->floatControlValue);
}

NodeP lFloatControlSetValue(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	double value=arg[2]->a_float(2);
	if throwing return Unbound;
//	GUI_ENTER
	floatControl->floatControl->SetValue(value);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFloatControlSetRange(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	double min=arg[2]->a_float(2);
	if throwing return Unbound;
	double max=arg[3]->a_float(3);
	if throwing return Unbound;
//	GUI_ENTER
	floatControl->floatControl->SetRange(min, max);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFloatControlOnChange(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	floatControl->setonChange(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFloatControlEnable(PairPP arg)
{
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	floatControl->floatControl->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lFloatControlDestroy(PairPP arg)
{
//	GUI_ENTER
	FloatControlNodeP floatControl=arg[1]->a_floatControl(1);
	if throwing return Unbound;
	{	WindowNodeP p=floatControl->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(floatControl));
			p->check_oldyoung(p->children);
		}
	}
	floatControl->floatControl->Destroy();
//	GUI_LEAVE
	return Unbound;
}
