#ifndef FiFo_included
	#include "FiFo.h"
#endif

class PtrToInt_;
class IntToPtr_;
class PtrToString;
class StringToPtr;
class DataInputStream;
class wxDataOutputStream;

extern void __cdecl logo_stop(int);
extern void __cdecl logo_pause(int);

class GC_		// The Garbage Collector
{
	protected:
		NodeP
			_next,			   // node list of one generation
			oldyoung_next;	// list containing nodes that have childs

		int _marked,  // in mark marked is set to the gc run number
			gen,        // node's Generation
			age;      // How many times to GC at this generation

	public:
#ifdef _DEBUG
		const char* newFilename;
		int newLine;

		#define operator_new	void *operator new(size_t size, int blocktype, const char* newFilename_, int newLine_)

		#define alloc_dbg_info	, int blockType, const char *filename, int linenumber
		#define  malloc_GC	 _malloc_dbg(                size_t size alloc_dbg_info)
		#define  calloc_GC	 _calloc_dbg(size_t num,     size_t size alloc_dbg_info)
		#define realloc_GC	_realloc_dbg(void *memblock, size_t size alloc_dbg_info)
#else

		#define operator_new	void *operator new(size_t size)

		#define  malloc_GC	 malloc(size_t size)
		#define  calloc_GC	 calloc(size_t num, size_t size)
		#define realloc_GC	realloc(void *memblock, size_t size)

#endif

#ifdef PTR_DEBUG
		GC_(): _next(_next), oldyoung_next(oldyoung_next) {}
#else
		GC_() {}
#endif
		virtual ~GC_() {}

		NodeP next_() {return _next;}
		int marked() {return _marked;}
		void setnext(NodeP nd) {_next=nd;}
		void setoldyoung_next(NodeP nd) {oldyoung_next=nd;}

		virtual void free()=0;
		virtual void inc_gc()=0;
		virtual int youngs()=0;
		virtual void check_oldyoungs()=0;

		static int run_numero() {return _run_numero;}

		void mark();	// Iterative mark procedure
		void markPtr(PtrToInt_ &ptrToInt);// mark and insert ptr into ptrToInt table

		void check_oldyoung(NodeP new_); 
		int gmark(NodeP child);
		int gmarkPtr(NodeP child, PtrToInt_ &ptrToInt);

		virtual DataInputStream &read(DataInputStream &istr)=0;
		virtual wxDataOutputStream &write(wxDataOutputStream &ostr)=0;
		virtual void setPtrs(IntToPtr_ &intToPtr)=0;
		virtual void aftersetPtrs()=0;
		virtual void recreate()=0;
		virtual void aftersetPtrs2()=0;
		virtual void aftersetPtrs3()=0;

		static void init();
		static void run(bool full);	// collect garbage

		static void * malloc_GC;
		static void * calloc_GC;
		static void *realloc_GC;
		static void free(void *mem);

		static void reserve_tank_fill ();
		static void	reserve_tank_check();
		static void reserve_tank_use  (); 

		static void set_C_Stack(int stacksize, int *stack, int b);

		static void load(charP name);
		static void save(charP name);

		static bool initialized;
		static const int generations;

		static int volatile is_running;
		static int volatile disabled;

		static PtrToInt_ *ptrToInt;

		static PtrToString *primitivesPtrToString;
		static StringToPtr *primitivesStringToPtr;

		static const int segment_size;
		static const int fifo_size;
		static const int 	min_free;

		static const unsigned int MemBlockSize;
		static const unsigned int MemBlockMask;

		static int mem_free;
		static int mem_used;
		static int mem_peak;

		static int mark_gen;
		static int max_gen;
		static int _run_numero;
		static int 	examined;	// for GC_DEBUG

		enum Interrupt {no_interrupt, pause, stop};
		static Interrupt interrupt;

		static NodeP free_node;  // first of the free nodes list
		static NodeP oldyoungs;	 // list of nodes that point to younger nodes
		static NodeP SpareNode;

		static PairP reserve_tank;	// list of nodes not in any generation
		static FiFo fifo;	// for resolving recursive mark() calls, which could be quite deep

		static void __cdecl mark_many(NodeP arg1, ...);
		static void __cdecl mark_manyPtr(PtrToInt_ &ptrToInt, NodeP arg1, ...);
		static void __cdecl find_many(IntToPtr_ &intToPtr, 
			intP manyPtrNrs, NodePP arg1, ...);

		static void mark_C_stack();	// Check C Stack for Node pointers and mark valid ones
		static void mark_C_stackPtr(PtrToInt_ &ptrToInt);	// Check C Stack for Node pointers and mark valid ones

		static NodePP mark_oldyoung(NodePP prev);	// Mark/traverse pointers to younger generations only
		static NodePP mark_oldyoungPtr(NodePP prev, PtrToInt_ &ptrToInt);	
		// Mark pointers for load

		static void mark_oldyoungs();	// mark pointers from old generations to young (the oldyoungs list)
		static void mark_oldyoungsPtr(PtrToInt_ &ptrToInt);	// mark oldyoungs list for load

		static void mark_phase(bool clean_all);	
		// mark the root (all accessable node pointers of the generations<=mark_gen)

		static void mark_phasePtr(PtrToInt_ &ptrToInt);	
		// mark the root (all accessable node pointers of the generations<=mark_gen)

		static int sweep_phase(int gen);	// promote marked to next gen, free the rest

		static void gc(bool full);	// Garbage Collector

		static void inc_mem_nodes();
		static NodeP newnode();

	friend class Segment;
	friend class IntToPtr_;
	friend class PtrToInt_;
	friend class MemoryWindow;

	friend NodeP lNodes(PairPP);
	friend NodeP lrunSecure(PairPP);

	friend void __cdecl logo_stop(int);
	friend void __cdecl logo_pause(int);
	friend int __cdecl _main(int argc, charP argv[]);
};

#define GC_included
