//___________________________________________________________________
//
//                           Class  Gauge
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/gauge.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Pair.h"
#include "WindowPair.h"
#include "Point.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "Color.h"
#include "Gauge.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyGauge : public wxGauge
{
public:
	MyGauge(wxWindow* parent, wxWindowID id, 
		int range, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = _("Gauge"));

	// MSW sends extra kill focus event
	virtual ~MyGauge()
	{
		if (parent) 
			parent->gauge = 0;
		parent=0;
	}

	Gauge *parent;
};

MyGauge::MyGauge(wxWindow *parent_, wxWindowID id,
							  int range,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxGauge(parent_, id, range, pos, size, style, validator, name)
{
	parent = (Gauge*)parent_;
}
//___________________________________________________________________

Gauge::Gauge(wxWindow *parent):
	label(0), gauge(0), bestSize(0,0)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
}

Gauge::Gauge(GaugeNodeP gaugeNode_, int style):
	label(0), gauge(0), bestSize(-999, -999)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=gaugeNode_;
	wxWindow *parent=(wxWindow*)console;
	if (gaugeNode_->parent)
		parent=gaugeNode_->parent->window;
	create(parent, -1, 
			1, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			wxDefaultValidator, 
			_("Gauge"));
}

Gauge::Gauge(GaugeNodeP gaugeNode_, 
			wxWindow* parent, wxWindowID id, 
			int range, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), gauge(0), bestSize(-999, -999)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=gaugeNode_;
	bestSize = wxSize(-999, -999);
	gaugeNode_->window=WindowP(this);
	gaugeNode_->style=style;
	create(parent, -1, 
			range, 
			wxDefaultPosition, 
			wxDefaultSize, 
			style, 
			wxDefaultValidator, 
			name);
}

bool Gauge::create(wxWindow* parent, wxWindowID id, 
			int range, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
	{	label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, _("staticText"));
		if (!label)
		{	error(OUT_OF_MEM); return false;
		}
	}else
		label=0;

	if (label)
		gauge = new MyGauge(this, id, range,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		gauge = new MyGauge(this, id,range,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);

	if (!gauge)
	{	error(OUT_OF_MEM); return false;
	}
	DoSetSize(pos.x, pos.y, width, height);
	SetInitialSize(wxSize(width, height));
	return true;
}

void Gauge::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (gauge)   
		gauge->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize Gauge::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxGauge *t=new wxGauge((wxWindow*)this, wxID_ANY, 0,
			wxPoint(0,0), wxDefaultSize, GetWindowStyle());
		if (!t)
		{	error(OUT_OF_MEM); return wxDefaultSize;
		}
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

Gauge::~Gauge()
{
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (gauge)
	{	delete gauge;
		gauge=0;
	}
}

wxString Gauge::wxstring()
{
	wxString str;
	if (gauge)
		str 
		<< _("Gauge") << " " << gauge->GetRange() << ' ' << gauge->GetValue() << ' ' 
		<< '[' << GetName() << "] "
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &Gauge::read(DataInputStream &istr) 
{	
	int range, value, style;
	wxPoint pos;

	wxSize size;
	wxString name;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> range >> value >> name
		>> pos.x >> pos.y >> size.x >> size.y >> style
		>> fr >> fg >> fb 
		>> br >> bg >> bb;

	gauge->SetRange(range);
	gauge->SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	return istr;
}
wxDataOutputStream &Gauge::write(wxDataOutputStream &ostr) 
{
	int range=0, value=0;
	if (gauge)
	{	range=gauge->GetRange();
		value=gauge->GetValue();
	}
	ostr
		<< range << value << GetName() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle());

	{	wxColour 
			f=gauge->GetForegroundColour(), 
			b=gauge->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	return ostr;
}

bool Gauge::SetBackgroundColour(const wxColour &c)
{
	if (gauge)
		return gauge->SetBackgroundColour(c);
	return false;
}

bool Gauge::SetForegroundColour(const wxColour &c)
{
	if (gauge)
		return gauge->SetForegroundColour(c);
	return false;
}

GaugeNode::GaugeNode(): 
	WindowNode(GAUGE),
	gauge((GaugeP&)window)
{}

GaugeNode::GaugeNode(WindowNodeP parent_): 
	WindowNode(GAUGE),
	gauge((GaugeP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void GaugeNode::free() 
{	
	if (gauge) 
	{	*gauge->d(Gauge::windowNode)=0;
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		gauge->Destroy(); 
		gauge=0;
	}
}

DataInputStream &GaugeNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr 
			>> (int&)parentNode 
			>> style;
		gauge=new Gauge(GaugeNodeP(this), style);
		if (!gauge)
		{	error(OUT_OF_MEM); return istr;
		}
		gauge->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &GaugeNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (gauge)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent) 
			<< style;
		gauge->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void GaugeNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
}

void GaugeNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	int range=0, value=0;
	if (GaugeP(gauge)->gauge)
	{	range=GaugeP(gauge)->gauge->GetRange();
		value=GaugeP(gauge)->gauge->GetValue();
	}
	GaugeP g=GaugeP(gauge), 
		g2=new Gauge(GaugeNodeP(this),
			p, -1, range, 
			g->GetPosition(), g->GetSize(), style, 
			wxDefaultValidator, g->GetName());
	if (!g2)
	{	error(OUT_OF_MEM); return;
	}
	g2->gauge->SetRange(range);
	g2->gauge->SetValue(value);
	g2->gauge->SetForegroundColour(g->gauge->GetForegroundColour());
	g2->gauge->SetBackgroundColour(g->gauge->GetBackgroundColour());
	gauge=g2;
	delete g;
	gauge=g2;
}

StringP GaugeNode::string()
{
	if (gauge)
		return new String(gauge->wxstring());
	return _Null_Word->name();
}

wxString GaugeNode::wxstring()
{
	if (gauge)
		return gauge->wxstring();
	return wxEmptyString;
}

void GaugeNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lGauge(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_("Gauge");
	if (arg[2])
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	int range;
	{	int r=arg[3]->a_int(3);
		if throwing return Unbound;
		range=r;	
	}
	int value=0;
	if (exist(4) && arg[4])
	{	int v=arg[4]->a_int(4);
		if throwing return Unbound;
		if (v > range)
		{	error(BAD_DATA, arg[4], 4);
			return Unbound;
		}
		value=v;	
	}
	long style = wxGA_HORIZONTAL;
	if (exist(5) && arg[5])
	{	int s=arg[5]->a_int(5);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(6) && arg[6])
	{	PairP p=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(7))
	{	PairP s=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	GaugeNodeP gaugeNode=new GaugeNode(parentNode);
	if throwing return Unbound;
	GaugeP gauge=new Gauge(gaugeNode,
			parent, id, range,
			pos, size, style, wxDefaultValidator, name);
	if (!gauge)
	{	error(OUT_OF_MEM); return Unbound;
	}
	gauge->gauge->SetValue(value);
//	GUI_LEAVE
	return gaugeNode;
}

NodeP lGaugeValue(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
//	GUI_ENTER
	IntP val=new Int(GaugeP(gauge->gauge)->gauge->GetValue());
//	GUI_LEAVE
	return val;
}

NodeP lGaugeSetValue(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	int value=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	gauge->gauge->gauge->SetValue(value);
//	GUI_LEAVE
	return Unbound;
}

NodeP lGaugeSetRange(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	int range=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	gauge->gauge->gauge->SetRange(range);
//	GUI_LEAVE
	return Unbound;
}

NodeP lGaugeSetBackgroundColor(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	gauge->gauge->SetBackgroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lGaugeSetColor(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	gauge->gauge->SetForegroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lGaugeDestroy(PairPP arg)
{
	GaugeNodeP gauge=arg[1]->a_gauge(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=gauge->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(gauge));
			p->check_oldyoung(p->children);
		}
	}
	gauge->gauge->Destroy();
//	GUI_LEAVE
	return Unbound;
}
