//___________________________________________________________________
//
//                          Class Graph
//
//                          Definitions
//___________________________________________________________________

#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
	#define DOSWIN32
	#include <windows.h>
	#include <vfw.h>
#endif

#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/dcmemory.h>
#include <wx/html/helpctrl.h>
#include <wx/log.h>

#include <wx/image.h>
#include <wx/datstrm.h>
#include <wx/txtstrm.h>
#include <wx/metafile.h>
#include <wx/aui/aui.h>

#ifndef WIN32
	#include <setjmp.h>
	#define __FT2_BUILD_UNIX_H__
	#define FREETYPE /usr/include/freetype2/freetype
	#define FT_FREETYPE_H <FREETYPE/freetype.h>
	#define FT_ERRORS_H <FREETYPE/fterrors.h>
	#define FT_CONFIG_CONFIG_H <FREETYPE/config/ftconfig.h>
	#define FT_CONFIG_OPTIONS_H <FREETYPE/config/ftoption.h>
	#define FT_CONFIG_STANDARD_LIBRARY_H <FREETYPE/config/ftstdlib.h>
	#define FT_MODULE_ERRORS_H <FREETYPE/ftmoderr.h>
	#define FT_BEGIN_HEADER
	#define FT_END_HEADER
	#define FT_ERROR_DEFINITIONS_H <FREETYPE/fterrdef.h>
	#define FT_TYPES_H <FREETYPE/fttypes.h>
	#define FT_SYSTEM_H <FREETYPE/ftsystem.h>
	#define FT_IMAGE_H <FREETYPE/ftimage.h>
	#define FT_GLYPH_H <FREETYPE/ftglyph.h>
#endif

#ifdef OPENGL
	#include <FTGL/FTGLPolygonFont.h>
	#include <gl2ps.h>
//	link with:
//	in Debug mode:
//	opengl32.lib glu32.lib freetype221MT_d.lib ftgl_static_MT_d.lib gl2psd.lib
// in Release mode:
// opengl32.lib glu32.lib freetype221MT.lib ftgl_static_MT.lib gl2ps.lib wxmsw28_gl.lib 
#endif

#ifndef WIN32
	#include <avifile-0.7/avifile.h>
	#include <avifile-0.7/avm_fourcc.h>
#endif

#include "Main.h"
#include "Error.h"
#include "File.h"
#include "Int.h"
#include "UInt8.h"
#include "Float.h"
#include "Pair.h"
#include "Line.h"
#include "List.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "Table.h"
#include "PtrTable.h"
#include "LogoObject.h"
#include "Frame.h"
#include "Init.h"
#include "NodePStack.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Mouse.h"
#include "FTGLFontManager.h"
#include "Font.h"
#include "Turtle.h"
#include "TurtleNode.h"
#include "Bitmap.h"
#include "Texture.h"

#include "Recorder.h"

#include "AVIGen.h"

#include "IntToPtr_.h"
#include "PtrToInt_.h"

#include "Event.h"
#include "Splitter.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Shadow.h"
#include "Security.h"
#include "Thread.h"
#include "Aui.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

#ifdef OPENGL
BEGIN_EVENT_TABLE(Graph, wxGLCanvas)
#else
BEGIN_EVENT_TABLE(Graph, wxWindow)
#endif
	EVT_SIZE					(Graph::OnSize					)
	EVT_PAINT				(Graph::OnPaint				)
	EVT_ERASE_BACKGROUND	(Graph::OnErase				) 
	EVT_SET_FOCUS			(Graph::OnSetFocus			)
	EVT_CHAR					(Graph::OnChar					)
	EVT_KEY_DOWN			(Graph::OnKeyDown				)
	EVT_KEY_UP				(Graph::OnKeyUp				)
	EVT_ENTER_WINDOW     (Graph::OnEnterWindow		)
	EVT_LEAVE_WINDOW     (Graph::OnLeaveWindow		)
	EVT_LEFT_DOWN			(Graph::OnMouseLeftDown		)
	EVT_RIGHT_DOWN			(Graph::OnMouseRightDown	)
	EVT_MIDDLE_DOWN		(Graph::OnMouseMiddleDown	)
	EVT_LEFT_UP				(Graph::OnMouseLeftUp		)
	EVT_RIGHT_UP			(Graph::OnMouseRightUp		)
	EVT_MIDDLE_UP			(Graph::OnMouseMiddleUp		)
	EVT_LEFT_DCLICK		(Graph::OnMouseLeftDClick	)	
	EVT_RIGHT_DCLICK		(Graph::OnMouseRightDClick	)
	EVT_MIDDLE_DCLICK		(Graph::OnMouseMiddleDClick)
	EVT_MOTION				(Graph::OnMouseMotion		)
	EVT_MOUSE_EVENTS		(Graph::OnMouse				)
END_EVENT_TABLE()

void mySwapBuffers()
{
	GraphP graph=Thread::This()->graph;
	if (graph)
	{	bool playOnSwap=graph->playOnSwap;
		graph->playOnSwap=false;
		graph->swapBuffers();
		graph->playOnSwap=playOnSwap;
	}
}

int Graph::attribs[]=
#ifdef OPENGL
#ifdef WIN32
	{	WX_GL_RGBA,
		WX_GL_DOUBLEBUFFER,
		WX_GL_DEPTH_SIZE,
		32,
		WX_GL_STENCIL_SIZE,
		1,
		0
	};
#else
	{	WX_GL_RGBA,
		WX_GL_DOUBLEBUFFER,
//		WX_GL_DEPTH_SIZE,
//		24,
		WX_GL_STENCIL_SIZE,
		1,
		0
	};
#endif
#else
	{0};
#endif

Graph::Graph(wxWindow *parent):
#ifdef OPENGL
	wxGLCanvas(parent, -1, 0, wxPoint(0,0), wxSize(0,0), 0)
#else
	wxWindow(parent, wxID_ANY, wxPoint(0,0), wxSize(0,0), wxNO_BORDER)
#endif
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	recorder=0;
	ftglFontManager=0;
	mouse=0;
	screen=0;
	shadowObject=0;
	csGraph=0;
	csWaitingForGraph=0;
#ifndef OPENGL
	mdc=0;
	bitmap=0;
#endif
}

Graph::Graph(GraphNodeP graphNode_):
#ifdef OPENGL
	wxGLCanvas((wxWindow*)//theApp->mainFrame->splitter
		theApp->mainFrame, wxID_ANY,
		wxDefaultPosition, wxDefaultSize, 0,
		wxGLCanvasName, attribs, wxNullPalette),
#else
	wxWindow(theApp->mainFrame, wxID_ANY, 
		wxDefaultPosition, wxDefaultSize, wxNO_BORDER),
#endif
	turtle(0)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=graphNode_;
	wxWindow *parent=0;
	if (*d(windowNode) && WindowNodeP(*d(windowNode))->parent)
		parent=WindowNodeP(*d(windowNode))->parent->window;
	else
		parent=theApp->graph;

#ifndef OPENGL
	mdc=0;
	bitmap=0;
#endif
	create(parent, -1, wxDefaultPosition, wxDefaultSize, 
		wxFULL_REPAINT_ON_RESIZE, _("Graph"), attribs,
		wxNullPalette, false);
	create2();
}

Graph::Graph(GraphNodeP graphNode_, 
				 wxWindow* parent, wxWindowID id, 
				 const wxPoint& pos, const wxSize& size, 
				 long style, const wxString& name, 
				 int* attribList, const wxPalette& palette,
				 bool pictureAutosave):
#ifdef OPENGL
#ifdef WIN32
	wxGLCanvas(parent, (wxGLCanvas*)0, id, pos, size, style, name, attribs)
#else
	wxGLCanvas(parent, (wxGLCanvas*)0, id, pos, size, style)
#endif
#else
	wxWindow(parent, wxID_ANY, pos, size, style)
#endif
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=graphNode_;
	ThreadP th=Thread::This();
	turtle=0;
	recorder=0;
#ifndef OPENGL
	mdc=0;
	bitmap=0;
#endif
	create(parent, id, pos, size, 
		style | wxFULL_REPAINT_ON_RESIZE, name, attribs,
		palette, pictureAutosave);

	turtle=new Turtle;
	if (!turtle)
	{	error(OUT_OF_MEM); return;
	}
	recorder=new Recorder(this);
	if (!recorder)
	{	error(OUT_OF_MEM); return;
	}
	recorderSave=recorder;
	turtle->create(recorder, screen);
	if (GC_::initialized)
	{	*d(turtleNode)=new TurtleNode(turtle, GraphNodeP(*d(windowNode)));
		if throwing return;
		th->turtleNode=TurtleNodeP(*d(turtleNode));
	}
	create2();
}

void Graph::create(wxWindow* parent, wxWindowID id, 
				 const wxPoint& pos, const wxSize& size, 
				 long style, const wxString& name, 
				 int* attribList, const wxPalette& palette,
				 bool pictureAutosave)
{
	ThreadP th=Thread::This();
	thread=th;
	inread=false;
	was1=printing=saving=false;
	oldPalette = NULL;
	doubleBuffer=theApp->doubleBuffer;
	if (!doubleBuffer)
	{
#ifdef OPENGL
		glReadBuffer(GL_FRONT);
		glDrawBuffer(GL_FRONT);
#endif
	}
	playOnSwap=theApp->playOnSwap;
	screen=new Screen_();
	if (!screen)
	{	error(OUT_OF_MEM); return;
	}
	mouse=new Mouse();
	if (!mouse)
	{	error(OUT_OF_MEM); return;
	}
	ftglFontManager=new FTGLFontManager;
	if (!ftglFontManager)
	{	error(OUT_OF_MEM); return;
	}

	wxSize s=GetClientSize();
	screen->create(s.GetWidth(), s.GetHeight());

	saveSize=wxSize(800,600);

	aviGen=0;
	autosave=pictureAutosave;
	pictureFileName=theApp->pictureFileName;
	shadowObject=new ShadowObject;
	if (!shadowObject)
	{	error(OUT_OF_MEM); return;
	}
	colorTolerance=0.3f;

	graphOwned=0;
	waitingForGraph=0;

	csWaitingForGraph=new DispatchMutex();
	if (!csWaitingForGraph)
	{	error(OUT_OF_MEM); return;
	}
	csGraph=new DispatchMutex();
	if (!csGraph)
	{	error(OUT_OF_MEM); return;
	}
	thread=th;
	inSwapBuffers=false;
	inOnPaint=false;
	inOnSize=false;
}

void Graph::create2()
{
	ThreadP th=Thread::This();
	th->gtTable->setitem(*d(windowNode), *d(turtleNode));
#ifndef OPENGL
	wxClientDC cdc(this);
	mdc=new wxMemoryDC(&cdc);
	bitmap=new wxBitmap(screen->winsizex, screen->winsizey, mdc);
	dc=mdc;
	pen.SetColour(*wxBLACK);
	brush.SetColour(*wxWHITE);
	dc->SetPen(pen);
	dc->SetBrush(brush);
#endif
	GraphP gr=th->graph;
	if (gr)
	{	gr->graphOwned=1;
		gr->leave();
	}
//	GraphP og=th->graph;
//	th->graph=this;
	enter();
#ifdef OPENGL
#ifndef WIN32
	th->dispatchMessages();
#endif
	if (GetContext())
		GetContext()->SetCurrent(*this);
	setOpenGLState();
#else
	dc->SetAxisOrientation(true, true);
	dc->SetDeviceOrigin(400,300);
#endif
	screen_clear(false);
//	if (og)
//		th->graph=og;
}

Graph::~Graph()
{
	ThreadP th=Thread::This();
	reset();
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
	if (recorder)
	{	recorder->refCount--;
		if (recorder->refCount==0)
			recorder->free();
	}
	if (ftglFontManager)
		delete ftglFontManager;
	if (screen)
		screen->destroy();
	if (mouse)
		delete mouse;
	if (screen)
		delete screen;
	if (shadowObject)
		delete shadowObject;

	if (csGraph)
	{	csGraph->Unlock();
		delete csGraph;
		csGraph=0;
	}
	if (csWaitingForGraph)
	{	delete csWaitingForGraph;
		csWaitingForGraph=0;
	}
	EventPP out=th->eventFiFo->out;
	EventP ev;
	while (*th->eventFiFo)
	{	*th->eventFiFo >> ev;
		switch (ev->type)
		{
		case Event::GraphChar:
		case Event::GraphKeyDown:
		case Event::GraphKeyUp:
			if (((EventGraph*)ev)->graph==this)
				((EventGraph*)ev)->graph=0;
			break;
		default:
			;
		}
	}
	th->eventFiFo->out=out;
#ifdef OPENGL
#else
	if (bitmap)
		delete bitmap;
	bitmap=0;
	if (mdc)
		delete mdc;
	mdc=0;
#endif
}

void Graph::reset()
{
	ThreadP th=Thread::This();
	GraphP &gr=th->graph;
	TurtleNodeP &tht=th->turtleNode;
	if (gr==this)
	{	if (gr 
#ifdef OPENGL
		&& gr->m_glContext 
#endif
		&&  theApp->mainFrame
		&&  gr != theApp->graph
		&&  theApp->mainFrame->initialized)
		{	gr=theApp->graph;
			if (graphOwned)
			{	graphOwned=1;
				leave();
				gr->enter();
			}
			tht=TurtleNodeP(*gr->d(turtleNode));
#ifdef OPENGL
#ifdef WIN32		
			wglMakeCurrent(
				(HDC)gr->GetHDC(), //m_glContext->m_hDC,
				gr->GetContext()->GetGLRC());//m_glContext->m_glContext);
#else
			gr->SetCurrent();
//			glXMakeCurrent((Display*) wxGetDisplay(),
//				Window(gr->m_Window->GetClientAreaWindow()),
//				gr->GetContext()->m_glContext);
#endif
#endif
		}else if (!theApp->mainFrame)
		{	gr=0;
			tht=0;
		}
	}
}

void Graph::enter()
{
	ThreadP th=Thread::This();
	if (graphOwned && thread==Thread::This())
	{	graphOwned++;
//c_out << "A" << graphOwned << " " << int(this) << endL;
		return;
	}
	if (!thread || thread==thread->threadFiber 
	|| thread->threadFiber != Thread::ThisThread())
	{	csWaitingForGraph->Lock();
		waitingForGraph++;
		csWaitingForGraph->Unlock();
		csGraph->Lock();
		csWaitingForGraph->Lock();
		waitingForGraph--;
#ifdef OPENGL
#ifdef WIN32
		wglMakeCurrent(HDC(GetHDC()), GetContext()->GetGLRC());
#else
		SetCurrent();
#endif
#endif
			//m_glContext->m_hDC, m_glContext->m_glContext);
		csWaitingForGraph->Unlock();
	}
	thread=Thread::This();
	graphOwned++;
	th->graph=this;
//c_out << "B" << graphOwned << " " << int(this) << endL;
	TurtleNodeP tn=TurtleNodeP(th->gtTable->item(*d(windowNode)));
	th->turtleNode=tn;
//c_out << int(tn) << endL; console->flush();
	if (!tn)
	{	turtle=new Turtle;
		if (!turtle)
		{	error(OUT_OF_MEM); return;
		}
		turtle->create(recorder, screen);
		if (GC_::initialized)
		{	tn=new TurtleNode(turtle, GraphNodeP(*d(windowNode)));
			if throwing return;
			ArrayP shapes=0;
			if (*d(turtleNode))
				shapes=TurtleNodeP(*d(turtleNode))->shapes;
			th->turtleNode=tn;
			tn->shapes=shapes;
			th->gtTable->setitem(*d(windowNode), *d(turtleNode));
		}
	}
	*d(turtleNode)=tn;
	tn->graphNode=GraphNodeP(*d(windowNode));
	turtle=tn->turtle;
	turtle->recorder=recorder;
	turtle->screen=screen;
	recorder->turtle=turtle;
	turtle->setColor(turtle->pen.color);
	if (recorder->recording)
		*recorder << new R_setPenColor(*turtle);
	Point p=*turtle;
	if (turtle->mode==Turtle::wrapmode)
		turtle->getInWindow();
	turtle->moveTo();
	*turtle=p;
}

void Graph::leave()
{
	if (!graphOwned)
		return;
	if (thread && thread->threadFiber != Thread::ThisThread())
		return;
	graphOwned--;
	wxASSERT(graphOwned>=0);
	if (!graphOwned)
	{
#ifdef OPENGL
		if (turtle)
			if (!turtle->polyMode && !turtle->tessMode && !turtle->surfaceMode)
				glFinish();
		ThreadP th=Thread::This();
		th->dispatchMessages();
#ifdef WIN32
		wglMakeCurrent(0,0);
#else
		glXMakeCurrent((Display*)wxGetDisplay(), 0,0);
#endif
#endif
		//		if (thread && thread->threadFiber == Thread::This())
			csGraph->Unlock();
	}
}

bool Graph::tryEnter()
{
/*	ThreadP th=Thread::This();
	if (graphOwned && thread==th)
		return true;
	csWaitingForGraph->Lock();
	if (!waitingForGraph)
	{	if (!graphOwned)
		{	waitingForGraph++;
			csGraph->Lock();
			waitingForGraph--;
			wglMakeCurrent((HDC) m_glContext->m_hDC, m_glContext->m_glContext);
			thread=th;
			graphOwned=true;
			turtleNode=th->turtleNode;
			turtle=turtleNode->turtle;
			turtle->recorder=recorder;
			turtle->screen=screen;
			TURTLE=turtle;
			RECORDER.turtle=turtle;
			turtle->setColor(turtle->pen.color);
			if (RECORDER.recording)
				RECORDER << new R_setPenColor(*TURTLE);
			Point p=*turtle;
			turtle->getInWindow();
			turtle->moveTo();
		}
		csWaitingForGraph->Unlock();
		return true;
	}
	csWaitingForGraph->Unlock();
*/	return false;
}

GraphLocker::GraphLocker()
{
	ThreadP th=Thread::This();
	theApp->prepareToDraw();
	if (th->graph)
		th->graph->enter();
}

GraphLocker::~GraphLocker()
{
	ThreadP th=Thread::This();
	if (th->graph)
		th->graph->leave();
}

void Graph::show(bool shown)
{
	Show();
	wxAuiPaneInfo &pane=theApp->aui->GetPane("Graph");
	if (pane.IsOk())
	{	pane.Show();
	}else
		theApp->aui->AddPane(this, 
			wxAuiPaneInfo().Name("Graph").Caption(_("Graph")).
			Top().
			BestSize(800,600).
			MaximizeButton().MinimizeButton().PinButton().PaneBorder(false)
		);
/*			wxAuiPaneInfo().Name("Graph").Caption(_("Graph")).
			Layer(2).Right().
			MinimizeButton(true).PinButton().Gripper().
			PaneBorder(false).DestroyOnClose());
*/	update();
	theApp->aui->Update();
}

void Graph::OnSetFocus(wxFocusEvent& event)
{
/*	if (this==theApp->graph && theApp->consoleFrame)
	{	theApp->mainFrame->SetFocus();
		theApp->consoleFrame->SetFocus();
		if (console)
			console->SetFocus();
	}
*/	event.Skip();
}

void Graph::OnChar(wxKeyEvent &event)				
{																
	ThreadP th=Thread::This();							
	EventP evt=new EventGraphChar(this, event);	
	if throwing return;										
	*th->eventFiFo << evt;
	event.Skip();												
}
void Graph::OnKeyDown(wxKeyEvent &event)				
{
	ThreadP th=Thread::This();								
	EventP evt=new EventGraphKeyDown(this, event);	
	if throwing return;										
	*th->eventFiFo << evt;									
	if (event.GetKeyCode()==WXK_TAB && event.ControlDown() && theApp->console)
	{	theApp->console->SetFocus();
		return;
	}
	event.Skip();												
}
void Graph::OnKeyUp(wxKeyEvent &event)				
{																	
	ThreadP th=Thread::This();								
	EventP evt=new EventGraphKeyUp(this, event);	
	if throwing return;									
	*th->eventFiFo << evt;								
	event.Skip();											
}

wxString Graph::wxstring()
{
	wxString str;
	str << _("Graph") << " [" 
		<< GetName() << "]["
		<< GetPosition().x << ' ' << GetPosition().y << "] [" 
		<< GetSize().x << ' ' << GetSize().y << "] " 
		<< GetWindowStyle() << " [" << GetName() << ']';
	return str;
}

DataInputStream &Graph::read(DataInputStream &istr) 
{	
	inread=true;
	wxString title, name;
	wxPoint pos;
	int style;
	
	recorder=new Recorder(this);
	if (!recorder)
	{	error(OUT_OF_MEM); return istr;
	}
	recorderSave=recorder;

	istr 
//		>> title 
		>> pos.x >> pos.y >> size.x >> size.y >> style >> name
#define READ_ON(NAME)	>> (int&)*d(on##NAME) >> (int&)*d(uon##NAME) >> (int&)*d(ton##NAME)
		READ_ON(Char)
		READ_ON(KeyDown)
		READ_ON(KeyUp)
		READ_ON(LeftDown)
		READ_ON(RightDown)
		READ_ON(MiddleDown)
		READ_ON(LeftUp)
		READ_ON(RightUp)
		READ_ON(MiddleUp)
		READ_ON(LeftDClick)
		READ_ON(RightDClick)
		READ_ON(MiddleDClick)
		READ_ON(Motion)
		>> (int&)*d(turtleNode)
		>> active
		>> direct
		>> doubleBuffer
		>> playOnSwap
		>> screen->eye >> screen->center >> screen->up
		>> screen->Min.x >> screen->Min.y >> screen->Min.z
		>> screen->Max.x >> screen->Max.y >> screen->Max.z
		>> *recorder;
//	turtle->read(istr);
//	turtle->refCount+=2;

//	SetTitle(title); 
	SetPosition(pos);
	SetWindowStyle(style);
	SetName(name); 
	return istr;
}

wxDataOutputStream &Graph::write(wxDataOutputStream &ostr) 
{
	ostr
//		<< GetTitle() 
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y << int(GetWindowStyle()) << GetName();

#define WRITE_ON(NAME)	\
	ostr << GC_::ptrToInt->find(*d(on##NAME));		\
	ostr << GC_::ptrToInt->find(*d(uon##NAME));	\
	ostr << GC_::ptrToInt->find(*d(ton##NAME));

	WRITE_ON(Char)
	WRITE_ON(KeyDown)
	WRITE_ON(KeyUp)
	WRITE_ON(LeftDown)
	WRITE_ON(RightDown)
	WRITE_ON(MiddleDown)

	WRITE_ON(LeftUp)
	WRITE_ON(RightUp)
	WRITE_ON(MiddleUp)

	WRITE_ON(LeftDClick)
	WRITE_ON(RightDClick)
	WRITE_ON(MiddleDClick)

	WRITE_ON(Motion)
	ostr << GC_::ptrToInt->find(*d(turtleNode));

	if (this==Thread::This()->graph)
		ostr << 1;
	else
		ostr << 0;
	ostr 
		<< direct
		<< doubleBuffer
		<< playOnSwap
		<< screen->eye << screen->center << screen->up
		<< screen->Min.x << screen->Min.y << screen->Min.z
		<< screen->Max.x << screen->Max.y << screen->Max.z;
	ostr << *recorder;
//	turtle->write(ostr);
	return ostr;
}

void Graph::setPtrs(IntToPtr_ &intToPtr)
{	
	*d(turtleNode)=(TurtleNode*)intToPtr.find(int(*d(turtleNode)));
	turtle=TurtleNodeP(*d(turtleNode))->turtle;
	recorder->turtle=turtle;
	turtle->create(recorder, screen);

	CommandListIterator i(recorder);
	for (i=recorder->begin();  i;  i++)
		i->setPtrs(intToPtr);

	if (active)
	{
#ifdef OPENGL
#ifdef WIN32
		wglMakeCurrent(HDC(GetHDC()), GetContext()->GetGLRC());
//			(HDC)m_glContext->m_hDC, m_glContext->m_glContext);
#else
//		glXMakeCurrent((Display*) wxGetDisplay(),
//			Window(m_glContext->m_window->GetClientAreaWindow()),
//			m_glContext->m_glContext);
		SetCurrent();
#endif
#endif
		Thread::This()->graph=this;
	}
	inread=false;
	SetSize(size);
	wxSizeEvent sev(size);
	OnSize(sev);
}

void Graph::setOpenGLState()
{
#ifdef OPENGL
	GLfloat fMaxObjSize, fAspect;

	glClearDepth(1.0f);
	glDepthFunc(GL_LEQUAL);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	GLfloat light_ambient[] = {0.1f,0.1f,0.1f,1.0f};
	GLfloat light_diffuse[] = {1,1,1,1};
	GLfloat light_specular[] = {1,1,1,1};
	/*  light_position is NOT default value */
	GLfloat light_position[] = {1,1,1,0};

	glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv (GL_LIGHT0, GL_POSITION, light_position);

//	glEnable (GL_LIGHTING);
//	glEnable (GL_LIGHT0);

//	glEnable(GL_COLOR_MATERIAL);
//	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

	if (oldRect.GetHeight())
		fAspect = (GLfloat)oldRect.GetWidth()/oldRect.GetHeight();
	else    // don't divide by zero, not that we should ever run into that...
		fAspect = 1.0f;
	nearPlane = 3.0f;
	farPlane = 10000;//7.0f;
	fMaxObjSize = (screen->Max.x-screen->Min.x)/2;//3.0f;
	fRadius = nearPlane + fMaxObjSize/2;

	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

//	glPushMatrix();
//	glTranslatef(0, 0, -fRadius);
#endif
	setFrustum(
		screen->Max.x - screen->Min.x, 
		screen->Max.y - screen->Min.y); 
	screen->getMatrices();
}

void Graph::OnPaint(wxPaintEvent& event)
{
//	if (recorder->inReplay) 
//		return;
	if (!(inOnPaint || printing || saving))
	{	inOnPaint=true;
		ThreadP th=Thread::This();
		if (this!=th->graph)
		{	GraphP og=th->graph;
			int oo=og->graphOwned;
			og->graphOwned=1;
			og->leave();
			enter();
			swapBuffers();
			leave();
			og->enter();
			og->graphOwned=oo;
		}else
			swapBuffers();
		inOnPaint=false;
	}
	event.Skip();
}

void Graph::drawTurtles(bool incNr)
{
//	turtle->draw();
//return;
//	turtle->shapeNr++;

	ThreadP th2=Thread::main->startFiber;
	ThreadP th0=th2;
	while (th2)
	{	if (th2->graph==this && th2->turtleNode)
		{	th2->turtleNode->turtle->draw();
			if (incNr)
				th2->turtleNode->turtle->shapeNr++;
		}
		if (th2->nextFiber==th0)
			break;
		else
			th2=th2->nextFiber;
	}
	for (ThreadList::Node *node = Thread::list.GetFirst();  node;  node = node->GetNext())
	{	ThreadP th2=node->GetData()->startFiber;
		ThreadP th0=th2;
		while (th2)
		{	if (th2->graph==this && th2->turtleNode)
			{	th2->turtleNode->turtle->draw();
				if (incNr)
					th2->turtleNode->turtle->shapeNr++;
			}
			if (th2->nextFiber==th0)
				break;
			else
				th2=th2->nextFiber;
		}
	}
}

void Graph::swapBuffers()
{
#ifdef OPENGL
	if (m_glContext && !inSwapBuffers)
	{	inSwapBuffers=true;
		ThreadP th=Thread::This();
		th->GuiEnter();
#ifdef WIN32
//		HDC oldDC=wglGetCurrentDC();
//		HGLRC oldContext=wglGetCurrentContext();
//		wglMakeCurrent((HDC) m_glContext->m_hDC, m_glContext->m_glContext);
#else
//		GLXDrawable oldDrawable=glXGetCurrentDrawable();
//		GLXContext oldContext=glXGetCurrentContext();
//		glXMakeCurrent((Display*) wxGetDisplay(),
//			Window(m_glContext->m_window->GetClientAreaWindow()),
//			m_glContext->m_glContext);
//		SetCurrent();
#endif
		if (doubleBuffer)
		{
			glReadBuffer (GL_BACK); 
			glDrawBuffer (GL_FRONT);

			GLboolean isat, isblend, isclop, iscm, isdi, isdt, isfog, 
				islop, isps, islight, istex;

			glGetBooleanv(GL_ALPHA_TEST, &isat);
			glGetBooleanv(GL_BLEND, &isblend);
			glGetBooleanv(GL_COLOR_LOGIC_OP, &isclop);
			glGetBooleanv(GL_COLOR_MATERIAL, &iscm);
			glGetBooleanv(GL_DEPTH_TEST, &isdt);
			glGetBooleanv(GL_DITHER, &isdi);
			glGetBooleanv(GL_FOG, &isfog);
			glGetBooleanv(GL_LIGHTING, &islight);
			glGetBooleanv(GL_LOGIC_OP, &islop);
			glGetBooleanv(GL_POINT_SMOOTH, &isps);
			glGetBooleanv(GL_TEXTURE_2D, &istex);

			glDisable(GL_ALPHA_TEST);
			glDisable(GL_BLEND);
			glDisable(GL_COLOR_LOGIC_OP);
			glDisable(GL_COLOR_MATERIAL);
			glDisable(GL_DEPTH_TEST);
			glDisable(GL_DITHER);
			glDisable(GL_FOG);
			glDisable(GL_LIGHTING);
			glDisable(GL_LOGIC_OP);
			glDisable(GL_POINT_SMOOTH);
			glDisable(GL_TEXTURE_2D);
			glLogicOp(GL_COPY);

			int mm;
			glGetIntegerv(GL_MATRIX_MODE, &mm);
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			glMatrixMode(GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity();
			Point p=Point(-1,-1,-1);
			glRasterPos3fv((float*)&p);
			glCopyPixels(0,0, screen->winsizex, screen->winsizey, GL_COLOR);
//			glClearDepth(1);
//			glClear(GL_DEPTH_BUFFER_BIT);
			glPopMatrix();
			glMatrixMode(GL_PROJECTION);
			glPopMatrix();
			glMatrixMode(mm);
			screen->getMatrices();

			if (isat)
				glEnable(GL_ALPHA_TEST);
			if (isblend)
				glEnable(GL_BLEND);
			if (isclop)
				glEnable(GL_COLOR_LOGIC_OP);
			if (iscm)
				glEnable(GL_COLOR_MATERIAL);
			if (isdt)
				glEnable(GL_DEPTH_TEST);
			if (isdi)
				glEnable(GL_DITHER);
			if (isfog)
				glEnable(GL_FOG);
			if (islight)
				glEnable(GL_LIGHTING);
			if (islop)
				glEnable(GL_LOGIC_OP);
			if (isps)
				glEnable(GL_POINT_SMOOTH);
			if (istex)
				glEnable(GL_TEXTURE_2D);

			glDrawBuffer (GL_FRONT);
			drawTurtles(true);
			glDrawBuffer (GL_BACK); 
			if (turtle)
				if (!turtle->polyMode && !turtle->tessMode && !turtle->surfaceMode)
				{	glFinish();
//					SwapBuffers();
//					Update();
				}
//#ifdef WIN32
			th->dispatchMessages();
//#endif
		}else
		{
			if (playOnSwap &&	!recorder->inReplay) 
				recorder->playAgain();
			drawTurtles(false);
			if (turtle)
				if (!turtle->polyMode && !turtle->tessMode && !turtle->surfaceMode)
				{	glFinish();
//					SwapBuffers();
//					Update();
				}
			glDrawBuffer(GL_FRONT); 
			glReadBuffer(GL_FRONT); 
			th->dispatchMessages();
			drawTurtles(true);
		}
#ifdef WIN32
//		wglMakeCurrent(oldDC, oldContext);
#else
//		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
		th->GuiLeave();
		inSwapBuffers=false;
	}
#else
	if (!saving)
	{	wxClientDC cdc(this);
		cdc.Blit(0, 0, bitmap->GetWidth(), bitmap->GetHeight(), mdc, 0,0, wxCOPY);
		double sx=1,sy=1;
		cdc.SetAxisOrientation(true, true);
		mdc->GetUserScale(&sx, &sy);
		cdc.SetUserScale(sx, sy);
		int x=0,y=0;
		mdc->GetDeviceOrigin(&x, &y);
		cdc.SetDeviceOrigin(x, y);
		dc=&cdc;
		drawTurtles(true);
		dc=mdc;
	}
#endif
}

void Graph::OnErase(wxEraseEvent &event)
{
//	InvalidateRect(HWND(GetHWND()), 0, FALSE);
//	swapBuffers();
}

void Graph::setMouse(wxMouseEvent &event, Mouse *mouse) 
{
	int x=event.GetX(), y=event.GetY();

	if (event.LeftUp())
		mouse->buttons&=~1;
	if (event.RightUp())
		mouse->buttons&=~2;
	if (event.MiddleUp())
		mouse->buttons&=~4;

	if (event.LeftDown())
		mouse->buttons|=1;
	if (event.RightDown())
		mouse->buttons|=2;
	if (event.MiddleDown())
		mouse->buttons|=4;

	if (event.ShiftDown())
		mouse->buttons|=8;
	else
		mouse->buttons&=~8;

	if (event.ControlDown())
		mouse->buttons|=16;
	else
		mouse->buttons&=~16;

	if (event.AltDown())
		mouse->buttons|=32;
	else
		mouse->buttons&=~32;

	if (event.MetaDown())
		mouse->buttons|=64;
	else
		mouse->buttons&=~64;

	wxSize s=GetClientSize();
	y=s.GetHeight()-y;
	if (screen->perspective)
	{	double p[3]={0,0,0};
#ifdef OPENGL
		gluUnProject(x, y, 0.9965, 
			screen->modelview_matrix, 
			screen->projection_matrix, 
			screen->viewport, 
			&p[0], &p[1], &p[2]);
#endif
		mouse->x=p[0];
		mouse->y=p[1];
		mouse->z=p[2];
	}else
	{//	gluUnProject(x, y, 0, 
	//		screen->modelview_matrix, 
	//		screen->projection_matrix, 
	//		screen->viewport, 
	//		&mouse->x, &mouse->y, &mouse->z);
			mouse->x=(double(x)/s.GetWidth ()-0.5)*(screen->Max.x-screen->Min.x)*screen->scale.x;
			mouse->y=(double(y)/s.GetHeight()-0.5)*(screen->Max.y-screen->Min.y)*screen->scale.y;
			mouse->z=0;
	}
	event.Skip();
}

void Graph::OnEnterWindow(wxMouseEvent &event) 
{
//	SetFocus();
	event.Skip();
}

void Graph::OnLeaveWindow(wxMouseEvent &event) 
{
/*	if (this==theApp->graph && theApp->consoleFrame)
	{	theApp->consoleFrame->SetFocus();
		if (console)
			console->SetFocus();
	}else
*/
	event.Skip();
}

void Graph::OnMouse(wxMouseEvent &event) 
{
	setMouse(event, mouse);
}

#define EVENT_ON_MOUSE(NAME)									\
void Graph::OnMouse##NAME(wxMouseEvent &event)			\
{																		\
	ThreadP th=Thread::This();									\
	setMouse(event, mouse);										\
	if (this==theApp->graph)	\
	{	EventP evt=new EventMouse##NAME(mouse);			\
		if throwing return;										\
		*th->eventFiFo << evt;									\
	}																	\
	EventP evt=new EventGraph##NAME(this, mouse);	\
	if throwing return;											\
	*th->eventFiFo << evt;										\
}

EVENT_ON_MOUSE(LeftDown)
EVENT_ON_MOUSE(RightDown)
EVENT_ON_MOUSE(MiddleDown)

EVENT_ON_MOUSE(LeftUp)
EVENT_ON_MOUSE(RightUp)
EVENT_ON_MOUSE(MiddleUp)

EVENT_ON_MOUSE(LeftDClick)
EVENT_ON_MOUSE(RightDClick)
EVENT_ON_MOUSE(MiddleDClick)

EVENT_ON_MOUSE(Motion)

bool Graph::OnMouseWheel(int nFlags, short zDelta, wxPoint pt) 
{
	mouse->wheel=pt.x;
	mouse->buttons=nFlags;
	return TRUE;
}

void Graph::setFrustum(double w, double h)
{
#ifdef OPENGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
#endif
	if (screen->perspective && !screen->orthographic)
	{
//		double y=screen->Min.y, x=screen->Min.x,
//				 Y=screen->Max.y, X=screen->Max.x;
//		double xm=(X+x)/2, ym=(Y+y)/2, W=X-x, H=Y-y;
/*		if (double(w)/h >= (X-x)/(Y-y))
		{
			glFrustum(((y-ym)*W)/H*screen->scale.x+xm, ((Y-ym)*W)/H*screen->scale.x+xm, 
				(y-ym)*screen->scale.y+ym, (Y-ym)*screen->scale.y+ym, -1, 1);
		}else
		{
			glFrustum((x-xm)*screen->scale.x+xm, (X-xm)*screen->scale.x+xm, 
				((x-xm)*H)/W*screen->scale.y+ym, ((X-xm)*H)/W*screen->scale.y+ym, -1, 1000);
		}
*/
#ifdef OPENGL
		glFrustum(-w/2.0, w/2.0,-h/2.0, h/2.0, -1, 1);
		if (h!=0)
//			gluPerspective(45.0f, (GLdouble)(X-x)/(Y-y), nearPlane, farPlane);
			gluPerspective(45.0f, (GLdouble)w/h, nearPlane, farPlane);
		glTranslated(0,0,-fRadius);
#endif
	}
	else
	{	if (h==0 || w==0)
		{
#ifdef OPENGL
			glOrtho(-1,-1,1,1,-1,1);
#endif
			winorigin=Point(0,0,0);
		}else
		{	double y=screen->Min.y, x=screen->Min.x,
					 Y=screen->Max.y, X=screen->Max.x,
					 xm=(X+x)/2, ym=(Y+y)/2;
			if (double(w)/h >= (X-x)/(Y-y))
			{
#ifdef OPENGL
//				glOrtho(((y-ym)*w)/h+xm, ((Y-ym)*w)/h+xm, y+ym, Y+ym, -1, 1);
//				glOrtho(((y-ym)*w)/h*screen->scale.x+xm, ((Y-ym)*w)/h*screen->scale.x+xm, 
//					y*screen->scale.y+ym, Y*screen->scale.y+ym, -1000, 1000);
				glOrtho(((y-ym)*w)/h*screen->scale.x+xm, ((Y-ym)*w)/h*screen->scale.x+xm, 
					(y-ym)*screen->scale.y+ym, (Y-ym)*screen->scale.y+ym, -1000, 1000);
#else
				dc->DestroyClippingRegion();
				dc->SetDeviceOrigin(w/2, h/2);
				dc->SetUserScale(h/(Y-y)*screen->scale.x, h/(Y-y)*screen->scale.y);
				screen->winrangex = int(((Y-y)*w)/h*screen->scale.x);
				screen->winrangey = int(((Y-y)*w)/h*screen->scale.y);
#endif
			}else
			{
#ifdef OPENGL
//				glOrtho(x+xm, X+xm, ((x-xm)*h)/w+ym, ((X-xm)*h)/w+ym, -1, 1);
				glOrtho((x-xm)*screen->scale.x+xm, (X-xm)*screen->scale.x+xm, 
					((x-xm)*h)/w*screen->scale.y+ym, ((X-xm)*h)/w*screen->scale.y+ym, -1000, 1000);
#else
				dc->DestroyClippingRegion();
				dc->SetDeviceOrigin(w/2, h/2);
				dc->SetUserScale(w/(X-x)*screen->scale.x, w/(X-x)*screen->scale.y);
				screen->winrangex = int(((X-x)*w)/h*screen->scale.x);
				screen->winrangey = int(((X-x)*w)/h*screen->scale.y);
#endif
			}
			screen->getMatrices();
#ifdef OPENGL
			double sx, sy, sz, p[3];
			gluProject(x, y, 0, 
				screen->modelview_matrix, 
				screen->projection_matrix, 
				screen->viewport, 
				&p[0], &p[1], &p[2]);
			winorigin=Point(p[0], p[1], 0);
			winorigin.x=floor(winorigin.x+0.5);
			winorigin.y=floor(winorigin.y+0.5);
			winorigin.z=0;
			gluProject(X, Y, 0, 
				screen->modelview_matrix, 
				screen->projection_matrix, 
				screen->viewport, 
				&sx, &sy, &sz);
			screen->winrangex = int(sx-winorigin.x);
			screen->winrangey = int(sy-winorigin.y);

			gluUnProject(0, 0, 0, 
				screen->modelview_matrix, 
				screen->projection_matrix, 
				screen->viewport, 
				&sx, &sy, &sz);
			screen->edge=Point(sx, sy, sz);
#else
#endif
		}
	}
#ifdef OPENGL
	//glOrtho(-w/2.0, w/2.0,-h/2.0, h/2.0, -1, 1);

	//glFrustum(-w/2.0, w/2.0,-h/2.0, h/2.0, -1, 1);
	//gluPerspective(45.0f, (GLdouble)cx/cy, nearPlane, farPlane);
	//glTranslated(0,0,-fRadius);
	glMatrixMode(GL_MODELVIEW);
#endif
}
void Graph::OnSize(wxSizeEvent &event)
{
	if (!inOnSize && 
		!inread
#ifdef OPENGL
	&&	m_glContext
#endif
	)
	{	inOnSize=true;
		ThreadP th=Thread::This();
		GraphP og=th->graph;
		int oo=og->graphOwned;
		if (this!=og)
		{	og->graphOwned=1;
			og->leave();
			enter();
		}
#ifdef WIN32
//		HDC oldDC=wglGetCurrentDC();
//		HGLRC oldContext=wglGetCurrentContext();
//		wglMakeCurrent((HDC) m_glContext->m_hDC, m_glContext->m_glContext);
#else
//		GLXDrawable oldDrawable=glXGetCurrentDrawable();
//		GLXContext oldContext=glXGetCurrentContext();
//		glXMakeCurrent((Display*) wxGetDisplay(),
//			Window(m_glContext->m_window->GetClientAreaWindow()),
//			m_glContext->m_glContext);
#endif
#ifdef OPENGL
		if (turtle && turtle->polyMode && turtle->polyPointNr > 1)
			glEnd();
#endif
		wxSize s=event.GetSize();
		int x=s.GetWidth(), y=s.GetHeight();
		
		if (screen && y > 0)
		{
#ifdef OPENGL
			glViewport(0, 0, x, y);
#else
			delete bitmap;
			bitmap=new wxBitmap(x,y,32);
			mdc->SelectObject(*bitmap);
			event.Skip();
#endif
			oldRect.SetWidth(x);
			oldRect.SetHeight(y);
			setFrustum(x, y);
			screen->getMatrices();
		}
		if (recorder && screen)
		{	recorder->stop();
			screen->destroy();
			screen->create(x,y);
			recorder->playAgain();
		}
#ifdef OPENGL
#ifdef WIN32
//		wglMakeCurrent(oldDC, oldContext);
#else
//		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
#endif
		if (this!=og)
		{	leave();
			og->enter();
			og->graphOwned=oo;
		}
		inOnSize=false;
	}
	event.Skip();
}

#ifdef WIN32
bool Graph::setDCPixelFormat(HDC hDC, DWORD dwFlags)
{
	PIXELFORMATDESCRIPTOR 
		pixelDesc = 
		{	sizeof(PIXELFORMATDESCRIPTOR), //nSize
			1,	//nVersion
			dwFlags,	//dwFlags
			PFD_TYPE_RGBA,	//iPixelType
			24, //cColorBits
			8, //cRedBits
			0, //cRedShift
			8, //cGreenBits
			8, //cGreenShift
			8, //cBlueBits
			16, //cBlueShift
			8, //cAlphaBits
			24, //cAlphaShift
			0, //cAccumBits
			0, //cAccumRedBits
			0, //cAccumGreenBits
			0, //cAccumBlueBits
			0, //cAccumAlphaBits
			32, //cDepthBits
			0, //cStencilBits

			0, //cAuxBuffers
			PFD_MAIN_PLANE, //iLayerType
			0, //bReserved
			0, //dwLayerMask
			0, //dwVisibleMask
			0, //dwDamageMask
		};
	
	int nPixelIndex = ChoosePixelFormat(hDC, &pixelDesc);
	if (nPixelIndex == 0) // Choose default
	{	nPixelIndex = 1;
		if (DescribePixelFormat(hDC, nPixelIndex, 
			sizeof(PIXELFORMATDESCRIPTOR), &pixelDesc) == 0)
			return false;
	}
	if (!SetPixelFormat(hDC, nPixelIndex, &pixelDesc))
		return false;

	return true;
}
#endif

void Graph::print(wxDC *dc, wxPoint displacement)
{
	printing=true;
#ifdef OPENGL
#ifdef WIN32
	HDC oldDC=wglGetCurrentDC();
	HGLRC oldContext=wglGetCurrentContext();
#else
	GLXDrawable oldDrawable=glXGetCurrentDrawable();
	GLXContext oldContext=glXGetCurrentContext();
#endif

//	int x0,y0;
	int w, h, d=displacement.x;
	dc->GetSize(&w, &h);
	double scaleX=(double)(screen->size.x/w);
	double scaleY=(double)(screen->size.y/h);
	double sc=scaleX/scaleY;
	double x0=(w-w*sc)/2-d;
	w=w*sc;

#ifdef WIN32
	BITMAPINFO bmi;
	memset(&bmi, 0, sizeof(BITMAPINFO));

	BITMAPINFOHEADER bmiHeader=
		{	sizeof(BITMAPINFOHEADER), //biSize
			w, //biWidth
			h, //biHeight
			1, //biPlanes
			24, //biBitCount
			BI_RGB, //biCompression
			w*h*4, //biSizeImage
		};
	bmi.bmiHeader=bmiHeader;

	HDC hDC = ::GetDC((HWND)m_hWnd);
	LPVOID bitmapBits;
	HBITMAP dib = ::CreateDIBSection(
		hDC, &bmi, DIB_RGB_COLORS, &bitmapBits, NULL, (DWORD)0);
	::ReleaseDC((HWND)m_hWnd, hDC);

	HDC mdc=::CreateCompatibleDC(0);
	HBITMAP oldbm = (HBITMAP)SelectObject(mdc, dib);

	setDCPixelFormat(mdc, 
		PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL);

	HGLRC rc=wglCreateContext(mdc), 
		oldrc=wglGetCurrentContext();
	wglMakeCurrent(mdc, rc);
#else
	Display     *dis = (Display*)wxGetDisplay();
	int       scrnum = DefaultScreen(dis);
	GLXContext  fbrc = glXGetCurrentContext();
	GLXDrawable fbdc = glXGetCurrentDrawable();

	if (fbdc == 0)
	{	error(FILE_ERROR, new String(_("unable to get device context in Graph::print")));
		return;
	}
	if (!fbrc)
	{	error(FILE_ERROR, new String(_("unable to get render context in Graph::print")));
		return;
	}
	int attrib[] =
	{	GLX_DOUBLEBUFFER,  False,
		GLX_RED_SIZE,      8,
		GLX_GREEN_SIZE,    8,
		GLX_BLUE_SIZE,     8,
		GLX_ALPHA_SIZE,    8,
//		GLX_STENCIL_SIZE,  8,
		GLX_DEPTH_SIZE,    24,
		GLX_RENDER_TYPE,   GLX_RGBA_BIT,
		GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT | GLX_WINDOW_BIT,
		None
	};

	int PBattrib[] =
	{	GLX_PBUFFER_WIDTH,   w,
		GLX_PBUFFER_HEIGHT,  h,
		GLX_LARGEST_PBUFFER, False,
		None
	};

// choose a pixel format that meets our minimum requirements
	int count = 0;
	GLXFBConfig *config = glXChooseFBConfig(dis, scrnum, attrib, &count);

	if(config == NULL || count == 0)
	{	error(FILE_ERROR, new String(_("no fitting pbuffer pixel format found in Graph::print")));
		return;
	}

	GLXDrawable pbdc = glXCreatePbuffer(dis, config[0], PBattrib);
	GLXContext  pbrc = glXCreateNewContext(dis, config[0], GLX_RGBA_TYPE, fbrc, true);
	XFree(config);
	glXMakeCurrent(dis, pbdc, pbrc);
#endif
	glViewport(0,0, w,h);
	setFrustum(w,h);
	setOpenGLState();
	
	screen->getMatrices();

	wxSize saveWinsize(screen->winsizex, screen->winsizey);
	screen->winsizex=w;
	screen->winsizey=h;

	ftglFontManager->clear();
	PairP oldt=Texture::clone_all();

	saving=true;
	recorder->playAgain();	// here the drawing takes place
	saving=false;

	int dummy;
	glReadPixels(0,0, 0,0, GL_RGB, GL_UNSIGNED_BYTE, &dummy);
#ifdef WIN32
	BOOL error=BitBlt((HDC)dc->GetHDC(), x0,0, w,h, mdc, 0,0, SRCCOPY);	

	wglMakeCurrent(oldDC, oldContext);
	wglDeleteContext(rc);
	DeleteObject(dib);
	DeleteDC(mdc);
#else
	if (w & 3)
		w=(w & ~3)+4;
	ColorP pixels=(ColorP)malloc(w*h*sizeof(ColorP));
	if (pixels)
	{	glReadPixels(0,0, w,h, GL_RGB, GL_UNSIGNED_BYTE, pixels);
		wxImage im(w,h, (unsigned char*)pixels, true);
		im=im.Mirror(false);
		wxBitmap bm=wxBitmap(im, -1);
		dc->DrawBitmap(bm, 0,0);
		GC_::free(pixels); pixels=0;
	}
	glXMakeCurrent(dis, oldDrawable, oldContext);
	glXDestroyPbuffer(dis, pbdc);
	glXDestroyContext(dis, pbrc);
#endif
	screen->winsizex=saveWinsize.GetWidth();
	screen->winsizey=saveWinsize.GetHeight();
	
	ftglFontManager->clear();
	Texture::unclone_all(oldt);
#endif //OPENGL
	printing=false;
}

NodeP Graph::saveScreen(charP filename)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	if (!th->security->filewrite)
	{	error(SECURITY_ERROR, new String(_("to write files"))); return Unbound;
	}
	int 
		x=saveSize.x, xx=(x & ~3)+4,
		y=saveSize.y;
//	if (x & 3)
//		x=(x & ~3)+4;

#ifdef WIN32
	HDC oldDC=wglGetCurrentDC();
	HGLRC oldContext=wglGetCurrentContext();
#else
	GLXDrawable oldDrawable=glXGetCurrentDrawable();
	GLXContext oldContext=glXGetCurrentContext();
#endif
#ifdef WIN32
	HDC olddc=wglGetCurrentDC();

	BITMAPINFO bmi;
	memset(&bmi, 0, sizeof(BITMAPINFO));

	BITMAPINFOHEADER bmiHeader=
		{	sizeof(BITMAPINFOHEADER), //biSize
			x, //biWidth
			y, //biHeight
			1, //biPlanes
			24, //biBitCount
			BI_RGB, //biCompression
			x*y*3, //biSizeImage
		};
	bmi.bmiHeader=bmiHeader;

	HDC hDC = ::GetDC((HWND)m_hWnd);
	void *bitmapbits;
	HBITMAP dib = ::CreateDIBSection(
		hDC, &bmi, DIB_RGB_COLORS, &bitmapbits, NULL, (DWORD)0);
	::ReleaseDC((HWND)m_hWnd, hDC);

	HDC mdc=::CreateCompatibleDC(0);
	HBITMAP oldbm = (HBITMAP)SelectObject(mdc, dib);

	setDCPixelFormat(mdc, 
		PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL);

	HGLRC rc=wglCreateContext(mdc), 
		oldrc=wglGetCurrentContext();
	wglMakeCurrent(mdc, rc);
#else
	Display     *dis = (Display*)wxGetDisplay();
	int       scrnum = DefaultScreen(dis);
	GLXContext  fbrc = glXGetCurrentContext();
	GLXDrawable fbdc = glXGetCurrentDrawable();

	if (fbdc == 0)
	{	error(FILE_ERROR, new String(_("unable to get device context in Graph::print")));
		return Unbound;
	}
	if (!fbrc)
	{	error(FILE_ERROR, new String(_("unable to get render context in Graph::print")));
		return Unbound;
	}
	int attrib[] =
	{	GLX_DOUBLEBUFFER,  False,
		GLX_RED_SIZE,      8,
		GLX_GREEN_SIZE,    8,
		GLX_BLUE_SIZE,     8,
		GLX_ALPHA_SIZE,    8,
//		GLX_STENCIL_SIZE,  8,
		GLX_DEPTH_SIZE,    24,
		GLX_RENDER_TYPE,   GLX_RGBA_BIT,
		GLX_DRAWABLE_TYPE, GLX_PBUFFER_BIT | GLX_WINDOW_BIT,
		None
	};

	int PBattrib[] =
	{	GLX_PBUFFER_WIDTH,   x,
		GLX_PBUFFER_HEIGHT,  y,
		GLX_LARGEST_PBUFFER, False,
		None
	};

// choose a pixel format that meets our minimum requirements
	int count = 0;
	GLXFBConfig *config = glXChooseFBConfig(dis, scrnum, attrib, &count);

	if(config == NULL || count == 0)
	{	error(FILE_ERROR, new String(_("no fitting pbuffer pixel format found in Graph::print")));
		return Unbound;
	}

	GLXDrawable pbdc = glXCreatePbuffer(dis, config[0], PBattrib);
	GLXContext  pbrc = glXCreateNewContext(dis, config[0], GLX_RGBA_TYPE, fbrc, true);
	XFree(config);
	glXMakeCurrent(dis, pbdc, pbrc);
#endif
	glViewport(0, 0, x, y);
	setFrustum(x, y);
	setOpenGLState();
#ifndef WIN32
	glColor4ubv((unsigned char*)&turtle->screen->color);
	glBegin(GL_QUADS);
	{	double x=screen->Max.x, y=screen->Max.y;
		glVertex2d(-x,-y);
		glVertex2d( x,-y);
		glVertex2d( x, y);
		glVertex2d(-x, y);
	}
	glEnd();
	glColor4ubv((unsigned char*)&turtle->pen.color);
#endif
	screen->getMatrices();

	wxSize saveWinsize(screen->winsizex, screen->winsizey);
	screen->winsizex=x;
	screen->winsizey=y;

	ftglFontManager->clear();
	PairP oldt=Texture::clone_all();

	saving=true;
	recorder->playAgain();	// here the drawing takes place
	saving=false;

	unsigned char *pixels=(unsigned char*)GC_::malloc(xx*y*3);
	if (pixels)
	{	glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glReadPixels(0,0, x,y, GL_RGB, GL_UNSIGNED_BYTE, (void*)pixels);
		wxImage im(x,y), im2;
		im.SetData(pixels, true);
		im2=im.Mirror(false);
		if (!im2.SaveFile(filename))
			error(FILE_ERROR, new List(
				new String(_("saveScreen to file")), 
				new String(filename),
				new String(_("failed")), _ListEnd));
		GC_::free(pixels); pixels=0;
	}
#ifdef WIN32
	wglMakeCurrent(olddc, oldrc);
	wglDeleteContext(rc);
	DeleteObject(dib);
	DeleteDC(mdc);	
#else
	glXMakeCurrent(dis, oldDrawable, oldContext);
	glXDestroyPbuffer(dis, pbdc);
	glXDestroyContext(dis, pbrc);
#endif
	screen->winsizex=saveWinsize.GetWidth();
	screen->winsizey=saveWinsize.GetHeight();
	screen->getMatrices();
	ftglFontManager->clear();
	Texture::unclone_all(oldt);
#else
	wxSize saveWinsize(screen->winsizex, screen->winsizey);
	int x=saveSize.x, y=saveSize.y;
	screen->winsizex=x;
	screen->winsizey=y;
	wxMemoryDC sdc(dc);
	wxBitmap bmp(x, y, sdc);
	sdc.SelectObject(bmp);
	wxDC *odc=dc;
	dc=&sdc;
	dc->SetAxisOrientation(true, true);
	setFrustum(x, y);
	saving=true;
	recorder->playAgain();	// here the drawing takes place
	drawTurtles(true);
	saving=false;
	dc=odc;
	wxImage im=bmp.ConvertToImage();
	if (!im.SaveFile(filename))
		error(FILE_ERROR, new List(
			new String(_("saveScreen to file")), 
			new String(filename),
			new String(_("failed")), _ListEnd));
	screen->winsizex=saveWinsize.GetWidth();
	screen->winsizey=saveWinsize.GetHeight();
#endif //OPENGL
	return Unbound;
}

void Graph::saveVector(wxString filename,
							  int sort, int options, int colornr)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	if (!th->security->filewrite)
	{	error(SECURITY_ERROR, new String(_("to write files"))); return;
	}
	int state = GL2PS_OVERFLOW, buffsize = 0;

	int x=saveSize.x,
		 y=saveSize.y;
	GLint viewport[4]={0,0, x,y};

	wxString ext=filename.AfterLast('.').Upper();
	int format=GL2PS_PS;
	if (ext=="EPS") format=GL2PS_EPS;
	if (ext=="PDF") format=GL2PS_PDF;
	if (ext=="TEX") format=GL2PS_TEX;
	if (ext=="SVG") format=GL2PS_SVG;
	if (ext=="PGF") format=GL2PS_PGF;

	FILE *fp=fopen(filename, "wb");
	if(!fp)
	{	error(FILE_ERROR, new String(
			wxString(_("Unable to open file ")
			+filename+_(" for writing")+endL)));
		return;
	}
	while (state == GL2PS_OVERFLOW)
	{	buffsize += 1024*1024;
		gl2psBeginPage(filename, aUCBLogoVersion, viewport, 
							format, sort, options,
							GL_RGBA, 0, NULL, 
							colornr, colornr, colornr, 
							buffsize, fp, filename);
		glViewport(0,0, x,y);
		setFrustum(x,y);
		setOpenGLState();
		screen->getMatrices();

		saving=true;
		recorder->playAgain();	// here the drawing takes place
		saving=false;

		state = gl2psEndPage();
	}
	fclose(fp);
	wxSizeEvent sev(GetSize());
	OnSize(sev);
#endif
}

void Graph::videoStart(charP filename, int rate)
{
	ThreadP th=Thread::This();
	if (!th->security->filewrite || !th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to create videos"))); return;
	}
	if (aviGen)
	{	error(FILE_ERROR, new String(_("Video output stream is already open (forgotten VideoEnd?)")));
		return;
	}
	aviGen=new AVIGen(filename, this, rate);
	int err=aviGen->InitEngine();
	if (err)
	{	error(FILE_ERROR, new String(aviGen->aviError));
		aviGen->ReleaseEngine();
		delete aviGen;
		aviGen=0;
	}
//	error(NOT_YET_SUPPORTED, new String(_("AVI generation not yet supported!")));
}

void Graph::videoFrame()
{
	if (!aviGen)
	{	error(FILE_ERROR, new String(_("Video output stream is not open (use VideoStart!)")));
		return;
	}
	int x=(screen->winsizex/4)*4, y=screen->winsizey;

	charP pixels=charP(GC_::malloc(x*y*3));
	if (pixels)
	{
#ifdef OPENGL
		glReadPixels(0,0, x,y, GL_BGR_EXT, GL_UNSIGNED_BYTE, (void*)pixels);
#endif
		aviGen->AddFrame(pixels);
		GC_::free(pixels); pixels=0;
	}
//	error(NOT_YET_SUPPORTED, new String(_("AVI generation not yet supported!")));
}

void Graph::videoEnd()
{
	if (!aviGen)
	{	error(FILE_ERROR, new String(_("Video output stream is not open (use VideoStart!)")));
		return;
	}
	aviGen->ReleaseEngine();
	delete aviGen;
	aviGen=0;
//	error(NOT_YET_SUPPORTED, new String(_("AVI generation not yet supported!")));
}

void Graph::clear() 
{
#ifdef OPENGL
	glClearColor(
		float(screen->color.r)/255, 
		float(screen->color.g)/255, 
		float(screen->color.b)/255, 
		float(screen->color.a)/255);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
#else
	brush.SetColour(wxColour(
		screen->color.r,
		screen->color.g,
		screen->color.b,
		screen->color.a));
	dc->SetBackground(brush);
	dc->Clear();
#endif
}

void Graph::clearScreen() 
{
	clear();
	if (direct)
	{	bool old=playOnSwap;
		playOnSwap=false;
		swapBuffers();
		playOnSwap=old;
	}
}

void Graph::screen_reset(int centerP)
{
	if (centerP) 
	{	turtle->x=turtle->y=turtle->z=0;
		turtle->outOfBounds=false;
		bool shown=turtle->shown;

		turtle->shown=false;
		turtle->setOrientation(Point(0,0,0));
		turtle->moveTo();
		turtle->shown=shown;
		turtle->reset3d();
	}
//	turtle->draw();
	recorder->clear();

	if (recorder->recording)
	{	*recorder << new R_setScreenColor(screen->color);
		if (turtle->x!=0.0 || turtle->y!=0.0) 
			*recorder << new R_MoveTo(*turtle, *recorder);
		if (!turtle->pen.down) 
			*recorder << new R_setPenDown(*turtle);
		if (!turtle->pen.roundLineEnds) 
			*recorder << new R_setPenRLE(*turtle);
		*recorder << new R_setCyLines(*turtle);
		if (turtle->pen.internalMode!=Pen::Paint) 
			*recorder << new R_setPenMode(*turtle);
		*recorder << new R_setPenColor(*turtle);
		*recorder << new R_setPenSize(*turtle);
		if (screen->perspective)
			*recorder << new R_Perspective();
		if (screen->orthographic)
			*recorder << new R_ortho();
	}
}

void Graph::screen_clear(int centerp) 
{
	clearScreen();
	screen_reset(centerp);
}

void Graph::setColor(Color c) 
{
	screen->color=c;
	clear();
}

void Graph::update()
{
//	Invalidate();
//	wxPaintEvent e;
//	OnPaint(e);
	bool old=playOnSwap;
	playOnSwap=false;
	swapBuffers();
	playOnSwap=old;
}

void Graph::updateGraph()
{
	if (direct)
	{	bool old=playOnSwap;
		playOnSwap=false;
		swapBuffers();
		playOnSwap=old;
	}
}

void Graph::autoload()
{
	ThreadP th=Thread::This();
	th->yield();
	if (autosave)
		if (wxFileExists(pictureFileName))
			recorder->load(pictureFileName);
}

void Graph::save()
{
	if (autosave)
		recorder->save(pictureFileName);
}
//___________________________________________________________________

TurtleNodeP GraphNode::turtleNode()
{
	if (graph)
		return TurtleNodeP(*graph->d(Graph::turtleNode));
	return 0;
}

void GraphNode::setturtleNode(TurtleNodeP tn)
{
	if (graph)
	{	*graph->d(Graph::turtleNode) = tn;
		check_oldyoung(tn);
	}
}

GraphNode::GraphNode(): 
	WindowNode(GRAPH), graph((GraphP&)window), textures(0)
{}

GraphNode::GraphNode(WindowNodeP parent_): 
	WindowNode(GRAPH), graph((GraphP&)window), textures(0)
{
	WindowNodeP p=parent_;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void GraphNode::free() 
{	
	if (graph) 
	{	ThreadP th=Thread::This();
//		GUI_ENTER
		GraphP(graph)->reset();
		*GraphP(graph)->d(Graph::windowNode)=0;
#ifdef OPENGL
#ifdef WIN32		
		HDC oldDC=wglGetCurrentDC();
		HGLRC oldContext=wglGetCurrentContext();
#else
		GLXDrawable oldDrawable=glXGetCurrentDrawable();
		GLXContext oldContext=glXGetCurrentContext();
#endif
#endif
		delete graph;	graph=0;
#ifdef OPENGL
#ifdef WIN32
		wglMakeCurrent(oldDC, oldContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
#endif
		WindowPairP c=children;
		while (c)
		{	if (c->first())
				c->first()->setParent(WindowNodeP(0));
			c=WindowPairP(c->tail());
		}				
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		PairP tl=textures;
		while (tl)
		{	TextureP(tl->first())->_graph=0;
			next(tl);
		}
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::GraphChar:
			case Event::GraphKeyDown:
			case Event::GraphKeyUp:
			case Event::GraphLeftDown:
			case Event::GraphRightDown:
			case Event::GraphMiddleDown:
			case Event::GraphLeftUp:
			case Event::GraphRightUp:
			case Event::GraphMiddleUp:
			case Event::GraphLeftDClick:
			case Event::GraphRightDClick:
			case Event::GraphMiddleDClick:
				if (((EventGraph*)ev)->graph==graph)
					((EventGraph*)ev)->graph=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		graph=0;
//		GUI_LEAVE
	}
}
void GraphNode::inc_gc() 
{
	WindowNode::inc_gc();
	fifo
		<< turtleNode()
		<< textures;
}
int GraphNode::youngs() 
{	
	return WindowNode::youngs()
		+gmark(turtleNode())
		+gmark(textures);
}
int GraphNode::youngsPtr(PtrToInt_ &ptrToInt)
{	return 
		WindowNode::youngsPtr(ptrToInt)
		+gmarkPtr(turtleNode(), ptrToInt)
		+gmarkPtr(textures, ptrToInt);
}
void GraphNode::check_oldyoungs() 
{	WindowNode::check_oldyoungs();
	check_oldyoung(turtleNode());
	check_oldyoung(textures);
}

DataInputStream &GraphNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	parent=theApp->graphNode;
		WindowNodeP parentNode=0;
		istr 
			>> (int&)parentNode 
			>> (int&)children
			>> (int&)textures;
		graph=new Graph(GraphNodeP(this));
		if (!graph)
		{	error(OUT_OF_MEM); return istr;
		}
		graph->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &GraphNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (graph)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent)
			<< GC_::ptrToInt->find(children)
			<< GC_::ptrToInt->find(textures);
		graph->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void GraphNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent   = ( FrameNode*)intToPtr.find(int(parent  ));
	children = (WindowPair*)intToPtr.find(int(children));
	textures = (      Pair*)intToPtr.find(int(textures));
	check_oldyoung(textures);
	if (graph)
	{
		if (!parent && theApp->mainFrame)
		{	//theApp->mainFrame->splitter->ReplaceWindow(
			//	theApp->graph, graph);
			theApp->mainFrame->RemoveChild(theApp->graph);
			theApp->mainFrame->AddChild(graph);
#ifdef OPENGL
#ifdef WIN32
			HDC oldDC=wglGetCurrentDC();
			HGLRC oldContext=wglGetCurrentContext();
#else
			GLXDrawable oldDrawable=glXGetCurrentDrawable();
			GLXContext oldContext=glXGetCurrentContext();
#endif
#endif
			delete theApp->graph;
#ifdef OPENGL
#ifdef WIN32
			wglMakeCurrent(oldDC, oldContext);
#else
			glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
#endif
			theApp->graph=GraphP(graph);
			theApp->graphNode=this;

			wxSizeEvent sev;
			GraphP(graph)->OnSize(sev);
		}

#define SETON(NAME)	\
	*graph->d(Graph:: on##NAME)=((      Pair*)intToPtr.find(int(*graph->d(Graph:: on##NAME))));	\
	*graph->d(Graph::uon##NAME)=((    Object*)intToPtr.find(int(*graph->d(Graph::uon##NAME))));	\
	*graph->d(Graph::ton##NAME)=((LogoObject*)intToPtr.find(int(*graph->d(Graph::ton##NAME))));

		SETON(Char)
		SETON(KeyDown)
		SETON(KeyUp)
		SETON(LeftDown)
		SETON(RightDown)
		SETON(MiddleDown)
		SETON(LeftUp)
		SETON(RightUp)
		SETON(MiddleUp)
		SETON(LeftDClick)
		SETON(RightDClick)
		SETON(MiddleDClick)
		SETON(Motion)
		GraphP(graph)->setPtrs(intToPtr);
	}
}


void GraphNode::aftersetPtrs() 
{
	if (parent)
		return;

	recreate();
}

void GraphNode::recreate()
{
	wxWindow *p=0;
	if (theApp->mainFrame)
		p=(wxWindow*)theApp->mainFrame;//->splitter;
	if (parent)
		p=parent->window;
	GraphP g=GraphP(graph), g2=GraphP(graph);
	if (parent)
	{	g2=new Graph(GraphNodeP(this),
			p, -1, g->GetPosition(), g->GetSize(), 
			g->GetWindowStyle(), g->GetName(), 0, wxNullPalette, false);
		if (!g2)
		{	error(OUT_OF_MEM); return;
		}
		delete g2->screen;
		delete g2->mouse;
		delete g2->recorder;
		delete g2->turtle;

		g2->active=g->active;
		g2->screen=g->screen;	g->screen=0;
		g2->mouse =g->mouse;		g->mouse=0;
		g2->turtle=g->turtle;	g->turtle=0;
		*g2->d(Graph::turtleNode)=*g->d(Graph::turtleNode);	*g->d(Graph::turtleNode)=0;
		g2->recorder=g->recorder; g->recorder=0;
		g2->recorder->graph=g2;
		g2->recorderSave=g2->recorder;

		graph=g2;

#define SET_ON(NAME) \
	*graph->d(Graph:: on##NAME)=*(graph->d(Graph:: on##NAME));	\
	*graph->d(Graph::uon##NAME)=*(graph->d(Graph::uon##NAME));	\
	*graph->d(Graph::ton##NAME)=*(graph->d(Graph::ton##NAME));	

		SET_ON(Char)
		SET_ON(KeyDown)
		SET_ON(KeyUp)

		SET_ON(LeftDown)
		SET_ON(RightDown)
		SET_ON(MiddleDown)
		SET_ON(LeftUp)
		SET_ON(RightUp)
		SET_ON(MiddleUp)
		SET_ON(LeftDClick)
		SET_ON(RightDClick)
		SET_ON(MiddleDClick)
		SET_ON(Motion)

		graph=g;
		delete g;
		graph=g2;
		if (GraphP(graph)->active)
		{	GraphP &gr=Thread::This()->graph;
			gr=GraphP(graph);
#ifdef OPENGL
#ifdef WIN32
			wglMakeCurrent(
				(HDC)gr->GetHDC(),
				gr->GetContext()->GetGLRC());
#else
			glXMakeCurrent((Display*) wxGetDisplay(),
				Window(gr->GetHandle()),
				gr->GetContext()->m_glContext);
#endif
#endif
		}
	}
	PairP c=children;
	while (c)
	{	c->first()->recreate();
		next(c);
	}
#ifdef OPENGL
#ifdef WIN32
	HDC oldDC=wglGetCurrentDC();
	HGLRC oldContext=wglGetCurrentContext();
#else
	GLXDrawable oldDrawable=glXGetCurrentDrawable();
	GLXContext oldContext=glXGetCurrentContext();
#endif
	wxGLContext *glContext=0;
	if (graph)
		glContext=GraphP(graph)->glContext();
	if (glContext)
#ifdef WIN32
		wglMakeCurrent((HDC)GraphP(graph)->GetHDC(), glContext->GetGLRC());
#else
//		glXMakeCurrent((Display*) wxGetDisplay(),
//			Window(glContext->m_window->GetClientAreaWindow()),
//			glContext->m_glContext);
		graph->SetCurrent();
#endif
#endif
	if (textures)
	{	textures = WindowPairP(textures->destructiveRemove(0));
		check_oldyoung(textures);
	}
	PairP t=textures;
	while (t)
	{	TextureP(t->first())->create();
		next(t);
	}
	if (parent)
	{
		GraphP(graph)->setOpenGLState();
		wxSizeEvent sev(graph->GetSize());
		GraphP(graph)->OnSize(sev);
		GraphP(graph)->OnSize(sev);
	}
#ifdef OPENGL
	if (glContext)
#ifdef WIN32
		wglMakeCurrent(oldDC, oldContext);
#else
		glXMakeCurrent((Display*) wxGetDisplay(), oldDrawable, oldContext);
#endif
#endif
}

StringP GraphNode::string()
{
	if (graph)
		return new String(graph->wxstring());
	return _Null_Word->name();
}

wxString GraphNode::wxstring()
{
	if (graph)
		return graph->wxstring();
	return wxEmptyString;
}

void GraphNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

int GraphNode::count()
{
	if (graph && graph->recorder)
		return graph->recorder->items;
	return 0;
}

NodeP lGraph(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	long style=wxFULL_REPAINT_ON_RESIZE;
	if (exist(2) && arg[2])
	{	int s=arg[2]->a_int(2);
		if throwing return Unbound;
		style=s;	
	}

	wxPoint pos=wxDefaultPosition;
	if (exist(3) && arg[3])
	{	PairP p=arg[3]->a_int_vector(3);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(4) && arg[4])
	{	PairP s=arg[4]->a_int_vector(4);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=_("Graph");
	if (exist(5))
	{	StringP n=arg[5]->a_string(5);
		if throwing return Unbound;
		name=n->ptr();
	}
//	GUI_ENTER
	GraphNodeP graphNode=new GraphNode(parentNode);
	if throwing return Unbound;
	GraphP graph=new Graph(graphNode,
			parent, id, pos, size, style, name, 0, wxNullPalette, false);
	if (!graph)
	{	error(OUT_OF_MEM); return Unbound;
	}
	graphNode->graph=graph;
	th->graph=graph;
#ifdef OPENGL
	if (graph && graph->GetContext())
	{
#ifdef WIN32
		wglMakeCurrent(
			(HDC)graph->GetHDC(),
			     graph->GetContext()->GetGLRC());
#else
//		glXMakeCurrent((Display*) wxGetDisplay(),
//			Window(graph->m_Window->GetClientAreaWindow()),
//			graph->GetContext()->m_glContext);
		th->dispatchMessages();
		graph->SetCurrent();
#endif
	}
#endif
//	graph->turtleNode=new TurtleNode(graph->turtle, graphNode);
//	if throwing return Unbound;
//	wxSizeEvent sev(parent->GetClientSize());
//	graph->OnSize(sev);
//	GUI_LEAVE
	return graphNode;
}

NodeP lGraphDestroy(PairPP arg)
{
	GraphNodeP graphNode=arg[1]->a_graph(1);
	if throwing return Unbound;
	if (graphNode != theApp->graphNode)
	{
//		GUI_ENTER
		WindowPairP c=graphNode->children;
		while (c)
		{	if (c->first())
				c->first()->setParent(WindowNodeP(0));
			c=WindowPairP(c->tail());
		}				
		WindowNodeP p=graphNode->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(graphNode));
			p->check_oldyoung(p->children);
		}
		graphNode->graph->Destroy();
		graphNode->graph=0;
//		GUI_LEAVE
	}
	return Unbound;
}

NodeP lGraphCurrent(PairPP arg)
{
	if (Thread::This()->graph)
		return *Thread::This()->graph->d(Graph::windowNode);
	else
		return 0;
}

NodeP lGraphSetCurrent(PairPP arg)
{
	GraphP &gr=Thread::This()->graph;
	if (gr)
	{	gr->graphOwned=1;
		gr->leave();
	}
	if (!arg[1] && theApp->mainFrame)
		gr=theApp->graph;
	else
	{	GraphNodeP graphNode=arg[1]->a_graph(1);
		if throwing return Unbound;
		gr=GraphP(graphNode->graph);
	}
	if (gr)
		gr->enter();
	Float::reset_FPU();
	return Unbound;
}

NodeP lGraphSetFocus(PairPP arg)
{
	GraphP &gr=Thread::This()->graph;
	if (!arg[1] && theApp->mainFrame)
		gr=theApp->graph;
	else
	{	GraphNodeP graphNode=arg[1]->a_graph(1);
		if throwing return Unbound;
		gr=GraphP(graphNode->graph);
	}
	gr->SetFocus();
	return Unbound;
}

#define _GRAPH_ON(NAME)								\
NodeP lGraphOn##NAME(PairPP arg)					\
{															\
	GraphNodeP graph=arg[1]->a_graph(1);		\
	if throwing return Unbound;					\
	PairP commandlist=arg[2]->a_runnable(2);	\
	if throwing return Unbound;					\
	graph->seton##NAME(commandlist);				\
	return Unbound;									\
}

_GRAPH_ON(Char)
_GRAPH_ON(KeyDown)
_GRAPH_ON(KeyUp)
_GRAPH_ON(LeftDown)
_GRAPH_ON(RightDown)
_GRAPH_ON(MiddleDown)

_GRAPH_ON(LeftUp)
_GRAPH_ON(RightUp)
_GRAPH_ON(MiddleUp)

_GRAPH_ON(LeftDClick)
_GRAPH_ON(RightDClick)
_GRAPH_ON(MiddleDClick)

_GRAPH_ON(Motion)

NodeP lsetPlayOnSwap(PairPP arg)
{
	GraphP graph=Thread::This()->graph;
	int playOnSwap=arg[1]->a_bool(1);
	if throwing return Unbound;
	theApp->playOnSwap = (playOnSwap==1);
	if (graph)
		graph->playOnSwap = (playOnSwap==1);
	return Unbound;
}

NodeP lMousePos(PairPP) 
{
	ThreadP th=Thread::This();
	th->dispatchMessages();
	GraphP graph=Thread::This()->graph;
	if (graph)
		return 
			new Pair(new Float(graph->mouse->getX()), 
			new Pair(new Float(graph->mouse->getY()), 
			new Pair(new Float(graph->mouse->getZ()), 0)));
	else
		return 
			new Pair(new Float(0), 
			new Pair(new Float(0), 
			new Pair(new Float(0), 0)));
}

NodeP lMouseX(PairPP)
{	GraphP graph=Thread::This()->graph;
	if (graph)
		return new Float(graph->mouse->getX());
	else
		return new Float(0);
}
NodeP lMouseY(PairPP)
{	GraphP graph=Thread::This()->graph;
	if (graph)
		return new Float(graph->mouse->getY());
	else
		return new Float(0);
}
NodeP lMouseZ(PairPP)
{	GraphP graph=Thread::This()->graph;
	if (graph)
		return new Float(graph->mouse->getZ());
	else
		return new Float(0);
}

NodeP lMouseButtons(PairPP)
{	ThreadP th=Thread::This();
	GraphP graph=th->graph;
	th->dispatchMessages();
	if (graph)
		return new Int(graph->mouse->button());
	else
		return new Int(0);
}

