#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/textctrl.h>
#include <wx/listctrl.h>
#include <wx/frame.h>
#include <wx/splitter.h>
#include <wx/dialog.h>
#include <wx/html/helpctrl.h>
#include <wx/image.h>
#include <wx/datstrm.h>
#include <wx/treectrl.h>

#include <math.h>
#include <stdio.h>

#ifdef OPENGL
	#include <GL/glu.h>
	#include <gl2ps.h>
#endif

#include "Main.h"
#include "File.h"
#include "PtrToInt_.h"
#include "IntToPtr_.h"
#include "Pair.h"
#include "Line.h"
#include "Object.h"
#include "List.h"
#include "Int.h"
#include "Int16.h"
#include "UInt8.h"
#include "Float.h"
#include "Array.h"
#include "IntArray.h"
#include "FloatArray.h"
#include "String.h"
#include "TurtleNode.h"
#include "GraphicNode.h"
#include "Bitmap.h"
#include "Texture.h"

#include "Error.h"
#include "NodePStack.h"
#include "Eval.h"
#include "Init.h"

#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Screen.h"
#include "Mouse.h"
#include "Font.h"
#include "Turtle.h"

#include "Recorder.h"

#include "Graph.h"
#include "Vars.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Shadow.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

//___________________________________________________________________

TurtleNode::TurtleNode(Turtle *turtle_, GraphNodeP graphNode_): 
	Node(TURTLE_), turtle(turtle_), graphNode(graphNode_), shapes(0)
{
	check_oldyoung(graphNode);
	if (turtle)
		turtle->turtleNode=this;
//	turtle->refCount++;
//	graphNode=theApp->mainFrame->graphNode;
}

void TurtleNode::free()
{
	if (!turtle)
	{	graphNode=0;
		return;
	}
/*	if (turtle->refCount > 0)
		--turtle->refCount;

	if (turtle == TURTLE)
		return;

	if (turtle->refCount==0)
	{
*/		turtle->destroy();
		if (turtle)
		{	delete turtle;	turtle=0;
			graphNode=0;
		}
//	}
}

void TurtleNode::inc_gc() 
{	fifo << shapes;
}
int TurtleNode::youngs() 
{	return gmark(shapes);
}
int TurtleNode::youngsPtr(PtrToInt_ &ptrToInt)
{	return gmarkPtr(shapes, ptrToInt);
}
void TurtleNode::check_oldyoungs() 
{	check_oldyoung(shapes);
}

DataInputStream &TurtleNode::read(DataInputStream &istr) 
{
	turtle=new Turtle;
	if (!turtle)
	{	error(OUT_OF_MEM); return istr;
	}
	turtle->read(istr);
	istr >> (int&)graphNode >> (int&)shapes;
	return istr;
}

wxDataOutputStream &TurtleNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	turtle->write(ostr);
	ostr << GC_::ptrToInt->find(graphNode);
	ostr << GC_::ptrToInt->find(shapes);
	return ostr;
}

void TurtleNode::setPtrs(IntToPtr_ &intToPtr)
{
	graphNode=(GraphNode*)intToPtr.find(int(graphNode));
	if (graphNode && graphNode->window)
	{	turtle->screen=GraphP(graphNode->window)->screen;
		turtle->recorder=GraphP(graphNode->window)->recorder;
	}else if (Thread::This()->graph)
	{	turtle->screen=Thread::This()->graph->screen;
		turtle->recorder=Thread::This()->graph->recorder;
	}else
	{	turtle->screen=0;
		turtle->recorder=0;
	}
	shapes=(Array*)intToPtr.find(int(shapes));
}

NodeP TurtleNode::copy(PtrToInt_ &ptrToInt)
{
	TurtleNodeP tn=new TurtleNode();
	if throwing return Unbound;
	tn->turtle=new Turtle(*turtle);
	tn->graphNode=GraphNodeP(ptrToInt.find(graphNode));
	tn->shapes=ArrayP(ptrToInt.find(shapes));
	return tn;
}

StringP TurtleNode::string()
{
	char format[32], s[256];
	int prec=Thread::This()->precision;
	sprintf(format, _("Turtle {%%-.%ig %%-.%ig %%-.%ig}"), 
		prec, prec, prec);

	sprintf(s, format, turtle->x, turtle->y, turtle->z);
	return new String(s, 0, strlen(s), STRING, String::copy);
}

void TurtleNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, _("Turtle {")); 
	::print(strm, width, turtle->x);	::print(strm, width, ' ');
	::print(strm, width, turtle->y);	::print(strm, width, ' ');
	::print(strm, width, turtle->z);	::print(strm, width, '}');
}
//___________________________________________________________________

GraphicNode::GraphicNode(Recorder *recorder_): Node(GRAPHIC), scale(1)
{
	recorder=recorder_;
}

void GraphicNode::free()
{
	if (recorder && !--recorder->refCount)
	{	recorder->free();
		recorder=0;
	}
}

void GraphicNode::setPtrs(IntToPtr_ &intToPtr)
{	
	CommandListIterator i(recorder);
	for (i=recorder->begin();  i;  i++)
		i->setPtrs(intToPtr);
}
DataInputStream &GraphicNode::read(DataInputStream &istr) 
{
	int nr;
	istr >> scale >> nr;
	recorder = (Recorder*)Recorder::intToPtr->find(nr);
	if (!recorder)
	{	recorder=new Recorder(Thread::This()->graph);
		if (!recorder)
		{	error(OUT_OF_MEM); return istr;
		}
		Recorder::intToPtr->insert(recorder);
		istr >> *recorder;
	}
	else
		recorder->refCount++;
	return istr;
}
wxDataOutputStream &GraphicNode::write(wxDataOutputStream &ostr) 
{
	ostr << type << scale;
	int nr;
	nr=Recorder::ptrToInt->find(recorder);
	if (!nr)
	{	Recorder::ptrToInt->insert(recorder);
		ostr << int(0) << *recorder;
	}
	else
		ostr << nr;
	return ostr;
}

StringP GraphicNode::string()
{
	return new String(_("Graphic"));
}

void GraphicNode::draw(Turtle *turtle)
{
	ThreadP th=Thread::This();
	bool rr=RECORDER.recording;
	RECORDER.recording=false;

//	TURTLE->draw();
	Turtle save(false);
	save=*turtle;
#ifdef OPENGL
	glPushMatrix();
#endif
	turtle->setTurtleMatrix();
#ifdef OPENGL
	GLboolean normalized;
	if (scale!=1)
	{	glScaled(scale, scale, scale);
		normalized=glIsEnabled(GL_NORMALIZE);
		glEnable(GL_NORMALIZE);
	}
#endif
	*turtle=Point(0,0,0);
	turtle->reset3d();
	turtle->heading=0;

	Turtle *saveRecT=recorder->turtle;
	recorder->turtle=turtle;
	recorder->screen=th->graph->screen;
	CommandListIterator i(recorder);
	i=recorder->begin();
	while (i)
	{	i->play(*recorder);
		++i;
	}
	recorder->turtle=saveRecT;
#ifdef OPENGL
	if (!normalized)
		glDisable(GL_NORMALIZE);
	glPopMatrix();
#endif
	*turtle=save;
#ifdef OPENGL
	glColor3ubv((GLubyte*)&save.pen.color);
	glTexCoord2f(save.texpos.x, save.texpos.y);
#endif
	RECORDER.recording=rr;
}
//___________________________________________________________________

NodeP lBitCopy(PairPP arg)
{
	ThreadP th=Thread::This();
	double 
		w=arg[1]->a_float(1),
		h=arg[2]->a_float(2);
	if throwing return Unbound;
	GraphLocker gl;
	return NodeP(TURTLE->bitCopy(w,h));
}

NodeP lBitPaste(PairPP arg)
{
	ThreadP th=Thread::This();
	BitmapP b=arg[1]->a_bitmap(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->bitPaste(b);
	if (RECORDER.recording)
		RECORDER << new R_BitPaste(b);
	return Unbound;
}

NodeP lBitPasteFast(PairPP arg)
{
	enum Params {_bitmap=1};
	ThreadP th=Thread::This();
	BitmapP bitmap=arg[_bitmap]->a_bitmap(_bitmap);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->bitPasteFast(bitmap);
//	if (RECORDER.recording)
//		RECORDER << new R_BitPaste(b);
	return Unbound;
}

NodeP lBitFill(PairPP arg)
{
	enum Params {_bitmap=1, _x, _y};
	ThreadP th=Thread::This();
	BitmapP bitmap=arg[_bitmap]->a_bitmap(_bitmap);
	int x=int(arg[_x]->a_float(_x)),
		 y=int(arg[_y]->a_float(_y));
	if throwing return Unbound;
	GraphLocker gl;
	Color fillc=TURTLE->pen.floodcolor, 
		rim=TURTLE->pen.color;
	if throwing return Unbound;
	bitmap->fill(x, y, fillc, rim);
	return Unbound;
}


NodeP lModelviewMatrix(PairPP)
{
	ThreadP th=Thread::This();
//	return SCREEN->matrix();
	GraphLocker gl;
	double *m;
	m=SCREEN->modelview_matrix;
	return new List
	(	new List(new Float(m[ 0]),new Float(m[ 1]),new Float(m[ 2]), _ListEnd),
		new List(new Float(m[ 4]),new Float(m[ 5]),new Float(m[ 6]), _ListEnd),
		new List(new Float(m[ 8]),new Float(m[ 9]),new Float(m[10]), _ListEnd),
		new List(new Float(m[12]),new Float(m[13]),new Float(m[14]), _ListEnd), _ListEnd);
}

NodeP lIdentityMatrix(PairPP)
{
	FloatP O=new Float(0), l=new Float(1);
	return new List
	(	new List( l,O,O, _ListEnd),
		new List( O,l,O, _ListEnd), 
		new List( O,O,l, _ListEnd), 
		new List( O,O,O, _ListEnd), _ListEnd);
}

NodeP lsetIdentityMatrix(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glLoadIdentity();
	glGetDoublev(GL_MODELVIEW_MATRIX, SCREEN->modelview_matrix);
#endif
	return Unbound;
}

NodeP lTurtleMatrix(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return TURTLE->matrix();
}
NodeP lsetTurtleMatrix(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
//	TURTLE->draw();
	TURTLE->setTurtleMatrix(); 
//	TURTLE->draw();
	return Unbound;
}

NodeP lsetModelviewMatrix(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP m=arg[1]->a_list(1);
	if throwing return Unbound;
	GraphLocker gl;
	if (m && m->first() 
	&& m->tail() && m->second() 
	&& m->tail()->tail() && m->third() 
	&& m->tail()->tail()->tail() && m->fourth()) 
	{	PairP 
			 basex=m-> first()->a_3dvector(1),
			 basey=m->second()->a_3dvector(1),
			 basez=m-> third()->a_3dvector(1),
			center=m->fourth()->a_3dvector(1);

		if (!throwing)
			TURTLE->setMatrix(basex,basey,basez,center);
	}
	else error(BAD_DATA, m);
	return Unbound;
}

NodeP lpushMatrix(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glPushMatrix(); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_pushMatrix();
	return Unbound;
}

NodeP lpopMatrix(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glPopMatrix(); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_popMatrix();
	return Unbound;
}

NodeP lsetScreenRange(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	GraphP graph=th->graph;
	if (!graph)
		return Unbound;
	double minx=arg[1]->a_float(1);
	if throwing return Unbound;
	double miny=arg[2]->a_float(2);
	if throwing return Unbound;
	double maxx=arg[3]->a_float(3);
	if throwing return Unbound;
	double maxy=arg[4]->a_float(4);
	if throwing return Unbound;
	graph->screen->Min.x=minx;
	graph->screen->Min.y=miny;
	graph->screen->Max.x=maxx;
	graph->screen->Max.y=maxy;

	graph->screen->size.x=maxx-minx;
	graph->screen->size.y=maxy-miny;

	int w,h;
	graph->GetClientSize(&w, &h);
#ifdef OPENGL
	glViewport(0, 0, w, h);
#endif
	graph->setFrustum(w, h);
	graph->screen->getMatrices();
	return Unbound;
}

NodeP lScreenRange(PairPP)
{
	GraphLocker gl;
	GraphP graph=Thread::This()->graph;
	if (graph)
		return new List(
			new Float(graph->screen->Min.x), 
			new Float(graph->screen->Min.y), 
			new Float(graph->screen->Max.x), 
			new Float(graph->screen->Max.y), _ListEnd);
	else
		return new List(
			new Float(0), 
			new Float(0), 
			new Float(0), 
			new Float(0), _ListEnd);
}

void Node::setPixel(NodeP color) {error(BAD_DATA, this);}
void Pair::setPixel(NodeP color)
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return;
	if (first()->islist() || first()->type==FLOATARRAY)
	{	PairP a=this;
		if (color->islist())
		{	PairP clist=PairP(color);
			while (a && clist)
			{	if (a->circcounter()==th->circcounter)
					break;
				else
					a->setcirccounter(th->circcounter);
				a->first()->setPixel( clist->first() );
				next(a);  
				next(clist);
			}
		}else
		{	while (a)
			{	if (a->circcounter()==th->circcounter)
					break;
				else
					a->setcirccounter(th->circcounter);
				a->first()->setPixel(color);
				next(a);
			}
		}
		th->setcirccounter(th->circcounter+1);
		if (!throwing) return;
	}else
	{	PairP vec=this;//a_vector();
		Color c=PairP(color)->a_color(2);
		if (!throwing)
		{	double x,y,z;
			if (vec->first()->numberP())
			{	NodeP X=vec->first();
				x = X->getnumber();
				if (vec->tail())
				{	NodeP Y=vec->second();
					y = Y->getnumber();
				}
				z=0;
				if (vec->tail()->tail())
					z=vec->third()->getnumber();
				TURTLE->setPixel(Point(x,y,z), c);
			}else if (vec->first()->type==FLOATARRAY)
			{	while (vec && vec->first()->type==FLOATARRAY)
				{	if (vec->circcounter()==th->circcounter)
						break;
					else
						vec->setcirccounter(th->circcounter);
					FloatArrayP fa=FloatArrayP(vec->first());
					doubleP d=fa->data();
					if (fa->length()>0)
						x=d[0];
					if (fa->length()>1)
						y=d[1];
					if (fa->length()>2)
						z=d[2];
					TURTLE->setPixel(Point(x,y,z), c);
					next(vec);
				}
			}
			th->setcirccounter(th->circcounter+1);
			return;
		}
	}
	error(BAD_DATA, this);
}
void FloatArray::setPixel(NodeP color)
{
	ThreadP th=Thread::This();
	Color c=PairP(color)->a_color(2);
	if throwing return;
	double x=0,y=0,z=0;
	doubleP d=data();
	if (length()>0)
		x=d[0];
	if (length()>1)
		y=d[1];
	if (length()>2)
		z=d[2];
	TURTLE->setPixel(Point(x,y,z), c);
}
NodeP lsetPixel(PairPP arg)
{
	if (arg[1]) 
	{	GraphLocker gl;
//		TURTLE->draw();
		arg[1]->setPixel(arg[2]);
//		TURTLE->draw();
	}
	return Unbound;
}

void Node::setPixelXY(NodeP yin, NodeP color) {error(BAD_DATA, this);}
void Int::setPixelXY(NodeP yin, NodeP color)
{
	ThreadP th=Thread::This();
	Color c=color->a_color(2);
	double
		x=int_,
		y=yin->getnumber(),
		z=0;
	TURTLE->setPixel(Point(x,y,z), c);
}
void Float::setPixelXY(NodeP yin, NodeP color)
{
	ThreadP th=Thread::This();
	Color c=color->a_color(2);
	double
		x=float_,
		y=yin->getnumber(),
		z=0;
	TURTLE->setPixel(Point(x,y,z), c);
}
void Pair::setPixelXY(NodeP yin, NodeP color)
{
	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return;
	if (!yin || !yin->is.list)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	PairP x=this, y=PairP(yin);
	if (color->islist())
	{	PairP c=PairP(color);
		while (x && y && c)
		{	if (x->circcounter()==th->circcounter)
				break;
			else
				x->setcirccounter(th->circcounter);
			x->first()->setPixelXY( y->first(), c->first() );
			next(x);
			next(y);
			next(c);
		}
	}else
	{	while (x && y)
		{	if (x->circcounter()==th->circcounter)
				break;
			else
				x->setcirccounter(th->circcounter);
			x->first()->setPixelXY( y->first(), color);
			next(x);
			next(y);
		}
	}
	th->setcirccounter(th->circcounter+1);
}
void FloatArray::setPixelXY(NodeP yin, NodeP color)
{	ThreadP th=Thread::This();
	if (type_of(yin) != FLOATARRAY)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	doubleP x=data();
	int i=length();
	FloatArrayP ya=FloatArrayP(yin);
	if (ya->length() != i)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	doubleP y=ya->data();
	if (color->type==INTARRAY  &&  IntArrayP(color)->length()==i)
	{	ColorP c=ColorP(IntArrayP(color)->data());
		while (i--)
		{	int ic=Pen::realcolor(*c++);
			TURTLE->setPixel(Point(*x++,*y++,0), ic);
		}
	}else
	{	Color c=color->a_color(3);
		while (i--)
		{	TURTLE->setPixel(Point(*x++,*y++,0), c);
		}
	}
}
NodeP lsetPixelXY(PairPP arg)
{
	if (arg[1]) 
	{	GraphLocker gl;
//		TURTLE->draw();
		arg[1]->setPixelXY(arg[2], arg[3]);
//		TURTLE->draw();
	}
	return Unbound;
}

void Node::setPixelXYZ(NodeP yin, NodeP zin, NodeP color) {error(BAD_DATA, this);}
void Int::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{
	ThreadP th=Thread::This();
	Color c=color->a_color(4);
	double
		x=int_,
		y=yin->getnumber(),
		z=zin->getnumber();
	TURTLE->setPixel(Point(x,y,z), c);
}
void Float::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{
	ThreadP th=Thread::This();
	Color c=color->a_color(4);
	double
		x=float_,
		y=yin->getnumber(),
		z=zin->getnumber();
	TURTLE->setPixel(Point(x,y,z), c);
}
void Pair::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{
	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return;
	if (!yin || !yin->is.list)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	if (!zin || !zin->is.list)
	{	error(BAD_DATA, zin, 3);
		return;
	}
	PairP x=this, y=PairP(yin), z=PairP(zin);
	if (color->islist())
	{	PairP c=PairP(color);
		while (x && y && z && c)
		{	if (x->circcounter()==th->circcounter)
				break;
			else
				x->setcirccounter(th->circcounter);
			x->first()->setPixelXYZ( y->first(), z->first(), c->first() );
			next(x);
			next(y);
			next(z);
			next(c);
		}
	}else
	{	while (x && y && z)
		{	if (x->circcounter()==th->circcounter)
				break;
			else
				x->setcirccounter(th->circcounter);
			x->first()->setPixelXYZ( y->first(), z->first(), color);
			next(x);
			next(y);
			next(z);
		}
	}
	th->setcirccounter(th->circcounter+1);
}
void FloatArray::setPixelXYZ(NodeP yin, NodeP zin, NodeP color)
{	ThreadP th=Thread::This();
	if (type_of(yin) != FLOATARRAY)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	doubleP x=data();
	int i=length();
	FloatArrayP ya=FloatArrayP(yin);
	if (ya->length() != i)
	{	error(BAD_DATA, yin, 2);
		return;
	}
	if (type_of(zin) != FLOATARRAY)
	{	error(BAD_DATA, zin, 3);
		return;
	}
	FloatArrayP za=FloatArrayP(zin);
	if (za->length() != i)
	{	error(BAD_DATA, zin, 3);
		return;
	}
	doubleP 
		y=ya->data(),
		z=za->data();
	if (color->type==INTARRAY  &&  IntArrayP(color)->length()==i)
	{	ColorP c=ColorP(IntArrayP(color)->data());
		while (i--)
		{	int ic= Pen::realcolor(*c++);
			TURTLE->setPixel(Point(*x++, *y++, *z++), ic);
		}
	}else
	{	Color c=color->a_color(4);
		while (i--)
		{	TURTLE->setPixel(Point(*x++, *y++, *z++), c);
		}
	}
}
NodeP lsetPixelXYZ(PairPP arg)
{
	if (arg[1]) 
	{	GraphLocker gl;
//		TURTLE->draw();
		arg[1]->setPixelXYZ(arg[2], arg[3], arg[4]);
//		TURTLE->draw();
	}
	return Unbound;
}

NodeP lPixel(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
//	TURTLE->draw();
	NodeP p=new Int(TURTLE->Pixel());
//	TURTLE->draw();
	return p;
}

void line_draw(NodeP a, NodeP color)
{	ThreadP th=Thread::This();
	PairP vec= PairP(a)->a_vector(1);
	IntP C=PairP(color)->a_integer(2);
	if (!throwing)
	{	NodeP X=vec->first(), Y=vec->second();
		double
			x=X->getnumber(),
			y=Y->getnumber();
		int c= C->int_;
		if (c ==-1)
			TURTLE->moveTo(x,y);
		else
		{	if (TURTLE->pen.rawcolor !=c)
				TURTLE->setColor(c);
			if (RECORDER.recording)
				RECORDER << new R_setPenColor(*TURTLE);
			TURTLE->lineTo(x,y);
			*TURTLE=Point(x,y,0);
		}

	}
}
void move_draw(NodeP a, NodeP color)
{	ThreadP th=Thread::This();
	PairP vec= PairP(a)->a_vector(1);
	IntP C=PairP(color)->a_integer(2);
	if (!throwing)
	{	NodeP X=vec->first(), Y=vec->second();
		double
			x=X->getnumber(),
			y=Y->getnumber();
		int c= C->int_;
		if (c ==-1)
			TURTLE->moveTo(x,y);
		else
		{	if (TURTLE->pen.rawcolor !=c)
				TURTLE->setColor(c);
			if (RECORDER.recording)
				RECORDER << new R_setPenColor(*TURTLE);
			TURTLE->moveTo(x,y);
		}
	}
}
void Node::line(NodeP color) {error(BAD_DATA, this);}
void Pair::line(NodeP color)
{
	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return;
	if (first()->islist())
	{	PairP a=this;
		if (a && a->first() && a->first()->first()->islist())
			if (color->islist())
			{	PairP clist=PairP(color);
				while (a && a->first() && clist) 
				{	if (a->circcounter()==th->circcounter)
						break;
					else
						a->setcirccounter(th->circcounter);
					a->first()->line(clist->first());
					next(a);
					next(clist);
				}
				th->setcirccounter(th->circcounter+1);
				if (!a && !clist && !throwing) return;
			}else
			{	while (a && a->first()) 
				{	if (a->circcounter()==th->circcounter)
						break;
					else
						a->setcirccounter(th->circcounter);
					a->first()->line(color);
					next(a);
				}
				th->setcirccounter(th->circcounter+1);
				if (!a && !throwing) return;
			}
		else
			if (color->islist())
			{	PairP clist=PairP(color);
				if (a && clist) 
					move_draw(a->first(), clist->first());
				while (a && clist)
				{	if (a->circcounter()==th->circcounter)
						break;
					else
						a->setcirccounter(th->circcounter);
					line_draw(a->first(), clist->first());
					next(a);  
					next(clist);
				}
				th->setcirccounter(th->circcounter+1);
				if (!a && !clist && !throwing) return;
			}else
			{	if (a) move_draw(a->first(), color);
				while (a)
				{	if (a->circcounter()==th->circcounter)
						break;
					else
						a->setcirccounter(th->circcounter);
					line_draw(a->first(), color);
					next(a);
				}
				th->setcirccounter(th->circcounter+1);
				if (!a && !throwing) return;
			}
	}else if (first()->type==ARRAY)
	{	PairP a=this;
		while (a)
		{	if (a->circcounter()==th->circcounter)
				break;
			else
				a->setcirccounter(th->circcounter);
			if (a->first() && a->first()->type==ARRAY)
				a->first()->line(color);
			else break;
			next(a);
		}
		th->setcirccounter(th->circcounter+1);
		if (!a && !throwing) return;
	}else
	{	move_draw(this, color);
		th->setcirccounter(th->circcounter+1);
		if (!throwing) return;
	}
	error(BAD_DATA, this);
}
void Array::line(NodeP color)
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return;
	else
		setcirccounter(th->circcounter);
	NodePP d=data();
	int i=length();
	if (color->type==ARRAY  &&  ArrayP(color)->length()==i)
	{	NodePP c=ArrayP(color)->data();
		if (*d) move_draw(*d, *c);
		while (i-- && *d)
		{	line_draw(*d++, *c++);
		}
		th->setcirccounter(th->circcounter+1);
		if (i==-1 && !throwing) return; 
	}else 
	{	if (*d) move_draw(*d, color);
		while (i-- && *d)
			line_draw(*d++, color);
		th->setcirccounter(th->circcounter+1);
		if (i==-1 && !throwing) return;
	}
	error(BAD_DATA, this);
}
void FloatArray::line(NodeP color)
{	error(BAD_DATA, this);
}
NodeP lLine(PairPP arg)
{
	ThreadP th=Thread::This();
	if (arg[1]) 
	{	GraphLocker gl;
//		TURTLE->draw();	
		Point t=*TURTLE;
		arg[1]->line(arg[2]);
		*TURTLE=t;
//		TURTLE->draw();
	}
	return Unbound;
}

NodeP lright(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->right(a);
	return Unbound;
}

NodeP lleft(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->right(-a);
	return Unbound;
}

NodeP lupPitch(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->upPitch(a);
	return Unbound;
}

NodeP ldownPitch(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->upPitch(-a);
	return Unbound;
}

NodeP lrightRoll(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->rightRoll(a);
	return Unbound;
}

NodeP lleftRoll(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->rightRoll(-a);
	return Unbound;
}

NodeP lspinX(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->spinX(a);
	return Unbound;
}

NodeP lspinY(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->spinY(a);
	return Unbound;
}

NodeP lspinZ(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	double a=val->getnumber();
	TURTLE->spinZ(a);
	return Unbound;
}

NodeP lforward(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	double d=val->getnumber();
	GraphLocker gl;
	TURTLE->forward(d);
	return Unbound;
}

NodeP lback(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP val = arg[1]->a_number(1);
	if throwing return Unbound;
	double d=val->getnumber();
	GraphLocker gl;
	TURTLE->forward(-d);
	return Unbound;
}

NodeP lshowTurtle(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->show(); return Unbound;
}
NodeP lhideTurtle(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->hide();  return Unbound;
}
NodeP lisshown(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return (TURTLE->shown ?  _True  :  _False);
}

NodeP lHeading(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->getHeading());
}
NodeP lpitch(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->pitch());
}
NodeP lroll(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->roll());
}

NodeP lOrientation(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return 
		new Pair(new Float(TURTLE->roll()), 
		new Pair(new Float(TURTLE->pitch()), 
		new Pair(new Float(TURTLE->getHeading()), 0)));
}

NodeP lsetHeading(PairPP arg) 
{
	ThreadP th=Thread::This();
	double heading = arg[1]->a_float(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setHeading(heading);
	return Unbound;
}

NodeP lsetRoll(PairPP arg) 
{
	ThreadP th=Thread::This();
	double roll = arg[1]->a_float(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setRoll(roll);
	return Unbound;
}

NodeP lsetPitch(PairPP arg) 
{
	ThreadP th=Thread::This();
	double pitch = arg[1]->a_float(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setPitch(pitch);
	return Unbound;
}

NodeP lsetOrientation(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP o=arg[1]->a_3dvector(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setOrientation(Point(
		o-> first()->getnumber(),
		o->second()->getnumber(),
		o-> third()->getnumber()));
	return Unbound;
}

NodeP ltowards(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_vector(1);
	if throwing return Unbound;
	GraphLocker gl;
	NodeP 
		nodeX = v->first(),
		nodeY = v->second();
	Point target(
			nodeX->type == FLOAT_ ?   FloatP(nodeX)->float_  :  IntP(nodeX)->int_,
			nodeY->type == FLOAT_ ?   FloatP(nodeY)->float_   :  IntP(nodeY)->int_,
			0);
	return new Float(TURTLE->towards_helper(target));
}

NodeP ltowardsXYZ(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_3dvector(1);
	if throwing return Unbound;
	GraphLocker gl;
	double
		x=v-> first()->getnumber(),
		y=v->second()->getnumber(),
		z=v-> third()->getnumber(),
		rx,ry,rz;

	Point target(x,y,z), diff=target-*TURTLE;
	diff=diff/diff.Norm();

   // Compute angle between Y Axis and Vector [DOT] (Rotation about Z)
   rz = acos(diff.y)/Float::deg2rad;

   // Project point onto X-Z plane and renormalize to measure roll
   diff.y = 0.0;
	double n=diff.Norm();
   if (n!=0)
		diff=diff/n;

   // Compute angle between X Axis and Projection [DOT] (Rotation about Y)
   ry = acos(diff.x)/Float::deg2rad;

   // Heading was limited to 0 to 180 but we must be able to roll -180 to 180
   // if point is in positive Z hemisphere then "roll" left to it.
   if (diff.z > 0.0) ry = -ry;

   // No Pitch needed
   rx = 0.0;

	return 
		new Pair(new Float(ry), 
		new Pair(new Float(rx), 
		new Pair(new Float(rz), 0)));
}

NodeP lDistance(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_vector(1);
	if throwing return Unbound;
	NodeP 
		nodeX = v->first(),
		nodeY = v->second();
	GraphLocker gl;
	Point target(
			nodeX->type == FLOAT_? FloatP(nodeX)->float_:IntP(nodeX)->int_,
			nodeY->type == FLOAT_? FloatP(nodeY)->float_:IntP(nodeY)->int_,
			0);
	return new Float(TURTLE->distance(target));
}

NodeP lDistanceXYZ(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_3dvector(1);
	if throwing return Unbound;
	NodeP 
		nodeX=v->first(),
		nodeY=v->second(),
		nodeZ=v->third();
	GraphLocker gl;
	Point target(
			nodeX->type == FLOAT_? FloatP(nodeX)->float_:IntP(nodeX)->int_,
			nodeY->type == FLOAT_? FloatP(nodeY)->float_:IntP(nodeY)->int_,
			nodeZ->type == FLOAT_? FloatP(nodeZ)->float_:IntP(nodeZ)->int_);
	return new Float(TURTLE->distance(target));
}

NodeP lxCor(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->x);
}
NodeP lyCor(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->y);
}
NodeP lzCor(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->z);
}

NodeP lPos(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return 
		new Pair(new Float(TURTLE->x), 
		new Pair(new Float(TURTLE->y), 0));
}

NodeP lPosXYZ(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return 
		new Pair(new Float(TURTLE->x), 
		new Pair(new Float(TURTLE->y), 
		new Pair(new Float(TURTLE->z), 0)));
}

NodeP lScrunch(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return new List
	(	new Float(TURTLE->screen->scale.x), 
		new Float(TURTLE->screen->scale.y), 
		new Float(TURTLE->screen->scale.z), _ListEnd
	);
}

NodeP lHome(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->home();
	return Unbound;
}

NodeP lclearScreen(PairPP arg) 
{	ThreadP th=Thread::This();
	if (th->graph)
	{	GraphLocker gl;
		th->graph->screen_clear(true); 
	}
	return Unbound;
}

NodeP lclean(PairPP arg) 
{	ThreadP th=Thread::This();
	if (th->graph)
	{	GraphLocker gl;
		th->graph->screen_clear(false); 
	}
	return Unbound;
}

void Node::setXYZ(NodeP ynode, NodeP znode)
{	error(BAD_DATA, this, 1);
}
void String::setXYZ(NodeP ynode, NodeP znode)
{	if (numberP())
		number()->setXYZ(ynode, znode);
	else 
		error(BAD_DATA, this, 1);
}
void Int::setXYZ(NodeP ynode, NodeP znode)
{	ThreadP th=Thread::This();
	if (ynode->numberP())
	{	if (znode)
		{	if (znode->numberP())
				TURTLE->setPos(this, ynode, znode);
			else 
			{	error(BAD_DATA, ynode, 2);	return;
			}
		}else
			TURTLE->setPos(this, ynode, 0);
	}else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}
void Float::setXYZ(NodeP ynode, NodeP znode)
{	ThreadP th=Thread::This();
	if (ynode->numberP())
	{	if (znode)
		{	if (znode->numberP())
				TURTLE->setPos(this, ynode, znode);
			else 
			{	error(BAD_DATA, ynode, 2);	return;
			}
		}else
			TURTLE->setPos(this, ynode, 0);
	}else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}
void Pair::setXYZ(NodeP ynode, NodeP znode)
{	ThreadP th=Thread::This();
	if (this && circcounter()==th->circcounter)
		return;
	PairP x=this;
	if (ynode->islist())
	{	PairP y=PairP(ynode);
		if (znode)
			if (znode->islist())
			{	PairP z=PairP(znode);
				while (x && y && z)
				{	if (x->circcounter()==th->circcounter)
						break;
					else
						x->setcirccounter(th->circcounter);
					if (x->first() && y->first() && z->first())
						x->first()->setXYZ(y->first(), z->first());
					else
					{	error(BAD_DATA, this, 2);	return;
					}
					next(x);
					next(y);
					next(z);
				}
			}else
			{	error(BAD_DATA, this, 2);	return;
			}
		else
		{	while (x && y)
			{	if (x->circcounter()==th->circcounter)
					break;
				else
					x->setcirccounter(th->circcounter);
				if (x->first() && y->first())
					x->first()->setXYZ(y->first(), 0);
				else
				{	error(BAD_DATA, this, 2);	return;
				}
				next(x);
				next(y);
			}
		}
	}
	else
	{	error(BAD_DATA, this, 2);	return;
	}
	th->setcirccounter(th->circcounter+1);
}
void FloatArray::setXYZ(NodeP ynode, NodeP znode)
{	ThreadP th=Thread::This();
	int l=length();
	if (ynode->type==FLOATARRAY && FloatArrayP(ynode)->length()==l)
	{	if (znode)
		{	if (znode->type==FLOATARRAY && FloatArrayP(znode)->length()==l)
			{	doubleP x=data(), 
					y=FloatArrayP(ynode)->data(),
					z=FloatArrayP(znode)->data();
				while (l--)
				{	TURTLE->setPos(*x++, *y++, *z++);
				}
			}
			else 
			{	error(BAD_DATA, znode, 3);	return;
			}
		}
		else
		{	doubleP x=data(), 
				y=FloatArrayP(ynode)->data();
			while (l--)
			{	TURTLE->setPos(*x++, *y++, 0);
			}
		}
	}
	else 
	{	error(BAD_DATA, ynode, 2);	return;
	}
}

NodeP lsetPos(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_vector(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setPos(v->first(), v->second(), NIL);
	return Unbound;
}

NodeP l_setPos(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_vector(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->_setPos(v->first(), v->second(), NIL);
	return Unbound;
}

NodeP lsetPosXYZ(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!arg[1])
	{	error(BAD_DATA, arg[1], 1); return Unbound;
	}
	if (arg[1]->is.list || arg[1]->type==ARRAY)
	{	PairP v = arg[1]->a_3dvector(1);
		if throwing return Unbound;
		GraphLocker gl;
		TURTLE->setPos(v->first(), v->second(), v->third());
	}else if (arg[1]->type==FLOATARRAY)
	{	FloatArrayP v = FloatArrayP(arg[1]);
		if (v->length() != 3)
		{	error(BAD_DATA, arg[1], 1); return Unbound;
		}
		doubleP d=v->data();
		GraphLocker gl;
		TURTLE->setPos(d[0], d[1], d[2]);
	}
	return Unbound;
}

NodeP l_setPosXYZ(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_3dvector(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->_setPos(v->first(), v->second(), v->third());
	return Unbound;
}

NodeP lsetXYZ(PairPP arg) 
{
 	NodeP
		x = arg[1],
		y = arg[2],
		z = arg[3];
	if (!x)
	{	error(BAD_DATA, x, 1); return Unbound;
	}
	if (!y)
	{	error(BAD_DATA, y, 2); return Unbound;
	}
	if (!z)
	{	error(BAD_DATA, z, 3); return Unbound;
	}
	if throwing return Unbound;
	GraphLocker gl;
	x->setXYZ(y, z);
	return Unbound;
}

NodeP lsetXY(PairPP arg) 
{
	NodeP 
		x = arg[1],
		y = arg[2];
	if (!x)
	{	error(BAD_DATA, x, 1);
		return Unbound;
	}
	if (!y)
	{	error(BAD_DATA, y, 2);
		return Unbound;
	}
	if throwing return Unbound;
	GraphLocker gl;
	x->setXYZ(y, 0);
	return Unbound;
}

NodeP lsetX(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP xnode = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setPos(xnode, NIL, NIL);
	return Unbound;
}

NodeP lsetY(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP ynode = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setPos(NIL, ynode, NIL);
	return Unbound;
}

NodeP lsetZ(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP znode = arg[1]->a_number(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setPos(NIL, NIL, znode);
	return Unbound;
}

NodeP lsetSpherePos(PairPP arg) 
{
	ThreadP th=Thread::This();
	double
		    r = arg[1]->a_float(1),
		  phi = arg[2]->a_float(2),
		theta = arg[3]->a_float(3);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setSpherePos(r, phi, theta);
	return Unbound;
}

NodeP lsetCylinderPos(PairPP arg) 
{
	ThreadP th=Thread::This();
	double
		    r = arg[1]->a_float(1),
		  phi = arg[2]->a_float(2),
		    y = arg[3]->a_float(3);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setCylinderPos(r, phi, y);
	return Unbound;
}

NodeP lwrap(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->wrap(); 
	if (RECORDER.recording)
		RECORDER << new R_wrap();
	return Unbound;
}

NodeP lFence(PairPP arg) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	lwrap(arg);	    // get turtle inside the fence
//	TURTLE->draw();
	TURTLE->mode=Turtle::fencemode;
//	TURTLE->draw();
	if (RECORDER.recording)
		RECORDER << new R_mode(*TURTLE);
	return Unbound;
}

NodeP lWindowMode(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
//	TURTLE->draw();
	TURTLE->mode=Turtle::windowmode;
//	TURTLE->draw();
	if (RECORDER.recording)
		RECORDER << new R_mode(*TURTLE);
	return Unbound;
}

NodeP lfill(PairPP arg) 
{
	enum Params {_mode=1};
	ThreadP th=Thread::This();
	int mode=0;
	if (exist(_mode))
	{	if (arg[_mode] && NodeP(arg[_mode])->equalP(_True))
			mode=1;
		else
			mode=arg[_mode]->a_int(_mode);
		if throwing return Unbound;
	}
	GraphLocker gl;
	TURTLE->fill(mode);
	if (RECORDER.recording)
		RECORDER << new R_fill(mode);
	return Unbound;
}

NodeP lsetFillTolerance(PairPP arg)
{
	ThreadP th=Thread::This();
	double tolerance=arg[1]->a_float(1);
	GraphLocker gl;
	theApp->graph->colorTolerance=tolerance;
	TURTLE->colorTolerance=tolerance;
	return Unbound;
}

NodeP lfillRect(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP v1=arg[1]->a_vector(1),
			v2=arg[2]->a_vector(2);
	if throwing return Unbound;
	NodeP X1=v1->first(), Y1=v1->second(),
			X2=v2->first(), Y2=v2->second();
	double x1=X1->getnumber(), y1=Y1->getnumber(),
			 x2=X2->getnumber(), y2=Y2->getnumber();
	Point p1=Point(x1,y1,0), p2=Point(x2,y2,0);
	GraphLocker gl;
	TURTLE->fillRect(p1, p2);
	if (RECORDER.recording)
		RECORDER << new R_fillRect(*TURTLE, p1,p2);
	return Unbound;
}

NodeP lfillEllipse(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP 
		radiusx = arg[1]->a_number(1),
		radiusy = arg[2]->a_number(2);
	
	if throwing return Unbound;
	double 
		rx=radiusx->getnumber(), 
		ry=radiusy->getnumber();
	Point r=Point(rx,ry,0);
	GraphLocker gl;
	TURTLE->fillEllipse(r);
	if (RECORDER.recording)
		RECORDER << new R_fillEllipse(*TURTLE, r);
	return Unbound;
}

NodeP lfillCircle(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP radius = arg[1]->a_number(1);
	if throwing return Unbound;
	double rx=radius->getnumber();
	Point r=Point(rx,rx,0);
	GraphLocker gl;
	TURTLE->fillEllipse(r);
	if (RECORDER.recording)
		RECORDER << new R_fillEllipse(*TURTLE, r);
	return Unbound;
}

NodeP lSphere(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP radius = arg[1]->a_number(1);
	if throwing return Unbound;
	double r=radius->getnumber();
	int slices=0;
	if (exist(2))
	{	slices=arg[2]->a_int(2);
		if throwing return Unbound;
	}
	int stacks=0;
	if (exist(3))
	{	stacks=arg[3]->a_int(3);
		if throwing return Unbound;
	}
	GraphLocker gl;
	if (TURTLE->shadows)
		TURTLE->partialEllipsoidSh(r, 0, 360, 40, 0, 180, 40);
	else
		TURTLE->sphere(r, slices, stacks);
	if (RECORDER.recording)
		RECORDER << new R_Sphere(*TURTLE, r, slices, stacks);
	return Unbound;
}

NodeP lEllipsoid(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP 
		rx = arg[1]->a_number(1),
		ry = arg[2]->a_number(2),
		rz = arg[3]->a_number(3);
	if throwing return Unbound;
	int slices=0;
	if (exist(4))
	{	slices=arg[4]->a_int(4);
		if throwing return Unbound;
	}
	int stacks=0;
	if (exist(5))
	{	stacks=arg[5]->a_int(5);
		if throwing return Unbound;
	}
	Point r=Point(rx->getnumber(),ry->getnumber(),rz->getnumber());
	GraphLocker gl;
	TURTLE->ellipsoid(r);
	if (RECORDER.recording)
		RECORDER << new R_Ellipsoid(*TURTLE, r, slices, stacks);
	return Unbound;
}

NodeP lpartialEllipsoid(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP 
		rx = arg[1]->a_number(1),
		ry = arg[2]->a_number(2),
		rz = arg[3]->a_number(3),
		slicesstart=arg[4]->a_number(4),
		slicesend=arg[5]->a_number(5),
		stacksstart=arg[7]->a_number(7),
		stacksend=arg[8]->a_number(8);
	int slices=arg[6]->a_int(6);
	int stacks=arg[9]->a_int(9);
	if throwing return Unbound;
	Point r=Point(rx->getnumber(),ry->getnumber(),rz->getnumber());
	double 
		sls=slicesstart->getnumber(),
		sle=slicesend->getnumber(),
		sts=stacksstart->getnumber(),
		ste=stacksend->getnumber();
	if throwing return Unbound;
	GraphLocker gl;
	if (TURTLE->shadows)
		TURTLE->partialEllipsoidSh(r, sls, sle, slices, sts, ste, stacks);
	else
		TURTLE->partialEllipsoid(r, sls, sle, slices, sts, ste, stacks);
	if (RECORDER.recording)
		RECORDER << new R_partialEllipsoid(*TURTLE, 
			r, sls, sle, slices, sts, ste, stacks);
	return Unbound;
}

NodeP lfillPie(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP 
		angle   = arg[1]->a_number(1),
		radiusx = arg[2]->a_number(2);
	if throwing return Unbound;
	double 
		ang=angle->getnumber(),
		rx=radiusx->getnumber();
	Point r=Point(rx,rx,0);
	GraphLocker gl;
	TURTLE->fillPie(ang,r);
	if (RECORDER.recording)
		RECORDER << new R_fillPie(*TURTLE, ang,r);
	return Unbound;
}

NodeP lLabel(PairPP arg) 
{
	ThreadP th=Thread::This();
	char textbuf[300];
	th->stringptr=textbuf;
	th->stringlen=300-1;
	stringbuffer << arg[1];
	*(th->stringptr)='\0';
	if throwing return Unbound;
	GraphLocker gl;
//	TURTLE->draw();
	TURTLE->Label(textbuf);
	if (RECORDER.recording)
		RECORDER << new R_Label(*TURTLE, textbuf);
//	TURTLE->draw();
	return Unbound;
}

NodeP lLabelSize(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (noarg)
	{	Font_ &f=TURTLE->font;
		return 
			new Pair(new Float(f.sizex),
			new Pair(new Float(f.sizey), 0));
	}
	char textbuf[300];
	th->stringptr=textbuf;
	th->stringlen=300-1;
	stringbuffer << arg[1];
	*(th->stringptr)='\0';
	if throwing return Unbound;
	GraphLocker gl;
	Point s=TURTLE->LabelSize(textbuf);
	return 
		new Pair(new Float(s.x), 
		new Pair(new Float(s.y), 
		new Pair(new Float(s.z), 0)));
	return Unbound;
}

NodeP lsetLabelSize(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP l=arg[1]->a_vector(1);
	if throwing return Unbound;
	NodeP xn=l->first(), yn=l->second();
	double x=xn->getnumber(), y=yn->getnumber();
	GraphLocker gl;
	TURTLE->setLabelSize(x,y);
	return Unbound;
}

NodeP lLabelFont(PairPP)
{
	ThreadP th=Thread::This();
	if throwing return Unbound;
	GraphLocker gl;
	StringP fontname=new String(TURTLE->getLabelFont());
	return fontname;
}

NodeP lsetLabelFont(PairPP arg)
{
	ThreadP th=Thread::This();
	StringP font=0;
	if (arg[1] && arg[1]->is.list 
	&& arg[1]->first() && arg[1]->first()->is.list)
		font=arg[1]->first()->a_string(1);
	else
		font=arg[1]->a_string(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setLabelFont(font->ptr());
	return Unbound;
}

NodeP lLabelWeight(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Float(TURTLE->font.weight);
}

NodeP lsetLabelWeight(PairPP arg)
{
	ThreadP th=Thread::This();
	int weight=arg[1]->a_int(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setLabelWeight(weight);
	return Unbound;
}

NodeP lLabelAlign(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	Font_ &f=TURTLE->font;
	return 
		new Pair(new Int(f.alignx),
		new Pair(new Int(f.aligny), 0));
}

NodeP lsetLabelAlign(PairPP arg)
{
	ThreadP th=Thread::This();
	int	alignx=arg[1]->a_int(1),
			aligny=arg[2]->a_int(2);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setLabelAlign(alignx,aligny);
	return Unbound;
}

NodeP lisPenDown  (PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return TURTLE->pen.down ?  _True  :  _False;
}
NodeP lPenMode    (PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return TURTLE->getNodePenMode();
}
NodeP lPenColor   (PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Int(TURTLE->pen.color);
}
NodeP lScreenColor(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Int(SCREEN->color);
}

NodeP lPenSize(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return new List
	(	new Float(TURTLE->pen.size.x),
		new Float(TURTLE->pen.size.y), _ListEnd
	);
}

NodeP lPenPattern(PairPP) {return 0;} // no Patterns yet!

NodeP lPenDown(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.down=true; 
	if (RECORDER.recording)
		RECORDER << new R_setPenDown(*TURTLE); 
	return Unbound;
}

NodeP lPenUp(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.down=false; 
	if (RECORDER.recording)
		RECORDER << new R_setPenDown(*TURTLE); 
	return Unbound;
}

NodeP lPenPaint(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.internalMode=Pen::Paint;
	TURTLE->penDown();
	if (RECORDER.recording)
		RECORDER << new R_setPenMode(*TURTLE);
	return lPenDown(PairPP(Thread::This()->argStack->index));
}

NodeP lPenErase(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.internalMode=Pen::Erase;
	TURTLE->penErase();
	if (RECORDER.recording)
		RECORDER << new R_setPenMode(*TURTLE);
	return lPenDown(PairPP(Thread::This()->argStack->index));
}

NodeP lPenReverse(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.internalMode=Pen::Reverse;
	TURTLE->penReverse();
	if (RECORDER.recording)
		RECORDER << new R_setPenMode(*TURTLE);
	return lPenDown(PairPP(Thread::This()->argStack->index));
}

NodeP lRGB(PairPP arg)	// RGB=Red, Green, Blue
{
	NodeP r,g,b;
	if (!exist(2) && arg[1] && arg[1]->is.list)
	{	PairP l=arg[1]->a_rgb(1);
		r=l->first();
		g=l->second();
		b=l->third();
	}else
	{	r=arg[1]->a_number(1), 
		g=arg[2]->a_number(2), 
		b=arg[3]->a_number(3);
	}
	if throwing return Unbound;
	Color c=Pen::rgb(r->getnumber(), g->getnumber(), b->getnumber());
	return new Int(c);
}

NodeP lRGBA(PairPP arg)	// RGBA=Red, Green, Blue, Alpha
{
	NodeP r,g,b,a;
	if (!exist(2) && arg[1] && arg[1]->is.list)
	{	PairP l=arg[1]->a_list(1);
		r=l->first();
		g=l->second();
		b=l->third();
		a=l->fourth();
	}else
	{	r=arg[1]->a_number(1), 
		g=arg[2]->a_number(2), 
		b=arg[3]->a_number(3);
		a=arg[4]->a_number(4);
	}
	if throwing return Unbound;
	Color c=Pen::rgba(r->getnumber(), g->getnumber(), b->getnumber(), a->getnumber());
	return new Int(c);
}

NodeP lreRGB(PairPP arg)
{
	Color c=arg[1]->a_int(1);
	if throwing return Unbound;
	return new List
	(	new Float(double(c.r)/0xff), 
		new Float(double(c.g)/0xff), 
		new Float(double(c.b)/0xff), _ListEnd
	);
}

NodeP lreRGBA(PairPP arg)
{
	Color c=arg[1]->a_int(1);
	if throwing return Unbound;
	return new List
	(	new Float(double(c.r)/0xff), 
		new Float(double(c.g)/0xff), 
		new Float(double(c.b)/0xff), 
		new Float(double(c.a)/0xff), _ListEnd
	);
}

NodeP lHSB(PairPP arg)	// HSB=Hue Saturation Brightness
{
	NodeP 
		h=arg[1]->a_number(1), 
		s=arg[2]->a_number(2), 

		b=arg[3]->a_number(3);
	if throwing return Unbound;
	Color c=Pen::hsb(h->getnumber(), s->getnumber(), b->getnumber());
	return new Int(c);
}

NodeP lHSBA(PairPP arg)	// HSB=Hue Saturation Brightness Alpha
{
	NodeP 
		h=arg[1]->a_number(1), 
		s=arg[2]->a_number(2), 
		b=arg[3]->a_number(3),
		a=arg[4]->a_number(4);
	if throwing return Unbound;
	Color c=Pen::hsba(
		h->getnumber(), 
		s->getnumber(), 
		b->getnumber(), 
		a->getnumber());
	return new Int(c);
}

NodeP lreHSB(PairPP arg)	// color to HSB
{
	Color	c=arg[1]->a_color(1);
	if throwing return Unbound;
	return Pen::rehsb(
		double(c.r)/0xff, 
		double(c.g)/0xff, 
		double(c.b)/0xff);
}

NodeP lreHSBA(PairPP arg)	// color to HSBA
{
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	return Pen::rehsba(
		double(c.r)/0xff, 
		double(c.g)/0xff, 
		double(c.b)/0xff, 
		double(c.a)/0xff);
}

NodeP laddColors(PairPP arg)
{
	Color
		c1=arg[1]->a_int(1),
		c2=arg[2]->a_int(2),
		c12;
	if (!throwing)
	{	int 
			r=c1.r+c2.r,
			g=c1.g+c2.g,
			b=c1.b+c2.b;
		if (r>255) r=255;
		if (g>255) g=255;
		if (b>255) b=255;
		c12.r=r;
		c12.g=g;
		c12.b=b;
		return new Int(c12);
	}
	return Unbound;
}

NodeP laddColorsMod(PairPP arg)
{
	Color
		c1=arg[1]->a_int(1),
		c2=arg[2]->a_int(2),
		c12;
	if (!throwing)
	{	int 
			r=(c1.r+c2.r) & 0xff,
			g=(c1.g+c2.g) & 0xff,
			b=(c1.b+c2.b) & 0xff,
			a=(c1.a+c2.a)/2;
		c12.r=r;
		c12.g=g;
		c12.b=b;
		c12.a=a;
		return new Int(c12);
	}
	return Unbound;
}

NodeP lisLikeColor(PairPP arg) 
{
	enum Paras {_color1=1, _color2, _tolerance};
	Color color1=arg[_color1]->a_color(_color1);
	if throwing return Unbound;
	Color color2=arg[_color2]->a_color(_color2);
	if throwing return Unbound;
	double tolerance=arg[_tolerance]->a_float(_tolerance);
	if throwing return Unbound;
	if (color1.isLike(color2, tolerance))
		return _True;
	return _False;
}

NodeP lgetColorDatabase(PairPP)
{
	wxStringToColourHashMap *cols=wxTheColourDatabase->GetHashMap();
	wxStringToColourHashMap::iterator it;
	PairP result=0;
	for(it=cols->begin();  it!=cols->end();  ++it)
	{	wxString name = it->first;
		wxColour *c = it->second;
		push(result, new List(
			new String(name), 
			new Float(c->Red()/255.0),
			new Float(c->Green()/255.0),
			new Float(c->Blue()/255.0), _ListEnd));
		if throwing return Unbound;
	}
	if (result)
		result=PairP(result->combSort());
	return result;
}

NodeP lsetPenColor(PairPP arg) 
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setColor(c);
	if (!TURTLE->surfaceMode && RECORDER.recording)
		RECORDER << new R_setPenColor(*TURTLE);
	return Unbound;
}

NodeP lsetFloodColor(PairPP arg) 
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->pen.floodcolor=c;
	if (RECORDER.recording)
		RECORDER << new R_setFloodColor(*TURTLE);
	return Unbound;
}

NodeP lFloodColor(PairPP arg) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Int(TURTLE->pen.floodcolor);
}

NodeP lsetShadowColor(PairPP arg) 
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->shadowcolor=c;
	if (RECORDER.recording)
		RECORDER << new R_setShadowColor(*TURTLE);
	return Unbound;
}

NodeP lShadowColor(PairPP arg) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return new Int(TURTLE->shadowcolor);
}

NodeP lsetScreenColor(PairPP arg) 
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setScreenColor(c);
	if (RECORDER.recording)
		RECORDER << new R_setScreenColor(SCREEN->color);
	if (Thread::This()->graph)
		Thread::This()->graph->screen_reset(true);
	return Unbound;
}

NodeP lsetPenSize(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!arg[1]) 
	{	error(BAD_DATA, arg[1]); return Unbound;
	}
	if (arg[1]->numberP())
	{	double s=arg[1]->getnumber();
		if throwing return Unbound;
		GraphLocker gl;
		TURTLE->setPenSize(s, s);
		if (RECORDER.recording)
			RECORDER << new R_setPenSize(*TURTLE);
		return Unbound;
	}
	PairP v = arg[1]->a_vector(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->setPenSize( 
		v-> first()->getnumber(),
		v->second()->getnumber());
	if (RECORDER.recording)
		RECORDER << new R_setPenSize(*TURTLE);
	return Unbound;
}

NodeP lsetPenPattern(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (arg[1]->isnolist() && !throwing)
		error(BAD_DATA, arg[1]);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->pen.setListPattern(arg[1]);
	if (RECORDER.recording)
		RECORDER << new R_setPenPattern(*TURTLE);
	return Unbound;
}

NodeP lsetScrunch(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP
		nodeX = arg[1]->a_number(1),
		nodeY = arg[2]->a_number(2),
		nodeZ = arg[3]->a_number(3);

	if throwing return Unbound;
	GraphLocker gl;
//	TURTLE->draw();
	TURTLE->screen->scale.x = nodeX->getnumber();
	TURTLE->screen->scale.y = nodeY->getnumber();
	TURTLE->screen->scale.z = nodeZ->getnumber();
	int w,h;
	if (th->graph)
	{	th->graph->GetClientSize(&w, &h);
//		TURTLE->draw();
#ifdef OPENGL
		glViewport(0, 0, w, h);
#endif
		th->graph->setFrustum(w, h);
//		TURTLE->draw();
	}
	return Unbound;
}

NodeP lTextMousePos(PairPP) 
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (th->graph)
		return 
			new Pair(new Int(th->graph->mouse->textx), 
			new Pair(new Int(th->graph->mouse->texty), 0));
	else
		return 
			new Pair(new Int(0), 
			new Pair(new Int(0), 0));
}

NodeP lTextMouseX(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	if (th->graph)
		return new Int(Thread::This()->graph->mouse->textx);
	else
		return new Int(0);
}
NodeP lTextMouseY(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	if (th->graph)
		return new Int(Thread::This()->graph->mouse->texty);
	else
		return new Int(0);
}

NodeP lArc(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP
		 angleNd = arg[1]->a_number(1),
		radiusNd = arg[2]->a_number(2);
	
	if throwing return Unbound;
	double 
		 angle= angleNd->getnumber(), 
		radius=radiusNd->getnumber();
		if (fabs(radius) > Int16::max)
			radius=Int16::max;
	if throwing return Unbound;
	TURTLE->arc(angle, radius);
	return Unbound;
}

NodeP lCircle(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP radiusNd = arg[1]->a_number(1);
	if throwing return Unbound;
	double radius=radiusNd->getnumber();
	if (fabs(radius) > Int16::max)
		radius=Int16::max;
	GraphLocker gl;
	TURTLE->arc(360, radius);
	return Unbound;
}

NodeP lEllipseArc(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP
		      angleNd = arg[1]->a_number(1),
		minorRadiusNd = arg[2]->a_number(2),
		majorRadiusNd = arg[3]->a_number(3),
		 startAngleNd = arg[4]->a_number(4);
	
	if throwing return Unbound;
	double 
		      angle=      angleNd->getnumber(), 
		minorRadius=minorRadiusNd->getnumber(),
		majorRadius=majorRadiusNd->getnumber(),
		 startAngle= startAngleNd->getnumber();
	if (fabs(minorRadius) > Int16::max)
		minorRadius=Int16::max;
	if (fabs(majorRadius) > Int16::max)
		majorRadius=Int16::max;
	GraphLocker gl;
	TURTLE->EllipseArc(angle, minorRadius, majorRadius, startAngle);
	return Unbound;
}

NodeP lEllipse(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP
		minorRadiusNd = arg[1]->a_number(1),
		majorRadiusNd = arg[2]->a_number(2);
	
	if throwing return Unbound;
	double 
		minorRadius=minorRadiusNd->getnumber(),
		majorRadius=majorRadiusNd->getnumber();
	GraphLocker gl;
	TURTLE->EllipseArc(360, minorRadius, majorRadius, 0);
	return Unbound;
}

NodeP lCylinder(PairPP arg)
{
	enum Params 
	{	_length=1, 
		_baseradius, _topradius, 
		_startangle, _endangle, 
		_scrunch, _slices
	};
	ThreadP th=Thread::This();
	double
		length = arg[_length]->a_float(_length),
		baseradius = arg[_baseradius]->a_float(_baseradius),
		topradius = baseradius;
	if (exist(_topradius))
		topradius = arg[_topradius]->a_float(_topradius);
	
	if throwing return Unbound;
	if (exist(_startangle))
	{	double startangle = arg[_startangle]->a_float(_startangle);
		double   endangle = arg[_endangle]->a_float(_endangle);
		double scrunch=1;
		if (exist(_scrunch))
			scrunch=arg[_scrunch]->a_float(_scrunch);
		int slices=30;
		if (exist(_slices))
			slices=arg[_slices]->a_int(_slices);
		if throwing return Unbound;
		GraphLocker gl;
		TURTLE->partialCylinder(length, baseradius, topradius, startangle, endangle, scrunch, slices);
		if (RECORDER.recording)
			RECORDER << new R_PartCylinder(*TURTLE, 
				length, baseradius, topradius, startangle, endangle, scrunch, slices);
	}else
	{	GraphLocker gl;
		TURTLE->cylinder(length, baseradius, topradius);
		if (RECORDER.recording)
			RECORDER << new R_Cylinder(*TURTLE, length, baseradius, topradius);
	}
	return Unbound;
}

NodeP lrefresh(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	RECORDER.on (); return Unbound;
}
NodeP lnoRefresh(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	RECORDER.off(); return Unbound;
}
NodeP lisRefresh(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	return Node::boolnode(RECORDER.recording);
}

NodeP lredraw(PairPP) 
{	ThreadP th=Thread::This();
	GraphLocker gl;
	RECORDER.playAgain(); return Unbound;
}

NodeP lsavePicture    (PairPP arg) 
{
	ThreadP th=Thread::This();
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GraphLocker gl;
	RECORDER.save    (name->ptr()); return Unbound;
}
NodeP lloadPicture    (PairPP arg) 
{
	ThreadP th=Thread::This();
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GraphLocker gl;
	RECORDER.load    (name->ptr()); return Unbound;
}
NodeP lsavePictureText(PairPP arg) 
{
	ThreadP th=Thread::This();
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GraphLocker gl;
	RECORDER.saveText(name->ptr()); return Unbound;
}
NodeP lloadPictureText(PairPP arg) 
{
	ThreadP th=Thread::This();
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GraphLocker gl;
	RECORDER.loadText(name->ptr()); return Unbound;
}
NodeP lsavePostScript (PairPP arg) 
{
	ThreadP th=Thread::This();
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	GraphLocker gl;
	RECORDER.writePS (name->ptr()); return Unbound;
}

NodeP lsetSaveSize(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP l=arg[1]->a_vector(1);
	if (!throwing && th->graph)
	{	GraphLocker gl;
		th->graph->saveSize.SetWidth ( int(l-> first()->getnumber()) );	
		th->graph->saveSize.SetHeight( int(l->second()->getnumber()) );
	}
	return Unbound;
}

NodeP lSaveSize(PairPP arg)
{
	ThreadP th=Thread::This();
	if (th->graph)
	{	GraphLocker gl;
		return new List(
			new Int(th->graph->saveSize.GetWidth()),
			new Int(th->graph->saveSize.GetHeight()), _ListEnd);
	}else
		return new List(
			new Int(0),
			new Int(0), _ListEnd);
}

NodeP lsaveScreen(PairPP arg)
{
	ThreadP th=Thread::This();
	StringP filename=arg[1]->a_string(1);
	filename=String::make(filename->ptr(), 0, filename->length(), STRING, String::copy_noparity);
	if (th->graph)
	{	GraphLocker gl;
		return th->graph->saveScreen(filename->ptr());
	}
	return Unbound;
}

NodeP lsaveScreenVector(PairPP arg)
{
#ifdef OPENGL
	ThreadP th=Thread::This();
	StringP filename=arg[1]->a_string(1);
	if throwing return Unbound;
	int sort=GL2PS_SIMPLE_SORT;
	if (exist(2) && arg[2])
	{	sort=arg[2]->a_int(2);
		if throwing return Unbound;
	}
	int options=GL2PS_DRAW_BACKGROUND;
	if (exist(3) && arg[3])
	{	options=arg[3]->a_int(3);
		if throwing return Unbound;
	}
	int colornr=0;
	if (exist(4))
	{	colornr=arg[4]->a_int(3);
		if throwing return Unbound;
	}
	GraphLocker gl;
	if (th->graph)
		th->graph->saveVector(filename->ptr(), sort, options, colornr);
#endif
	return Unbound;
}

NodeP lupdateGraph(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (th->graph)
		th->graph->update();
	return Unbound;
}

NodeP lsetUpdateGraph(PairPP arg)
{
	ThreadP th=Thread::This();
	int b=arg[1]->a_bool(1);
	if throwing return Unbound;
	GraphLocker gl;
	th->graph->direct= b==1 ? true : false;
	return Unbound;
}

NodeP lscroll(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP 
		size=arg[1]->a_vector(1),
		displacement=arg[2]->a_vector(2);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->scroll(
		int(size-> first()->getnumber()),
		int(size->second()->getnumber()),
		displacement-> first()->getnumber(),
		displacement->second()->getnumber());
	return Unbound;
}

NodeP lscrollCalibrate(PairPP arg)
{
	ThreadP th=Thread::This();
	PairP 
		displacement=arg[1]->a_vector(1);
	if throwing return Unbound;
	GraphLocker gl;
	Point s=TURTLE->scrollCalibrate(			
		displacement-> first()->getnumber(),
		displacement->second()->getnumber());
	return new List(
		new Float(s.x), new Float(s.y), _ListEnd);
	return Unbound;
}

NodeP lTurtle(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return *th->graph->d(Graph::turtleNode);
}

NodeP lnewTurtle(PairPP)
{
	ThreadP th=Thread::This();
	Turtle *t=new Turtle;
	if (!t)
	{	error(OUT_OF_MEM); return Unbound;
	}
	if throwing return Unbound;
	GraphLocker gl;
	ArrayP shapes=th->turtleNode->shapes;
	t->create(th->graph->recorder, SCREEN);
	TurtleNodeP tn=new TurtleNode(t, GraphNodeP(*th->graph->d(Graph::windowNode)));
	if throwing return Unbound;
	tn->shapes=shapes;
	return tn;
}

NodeP lsetTurtle(PairPP arg)
{
	ThreadP th=Thread::This();
	TurtleNodeP tn=arg[1]->a_turtleNode(1);
	if throwing return Unbound;
	GraphLocker gl;
	GraphP graph=th->graph;
	if (!graph)
		return Unbound;
	*graph->d(Graph::turtleNode)=tn;
	Turtle *t=tn->turtle;
	t->recorder=graph->recorder;
	t->screen=graph->screen;
	TURTLE=t;
	RECORDER.turtle=t;
	t->setColor(t->pen.color);
	if (RECORDER.recording)
		RECORDER << new R_setPenColor(*TURTLE);
	Point p=*t;
	t->getInWindow();
	t->moveTo();
	*t=p;
	th->turtleNode=tn;
	return Unbound;
}

NodeP lTurtleShapes(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	return th->turtleNode->shapes;
}

NodeP lsetTurtleShapes(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP shapes;
	if (arg[1])	
		shapes=arg[1]->a_array(1);
	else
		shapes=0;
	if throwing return Unbound;
	GraphLocker gl;
	th->turtleNode->shapes=ArrayP(shapes);
	th->turtleNode->turtle->shapeNr=0;
	return Unbound;
}

NodeP lperspective(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
//	TURTLE->draw();
	SCREEN->perspective=true;
	SCREEN->orthographic=false;
	GraphP graph=Thread::This()->graph;
	if (!graph)
		return Unbound;
	graph->setFrustum(graph->GetSize().x,graph->GetSize().y);

	SCREEN->eye=Point(400,400,600);
	SCREEN->center=Point(0,0,0);
	SCREEN->up=Point(0,1,0);
	SCREEN->setEye();
	SCREEN->getMatrices();

#ifdef OPENGL
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);

	glFogf(GL_FOG_START, 300);
	glFogf(GL_FOG_END, 1000);
	glFogf(GL_FOG_DENSITY, 0.04f);
	float c[4]={0,0,0,0};
	glFogfv(GL_FOG_COLOR, (float*)&c);
	glFogi(GL_FOG_MODE, GL_LINEAR);
#endif
	if (RECORDER.recording)
		RECORDER << new R_Perspective();

	TURTLE->reset3d();
	TURTLE->mode=Turtle::windowmode;
//	TURTLE->draw();
	return Unbound;
}

NodeP lunPerspective(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
//	TURTLE->draw();
	SCREEN->perspective=false;
	SCREEN->orthographic=false;

	th->graph->setFrustum(SCREEN->winsizex, SCREEN->winsizey);
#ifdef OPENGL
	glLoadIdentity();
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
#endif
	SCREEN->getMatrices();

	if (RECORDER.recording)
		RECORDER << new R_unPerspective();

	TURTLE->reset3d();
//	TURTLE->draw();
	return Unbound;
}

NodeP lorthographic(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
//	TURTLE->draw();
	SCREEN->perspective=true;
	SCREEN->orthographic=true;
	th->graph->setFrustum(SCREEN->winsizex, SCREEN->winsizey);
#ifdef OPENGL
	glLoadIdentity();
#endif
	SCREEN->getMatrices();
	
	if (RECORDER.recording)
		RECORDER << new R_ortho();

	TURTLE->reset3d();
//	TURTLE->draw();
	return Unbound;
}


NodeP lsetEye(PairPP arg)
{
	ThreadP th=Thread::This();
	Point 
		   eye=arg[1]->a_3darray(1),
		center=arg[2]->a_3darray(2),
		    up=arg[3]->a_3darray(3);

	if throwing return Unbound;
	GraphLocker gl;
//	TURTLE->draw();
	SCREEN->eye=eye;
	SCREEN->center=center;
	SCREEN->up=up;
	SCREEN->setEye();
	SCREEN->getMatrices();
//	TURTLE->draw();
//	if (RECORDER.recording)
//		RECORDER << new R_setEye(eye, center, up);
	return Unbound;
}

NodeP lsetLightPos(PairPP arg)
{
	ThreadP th=Thread::This();
	Point pos=arg[1]->a_3darray(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->lightPos=pos;
	float p[4]={pos.x,pos.y,pos.z,0};
#ifdef OPENGL
	glLightfv(GL_LIGHT0, GL_POSITION, p);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setLightPos(pos);
	return Unbound;
}

NodeP lsetLightAmbient(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,
		float(c.a)/255
	};
#ifdef OPENGL
	glLightfv(GL_LIGHT0, GL_AMBIENT, cf);
	if (RECORDER.recording)
		RECORDER << new R_Light(GL_AMBIENT, c);
#endif
	return Unbound;
}

NodeP lsetLightDiffuse(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,

		float(c.a)/255
	};
#ifdef OPENGL
	glLightfv(GL_LIGHT0, GL_DIFFUSE, cf);
	if (RECORDER.recording)
		RECORDER << new R_Light(GL_DIFFUSE, c);
#endif
	return Unbound;
}

NodeP lsetLightSpecular(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,
		float(c.a)/255
	};
#ifdef OPENGL
	glLightfv(GL_LIGHT0, GL_SPECULAR, cf);
	if (RECORDER.recording)
		RECORDER << new R_Light(GL_SPECULAR, c);
#endif
	return Unbound;
}

NodeP lsetLightSpotExponent(PairPP arg)
{
	double e=arg[1]->a_float(1);
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	glLightf(GL_LIGHT0, GL_SPOT_EXPONENT, e);
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_Lightf(GL_SPOT_EXPONENT, e);
	return Unbound;
}

NodeP lsetMaterialAmbient(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,
		float(c.a)/255
	};
#ifdef OPENGL
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, cf);
	if (RECORDER.recording)
		RECORDER << new R_Material(GL_AMBIENT, c);
#endif
	return Unbound;
}

NodeP lsetMaterialDiffuse(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,
		float(c.a)/255
	};
#ifdef OPENGL
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, cf);
	if (RECORDER.recording)
		RECORDER << new R_Material(GL_DIFFUSE, c);
#endif
	return Unbound;
}

NodeP lsetMaterialSpecular(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,
		float(c.a)/255
	};
#ifdef OPENGL
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, cf);
	if (RECORDER.recording)
		RECORDER << new R_Material(GL_SPECULAR, c);
#endif
	return Unbound;
}

NodeP lsetMaterialEmission(PairPP arg)
{
	ThreadP th=Thread::This();
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	GraphLocker gl;
	float cf[4]=
	{	float(c.r)/255,
		float(c.g)/255,
		float(c.b)/255,
		float(c.a)/255
	};
#ifdef OPENGL
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, cf);
	if (RECORDER.recording)
		RECORDER << new R_Material(GL_EMISSION, c);
#endif
	return Unbound;
}

NodeP lsetMaterialShininess(PairPP arg)
{
	ThreadP th=Thread::This();
	float s=arg[1]->a_float(1);
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, s);
#endif
	if (RECORDER.recording)
		RECORDER << new R_Shininess(s);
	return Unbound;
}

NodeP lsetFogDensity(PairPP arg)
{
	float density=float(arg[1]->a_float(1));
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	glFogf(GL_FOG_DENSITY, density);
#endif
	return Unbound;
}

NodeP lsetFogRange(PairPP arg)
{
	float start=float(arg[1]->a_float(1));
	if throwing return Unbound;
	float end=float(arg[2]->a_float(2));
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	glFogf(GL_FOG_START, start);
	glFogf(GL_FOG_END, end);
#endif
	return Unbound;
}

NodeP lsetFogColor(PairPP arg)
{
	Color c=arg[1]->a_color(1);
	if throwing return Unbound;
	float col[4]={c.r/255,c.g/255,c.b/255,c.a/255};
	GraphLocker gl;
#ifdef OPENGL
	glFogfv(GL_FOG_COLOR, (float*)&col);
#endif
	return Unbound;
}

NodeP lsetFogMode(PairPP arg)
{
	int mode=arg[1]->a_int(1);
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	glFogi(GL_FOG_MODE, mode);
#endif
	return Unbound;
}

NodeP lPolyStart(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->polyStart();
	return Unbound;
}

NodeP lPolyEnd(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->polyEnd();
	return Unbound;
}

NodeP lPolygon(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP instructions=arg[1]->a_runnable(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->polyStart();
	return new ToDoNode(__polygon_continuation, instructions);
}


NodeP lTessStart(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->tessStart();
	return Unbound;
}

NodeP lTessContour(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->tessContour();
	return Unbound;
}

NodeP lTessEnd(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->tessEnd();
	return Unbound;
}

NodeP lTesselation(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP instructions=arg[1]->a_runnable(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->tessStart();
	return new ToDoNode(__tess_continuation, instructions);
}

NodeP lsetTessWindingRule(PairPP arg)
{
	ThreadP th=Thread::This();
	int rule=arg[1]->a_int(1);
	if throwing return Unbound;
	if (rule < 0 || rule > 4)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	GraphLocker gl;
#ifdef OPENGL
	TURTLE->tessWindingRule=GLU_TESS_WINDING_ODD+rule;
#endif
	if (RECORDER.recording)
		RECORDER << new R_TessWindingRule(rule);
	return Unbound;
}

NodeP lSurfaceStart(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->surfaceStart();
	return Unbound;
}

NodeP lSurfaceColumn(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->surfaceColumn();
	return Unbound;
}

NodeP lSurfaceEnd(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->surfaceEnd();
	return Unbound;
}

NodeP lSurface(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP instructions=arg[1]->a_runnable(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->surfaceStart();
	return new ToDoNode(__surface_continuation, instructions);
}

NodeP lenablePolySmooth(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_POLYGON_SMOOTH);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setPolySmooth(true);
	return Unbound;
}
NodeP ldisablePolySmooth(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_POLYGON_SMOOTH); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setPolySmooth(false); 
	return Unbound;
}

NodeP lenableLineSmooth(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_LINE_SMOOTH);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setLineSmooth(true);
	return Unbound;
}
NodeP ldisableLineSmooth(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_LINE_SMOOTH);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setLineSmooth(false);
	return Unbound;
}

NodeP lenablePointSmooth(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_POINT_SMOOTH);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setPointSmooth(true);
	return Unbound;
}
NodeP ldisablePointSmooth(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_POINT_SMOOTH);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setPointSmooth(false);
	return Unbound;
}

NodeP lenableDepthTest(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_DEPTH_TEST);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setDepthTest(true);
	return Unbound;
}
NodeP ldisableDepthTest(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_DEPTH_TEST); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setDepthTest(false);
	return Unbound;
}

NodeP lenableLighting(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_LIGHTING);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setLighting(true);
	return Unbound;
}
NodeP ldisableLighting(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_LIGHTING); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setLighting(false);
	return Unbound;
}

NodeP lenableDither(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_DITHER);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setDither(true);
	return Unbound;
}
NodeP ldisableDither(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_DITHER); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setDither(false);
	return Unbound;
}

NodeP lenableTexture(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_TEXTURE_2D); SCREEN->textured=true; 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setTextured(true); 
	return Unbound;
}

NodeP ldisableTexture(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_TEXTURE_2D); SCREEN->textured=false; 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setTextured(false); 
	return Unbound;
}

NodeP lenableTextureFont(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	SCREEN->textureFont=true; 
	if (RECORDER.recording)
		RECORDER << new R_setTextureFont(true); 
	return Unbound;
}

NodeP ldisableTextureFont(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	SCREEN->textureFont=false; 
	if (RECORDER.recording)
		RECORDER << new R_setTextureFont(false); 
	return Unbound;
}

NodeP lenableRoundLineEnds(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.roundLineEnds=true; 
	if (RECORDER.recording)
		RECORDER << new R_setPenRLE(*TURTLE); 
	return Unbound;
}
NodeP ldisableRoundLineEnds(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->pen.roundLineEnds=false; 
	if (RECORDER.recording)
		RECORDER << new R_setPenRLE(*TURTLE); 
	return Unbound;
}

NodeP lenableCylinderLines(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->cylinderLines=true; 
	if (RECORDER.recording)
		RECORDER << new R_setCyLines(*TURTLE); 
	return Unbound;
}
NodeP ldisableCylinderLines(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->cylinderLines=false; 
	if (RECORDER.recording)
		RECORDER << new R_setCyLines(*TURTLE); 
	return Unbound;
}

NodeP lenableShadows(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->shadows=true; 
	if (RECORDER.recording)
		RECORDER << new R_setShadows(*TURTLE); 
	return Unbound;
}
NodeP ldisableShadows(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
	TURTLE->shadows=false; 
	if (RECORDER.recording)
		RECORDER << new R_setShadows(*TURTLE); 
	return Unbound;
}

NodeP lenableFog(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_FOG); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setFogEnabled(true); 
	return Unbound;
}
NodeP ldisableFog(PairPP)
{	ThreadP th=Thread::This();
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_FOG); 
#endif
	if (RECORDER.recording)
		RECORDER << new R_setFogEnabled(false); 
	return Unbound;
}

NodeP lenableEdgeFlag(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
	glEdgeFlag(true); 
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_SetEdgeFlag(true); 
	return Unbound;
}
NodeP ldisableEdgeFlag(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
	glEdgeFlag(false); 
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_SetEdgeFlag(false); 
	return Unbound;
}

NodeP lenableBlend(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
	glEnable(GL_BLEND); 
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_SetBlend(true); 
	return Unbound;
}
NodeP ldisableBlend(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_BLEND); 
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_SetBlend(false); 
	return Unbound;
}

NodeP lenableAlphaTest(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
	glAlphaFunc(GL_GEQUAL, 1.0f/256);
	glEnable(GL_ALPHA_TEST); 
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_SetBlend(true); 
	return Unbound;
}
NodeP ldisableAlphaTest(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
	glDisable(GL_ALPHA_TEST); 
#endif
//	if (RECORDER.recording)
//		RECORDER << new R_SetBlend(false); 
	return Unbound;
}

NodeP lsetTexWrapping(PairPP arg)
{
	int wrapping=arg[1]->a_bool(1);
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	if (wrapping)
	{	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	}
	else
	{	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}
#endif
	return Unbound;
}

NodeP ldoubleBuffer(PairPP)
{	
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
	th->graph->doubleBuffer=true; 
	theApp->doubleBuffer=true; 
#ifdef OPENGL
	glReadBuffer(GL_BACK);
	glDrawBuffer(GL_BACK);
#endif
	return Unbound;
}
NodeP lsingleBuffer(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
	th->graph->doubleBuffer=false; 
	theApp->doubleBuffer=false; 
#ifdef OPENGL
	glReadBuffer(GL_FRONT);
	glDrawBuffer(GL_FRONT);
#endif
	return Unbound;
}

NodeP lsetDepthFunc(PairPP arg)
{
	ThreadP th=Thread::This();
	int f=arg[1]->a_int(1);
	if (f < 0 || f > 7)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	glDepthFunc(GL_NEVER+f);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setDepthFunc(f); 
	return Unbound;
}

NodeP lsetPolygonMode(PairPP arg)
{
	ThreadP th=Thread::This();
	int mode=arg[1]->a_int(1);
	if throwing return Unbound;
	if (mode < 0 || mode > 2)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	GraphLocker gl;
#ifdef OPENGL
	glPolygonMode(GL_FRONT_AND_BACK, GL_POINT+mode);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setPolygonMode(mode); 
	return Unbound;
}

NodeP lColorLogicOp(PairPP arg)
{
	ThreadP th=Thread::This();
	IntP mode=0;
	int m=0;
	GraphLocker gl;
#ifdef OPENGL
	glGetIntegerv(GL_LOGIC_OP_MODE, &m);
#else
	m=0x1503;
#endif
	mode=new Int(m);
	return mode;
}

NodeP lsetColorLogicOp(PairPP arg)
{
	ThreadP th=Thread::This();
	int mode=arg[1]->a_int(1);
	if throwing return Unbound;
#ifdef OPENGL
	mode-=GL_CLEAR;
#else
	mode-=0x1500;
#endif
	if (mode < 0 || mode > 0xf)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
#ifdef OPENGL
	GraphLocker gl;
	if (mode)
	{	glLogicOp(GL_CLEAR+mode);
		glEnable(GL_COLOR_LOGIC_OP);
	}else
		glDisable(GL_COLOR_LOGIC_OP);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setColorLogicOp(mode); 
	return Unbound;
}

NodeP lsetPointSize(PairPP arg)
{
	ThreadP th=Thread::This();
	int size=arg[1]->a_int(1);
	if throwing return Unbound;
	GraphLocker gl;
#ifdef OPENGL
	double range[2];
	glGetDoublev(GL_POINT_SIZE_RANGE, range);
	if (size < range[0] || size > range[1])
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	glPointSize(size);
#endif
	if (RECORDER.recording)
		RECORDER << new R_setPointSize(size); 
	return Unbound;
}

NodeP lTexture(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!arg[1])
	{	ldisableTexture(0);
	}
	else if (arg[1]->type==BITMAP_)
	{	BitmapP b=BitmapP(arg[1]);
		GraphLocker gl;
		TextureP tex=new Texture(b);
		if throwing return Unbound;
		SCREEN->textured=true; 
		if (RECORDER.recording)
			RECORDER << new R_Texture(tex); 
		return tex;
	}
	else if (arg[1]->type==TEXTURE)
	{	TextureP tex=TextureP(arg[1]);
		GraphLocker gl;
#ifdef OPENGL
		glBindTexture(GL_TEXTURE_2D, tex->_texnr);
#endif
		SCREEN->textured=true; 
		if (RECORDER.recording)
			RECORDER << new R_Texture(tex); 
	}
	else
		error(BAD_DATA, arg[1]);
	return Unbound;
}

NodeP lsetTexXY(PairPP arg) 
{
	ThreadP th=Thread::This();
	double 
		x = arg[1]->a_float(1),
		y = arg[2]->a_float(2);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->texpos=Point(x, y, 0);
	return Unbound;
}

NodeP lsetTexPos(PairPP arg) 
{
	ThreadP th=Thread::This();
	PairP v = arg[1]->a_vector(1);
	if throwing return Unbound;
	GraphLocker gl;
	TURTLE->texpos=Point(v->first()->getnumber(), v->second()->getnumber(), 0);
	return Unbound;
}

NodeP ldeleteTextures(PairPP)
{
	GraphLocker gl;
	if (R_Texture::maxnr)
	{	unsigned *textures=new unsigned[R_Texture::maxnr];
		if (textures)
		{	for (int i=1; i<=R_Texture::maxnr; i++)
				textures[i-1]=i;
#ifdef OPENGL
			glDeleteTextures(R_Texture::maxnr, textures);
#endif
			R_Texture::maxnr=0;
			delete textures;
		}
	}
	R_Texture::textures=0;
	return Unbound;
}

NodeP lVideoStart(PairPP arg)
{
	ThreadP th=Thread::This();
	StringP filename=arg[1]->a_string(1);
	if throwing return Unbound;
	int rate;
	if (exist(2))
		rate=arg[2]->a_int(2);
	else
		rate=30;
	if throwing return Unbound;
	GraphLocker gl;
	if (!th->graph)
		return Unbound;

	th->graph->videoStart(filename->ptr(), rate);
	return Unbound;
}

NodeP lVideoFrame(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
	th->graph->videoFrame();
	return Unbound;
}

NodeP lVideoEnd(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
	th->graph->videoEnd();
	return Unbound;
}


NodeP lGraphicStart(PairPP)
{
	GraphLocker gl;
	GraphP graph=Thread::This()->graph;
	if (!graph || graph->recorder != graph->recorderSave)
	{	error(BAD_DATA, new String(_("You must call GraphicEnd before a new GraphicStart!")));
		return Unbound;
	}
	graph->recorderSave=graph->recorder;
	Recorder *rec=new Recorder(graph);
	if throwing return Unbound;
	graph->recorder=rec;
	graph->turtle->recorder=graph->recorder;
	return Unbound;
}

NodeP lGraphicEnd(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	GraphP graph=Thread::This()->graph;
	if (!graph || graph->recorder == graph->recorderSave)
	{	error(BAD_DATA, new String(_("You must call GraphicStart before GraphicEnd!")));
		return Unbound;
	}
	Recorder *r=graph->recorder;
	graph->recorder=graph->recorderSave;
	graph->turtle->recorder=graph->recorder;

	if (RECORDER.recording)
		RECORDER << new R_Graphic(*TURTLE, r, 1);

	return new GraphicNode(r);
}

NodeP lGraphic(PairPP arg)
{
	NodeP instructions=arg[1]->a_runnable(1);
	if throwing return Unbound;
	GraphLocker gl;
	lGraphicStart(0);
	return new ToDoNode(__graphic_continuation, instructions);
}

NodeP ldrawGraphic(PairPP arg)
{
	ThreadP th=Thread::This();
	GraphicNodeP g=arg[1]->a_graphicNode(1);
	double scale=1;
	if (exist(2))
	{	scale=arg[2]->a_float(2);
	}
	if throwing return Unbound;
	GraphLocker gl;
	g->scale=scale;
	g->draw(TURTLE);
	if (RECORDER.recording)
		RECORDER << new R_Graphic(*TURTLE, g->recorder, scale);
	return Unbound;
}

NodeP lcastShadows(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	ShadowObject *so=Thread::This()->graph->shadowObject;
	for (int i=0;  i<so->nPlanes;  i++)	// Loop Through All Object Planes
		so->calcPlane(&(so->planes[i]));	// Compute Plane Equations For All Faces

//	so->setConnectivity();	// O(n^2), so don't use it

#ifdef OPENGL
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);
#endif
	so->castShadows(TURTLE->lightPos, 1);
#ifdef OPENGL
	glDisable(GL_CULL_FACE);
#endif
	return Unbound;
}

NodeP lclearShadows(PairPP)
{
	GraphLocker gl;
	Thread::This()->graph->shadowObject->clear();
	return Unbound;
}

NodeP lglReleaseContext(PairPP)
{
	GraphLocker gl;
#ifdef OPENGL
#ifdef WIN32
	wglMakeCurrent(0,0);
#else
	glXMakeCurrent((Display*)wxGetDisplay(), 0,0);
#endif
#endif
	return Unbound;
}

NodeP lglMakeCurrent(PairPP arg)
{
	GraphLocker gl;
	GraphP gr=0;
	if (!arg[1])
		gr=Thread::This()->graph;
	else
	{	GraphNodeP graphnd;
		graphnd=arg[1]->a_graph(1);
		if throwing return Unbound;
		gr=GraphP(graphnd->window);
	}
#ifdef OPENGL
#ifdef WIN32
	wglMakeCurrent(
		(HDC)gr->GetHDC(), 
		gr->GetContext()->GetGLRC());
#else
	gr->SetCurrent();
#endif
#endif
	return Unbound;
}

NodeP lGraphEnter(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
	Thread::This()->graph->enter();
	return Unbound;
}

NodeP lGraphLeave(PairPP)
{
	ThreadP th=Thread::This();
	GraphLocker gl;
	if (!th->graph)
		return Unbound;
	Thread::This()->graph->leave();
	return Unbound;
}
