#include "stdafx.h"

#include <wx/listbox.h>
#include <wx/textctrl.h>
#include <wx/frame.h>

#include "Main.h"
#include "UInt8.h"
#include "File.h"
#include "Color.h"
#include "Console.h"
#include "Consolew.h"
#include "Error.h"
#include "History.h"

class HistoryBox: public wxListBox
{
public:
	HistoryBox(wxWindow *parent, wxWindowID id, 
		const wxPoint& pos, const wxSize& size, 
		const wxArrayString& choices, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = _("comboBox")):
	wxListBox(parent, id, pos, size, choices, 
		style, validator, name)
	{
	}

	void OnChar   (wxKeyEvent &event);	// pass chars to ComboBox
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp  (wxKeyEvent &event);
	void OnSelect(wxCommandEvent &event);
	void OnChange(wxCommandEvent &event);
	void OnEnter (wxCommandEvent &event);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(HistoryBox,wxListBox)
	EVT_CHAR		(HistoryBox::OnChar		)
END_EVENT_TABLE()

void HistoryBox::OnChar(wxKeyEvent &event)
{
	int key=event.GetKeyCode();
	switch (key)
	{
		case WXK_ESCAPE:
			GetParent()->Show(false); 
			if (console)
				console->SetFocus();
			else if (consolew)
				consolew->SetFocus();
			break;

		case WXK_RETURN:
		{	wxString val=GetStringSelection();
			GetParent()->Show(false);
			if (console)
			{	console->SetFocus();
				int index=console->GetInsertionPoint();
				long x,y;
				console->PositionToXY(index, &x, &y);
				index=console->XYToPosition(0,y);
				console->SetInsertionPoint(index);
				console->WriteText(val+'\n');
				console->finishLine(val+'\n');
			}else if (consolew)
			{	consolew->SetFocus();
				COORD coord=consolew->coord0;
				consolew->SetInsertionCoord(0, coord.Y);
				consolew->WriteText(val);
				consolew->finishLine(val);
			}
			break;
		}
		case WXK_SPACE:
		{	wxString val=GetStringSelection();
			GetParent()->Show(false);
			if (console)
			{	console->SetFocus();
				int index=console->GetInsertionPoint();
				long x,y;
				console->PositionToXY(index, &x, &y);
				index=console->XYToPosition(0,y);
				console->SetInsertionPoint(index);

				console->WriteText(val+'\n');
				console->SetInsertionPoint(index);
			}else if (consolew)
			{	consolew->SetFocus();
				COORD coord=consolew->coord0;
				consolew->SetInsertionCoord(0, coord.Y);
				consolew->WriteText(val);
			}
			break;
		}
		default:	
			event.Skip();
	}
}

History::History(Console *parent, wxString filename_)
{
	filename=filename_;
	win=new wxFrame(parent, -1, _("History"),
		wxDefaultPosition, wxSize(600,200), 
		wxNO_BORDER, _("History"));
	if (!win)
	{	error(OUT_OF_MEM); return;
	}
	win->Show(false);
	box=new HistoryBox(win, -1, 
		wxDefaultPosition, wxSize(600,200), 0, 
		wxLB_SINGLE, 
		wxDefaultValidator, _("HistoryBox"));
	if (!box)
	{	error(OUT_OF_MEM); return;
	}
#ifdef WIN32
	int fontSize=10;
#else
	int fontSize=10;
#endif
	box->SetFont(wxFont(fontSize, wxMODERN, wxNORMAL, wxNORMAL, false, "Courier New"));
	load();
}

History::~History()
{
	save();
}

void History::load()
{
	if (!wxFileExists(filename))
		return;
	wxFileInputStream fin(filename);
	TextInputStream in(fin);
	wxString line;
	while (!fin.Eof())
	{	line=in.ReadLine();
		if (line != wxEmptyString)
			box->Append(line);
	}
}

void History::save()
{
	wxFileOutputStream fout(filename);
	TextOutputStream out(fout);
	unsigned int i;
	for (i=0; i<box->GetCount(); i++)
		out << box->GetString(i) << endL;
}

void History::show(int y, wxString line)
{
#ifdef WIN32
	int fontSize=18;
#else
	int fontSize=24;
#endif
	int h=box->GetCount()*fontSize, x=0;
	if (h > 200)
		h=200;
	if (console)
		console->ClientToScreen(&x,&y);
	else if (consolew)
		consolew->ClientToScreen(&x,&y);

	y=y-h+4;
	if (y < 0)
		y=0;
	win->Move(x-1, y);
	if (console)
		win->SetSize(console->GetSize().GetWidth()-2, h);
	else if (consolew)
		win->SetSize(consolew->GetSize().GetWidth()-2, h);//TODO
	if (box->GetCount() != 0)
		box->SetSelection(box->GetCount()-1);
	win->Show();
	box->SetFocus();
}

void History::append(wxString line)
{
	line=line.Before('\n');
	if (line==wxEmptyString)
		return;
	int count=box->GetCount();
	if (count == 0)
		box->Append(line);
	else 
	{	wxString old=box->GetString(count-1);
		if (line!=old)
		{	if (count >= 2)
			{	old=box->GetString(count-2);
				if (line!=old)
					box->Append(line);
			}else
				box->Append(line);
		}
	}
}
