#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "String.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Pair.h"
#include "File.h"
#include "Init.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

Int16Array::Int16Array(int length_, int origin_): Node(INT16ARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (shortP) GC_::calloc(length_+1, sizeof(short));
	if (!_data) 
		return;
	_length=length_;
}

Int16Array::Int16Array(Int16ArrayP array_): Node(INT16ARRAY)
{
	_length=0;
	_origin = array_->_origin;
	_data = (shortP) GC_::calloc(array_->_length+1, sizeof(short));
	if (!_data) 
		return;
	_length = array_->_length;
	shortP source = array_->_data, 
		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void Int16Array::free() {GC_::free(_data); _data=0;}

DataInputStream &Int16Array::read(DataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (shortP) GC_::calloc(sizeof(short), length+1);
	if (!_data) 
		return istr;
	_length=length;
	shortP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (short&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Int16Array::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	shortP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

NodeP Int16Array::copy(PtrToInt_ &ptrToInt)
{
	Int16ArrayP arr=new Int16Array(this);
	return arr;
}

StringP Int16Array::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*(Thread::This()->stringptr)='\0';
	return new String(textbuf);
}

void Int16Array::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	int i=0, wid=0;
	shortP pp=data();

	if (width >= 2*_length-10)
		wid=width/_length;

	if (noBrackets <= 0)
		::print(strm, width, "{");

	while (i < _length) 
	{	int w=wid;
		::print(strm, w, int(*pp++));	
		if (++i < length())
			::print(strm, w, ' ');
		width-=wid-w;
	}
	if (i < length()) 
		::print(strm, width, "...");
	if (noBrackets <= 0)
		::print(strm, width, '}');

	if (origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", origin());
		::print(strm, width, org);
	}
}

void Int16Array::printBin(FILEP strm)
{
	ThreadP th=Thread::This();
	int wid=length();
	shortP pp=data();
	if (!strm)
	{	if (th->stringlen > 0)
			th->stringlen-=sizeof(short)*wid;
		if (th->stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*shortP(th->stringptr) = *pp++;
				th->stringptr+=sizeof(short);
			}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print int16array to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(data());
		int width=Int::max;	
		for (int l=sizeof(int)*wid; l; l--)
			::print(strm, width, *cp++);
		return;
	}
	fwrite(pp, sizeof(short), wid, strm);
}

int Int16Array::count() {return length();}

int Int16Array::_sizeof() {return length()*sizeof(short);}

int Int16Array::compare(NodeP nd)
{	if (nd->type != INT16ARRAY) 
		return Node::boolerror;
	if (this==nd)
		return 0;

	Int16ArrayP b=Int16ArrayP(nd);
	shortP dp=_data, dpb=b->_data;
	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{	int c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  1 : -1);
	}
	if (!i & !j)
		return 0;
	if (i)
		return 1;
	return -1;
}

bool Int16Array::equalP(NodeP b)
{
	if (b->type != INT16ARRAY) 
		return false;
	if (this==b)
		return true;

	Int16ArrayP ba=Int16ArrayP(b);
	shortP dp=_data, dpb=ba->_data;
	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	if (!i & !j)
		return true;
	return false;
}

NodeP Int16Array::first_() 
{
	if (length() >= 1) 
		return new Int16(*(Int16ArrayP(this)->data()));
	else
		return 0;
}

NodeP Int16Array::butFirst()
{
	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l, o);
	if throwing return Unbound;
	shortP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Int16Array::last()
{
	return new Int16(*(data()+length()-origin()));
}

NodeP Int16Array::butLast()
{
	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l, o);
	if throwing return Unbound;
	shortP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Int16Array::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l+1, o);
	if throwing return Unbound;
	shortP source=data(), dest=result->data();
	*dest++=Int16::clip(nd->getnumber());
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP Int16Array::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	Int16ArrayP result=new Int16Array(l+1, o);
	shortP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=Int16::clip(nd->getnumber());
	return result;
}

double Int16Array::_min(double d)
{	shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	short n=*dp++;
		if (n < d)
			d=n;
	}
	return d;
}

double Int16Array::_max(double d)
{	shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	short n=*dp++;
		if (n > d)
			d=n;
	}
	return d;
}

NodeP Int16Array::Norm()
{	double d=0;  shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	int h = (*dp++);
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}

NodeP Int16Array::maxNorm()
{	int d=0;  shortP dp=_data;
	for (int i=_length;  i;  i--)
	{	short h = (*dp++);
		if (throwing) return Unbound;
		short ha=::abs(h);
		if (ha > d) d = ha;
	}
	return new Int(d);
}

NodeP Int16Array::resize(int newsize)
{	Int16ArrayP r=new Int16Array(newsize);
	if throwing return Unbound;
	shortP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	int diff = d[i+1]-d[i];
					double slope = double(diff)/di;

					for (int ii=di-1;  ii;  ii--)
						r->_data[ni+ii] = short(d[i]+slope*ii);
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=short(d[int(i)]+frac*diff+0.5);
			}
		}
	}
	return r;
}

NodeP Int16Array::lowPassFilter(double weight)
{
	Int16ArrayP result=new Int16Array(length());
	if throwing return Unbound;
	shortP source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=short(median);
	}
	Float::reset_FPU();
	return result;
}

NodeP Int16Array::saturateAbove(double limit) 
{
	Int16ArrayP result=new Int16Array(length());
	if throwing return Unbound;
	shortP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	short s=*source++;
		*dest++= s>limit? short(limit) : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP Int16Array::saturateBelow(double limit) 
{
	Int16ArrayP result=new Int16Array(length());
	if throwing return Unbound;
	shortP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	short s=*source++;
		*dest++= s<limit? short(limit) : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP Int16Array::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Int16(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}

NodeP Int16Array::items(NodeP start, NodeP end)
{
	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	Int16ArrayP newar=new Int16Array(e-s+1);
					if throwing return Unbound;
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP Int16Array::setitem(NodeP index, NodeP newitem, bool safe)
{
	int i=index->a_int(1);
	short ni=newitem->a_int16(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}

NodeP Int16Array::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	shortP d=_data;   int l=_length;
	Int16P i=new Int16(0);
	if throwing return Unbound;

	while (--l >= 0 && !throwing) 
	{	i->int_=(*d++);
		if (subobj->equalP(i))
			return _True;
	}
	return _False;
}

NodeP Int16Array::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 

	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					shortP d=&_data[i];
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =  Int16::clip(IntP(h)->int_);	break;
							case FLOAT_:*d++ =Int16::clip(FloatP(h)->float_);break;
							case NINT16:	*d++ =            Int16P(h)->int_;	break;
							case NINT8:	*d++ =             Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = Int16::clip(h->getnumber());
											else
											{	error(BAD_DATA, n, 3); 
												return IntArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);	
								return IntArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->data();
					shortP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int16::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	shortP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)

			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					shortP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int16::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP Int16Array::reverse()
{
	Int16ArrayP result=new Int16Array(_length);
	if throwing return Unbound;

	shortP d=_data, rev=result->_data+_length-1;
	for (int i=_length; i>0; i--)
		*rev-- = *d++;
	return result;
}

NodeP Int16Array::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	Int16ArrayP val = new Int16Array(length(), origin());
	if throwing return Unbound;
	shortP
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}

NodeP Int16Array::shuffle()
{	Int16ArrayP result=new Int16Array(Int16ArrayP(this));
	if throwing return Unbound;
	shortP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		short tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

PairP Int16Array::list()
{
	shortP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
	{	p = new Pair(new Int16(*--d), p);
		if throwing return PairP(Unbound);
	}
	return p;
}

ArrayP Int16Array::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	if throwing return ArrayP(Unbound);
	NodePP d = array->data();
	shortP s=_data;

	for (int i=_length;  i;  i--)
	{	*d++ = new Int16(*s++);
		if throwing return ArrayP(Unbound);
	}
	return array;
}

IntArrayP Int16Array::intarray(int origin) 
{
	IntArrayP intarray = new IntArray(_length, _origin);
	if throwing return IntArrayP(Unbound);
	intP d = intarray->data();
	shortP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return intarray;
}

Int16ArrayP Int16Array::int16array(int origin) 
{
	return new Int16Array(Int16ArrayP(this));
}

FloatArrayP Int16Array::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	if throwing return FloatArrayP(Unbound);
	doubleP d = floatarray->data();
	shortP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return floatarray;
}

void Int16Array::combSortInPlace()
{
	double gap=_length; //initialize gap size
	int swaps=0; //see bubblesort for an explanation
    
	while (!(gap<=1 && swaps==0))
	{	//update the gap value for a next comb
		if (gap > 1)
		{	gap/=1.3;
			if (gap==10 || gap==9)
				gap=11;
		}
		int i=0;
		//a single "comb" over the input list
		while (i+gap<_length)
		{	int j=int(i+gap);
			if (_data[i] > _data[j])
			{	short tmp=_data[i];
				_data[i]=_data[j];
				_data[j]=tmp;
			}
			i++;
		}
	}
}

NodeP Int16Array::combSort()
{
	Int16ArrayP a=new Int16Array(this);
	if throwing return Unbound;
	a->combSortInPlace();
	return a;
}

NodeP Int16Array::readbin(FileP file)
{
	ThreadP th=Thread::This();
	shortP ip=data();

	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read int16array binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(short), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(short)*_length;
			if (th->rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *shortP(th->rstringptr);	// reading from string
					th->rstringptr+=sizeof(short);
				}
			else
				error(FILE_ERROR, new String
					(_("memory block too small to read int16array")));
		}
	return this;
}
