#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "String.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Pair.h"
#include "File.h"
#include "Bitmap.h"
#include "Color.h"
#include "Init.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

IntArray::IntArray(int length_, int origin_): Node(INTARRAY)
{
	_origin=origin_;
	_length=0;
	_data = (intP) GC_::calloc(length_+1, sizeof(int));
	if (_data) 
		_length=length_;
}

IntArray::IntArray(IntArrayP array_): Node(INTARRAY)
{
	_length= 0;
	_origin = array_->_origin;
	_data = (intP) GC_::calloc(array_->_length+1, sizeof(int));
	if (!_data) 
	{	_length=0;
		return;
	}
	_length= array_->_length;
	intP source = array_->_data, 

		  destination = _data;

	for (int i=_length;  i;  i--)
		*destination++ = *source++;
}

void IntArray::free() {GC_::free(_data); _data=0;}

DataInputStream &IntArray::read(DataInputStream &istr) 
{	
	int length;
	istr >> length >> _origin;
	_length=0;
	_data = (intP) GC_::calloc(sizeof(int), length+1);
	if (!_data) 
		return istr;
	_length=length;
	intP itemptr=_data;
	for (int i=_length;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &IntArray::write(wxDataOutputStream &ostr) 
{
	ostr << type << _length << _origin;
	intP itemptr=_data;
	for (int i=_length;  i;  i--)
		ostr << *itemptr++;
	return ostr;
}

NodeP IntArray::copy(PtrToInt_ &ptrToInt)
{
	IntArrayP arr=new IntArray(this);
	return arr;
}

StringP IntArray::string()
{
 	const int maxlength=65536;
	char textbuf[maxlength];
	Thread::This()->stringptr=textbuf;
	Thread::This()->stringlen=maxlength-1;
	stringbuffer << NodeP(this);
	*(Thread::This()->stringptr)='\0';
	return new String(textbuf);
}

void IntArray::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	int i=0, wid=0;
	intP pp=data();
						
	if (width >= 2*_length-10)
		wid=width/_length;

	if (noBrackets <= 0)
		::print(strm, width, '{');

	while (i < _length) 
	{	int w=wid;
		::print(strm, w, *pp++);	
		if (++i < _length)
			::print(strm, w, ' ');
		width-=wid-w;
	}
	if (i < _length) 
		::print(strm, width, "...");
	if (noBrackets <= 0)
		::print(strm, width, '}');

	if (origin() != 1) 
	{	char org[]="@            ";
		sprintf(&org[1], "%d", origin());
		::print(strm, width, org);
	}
}

void IntArray::printBin(FILEP strm)
{
	ThreadP th=Thread::This();
	int wid=length();
	intP pp=data();
	if (!strm)
	{	if (th->stringlen > 0)     // printing to string
			th->stringlen-=sizeof(int)*wid;
		if (th->stringlen >= 0)     // printing to string
			for (int l=wid;  l;  l--)
			{	*intP(th->stringptr) = *pp++;
				th->stringptr+=sizeof(int);
			}
		else
			error(FILE_ERROR, new String
				(_("memory block too small to print intarray to")));
		return;
	}
	if (th->interactive && strm==File::conout)
	{	charP cp=charP(data());
		int width=Int::max;
		for (int l=sizeof(int)*wid; l; l--)
			::print(strm, width, *cp++);
		return;
	}
	fwrite(pp, sizeof(int), wid, strm);
}

int IntArray::count() {return length();}

int IntArray::_sizeof() {return length()*sizeof(int);}

int IntArray::compare(NodeP nd)
{	if (nd->type != INTARRAY) 
		return Node::boolerror;

	if (this==nd)
		return 0;

	IntArrayP b=IntArrayP(nd);
	intP dp=_data, dpb=b->_data;

	int i,j;
	for (i=length(), j=b->length();  i && j;  i--, j--)
	{
		int c=(*dp++) - (*dpb++);
		if (c)
			return (c > 0 ?  1 : -1);
	}
	if (!i & !j)
		return 0;

	if (i)

		return 1;

	return -1;
}

bool IntArray::equalP(NodeP b)
{	if (b->type != INTARRAY) 
		return false;

	if (this==b)
		return true;

	IntArrayP ba=IntArrayP(b);
	intP dp=_data, dpb=ba->_data;

	int i,j;
	for (i=length(), j=ba->length();  i && j;  i--, j--)
		if ((*dp++) - (*dpb++))
			return false;
	
	if (!i & !j)
		return true;

	return false;
}

NodeP IntArray::first_() 
{	if (length() >= 1) 
		return new Int(*(IntArrayP(this)->data()));
	else
		return 0;
}

NodeP IntArray::butFirst()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l, o);
	if throwing return Unbound;
	intP source=data()+1, dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP IntArray::last()
{	return new Int(*(data()+length()-origin()));
}

NodeP IntArray::butLast()
{	int l=length()-1, o=origin();
	if (l < 1)
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l, o);
	if throwing return Unbound;
	intP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP IntArray::fPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l+1, o);
	if throwing return Unbound;
	intP source=data(), dest=result->data();
	*dest++=Int::clip(nd->getnumber());
	while (l--)
	{	*dest++=*source++;
	}
	return result;
}

NodeP IntArray::lPut(NodeP nd)
{	int l=length(), o=origin();
	if (l < 1 || !nd || !nd->numberP())
	{	error(BAD_DATA, this); return Unbound;
	}
	IntArrayP result=new IntArray(l+1, o);
	if throwing return Unbound;
	intP source=data(), dest=result->data();
	while (l--)
	{	*dest++=*source++;
	}
	*dest=Int::clip(nd->getnumber());
	return result;
}

double IntArray::_min(double d)
{	intP dp=_data;
	for (int i=_length;  i;  i--)
	{	int n=*dp++;
		if (n < d)
			d=n;
	}
	return d;
}

double IntArray::_max(double d)
{	intP dp=_data;
	for (int i=_length;  i;  i--)
	{	int n=*dp++;
		if (n > d)
			d=n;
	}
	return d;
}

NodeP IntArray::Norm()
{	double d=0;  intP dp=_data;
	for (int i=_length;  i;  i--)
	{	double h = (*dp++);
		if (throwing) return Unbound;
		d+= h*h;
	}
	return new Float(::sqrt(d));
}

NodeP IntArray::maxNorm()
{	int d=0;  intP dp=_data;
	for (int i=_length;  i;  i--)
	{	int h = (*dp++);
		if (throwing) return Unbound;
		int ha=::abs(h);
		if (ha > d) d = ha;
	}
	return new Int(d);
}

NodeP IntArray::resize(int newsize)
{	IntArrayP r=new IntArray(newsize);
	if throwing return Unbound;
	intP d=_data;
	int i=length();
	r->_data[newsize-1]=d[i-1];
	if (i > 1)
	{	if (newsize >= i)
		{	int oi=newsize-1;
			i--;
			double f=double(newsize)/i;
			while (i--)
			{	int ni=int(i*f-0.5),  di=oi-ni;
				r->_data[ni]=d[i];
				if (d[i] && di-1 > 0)
				{	int diff = d[i+1]-d[i];
					double slope = double(diff)/di;

					for (int ii=di-1;  ii;  ii--)


						r->_data[ni+ii] = int(d[i]+slope*ii);
				}
				oi=ni;
			}
		}else
		{	int ni=newsize-1;
			i--;
			double f=double(newsize-1)/i;
			while (ni--)
			{	double i=ni/f,
					diff=d[int(i-0.5)]-d[int(i-0.5)-1],
					frac=fmod(i,1);
				r->_data[ni]=int(d[int(i)]+frac*diff+0.5);
			}
		}

	}
	return r;
}

NodeP IntArray::lowPassFilter(double weight)
{
	IntArrayP result=new IntArray(length());
	if throwing return Unbound;
	intP source=data(), dest=result->data();
	double median=*source;
	for (int i=length(); i>0; i--)
	{	median=(median*weight+(*source++))/(weight+1);
		*dest++=int(median);
	}
	Float::reset_FPU();
	return result;
}

NodeP IntArray::saturateAbove(double limit) 
{
	IntArrayP result=new IntArray(length());
	if throwing return Unbound;
	intP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	int s=*source++;
		*dest++= s>limit? int(limit) : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP IntArray::saturateBelow(double limit) 
{
	IntArrayP result=new IntArray(length());
	if throwing return Unbound;
	intP source=data(), dest=result->data();
	for (int i=length(); i>0; i--)
	{	int s=*source++;
		*dest++= s<limit? int(limit) : s;
	}
	Float::reset_FPU();
	return result;
}

NodeP IntArray::item(int i)
{	if (i >= 0  &&  i < _length) 
		return new Int(_data[i]);
	error(BAD_DATA, new Int(i)); return Unbound;
}

NodeP IntArray::items(NodeP start, NodeP end)
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_-_origin, e=IntP(end)->int_-_origin;
			if (s >= 0  &&  s < _length)
				if (e >= 0  &&  e < _length) 
				{	IntArrayP newar=new IntArray(e-s+1);
					if throwing return Unbound;
					for (int i=s, j=0;  i<=e;  i++, j++)
						newar->_data[j]=_data[i];
					return newar;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP IntArray::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1),
		ni=newitem->a_int(3);
	if throwing return Unbound;
	i-=origin();

	if ((i < 0 || i >= length()) && !throwing) 
		error(BAD_DATA, index);

	if (!throwing) 
	{	_data[i] = ni;
	}
	return Unbound;
}

NodeP IntArray::memberP(NodeP subobj, bool notp, bool substr)
{
	if (notp) error(BAD_DATA_UNREC, this);
	if (substr) return _False;

	intP d=_data;   int l=_length;
	IntP i=new Int(0);
	if throwing return Unbound;

	while (--l >= 0 && !throwing) 
	{	i->int_=(*d++);
		if (subobj->equalP(i))
			return _True;
	}
	return _False;
}

NodeP IntArray::setitems(NodeP start, NodeP newitems)
{	int i=start->a_int(1);
	if throwing return Unbound;
	if (!throwing) 
	{	i-=origin();

		if ((i < 0 || i >= length()) && !throwing) 
			error(BAD_DATA, start);

		if (!throwing) 
		{	if (type_of(newitems)==ARRAY)
			{	ArrayP n=ArrayP(newitems);
				if (i+n->length() <= _length)
				{	NodePP s=n->data();
					intP d = _data+i;
					
					for (int j=n->length();  j;  j--)
					{	NodeP h=*s;
						switch (type_of(h))
						{	case INT_:	*d++ =             IntP(h)->int_;	break;
							case FLOAT_:*d++ = Int::clip(FloatP(h)->float_);break;
							case NINT16:	*d++ =           Int16P(h)->int_;	break;
							case NINT8:	*d++ =            Int8P(h)->int_;	break;
							case STRING:if (h->numberP())
												*d++ = Int::clip(h->getnumber());
											else
											{	error(BAD_DATA, n, 3); 
												return IntArrayP(Unbound);
											}	
											break;
							default:
								error(BAD_DATA, n, 3);
								return IntArrayP(Unbound);
						}
						s++;
					}
				}else
					error(BAD_DATA, newitems, 3);
			}
			else if (type_of(newitems)==INTARRAY)
			{	IntArrayP n=IntArrayP(newitems);
				if (i+n->length() <= _length)
				{	intP s=n->_data, d=&_data[i];
					for (int j=n->_length;  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==INT16ARRAY)
			{	Int16ArrayP n=Int16ArrayP(newitems);
				if (i+n->length() <= _length)
				{	shortP s=n->data();
					intP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = *s++;
				}else
					error(BAD_DATA, newitems, 3);
			}else if (type_of(newitems)==FLOATARRAY)
			{	FloatArrayP n=FloatArrayP(newitems);
				if (i+n->length() <= _length)
				{	doubleP s=n->data();
					intP d=&_data[i];
					for (int j=n->length();  j;  j--) 
						*d++ = Int::clip(*s++);
				}else
					error(BAD_DATA, newitems, 3);
			}else
				error(BAD_DATA, newitems, 3);
		}
	}
	return Unbound;
}

NodeP IntArray::reverse()
{
	IntArrayP result=new IntArray(_length);
	if throwing return Unbound;

	intP d=_data, rev=result->_data+_length-1;
	for (int i=_length; i>0; i--)
		*rev-- = *d++;
	return result;
}

NodeP IntArray::rotate(int places)
{
	places=Int::mod(places, length());
	if (!places) return this;

	IntArrayP val = new IntArray(length(), origin());
	if throwing return Unbound;
	intP
		a= _data,
		b= & val->_data[places];

	int i;
	for (i=length()-places;  i;  i--)
		*b++ = *a++;

	b=val->_data;
	for (i=places;  i;  i--)
		*b++ = *a++;

	return val;
}

NodeP IntArray::shuffle()
{
	IntArrayP result=new IntArray(IntArrayP(this));
	if throwing return Unbound;
	intP d=result->data();
	int l=length();
	for (int i=l;  i;  i--)
	{	int
			j=Int::random(l), 
			k=Int::random(l);
		int tmp=d[j];
		d[j]=d[k];
		d[k]=tmp;
	}
	return result;
}

PairP IntArray::list()
{
	intP d = &_data[_length];
	PairP p=0;

	for (int i=_length;  i;  i--)
	{	p = new Pair(new Int(*--d), p);
		if throwing return PairP(Unbound);
	}
	return p;
}

ArrayP IntArray::array(int origin) 
{
	ArrayP array = new Array(_length, _origin);
	if throwing return ArrayP(Unbound);
	NodePP d = array->data();
	intP s=_data;

	for (int i=_length;  i;  i--)
	{	*d++ = new Int(*s++);
		if throwing return ArrayP(Unbound);
	}
	return array;
}

IntArrayP IntArray::intarray(int origin) 
{
	return new IntArray(IntArrayP(this));
}

Int16ArrayP IntArray::int16array(int origin) 
{
	Int16ArrayP int16array = new Int16Array(_length, _origin);
	if throwing return Int16ArrayP(Unbound);
	shortP d = int16array->data();
	intP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = Int16::clip(*s++);

	return int16array;
}

FloatArrayP IntArray::floatarray(int origin) 
{
	FloatArrayP floatarray = new FloatArray(_length, _origin);
	if throwing return FloatArrayP(Unbound);
	doubleP d = floatarray->data();
	intP s=_data;

	for (int i=_length;  i;  i--)
		*d++ = *s++;

	return floatarray;
}

BitmapP IntArray::bitmap(int width)
{
	if (width==0 || width>_length)
	{	error(BAD_DATA, new Int(width), 2);	return BitmapP(Unbound);
	}
	int height=_length/width;
	BitmapP bm=new Bitmap(width, height, width, height, 0);
	if throwing return BitmapP(Unbound);
	bm->pixels=ColorP(GC_::malloc(width*height*sizeof(Color)));
	if throwing return BitmapP(Unbound);
	ColorP d=bm->pixels;
	intP s=_data;

	for (int i=width*height;  i;  i--)
		*d++ = *s++;
	return bm;
}

void IntArray::combSortInPlace()
{
	double gap=_length; //initialize gap size
	int swaps=0; //see bubblesort for an explanation
    
	while (!(gap<=1 && swaps==0))
	{	//update the gap value for a next comb
		if (gap > 1)
		{	gap/=1.3;
			if (gap==10 || gap==9)
				gap=11;
		}
		int i=0;
		//a single "comb" over the input list
		while (i+gap<_length)
		{	int j=int(i+gap);
			if (_data[i] > _data[j])
			{	int tmp=_data[i];
				_data[i]=_data[j];
				_data[j]=tmp;
			}
			i++;
		}
	}
}

NodeP IntArray::combSort()
{
	IntArrayP a=new IntArray(this);
	if throwing return Unbound;
	a->combSortInPlace();
	return a;
}

NodeP IntArray::readbin(FileP file)
{
	ThreadP th=Thread::This();
	intP ip=data();

	if (th->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read intarray binary from keyboard")));
		return Unbound;
	}
	if (file->file())
	{	fread(ip, sizeof(int), _length, file->file());
		if (file->eof()) return 0;
	}else
		if (th->rstringlen > 0)
		{	th->rstringlen-=sizeof(int)*_length;
			if (th->rstringlen >= 0)
				for (int i=_length;  i;  i--)
				{	*ip++ = *intP(th->rstringptr);	// reading from string
					th->rstringptr+=sizeof(int);
				}
			else
				error(FILE_ERROR, new String
					(_("memory block too small to read intarray")));
		}
	return this;
}
