//___________________________________________________________________
//
//                         Class IntToPtr
//
//                           Definitions
//___________________________________________________________________

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <stdio.h>
#include <malloc.h>

#include <wx/datstrm.h>
#include <wx/slider.h>
#include <wx/button.h>
#include <wx/tglbtn.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>
#include <wx/frame.h>
#include <wx/sizer.h>
#include <wx/glcanvas.h>

#include <m_apm.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"
#include "Node.h"
#include "String.h"
#include "Float.h"
#include "Complex.h"
#include "File.h"
#include "Pair.h"
#include "Procedur.h"
#include "Object.h"
#include "Table.h"
#include "Struct.h"
#include "List.h"
#include "Quote.h"
#include "Colon.h"
#include "Tree.h"
#include "Runparsd.h"
#include "WindowPair.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "BigFloat.h"
#include "Point.h"
#include "Array.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"
#include "Slider.h"
#include "Gauge.h"
#include "IntControl.h"
#include "FloatControl.h"
#include "Button.h"
#include "ToggleButton.h"
#include "CheckBox.h"
#include "RadioButton.h"
#include "StaticText.h"
#include "TextControl.h"
#include "ListBox.h"
#include "ChoiceBox.h"
#include "ComboBox.h"
#include "ListControl.h"
#include "Frame.h"
#include "Graph.h"
#include "BoxSizer.h"

#include "Color.h"
#include "Bitmap.h"
#include "Texture.h"
#include "TurtleNode.h"
#include "GraphicNode.h"
#include "DL.h"
#include "Eval.h"
#include "IntToPtr.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)

	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

IntToPtr::IntToPtr(int size_):
	size(0), nr(0)
{
	data = (PtrIntPairP*)(GC_::calloc(size_, sizeof(PtrIntPairP)));
	if (data)
		size=size_;
}

IntToPtr::~IntToPtr()
{
	for (int i=0; i < size; i++)
	{	if (data[i])
			delete data[i];
	}
	GC_::free(data);
}

int IntToPtr::hash(int i)
{
	unsigned int index = (unsigned int)i;
	if (size)
		index %= size;
	else
		index=0;
	return index;
}

void *IntToPtr::find(int i)
{
	PtrIntPairP obj;
	for (obj=data[hash(i)];  obj;  obj=obj->tail)
		if (obj->i==i)
			return obj->p;
	return 0;
}

void IntToPtr::insert(void *p)
{
	nr++;
	PtrIntPairP &obj=data[hash(nr)];
	obj = new PtrIntPair(p, nr, obj);
}

wxDataInputStream &IntToPtr::readNodes(wxDataInputStream &istr)
{
	GC_::disabled=true;
	int k=0;
	while (!istr.Eof())
	{	int i;
		istr >> i;
//		if (istr.Eof())
//			break;
/*		if (!istr.CanRead()) 
		{	//error(FILE_ERROR, new String(
			//	"File bad format or some strange bug in Recorder::load"));
			return istr;
		}
*/		NodeType t=NodeType(i);
		NodeP n;
		switch(t)
		{

		case UNBOUND:		n=Unbound;					break;//
		case SPARE_NODE:	n=new Node;					break;//
		case STRING:
		case BACKSLASHED:
		case VBAR_STRING:	n=new String;				break;//
		case INT_:			n=new Int;					break;//
		case FLOAT_:		n=new Float;				break;//
		case BIGFLOAT:		n=new BigFloat;			break;//
		case COMPLEX:		n=new Complexd;			break;//
		case INT16:			n=new Int16;				break;//
		case INT8:			n=new Int8;					break;//
		case UINT8:			n=new UInt8;				break;//
		case QUOTE:			n=new Quote;				break;//
		case COLON:			n=new Colon;				break;//
		case OBJECT:		n=new Object;				break;//
		case PAIR:			n=new Pair;					break;//
		case WINDOWPAIR:	n=new WindowPair;			break;//
		case LINE:			n=new Line;					break;//
		case RUNPARSED:	n=new Runparsed;			break;//
		case TREE:			n=new Tree;					break;//
		case TODO:			n=new ToDoNode;			break;//
		case FILE_:			n=new File;					break;//
		case PROC_:			n=new Procedure;			break;//
		case ARRAY:			n=new Array;				break;//
		case BITMAP_:		n=new Bitmap;				break;//
		case TEXTURE:		n=new Texture;				break;//
		case TURTLE_:		n=new TurtleNode;			break;//
		case TABLE:			n=new Table;				break;//
		case STRUCT:		n=new Struct;				break;//
		case GRAPHIC:		n=new GraphicNode;		break;//
		case DYNAMIC_LIB:	n=new DynamicLibrary;	break;//
		case INTARRAY:		n=new IntArray;			break;//
		case INT16ARRAY:	n=new Int16Array;			break;//
		case FLOATARRAY:	n=new FloatArray;			break;//
		case SLIDER:		n=new SliderNode;			break;//
		case GAUGE:			n=new GaugeNode;			break;//
		case INTCONTROL:	n=new IntControlNode;	break;//
		case FLOATCONTROL:n=new FloatControlNode;	break;//
		case FRAME:			n=new FrameNode;			break;//
		case GRAPH:			n=new GraphNode;			break;//
		case BUTTON:		n=new ButtonNode;			break;//
		case TOGGLEBUTTON:n=new ToggleButtonNode;	break;//
		case CHECKBOX:		n=new CheckBoxNode;		break;//
		case RADIOBUTTON:	n=new RadioButtonNode;	break;//
		case STATICTEXT:	n=new StaticTextNode;	break;//
		case TEXTCONTROL:	n=new TextControlNode;	break;//
		case LISTBOX:		n=new ListBoxNode;		break;//
		case CHOICEBOX:	n=new ChoiceBoxNode;		break;//
		case COMBOBOX:		n=new ComboBoxNode;		break;//
		case LISTCONTROL:	n=new ListControlNode;	break;//
		case BOXSIZER:		n=new BoxSizerNode;		break;//
		case PRIM:
		case MACRO:
		case TAILFORM:
		case INFIX:			n=new Primitive;			break;//
		default: 
//			_pause(0);
					n=new Node;	// should never be reached if file format is correct!
					t=PNIL;
		}
		if (t==PNIL)
			break;
		n->type=t;
		n->read(istr);
		GC_::intToPtr->insert(n);
		k++;
	}
	int i;
	for (i=0; i < size; i++)
	{	PtrIntPairP obj;
		for (obj=data[i];  obj;  obj=obj->tail)
			if (obj && obj->p) 
				((Node*)(obj->p))->setPtrs();
	}
	for (i=0; i < size; i++)
	{	PtrIntPairP obj;
		for (obj=data[i];  obj;  obj=obj->tail)
			if (obj && obj->p) 
				((Node*)(obj->p))->aftersetPtrs();
	}
	for (i=0; i < size; i++)
	{	PtrIntPairP obj;
		for (obj=data[i];  obj;  obj=obj->tail)
			if (obj && obj->p) 
				((Node*)(obj->p))->aftersetPtrs2();
	}
	for (i=0; i < size; i++)
	{	PtrIntPairP obj;
		for (obj=data[i];  obj;  obj=obj->tail)
			if (obj && obj->p) 
				((Node*)(obj->p))->aftersetPtrs3();
	}
	GC_::disabled=false;
	return istr;
}
