//___________________________________________________________________
//
//                           Class ListBox
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/listbox.h>

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Object.h"
#include "LogoObject.h"
#include "Int.h"
#include "UInt8.h"
#include "Point.h"
#include "Error.h"
#include "NodePStack.h"

#include "Event.h"
#include "ListBox.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Color.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyListBox : public wxListBox
{
public:
	MyListBox(wxWindow* parent, wxWindowID id, 
		const wxArrayString& choices, 
		const wxPoint& pos=wxDefaultPosition, 
		const wxSize& size=wxDefaultSize, 
		long style = 0, 
		const wxValidator& validator = wxDefaultValidator, 
		const wxString& name = _("ListBox"));

	// MSW sends extra kill focus event
	virtual ~MyListBox()
	{
		if (parent) 
			parent->listBox = 0;
		parent=0;
	}

	ListBox_ *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to ListBox
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);
	void OnSelect(wxCommandEvent &event);
	void OnDClick(wxCommandEvent &event);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyListBox,wxListBox)
	EVT_CHAR		(MyListBox::OnChar		)
	EVT_KEY_DOWN(MyListBox::OnKeyDown	)
	EVT_KEY_UP	(MyListBox::OnKeyUp		)
	EVT_LISTBOX			(-1,	MyListBox::OnSelect	)
	EVT_LISTBOX_DCLICK(-1,	MyListBox::OnDClick	)
END_EVENT_TABLE()

MyListBox::MyListBox(wxWindow *parent_, wxWindowID id,
							  const wxArrayString &choices_,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxListBox(parent_, id, pos, size, choices_, style, validator, name)
{
	parent = (ListBox_*)parent_;
}

void MyListBox::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyListBox::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyListBox::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}

void MyListBox::OnSelect(wxCommandEvent &event)
{
	if (parent) 
		parent->OnSelect(event);
}

void MyListBox::OnDClick(wxCommandEvent &event)
{
	if (parent) 
		parent->OnDClick(event);
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(ListBox_, wxControl)
	EVT_CHAR				(		ListBox_::OnChar		)
	EVT_KEY_DOWN		(		ListBox_::OnKeyDown	)
	EVT_KEY_UP			(		ListBox_::OnKeyUp		)
	EVT_LISTBOX			(-1,	ListBox_::OnSelect		)
	EVT_LISTBOX_DCLICK(-1,	ListBox_::OnDClick		)
END_EVENT_TABLE()

ListBox_::ListBox_(wxWindow *parent)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	label=0;
	listBox=0;
}

ListBox_::ListBox_(ListBoxNodeP listBoxNode_, int style):
label(0), listBox(0), bestSize(-999, -999)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=listBoxNode_;
	wxWindow *parent=(wxWindow*)console;
	if (listBoxNode_->parent)
		parent=listBoxNode_->parent->window;
	create(parent, -1, wxArrayString(), 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, _("ListBox"));
}

ListBox_::ListBox_(ListBoxNodeP listBoxNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			PairP onSelect_,
			PairP onDClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), listBox(0), bestSize(-999, -999), choices(choices_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=listBoxNode_;
	bestSize = wxSize(-999, -999);
	listBoxNode_->window=WindowP(this);
	listBoxNode_->style=style;
	create(parent, id, choices, pos, size, style, validator, name);
	listBoxNode_->setonSelect(onSelect_);
	listBoxNode_->setonDClick(onDClick_);
}

bool ListBox_::create(wxWindow* parent, wxWindowID id, 
			const wxArrayString &choices_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
	{	label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, _("staticText"));
		if (!label)
		{	error(OUT_OF_MEM); return false;
		}
	}else
		label=0;

	if (label)
		listBox = new MyListBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		listBox = new MyListBox(this, id, choices_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);
	if (!listBox)
	{	error(OUT_OF_MEM); return false;
	}
	choices.Clear();

	DoSetSize(pos.x, pos.y, width, height);
	SetInitialSize(wxSize(width, height));
	return true;
}

void ListBox_::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (listBox)   
		listBox->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize ListBox_::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxListBox *t=new wxListBox((wxWindow*)this, wxID_ANY, 
			wxPoint(0,0), wxDefaultSize, choices, GetWindowStyle());
		if (!t)
		{	error(OUT_OF_MEM); return wxDefaultSize;
		}
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

ListBox_::~ListBox_()
{
	ThreadP th=Thread::This();
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
	if (label)
	{	delete label;
		label=0;
	}
	if (listBox)
	{	delete listBox;
		listBox=0;
	}

	EventPP out=th->eventFiFo->out;
	EventP ev;
	while (*th->eventFiFo)
	{	*th->eventFiFo >> ev;
		switch (ev->type)
		{
		case Event::ListBoxSelect:
		case Event::ListBoxDClick:
		case Event::ListBoxChar:
		case Event::ListBoxKeyDown:
		case Event::ListBoxKeyUp:
			if (((EventListBox*)ev)->listBox==this)
				((EventListBox*)ev)->listBox=0;
			break;
		default:
			;
		}
	}
	th->eventFiFo->out=out;
}

void ListBox_::Set(wxArrayString &choices)
{
	if (listBox)
		listBox->Set(choices);
}

#define LB_ON(NAME)															\
void ListBox_::On##NAME(wxKeyEvent &event)							\
{																					\
	ThreadP th=Thread::This();												\
	EventP evt=new EventListBox##NAME(this, event);\
	if throwing return;														\
	*th->eventFiFo << evt;													\
	event.Skip();																\
}

LB_ON(Char)
LB_ON(KeyDown)
LB_ON(KeyUp)

void ListBox_::OnSelect(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventListBoxSelect(this);
	if throwing return;
	*th->eventFiFo << evt;
}

void ListBox_::OnDClick(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventListBoxDClick(this);
	if throwing return;
	*th->eventFiFo << evt;
}

wxString ListBox_::wxstring()
{
	wxString str;
	str << _("ListBox") << " " 
		<< '[' << GetName() << "] ";
	if (*d(onSelect))
		str << (*d(onSelect))->string()->ptr() << ' ';
	else
		str << "[] ";
	if (*d(onDClick))
		str << (*d(onDClick))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &ListBox_::read(DataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString name, fontName;
	int fontSize, fontStyle, fontWeight;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> name 
		>> (int&)*d(onSelect )
		>> (int&)*d(onDClick )
		>> pos.x >> pos.y >> size.x >> size.y
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	int nr, i;
	istr >> nr;
	choices.Clear();
	for (i=0; i<nr; i++)
	{	wxString item;
		istr >> item;
		choices.Add(item);
	}
	listBox->Set(choices);

	istr >> nr;
	for (i=0; i<nr; i++)
	{	int sel;
		istr >> sel;
		listBox->SetSelection(sel);
	}
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &ListBox_::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetName() 
		<< GC_::ptrToInt->find(*d(onSelect))
		<< GC_::ptrToInt->find(*d(onDClick))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y;

	{	wxColour 
			f=listBox->GetForegroundColour(), 
			b=listBox->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=listBox->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();

	int nr=listBox->GetCount(), i;
	ostr << nr;
	for (i=0; i<nr; i++)
		ostr << listBox->GetString(i);

	wxArrayInt sel;
	nr=listBox->GetSelections(sel);
	ostr << nr;
	for (i=0; i<nr; i++)
		ostr << sel[i];
	return ostr;
}

bool ListBox_::SetBackgroundColour(const wxColour &c)
{
	if (listBox)
		return listBox->SetBackgroundColour(c);
	return false;
}

bool ListBox_::SetForegroundColour(const wxColour &c)
{
	if (listBox)
		return listBox->SetForegroundColour(c);
	return false;
}

bool ListBox_::SetFont(const wxFont &f)
{
	if (listBox)
		return listBox->SetFont(f);
	return false;
}

//___________________________________________________________________

PairP ListBoxNode::onSelect()
{
	if (listBox)
		return PairP(*listBox->d(ListBox_::onSelect));
	return 0;
}

void ListBoxNode::setonSelect(NodeP nd)
{
	ThreadP th=Thread::This();
	if (listBox)
	{	*listBox->d(ListBox_::onSelect) = PairP(nd);	check_oldyoung(nd);
		*listBox->d(ListBox_::uonSelect) = th->lexufun;
		*listBox->d(ListBox_::tonSelect) = th->thisobj;
	}
}

PairP ListBoxNode::onDClick()
{
	if (listBox)
		return PairP(*listBox->d(ListBox_::onDClick));
	return 0;
}

void ListBoxNode::setonDClick(NodeP nd)
{
	ThreadP th=Thread::This();
	if (listBox)
	{	*listBox->d(ListBox_::onDClick) = PairP(nd);	check_oldyoung(nd);
		*listBox->d(ListBox_::uonDClick) = th->lexufun;
		*listBox->d(ListBox_::tonDClick) = th->thisobj;
	}
}

ListBoxNode::ListBoxNode(): 
	WindowNode(LISTBOX),
	listBox((ListBoxP&)window)
{}

ListBoxNode::ListBoxNode(WindowNodeP parent_): 
	WindowNode(LISTBOX, parent_),
	listBox((ListBoxP&)window)
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void ListBoxNode::free() 
{	
	if (listBox) 
	{	ThreadP th=Thread::This();
		*listBox->d(ListBox_::windowNode)=0;
		listBox->Destroy(); 
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::ListBoxSelect:
			case Event::ListBoxDClick:
			case Event::ListBoxChar:
			case Event::ListBoxKeyDown:
			case Event::ListBoxKeyUp:
				if (WindowP(((EventListBox*)ev)->listBox)==window)
					((EventListBox*)ev)->listBox=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		listBox=0;
	}
}
void ListBoxNode::inc_gc() 
{	fifo << onSelect() << onDClick()
		 << onChar() << onKeyDown() << onKeyUp();
}
int ListBoxNode::youngs() 
{	return gmark(onSelect())+gmark(onDClick())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int ListBoxNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return 
		 gmarkPtr(onSelect(), ptrToInt)
		+gmarkPtr(onDClick(), ptrToInt)
		+gmarkPtr(onChar(), ptrToInt)
		+gmarkPtr(onKeyDown(), ptrToInt)
		+gmarkPtr(onKeyUp(), ptrToInt);
}
void ListBoxNode::check_oldyoungs() 
{	check_oldyoung(onSelect());
	check_oldyoung(onDClick());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

DataInputStream &ListBoxNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode >> style;
		listBox=new ListBox_(ListBoxNodeP(this), style);
		if (!listBox)
		{	error(OUT_OF_MEM); return istr;
		}
		listBox->read(istr);
		parent=parentNode;
	}
	return istr;
}
wxDataOutputStream &ListBoxNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (listBox)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent) 
			<< style;
		listBox->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ListBoxNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (listBox)
	{	ListBoxP t=ListBoxP(listBox);
		setonSelect ((Pair*)intToPtr.find(int(t->onSelect	)));
		setonDClick ((Pair*)intToPtr.find(int(t->onDClick	)));
	}
}

void ListBoxNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	ListBoxP t=listBox, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ListBox_(ListBoxNodeP(this),
			p, -1, t->choices, 
			PairP(*t->d(ListBox_::onSelect)), 
			PairP(*t->d(ListBox_::onDClick)), 
			pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	if (!t2)
	{	error(OUT_OF_MEM); return;
	}
	listBox=t2;
	wxArrayInt sel;
	int nr=t->listBox->GetSelections(sel);
	for (int i=0; i<nr; i++)
		t2->listBox->SetSelection(sel[i]);
	setonSelect (*t->d(ListBox_::onSelect ));
	setonDClick (*t->d(ListBox_::onDClick ));
	t2->listBox->SetForegroundColour(t->listBox->GetForegroundColour());
	t2->listBox->SetBackgroundColour(t->listBox->GetBackgroundColour());
	wxFont f=t->listBox->GetFont();
	t2->listBox->SetFont(f);
	delete t;
	listBox=t2;
}

StringP ListBoxNode::string()
{
	if (listBox)
		return new String(listBox->wxstring());
	return _Null_Word->name();
}

wxString ListBoxNode::wxstring()
{
	if (listBox)
		return listBox->wxstring();
	return wxEmptyString;
}

void ListBoxNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lListBox(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_("ListBox");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxArrayString choices;
	{	PairP list=arg[3]->a_list(3);
		if throwing return Unbound;
		while (list)
		{	StringP s=list->first()->a_string(3); 
			if throwing return Unbound;
			choices.Add(s->ptr());
			next(list);
		}

	}
	PairP onSelect=0;
	if (exist(4))
	{	PairP os=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onSelect=os;
	}
	PairP onDClick=0;
	if (exist(5))
	{	PairP odc=arg[5]->a_runnable(5);
		if throwing return Unbound;
		onDClick=odc;
	}
	long style = 0;
	if (exist(6))
	{	int s=arg[6]->a_int(6);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(7))
	{	PairP p=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(8))
	{	PairP s=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	ListBoxNodeP listBoxNode=new ListBoxNode(parentNode);
	if throwing return Unbound;
	ListBox_ *lb=new ListBox_(listBoxNode,
			parent, id, choices, 
			onSelect, onDClick, 
			pos, size, style, wxDefaultValidator, name);
	if (!lb)
	{	error(OUT_OF_MEM); return Unbound;
	}
//	GUI_LEAVE
	return listBoxNode;
}

NodeP lListBoxSelections(PairPP arg)
{
	ThreadP th=Thread::This();
	ListBox_ *listBox=0;
	if (exist(1))
	{	ListBoxNodeP listBoxNode=arg[1]->a_listBox(1);
		if throwing return Unbound;
		listBox=ListBoxP(listBoxNode->listBox);
	}
	else
	{	listBox=th->listBoxValue;
	}
	if (!listBox) return 0;
	wxArrayInt sel;
//	GUI_ENTER
	int nr=listBox->listBox->GetSelections(sel);
//	GUI_LEAVE
	PairP result=0, last=0, tnode=0;
	for (int i=0; i<nr; i++)
	{	int s=sel[i];
		tnode=new Pair(new Int(s), 0);
		if throwing return Unbound;
		if (!last)
		{	result=tnode;
		}
		else
		{	last->settail(tnode);
		}
		last=tnode;
	}
	return result;
}

NodeP lListBoxSetSelections(PairPP arg)
{
	ThreadP th=Thread::This();
	ListBox_ *listBox=0;
	if (exist(1))
	{	ListBoxNodeP listBoxNode=arg[1]->a_listBox(1);
		if throwing return Unbound;
		listBox=ListBoxP(listBoxNode->listBox);
	}
	else
	{	listBox=th->listBoxValue;
	}
	if (!listBox) return 0;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
//	GUI_ENTER
	int nr=listBox->listBox->GetCount();
	for (int i=0; i<nr; i++)
		listBox->listBox->Deselect(i);
	while (list)
	{	int sel=list->first()->a_int(2);
		if throwing return Unbound;
		if (sel < 0 || sel >= nr)
		{	error(BAD_DATA, arg[2]);
			return Unbound;
		}
		listBox->listBox->SetSelection(sel);
		next(list);
	}
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetChoices(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	wxArrayString choices;
	PairP list=arg[2]->a_list(2);
	if throwing return Unbound;
	while (list)
	{	StringP s=list->first()->a_string(2);
		if throwing return Unbound;
		choices.Add(s->ptr());
		next(list);
	}
//	GUI_ENTER
	ListBoxP(listBox->listBox)->listBox->Set(choices);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxAppend(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	ListBoxP(listBox->listBox)->listBox->Append(value->ptr());
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetItem(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
	StringP str=arg[3]->a_string(3);
	if throwing return Unbound;
//	GUI_ENTER
	if (n < 0 || n >= int(ListBoxP(listBox->listBox)->listBox->GetCount()))
	{	error(BAD_DATA, arg[2]);
//		GUI_LEAVE
		return Unbound;
	}
	ListBoxP(listBox->listBox)->listBox->SetString(n, str->ptr());
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxRemoveItem(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int n=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	if (n < 0 || n >= int(ListBoxP(listBox->listBox)->listBox->GetCount()))
	{	error(BAD_DATA, arg[2]);
//		GUI_LEAVE
		return Unbound;
	}
	ListBoxP(listBox->listBox)->listBox->Delete(n);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxCount(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
//	GUI_ENTER
	int count=ListBoxP(listBox->listBox)->listBox->GetCount();
//	GUI_LEAVE
	return new Int(count);
}

#define _TC_ON(NAME)									\
NodeP lListBoxOn##NAME(PairPP arg)			\
{															\
	ListBoxNodeP listBox=arg[1]->a_listBox(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	listBox->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Select)
_TC_ON(DClick)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lListBoxSetBackgroundColor(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	listBox->listBox->SetBackgroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetColor(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	listBox->listBox->SetForegroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetFontSize(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=listBox->listBox->GetFont();
	f.SetPointSize(s);
	listBox->listBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetFontName(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=listBox->listBox->GetFont();
	f.SetFaceName(s->ptr());
	listBox->listBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetFontStyle(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=listBox->listBox->GetFont();
	f.SetStyle(s);
	listBox->listBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxSetFontWeight(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=listBox->listBox->GetFont();
	f.SetWeight(s);
	listBox->listBox->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxEnable(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	listBox->listBox->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lListBoxDestroy(PairPP arg)
{
	ListBoxNodeP listBox=arg[1]->a_listBox(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=listBox->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(listBox));
			p->check_oldyoung(p->children);
		}
	}
	listBox->listBox->Destroy();
//	GUI_LEAVE
	return Unbound;
}
