//___________________________________________________________________
//
//                       Class MainFrame
//
//                          Definitons
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/textctrl.h>
#include <wx/listctrl.h>
#include <wx/frame.h>
#include <wx/splitter.h>
#include <wx/menu.h>
#include <wx/dialog.h>
#include <wx/filedlg.h>
#include <wx/treectrl.h>
#include <wx/html/helpctrl.h>
#include <wx/msgdlg.h>
#include <wx/dc.h>
#include <wx/aui/aui.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>

#include <math.h>

#ifdef OPENGL
	#include <GL/glu.h>
#endif

#include "Main.h"

#include "File.h"
#include "String.h"
#include "Pair.h"
#include "Object.h"

#include "NodePStack.h"
#include "Error.h"
#include "Files.h"
#include "Init.h"
#include "Workspace.h"

#include "UInt8.h"
#include "Color.h"
#include "Point.h"
#include "Pen.h"
#include "Font.h"
#include "Screen.h"
#include "Turtle.h"
#include "Console.h"
#include "Graph.h"
#include "Recorder.h"
#include "Vars.h"
#include "Calls.h"
#include "Splitter.h"
#include "MainFrame.h"
#include "AboutDlg.h"
#include "MemoryWindow.h"
#include "LogoApp.h"
#include "Thread.h"
#include "Aui.h"

#ifdef _DEBUG
	#undef THIS_FILE
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

static char THIS_FILE[]= __FILE__;

BEGIN_EVENT_TABLE(MainFrame, wxFrame)
	EVT_CLOSE(								MainFrame::OnClose			)
	EVT_CHAR		(							MainFrame::OnChar				)
	EVT_KEY_DOWN(							MainFrame::OnKeyDown			)
	EVT_MENU(wxID_HELP_CONTENTS	, MainFrame::OnHelp				)
	EVT_ERASE_BACKGROUND(           MainFrame::OnErase				) 
	EVT_MOVE (                      MainFrame::OnMove				)
//	EVT_SET_FOCUS(						  MainFrame::OnSetFocus			)
	EVT_MENU (wxID_EXIT				, MainFrame::OnExit				)
	EVT_MENU (wxID_ABOUT				, MainFrame::OnAppAbout			)
	EVT_MENU (wxID_NEW				, MainFrame::OnFileNew			)
	EVT_MENU (wxID_OPEN				, MainFrame::OnFileOpen			)
	EVT_MENU (wxID_SAVE				, MainFrame::OnFileSave			)
	EVT_MENU (ID_CONSOLE_OPEN		, MainFrame::OnConsoleOpen		)
	EVT_MENU (ID_CONSOLE_SAVE		, MainFrame::OnConsoleSave		)
	EVT_MENU (ID_CONSOLE_AUTOSAVE	, MainFrame::OnConsoleAutoSave)
	EVT_MENU (ID_PICTURE_OPEN		, MainFrame::OnPictureOpen		)
	EVT_MENU (ID_PICTURE_SAVE		, MainFrame::OnPictureSave		)
	EVT_MENU (ID_PICTURE_TEXT_OPEN, MainFrame::OnPictureTextOpen)
	EVT_MENU (ID_PICTURE_TEXT_SAVE, MainFrame::OnPictureTextSave)
	EVT_MENU (ID_PICTURE_AUTOSAVE	, MainFrame::OnPictureAutoSave)
	EVT_MENU (ID_NODES_OPEN			, MainFrame::OnNodesOpen		)
	EVT_MENU (ID_NODES_SAVE			, MainFrame::OnNodesSave		)
	EVT_MENU (ID_NODES_AUTOSAVE	, MainFrame::OnNodesAutoSave	)
	EVT_MENU (ID_EDIT_REPEAT		, MainFrame::OnEditRepeat		)
	EVT_MENU (ID_EDIT_RESET			, MainFrame::OnEditReset		)
	EVT_MENU (ID_EDIT_GOTO_CURSOR_LINE, MainFrame::OnEditGotoCursorLine	)
	EVT_MENU (ID_EDIT_WRAP_LINES	, MainFrame::OnEditWrapLines	)
	EVT_MENU (ID_EDIT_LANGUAGE    , MainFrame::OnEditChooseLanguage		)
	EVT_MENU (ID_VIEW_CONSOLE		, MainFrame::OnViewConsole		)
	EVT_MENU (ID_VIEW_EDITOR		, MainFrame::OnViewEditor		)
	EVT_MENU (ID_VIEW_GRAPH			, MainFrame::OnViewGraph		)
	EVT_MENU (ID_VIEW_STACKVARS	, MainFrame::OnViewStackVars	)
	EVT_MENU (ID_VIEW_VARS			, MainFrame::OnViewVars			)
	EVT_MENU (ID_VIEW_CALLS			, MainFrame::OnViewCalls		)
	EVT_MENU (ID_VIEW_MEMORY		, MainFrame::OnViewMemory		)
	EVT_MENU_RANGE(ID_LAYOUT_LOAD1, ID_LAYOUT_LOAD9, MainFrame::OnLayoutLoad)
	EVT_MENU_RANGE(ID_LAYOUT_SAVE1, ID_LAYOUT_SAVE9, MainFrame::OnLayoutSave)
	EVT_MENU(wxID_PRINT				, MainFrame::OnPrint				)
	EVT_MENU(wxID_PREVIEW			, MainFrame::OnPrintPreview	)
	EVT_MENU(wxID_PRINT_SETUP		, MainFrame::OnPrintSetup		)
	EVT_MENU(ID_PAGE_SETUP			, MainFrame::OnPageSetup		)
END_EVENT_TABLE()

#define NEED_MEM(x)											\
	if (!x)														\
	{	delete theApp->heapReserve;						\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x << _(": Insufficient Memory");			\
		wxMessageBox(msg,	aUCBLogo+" "+_("Error"));		\
		return;													\
	}

class TestFrame: public wxPanel
{
public:
	enum
	{	ID_TESTMENU_ITEM1=2000
	};

	wxFlatMenuBar *menubar;
	wxFlatMenu *testmenu, *testmenu2;
	wxTextCtrl *textctrl;

	TestFrame(wxWindow *parent);
};

TestFrame::TestFrame(wxWindow *parent):
	wxPanel(parent, wxID_ANY, wxDefaultPosition, parent->GetClientSize(),
		wxNO_BORDER, _T("TestFrame"))
{
Show();
	// Init resources and add the PNG handler
 	wxImage::AddHandler( new wxPNGHandler ); 
	wxImage::AddHandler( new wxCURHandler ); 
	wxImage::AddHandler( new wxBMPHandler );
	wxImage::AddHandler( new wxXPMHandler );

	testmenu=new wxFlatMenu;
	if (!testmenu)
		return;
	testmenu->Append(ID_TESTMENU_ITEM1, _("Item1"));

	testmenu2=new wxFlatMenu;
	if (!testmenu2)
		return;
	testmenu2->Append(ID_TESTMENU_ITEM1+1, _("Item1"));

	menubar=new wxFlatMenuBar(this, wxID_ANY, false, SmallIcons);
	if (!menubar)
		return;
	menubar->Append(testmenu, _("&TestMenu"));
	menubar->Append(testmenu2, _("TestMenu&2"));

	textctrl=new wxTextCtrl(this, wxID_ANY, 
		_("Hallo, I'm a test wxFrame with a menubar inside a wxAui managed window."), 
		wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE);

	wxBoxSizer *sizer=new wxBoxSizer(wxVERTICAL);
	// Set the main frame sizer
	sizer->Add(menubar, 0, wxEXPAND);
	sizer->Add(textctrl, 1, wxEXPAND);
	SetSizer(sizer);

	Aui *aui=theApp->aui;
	aui->AddPane(this,
		wxAuiPaneInfo().
		Name(wxT("TestFrame")).Caption(wxT("TestFrame Caption")).
		Top());
}


MainFrame::MainFrame(wxWindow* parent, wxWindowID id, 
	const wxString& title, 
	const wxPoint& pos, 
	const wxSize& size, 
	long style, 
	const wxString& name,
	bool consoleAutosave, int caretpos, 
	bool pictureAutosave,
	bool nodesAutosave_):
wxFrame(parent, id, title, pos, size, style, name)
{
	ThreadP th=Thread::This();
	initialized=false;
	theApp->aui->SetManagedWindow(this);
	theApp->aui->init();
	theApp->mainFrame=this;

	wxArtManagerST::Get()->DrawMenuBarBorder(false);

//new TestFrame(this);

	printData=0;
	pageSetupData=0;
	nodesAutosave=nodesAutosave_;

	viewMenu=0;
	viewMenuConsole=0;
	viewMenuEditor=0;
	viewMenuGraph=0;
	viewMenuStackVars=0;
	viewMenuVars=0;
	viewMenuCalls=0;
	viewMenuMemory=0;
	layoutMenu=0;


	viewMenu = new wxFlatMenu;
	NEED_MEM(viewMenu);
	viewMenuConsole=viewMenu->AppendCheckItem(ID_VIEW_CONSOLE,_("&Console"),_("View Console"));
	viewMenuEditor=viewMenu->AppendCheckItem(ID_VIEW_EDITOR,_("&Editor"),_("View Editor"));
	viewMenuGraph=viewMenu->AppendCheckItem(ID_VIEW_GRAPH,_("&Graph"),_("View Graph"));
	viewMenuStackVars=viewMenu->AppendCheckItem(ID_VIEW_STACKVARS ,_("&StackVars") ,_("View StackVars"));
	viewMenuVars=viewMenu->AppendCheckItem(ID_VIEW_VARS ,_("&Vars") ,_("View Vars"));
	viewMenuCalls=viewMenu->AppendCheckItem(ID_VIEW_CALLS,_("&Calls"),_("View Calls"));
	viewMenuMemory=viewMenu->AppendCheckItem(ID_VIEW_MEMORY, _("&Memory"));

	layoutMenu = new wxFlatMenu;
	NEED_MEM(layoutMenu);
	layoutMenu->Append(ID_LAYOUT_LOAD1,_("Load &SplitScreen\tCtrl+1"),_("Load SplitScreen"));
	layoutMenu->Append(ID_LAYOUT_SAVE1,_("Save SplitScreen"),_("Save SplitScreen"));
	layoutMenu->Append(ID_LAYOUT_LOAD2,_("Load &TextScreen\tCtrl+2"),_("Load TextScreen"));
	layoutMenu->Append(ID_LAYOUT_SAVE2,_("Save TextScreen"),_("Save TextScreen"));
	layoutMenu->Append(ID_LAYOUT_LOAD3,_("Load &FullScreen\tCtrl+3"),_("Load FullScreen"));
	layoutMenu->Append(ID_LAYOUT_SAVE3,_("Save FullScreen"),_("Save FullScreen"));
	layoutMenu->Append(ID_LAYOUT_LOAD4,_("Load &EditScreen\tCtrl+4"),_("Load EditScreen"));
	layoutMenu->Append(ID_LAYOUT_SAVE4,_("Save EditScreen"),_("Save EditScreen"));
	layoutMenu->Append(ID_LAYOUT_LOAD5,_("Load &Layout 5\tCtrl+5"),_("Load Layout 5"));
	layoutMenu->Append(ID_LAYOUT_SAVE5,_("Save Layout 5"),_("Save Layout 5"));
	layoutMenu->Append(ID_LAYOUT_LOAD6,_("Load &Layout 6\tCtrl+6"),_("Load Layout 6"));
	layoutMenu->Append(ID_LAYOUT_SAVE6,_("Save Layout 6"),_("Save Layout 6"));
	layoutMenu->Append(ID_LAYOUT_LOAD7,_("Load &Layout 7\tCtrl+7"),_("Load Layout 7"));
	layoutMenu->Append(ID_LAYOUT_SAVE7,_("Save Layout 7"),_("Save Layout 7"));
	layoutMenu->Append(ID_LAYOUT_LOAD8,_("Load &Layout 8\tCtrl+8"),_("Load Layout 8"));
	layoutMenu->Append(ID_LAYOUT_SAVE8,_("Save Layout 8"),_("Save Layout 8"));
	layoutMenu->Append(ID_LAYOUT_LOAD9,_("Load &Layout 9\tCtrl+9"),_("Load Layout 9"));
	layoutMenu->Append(ID_LAYOUT_SAVE9,_("Save Layout 9"),_("Save Layout 9"));

	wxPanel *menu=new wxPanel(this, wxID_ANY, 
		wxDefaultPosition, wxSize(40,20), 
		wxNO_BORDER, "ViewMenu");
	NEED_MEM(menu);
	menuBar=new wxFlatMenuBar(menu);
	NEED_MEM(menuBar);
	menuBar->Append(viewMenu, _("&View"));
	menuBar->Append(layoutMenu, _("&Layout"));
	wxBoxSizer *sizer=new wxBoxSizer(wxVERTICAL);
	sizer->Add(menuBar, 0, wxEXPAND);
	menu->SetSizer(sizer);
	menu->Show();		
	theApp->aui->AddPane(menu, 
		wxAuiPaneInfo().
		Name(wxT("ViewMenu")).Caption(_("ViewMenu")).
		Layer(1).Top().
		MinSize(40,20).BestSize(40,20).MaxSize(40,20).
		CloseButton(false)
		);

//	int sstyle=wxSP_NOBORDER | wxSP_3DSASH;
//	splitter=new Splitter(this, -1, wxDefaultPosition, 
//		wxDefaultSize,	sstyle | wxFULL_REPAINT_ON_RESIZE | wxSP_LIVE_UPDATE, "Splitter");
//	NEED_MEM(splitter);

//	splitter2=new Splitter(splitter, -1, wxDefaultPosition, 
//		wxDefaultSize,	sstyle | wxFULL_REPAINT_ON_RESIZE | wxSP_LIVE_UPDATE, 
//		"Splitter2");
//	NEED_MEM(splitter2);
//	splitter->SplitHorizontally(graph, splitter2, 1);
//	splitter->SetMinimumPaneSize(1);
//	double ratio;
//	theApp->config->Read("SplitterRatio", &ratio, double(0)); 
//	splitter->setRatio(ratio);
//	if (ratio > 0.01)
//		inGraphicsMode=inSplitScreen=true;
//	else
//		inGraphicsMode=inSplitScreen=false;

	theApp->consoleFrame=new ConsoleFrame(this);
	NEED_MEM(theApp->consoleFrame);

	theApp->aui->Update();

	theApp->graph=new Graph(GraphNodeP(0), this, ID_GRAPH, 
		wxDefaultPosition, wxSize(800,600), 
		wxFULL_REPAINT_ON_RESIZE, _("Graph"),
		Graph::attribs,
		wxNullPalette, pictureAutosave);
	NEED_MEM(theApp->graph);
	graphAutosave=pictureAutosave;
	th->graph=theApp->graph;
	theApp->graphNode=0;

#ifdef WIN32
	if (!theApp->graphShown)
		theApp->graph->Hide();
	theApp->aui->AddPane(theApp->graph, 
		wxAuiPaneInfo()
		.Name("Graph").Caption(_("Graph"))
		.Top()
		.BestSize(800,600)
		.MaximizeButton().MinimizeButton().PinButton()
	);
#else
	theApp->graph->show();
#endif
//	splitter3=new Splitter(splitter2, -1, wxDefaultPosition, 
//		wxDefaultSize,	sstyle | wxSP_LIVE_UPDATE, "Splitter3");
//	NEED_MEM(splitter3);
//	splitter2->SplitHorizontally(console, splitter3, int(theApp->max.y*0.9*3/4));
//	splitter2->SetMinimumPaneSize(1);
//	theApp->config->Read("SplitterRatio2", &ratio, double(1)); 
//	splitter2->setRatio(ratio);

	if (theApp->stackvarsShown)
	{	theApp->stackvars=new StackVars(this, ID_STACKVARS);
		NEED_MEM(theApp->stackvars);
	}
//	splitter4=new Splitter(splitter3, -1, wxDefaultPosition, 
//		wxDefaultSize,	sstyle | wxSP_LIVE_UPDATE, "Splitter4");
//	NEED_MEM(splitter4);
//	splitter3->SplitHorizontally(stackvars, splitter4, int(theApp->max.y*0.9*3/4));
//	splitter3->SetMinimumPaneSize(1);
//	theApp->config->Read("SplitterRatio3", &ratio, double(1)); 
//	splitter3->setRatio(ratio);

	if (theApp->varsShown)
	{	theApp->vars=new Vars(this, ID_VARS);
		NEED_MEM(theApp->vars);
	}
	if (theApp->callsShown)
	{	theApp->calls=new Calls(this, ID_CALLS);
		NEED_MEM(theApp->calls);
	}
//	splitter4->SplitHorizontally(vars, calls, 10);
//	splitter4->SetMinimumPaneSize(1);
//	theApp->config->Read("SplitterRatio4", &ratio, double(1)); 
//	splitter4->setRatio(ratio);

	SetFocus();
	if (theApp->consoleFrame)
		theApp->consoleFrame->SetFocus();

#ifdef WIN32
	SetIcons(wxIconBundle(wxIcon("ICON_LOGO", wxBITMAP_TYPE_ICO_RESOURCE)));
#else
	SetIcons(wxIconBundle(wxIcon(
		theApp->runDirectory+"/logo.ico", wxBITMAP_TYPE_ICO)));
#endif

//	theApp->aui->Update();

	printData = new wxPrintData;
	NEED_MEM(printData);
	printData->SetOrientation(wxLANDSCAPE);
	pageSetupData = new wxPageSetupDialogData;
	NEED_MEM(pageSetupData);
	initialized=true;
	theApp->mainShown=true;
}

MainFrame::~MainFrame() 
{
	ThreadP th=Thread::This();
	if (theApp)
		theApp->saveMainConfig();
	if (th)
		th->graph=0;
	if (printData)
		delete printData;
	if (pageSetupData)
		delete pageSetupData;
	::console=0;
	if (theApp)
	{	theApp->mainFrame=0;
		theApp->mainShown=false;
		theApp->calls=0;
		theApp->vars=0;
		theApp->stackvars=0;
		theApp->aui->UnInit();
	}
	wxArtManagerST::Free();
}

void MainFrame::setSplitter(double y)
{
	if (y==0)
		inGraphicsMode=inSplitScreen=false;
	else
		inGraphicsMode=inSplitScreen=true;
//	splitter->setRatio(y);
}

void MainFrame::splitScreen(){loadLayout("1");}
void MainFrame:: TextScreen(){loadLayout("2");}
void MainFrame:: fullScreen(){loadLayout("3");}
void MainFrame::allFullScreen(){ShowFullScreen(true, wxFULLSCREEN_ALL);}
void MainFrame::notFullScreen(){ShowFullScreen(false);}

NodeP lTextScreen (PairPP) 
{
	if (theApp->mainFrame)
		theApp->mainFrame->TextScreen(); 
	return Unbound;
}

NodeP lsplitScreen(PairPP arg) 
{
	if (theApp->mainFrame)
	{	if (exist(1))
		{
//			double ratio=arg[1]->a_float(1);
			if throwing return Unbound;
//			theApp->mainFrame->setSplitter(ratio); 
		}else
			theApp->mainFrame->splitScreen(); 
	}
	return Unbound;
}

NodeP lfullScreen (PairPP) 
{
	if (theApp->mainFrame)
		theApp->mainFrame-> fullScreen(); 
	return Unbound;
}

NodeP lsetVarsSplitter(PairPP arg) 
{
//	double ratio=arg[1]->a_float(1);
	if throwing return Unbound;
	if (theApp->mainFrame)
	{
//		theApp->mainFrame->splitter2->setRatio(ratio); 
	}
	return Unbound;
}

NodeP lsetCallsSplitter(PairPP arg) 
{
//	double ratio=arg[1]->a_float(1);
	if throwing return Unbound;
	if (theApp->mainFrame)
	{
//		theApp->mainFrame->splitter3->setRatio(ratio); 
	}
	return Unbound;
}

NodeP lallFullScreen(PairPP)
{
	if (theApp->mainFrame)
		theApp->mainFrame->allFullScreen(); 
	return Unbound;
}

NodeP lnotFullScreen(PairPP) 
{
	if (theApp->mainFrame)
		theApp->mainFrame->notFullScreen(); 
	return Unbound;
}

void MainFrame::OnMove(wxMoveEvent& event)
{
	if (initialized && theApp->graph)
		theApp->graph->swapBuffers();
}
void MainFrame::OnErase(wxEraseEvent& event){}

void MainFrame::OnSetFocus(wxFocusEvent& event)
{
//	SetFocus();
/*	if (theApp->consoleFrame)
		theApp->consoleFrame->SetFocus();
	if (console)
		console->SetFocus();
*/	event.Skip();
}

void MainFrame::printPrompt(charP str) 
{
	if (inGraphicsMode && !inSplitScreen)
		lsplitScreen(PairPP(Thread::This()->argStack->index));
	c_out << str;
	if (console)
		console->flush();
}
void File::printPrompt(charP str) 
{
	if (theApp->mainFrame)
		theApp->mainFrame->printPrompt(str);
	else
		c_out << str;
}

void MainFrame::OnClose(wxCloseEvent &event) 
{
	theApp->action=EXIT_LOGO;
}

void MainFrame::OnExit(wxCommandEvent &event) 
{
	wxCloseEvent cev;
	OnClose(cev);
}

void MainFrame::OnChar(wxKeyEvent &event)				
{																
	if (theApp->console)
	{	theApp->console->SetFocus();
		theApp->console->OnChar(event);
	}
	event.Skip();												
}

void MainFrame::OnKeyDown(wxKeyEvent &event)				
{
	if (theApp->console)
	{	theApp->console->SetFocus();
		theApp->console->OnKeyDown(event);
	}
	event.Skip();												
}

void MainFrame::OnViewConsole(wxCommandEvent &event) 
{
	if (theApp->consoleFrame)
		theApp->consoleFrame->show();
}

void MainFrame::OnViewGraph(wxCommandEvent &event) 
{
	if (theApp->graph)
		theApp->graph->show();
}

void MainFrame::OnViewStackVars(wxCommandEvent &event) 
{
	if (theApp->stackvars)
	{	theApp->aui->DetachPane(theApp->stackvars);
		delete theApp->stackvars;
		theApp->aui->Update();
	}else
		theApp->stackvars=new StackVars(this, ID_STACKVARS);
}

void MainFrame::OnViewVars(wxCommandEvent &event) 
{
	if (theApp->vars)
	{	theApp->aui->DetachPane(theApp->vars);
		delete theApp->vars;
		theApp->aui->Update();
	}else
		theApp->vars=new Vars(this, ID_VARS);
}

void MainFrame::OnViewCalls(wxCommandEvent &event) 
{
	if (!theApp->calls)
		theApp->callsShow();
	else
		theApp->callsHide();
}

void MainFrame::OnViewEditor(wxCommandEvent &event) 
{
	if (!theApp->edFrame)
		theApp->editorShow();
	else
		theApp->editorHide();
}

void MainFrame::OnViewMemory(wxCommandEvent &event) 
{
	if (!memoryWindow)
		theApp->memoryShow();
	else
		theApp->memoryHide();
}

void MainFrame::loadLayout(wxString name)
{
	wxString fullname("Layout");
	fullname << name;
	wxString perspective=theApp->aui->SavePerspective();
	perspective=theApp->config->Read(fullname, perspective);
	theApp->aui->LoadPerspective(perspective);
	theApp->aui->Update();
}

void MainFrame::OnLayoutLoad(wxCommandEvent &event) 
{
	int i=event.GetId()-ID_LAYOUT_LOAD1+1;
	wxString name;
	name<<i;
	loadLayout(name);
}

NodeP lloadLayout(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	theApp->mainFrame->loadLayout(name->ptr());
	return Unbound;
}

void MainFrame::saveLayout(wxString name) 
{
	wxString fullname("Layout");
	fullname << name;
	wxString perspective=theApp->aui->SavePerspective();
	bool result=theApp->config->Write(fullname, perspective);
	result=result;
}

void MainFrame::OnLayoutSave(wxCommandEvent &event) 
{
	int id=event.GetId()-ID_LAYOUT_SAVE1+1;
	wxString name;
	name<<id;
	saveLayout(name);
}

NodeP lsaveLayout(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	theApp->mainFrame->saveLayout(name->ptr());
	return Unbound;
}

void MainFrame::OnHelp(wxCommandEvent &event) 
{
	if (console)
	{	wxHelpEvent hev;
		console->OnHelp(hev);
	}
}

void MainFrame::OnAppAbout(wxCommandEvent &event)
{
	AboutDlg aboutDlg;
	aboutDlg.ShowModal();
}

void MainFrame::OnFileNew(wxCommandEvent &event)
{
	if (_True)
		leraseAll(PairPP(Thread::This()->argStack->index));
}

void MainFrame::OnFileOpen(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	wxFileDialog fd(this, _("Choose a Logo file"),"","",
		_("Logo Files (*.lg)|*.lg|All Files (*.*)|*.*"),
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString cstr=fd.GetPath();
		NodeP pn=new String(cstr);
		if throwing return;
		PairPP asp=PairPP(th->argStack->index);
		th->argStack->push(pn);
		lload(asp);
		th->argStack->pop_till(asp);
	}
}

void MainFrame::OnFileSave(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	wxFileDialog fd(this, _("Choose a filename"),"","",
		_("Logo Files (*.lg)|*.lg|All Files (*.*)|*.*"),
		wxSAVE|wxOVERWRITE_PROMPT|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString cstr=fd.GetPath();
		NodeP pn=new String(cstr);
		if throwing return;
		PairPP asp=PairPP(th->argStack->index);
		th->argStack->push(pn);
		lsave(asp);
		th->argStack->pop_till(asp);
	}
}

void MainFrame::OnConsoleOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, _("Choose a Console file"),"","",
		_("Logo Files (*.txt)|*.txt|All Files (*.*)|*.*"),
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString name=fd.GetPath();
		console->LoadFile(name);
	}
}

void MainFrame::OnConsoleSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, _("Choose a Console file"),"","",
		_("Logo Files (*.txt)|*.txt|All Files (*.*)|*.*"),
		wxSAVE|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK)
	{	wxString name=fd.GetPath();
		console->SaveFile(name);
	}
}

void MainFrame::OnConsoleAutoSave(wxCommandEvent &event)
{
	console->autosave = !console->autosave;
}

void MainFrame::OnPictureOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, _("Choose a Picture file"),"","",
		_("Picture Files (*.zip)|*.zip|All Files (*.*)|*.*"),
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK && theApp->graph)
		theApp->graph->recorder->load(fd.GetPath().c_str());
}

void MainFrame::OnPictureSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, _("Choose a Picture file"),"","",
		_("Picture Files (*.zip)|*.zip|All Files (*.*)|*.*"),
		wxSAVE|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK && theApp->graph)
		theApp->graph->recorder->save(fd.GetPath().c_str());
}

void MainFrame::OnPictureTextOpen(wxCommandEvent &event)
{
	wxFileDialog fd(this, _("Choose a Picture Text file"),"","",
		_("Picture Files (*.dat)|*.dat|All Files (*.*)|*.*"),
		wxOPEN|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK && theApp->graph)
		theApp->graph->recorder->loadText(fd.GetPath().c_str());
}

void MainFrame::OnPictureTextSave(wxCommandEvent &event)
{
	wxFileDialog fd(this, _("Choose a Picture Text file"),"","",
		_("Picture Files (*.dat)|*.dat|All Files (*.*)|*.*"),
		wxSAVE|wxCHANGE_DIR);

	if (fd.ShowModal()==wxID_OK && theApp->graph)
		theApp->graph->recorder->saveText(fd.GetPath().c_str());
}

void MainFrame::OnPictureAutoSave(wxCommandEvent &event)
{
	graphAutosave = !graphAutosave;
}

void MainFrame::OnNodesOpen(wxCommandEvent &event)
{
	theApp->action=LOAD_NODES;
}

void MainFrame::OnNodesSave(wxCommandEvent &event)
{
	theApp->action=SAVE_NODES;
}

void MainFrame::OnNodesAutoSave(wxCommandEvent &event)
{
	nodesAutosave = !nodesAutosave;
}

void MainFrame::OnEditRepeat(wxCommandEvent &event)
{
	console->OnEditRepeat(event);
}

void MainFrame::OnEditReset(wxCommandEvent &event)
{
	console->OnEditReset(event);
}

void MainFrame::OnEditGotoCursorLine(wxCommandEvent &event)
{
	console->OnGotoLine(event);
}

void MainFrame::OnEditWrapLines(wxCommandEvent &event)
{
	if (console)
	{	wxSize size=console->GetSize();
		delete console; console=0;
		theApp->consoleHSCROLL=!theApp->consoleHSCROLL;
		console=new Console(theApp->consoleFrame, ID_CONSOLE, "", 
			wxDefaultPosition, size, 
			wxTE_MULTILINE | wxTE_PROCESS_TAB | wxTE_RICH2 | wxTE_NOHIDESEL 
			| wxHSCROLL*theApp->consoleHSCROLL,
			wxDefaultValidator, _("Console"), theApp->consoleAutosave, theApp->caretpos);
	}
}

void MainFrame::OnEditChooseLanguage(wxCommandEvent &event)
{
	wxArrayString dirs, names;
	wxArrayLong identifiers;
	theApp->getInstalledLanguages(dirs, names, identifiers);
	theApp->askUserForLanguage(dirs, names, identifiers);    
}


class MyPrintout: public wxPrintout
{
public:
	MyPrintout(const wxChar *title = _("My printout")):
		wxPrintout(title) {}
	bool OnPrintPage(int page);
	bool HasPage(int page);
	bool OnBeginDocument(int startPage, int endPage);
	void GetPageInfo(int *minPage, int *maxPage, int *selPageFrom, int *selPageTo);
};

bool MyPrintout::OnPrintPage(int page)
{
	ThreadP th=Thread::This();
	if (!th->graph)
		return false;
	wxDC *dc = GetDC();
	if (dc)
	{	if (page == 1)
		{	wxRect p=GetPaperRectPixels();
			int w=0,h=0;
			GetPageSizePixels(&w, &h);
			wxSize s=dc->GetSize();
			wxPoint d(-((p.width-w)/2+p.x)*s.GetWidth()/w, 0);
			th->graph->print(dc, d);
		}
		return true;
	}
	else
		return false;
}

bool MyPrintout::OnBeginDocument(int startPage, int endPage)
{
	if (!wxPrintout::OnBeginDocument(startPage, endPage))
		return false;

	return true;
}

void MyPrintout::GetPageInfo(int *minPage, int *maxPage, 
									  int *selPageFrom, int *selPageTo)
{
	*minPage=1;
	*maxPage=1;
	*selPageFrom=1;
	*selPageTo  =1;
}

bool MyPrintout::HasPage(int pageNum)
{
	return pageNum==1;
}


void MainFrame::OnPrint(wxCommandEvent &event)
{
	wxPrintDialogData printDialogData(*printData);

	wxPrinter printer(&printDialogData);
	MyPrintout printout("aUCBLogo");
	if (!printer.Print(this, &printout, true))
	{	if (wxPrinter::GetLastError() == wxPRINTER_ERROR)
			wxMessageBox(_("There was a problem printing.\n"
				"Perhaps your current printer is not set correctly?"), 
				_("Printing"), wxOK);
	}else
	{	*printData = printer.GetPrintDialogData().GetPrintData();
	}
}

void MainFrame::OnPrintPreview(wxCommandEvent& WXUNUSED(event))
{
	// Pass two printout objects: for preview, and possible printing.
	wxPrintDialogData printDialogData(*printData);
	wxPrintPreview *preview = new wxPrintPreview(
		new MyPrintout, new MyPrintout, &printDialogData);
	if (!preview)
	{	error(OUT_OF_MEM); return;
	}
	if (!preview->Ok())
	{	delete preview;
		wxMessageBox(_("There was a problem previewing.\n"
			"Perhaps your current printer is not set correctly?"),
			_("Previewing"), wxOK);
		return;
	}

	wxPreviewFrame *frame = new wxPreviewFrame(
		preview, this, _("Print Preview"), 
		wxPoint(100, 100), wxSize(800, 1024));
	if (!frame)
	{	error(OUT_OF_MEM); return;
	}
	frame->Centre(wxBOTH);
	frame->Initialize();
	frame->Show(TRUE);
}

void MainFrame::OnPrintSetup(wxCommandEvent& WXUNUSED(event))
{
	wxPrintDialogData printDialogData(*printData);
	wxPrintDialog printerDialog(this, &printDialogData);

//	printerDialog.GetPrintDialogData().SetSetupDialog(true);
	printerDialog.ShowModal();

	*printData = printerDialog.GetPrintDialogData().GetPrintData();
}

void MainFrame::OnPageSetup(wxCommandEvent& WXUNUSED(event))
{
	*pageSetupData = *printData;

	wxPageSetupDialog pageSetupDialog(this, pageSetupData);
	pageSetupDialog.ShowModal();

	*printData = pageSetupDialog.GetPageSetupData().GetPrintData();
	*pageSetupData = pageSetupDialog.GetPageSetupData();
}
