//___________________________________________________________________
//
//                       Class MemoryWindow
//
//                           Definitons
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/frame.h>
#include <wx/dcclient.h>
#include <wx/dcmemory.h>
#include <wx/textctrl.h>
#include <wx/tooltip.h>
#include <wx/tipwin.h>
#include <wx/scrolwin.h>
#include <wx/app.h>
#include <wx/menu.h>
#include <wx/numdlg.h>
#include <wx/config.h>
#include <wx/aui/aui.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>

#include "Main.h"
#include "Node.h"
#include "Object.h"
#include "UInt8.h"
#include "String.h"
#include "Pair.h"
#include "File.h"
#include "Color.h"
#include "Segment.h"
#include "Error.h"
#include "MainFrame.h"
#include "MemoryWindow.h"
#include "LogoApp.h"
#include "NodePStack.h" 
#include "Thread.h"
#include "Aui.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#else
	static char THIS_FILE[]= __FILE__;
#endif

#define NEED_MEM(x)											\
	if (!x)														\
	{	delete theApp->heapReserve;						\
		wxString msg;											\
		msg << THIS_FILE << "(" << __LINE__ << "):"	\
			#x << _(": Insufficient Memory");			\
		wxMessageBox(msg,	aUCBLogo+" "+_("Error"));		\
		return;													\
	}

IMPLEMENT_DYNAMIC_CLASS(MemoryFrame, wxPanel)

BEGIN_EVENT_TABLE(MemoryFrame, wxPanel)
	EVT_MENU (ID_MEMORY_CHANGE_BLOCK_SIZE, MemoryFrame::OnChangeBlockSize)
	EVT_MENU (ID_MEMORY_CHANGE_BLOCK_BORDER, MemoryFrame::OnChangeBlockBorder)
END_EVENT_TABLE()

MemoryFrame::MemoryFrame(wxWindow* parent_, wxWindowID id, 
						const wxString& title, 
						const wxPoint& pos, 
						const wxSize& size, 
						long style, 
						const wxString& name,
						int blockx, int blocky, int border):
wxPanel(parent_, id, pos, size, style, name)
{
	optionsMenu = new wxFlatMenu;
	NEED_MEM(optionsMenu);
	optionsMenu->Append(ID_MEMORY_CHANGE_BLOCK_SIZE, _("Change Block&size"), _("Change Blocksize"));
	optionsMenu->Append(ID_MEMORY_CHANGE_BLOCK_BORDER, _("Change Block &Border"), _("Change Block Border"));
	wxFlatMenuBar *menuBar=new wxFlatMenuBar(this);
	NEED_MEM(menuBar);
	menuBar->Append(optionsMenu  ,_("&Options"));

	memoryWindow=new MemoryWindow(this, -1, _("Memory"), 
		wxPoint(0, 0), wxSize(size.x-20,size.y),
		wxVSCROLL, _("Memory"),
		blockx, blocky, border);

	wxBoxSizer *sizer=new wxBoxSizer(wxVERTICAL);
	sizer->Add(menuBar, 0, wxEXPAND);
	sizer->Add(memoryWindow, 1, wxEXPAND);
	SetSizer(sizer);

	::memoryWindow=memoryWindow;
	show();
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuMemory->Check(true);
}

MemoryFrame::~MemoryFrame()
{
	GetPosition(&theApp->memposx , &theApp->memposy );
	GetSize    (&theApp->memsizex, &theApp->memsizey);
	if (memoryWindow->tip)
		delete memoryWindow->tip;
	theApp->memoryFrame=0;
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuMemory->Check(false);
}

void MemoryFrame::show(bool shown)
{
	wxAuiPaneInfo &pane=theApp->aui->GetPane(this);
	if (pane.IsOk())
	{	pane.Show();
	}else
	{	wxString pi;
		if (theApp->config->Read("MemoryPaneInfo", &pi))
			theApp->aui->LoadPaneInfo(pi, pane);
		else
			pane=wxAuiPaneInfo()
				.Name("Memory").Caption(_("Memory"))
				.Layer(1).Bottom()
				.MaximizeButton().MinimizeButton().PinButton()
				.DestroyOnClose();
		theApp->aui->AddPane(this, pane);
	}
	theApp->aui->Update();
}


void MemoryFrame::OnChangeBlockSize(wxCommandEvent& WXUNUSED(event))
{
	int bx=wxGetNumberFromUser( 
		_("Enter the new width for the visible blocks"), _("1..100"), _("Change BlocksizeX"), 
		memoryWindow->blockx-memoryWindow->border, 1, 100, this);
	if (bx)
		memoryWindow->blockx = bx+memoryWindow->border;

	int by=wxGetNumberFromUser( 
		_("Enter the new height for the visible blocks"), _("1..100"), _("Change BlocksizeY"), 
		memoryWindow->blocky-memoryWindow->border, 1, 100, this);
	if (by)
		memoryWindow->blocky = by+memoryWindow->border;
	Refresh();
}

void MemoryFrame::OnChangeBlockBorder(wxCommandEvent& WXUNUSED(event))
{
	int max=0;
	if (memoryWindow->blockx < memoryWindow->blocky)
		max=memoryWindow->blockx;
	else
		max=memoryWindow->blocky;
	memoryWindow->border = wxGetNumberFromUser( 
		_("Enter the new width for the visible blocks"), _("0..10"), _("Change BlocksizeX"), 
		memoryWindow->border, 0, max, this);
	Refresh();
}

IMPLEMENT_DYNAMIC_CLASS(MemoryWindow, wxScrolledWindow)

BEGIN_EVENT_TABLE(MemoryWindow, wxScrolledWindow)
	EVT_ERASE_BACKGROUND(MemoryWindow::OnErase) 
	EVT_SIZE		(MemoryWindow::OnSize		)
	EVT_PAINT	(MemoryWindow::OnPaint		)
	EVT_MOTION	(MemoryWindow::OnMouseMove	)
END_EVENT_TABLE()

MemoryWindow::MemoryWindow(wxWindow* parent_, wxWindowID id, 
	const wxString& title, 
	const wxPoint& pos, 
	const wxSize& size, 
	long style, 
	const wxString& name,
	int blockx_, int blocky_, int border_):
//wxFrame(parent, id, title, pos, size, style, name)
wxScrolledWindow(parent_, id, pos, size, style, name),
parent(parent_)
{
	width=size.x;
	height=size.y*2;
	SetVirtualSize(width, height);
	GetClientSize(&winwidth, &winheight);
	blockx=blockx_;
	blocky=blocky_;
	border=border_;
	SetScrollRate(0, blocky);
	EnableScrolling(false, true);

	inMouseMove=false;
	inOnPaint=false;

	pen=new wxPen();
	brush=new wxBrush();
#ifdef WIN32
	wxToolTip::SetDelay(0);
#endif
	tip=0;

	palsize=0x1000;
	palette   =new Color[palsize];
	paletteBGR=new Color[palsize];
	int i;
	for (i=0; i< palsize; i++)
	{	Color &c=palette[i];
		switch (i)
		{
		case nt_ENUM:		c=0;			break;
		case UNBOUND:		c=0xffff00;	break;
		case STRING:
		case BACKSLASHED:
		case VBAR_STRING:	c=0xff;		break;
		case INT_:			c=0xff00;	break;
		case FLOAT_:		c=0xff00;	break;
		case COMPLEX:		c=0xff00;	break;
		case NINT16:		c=0xff00;	break;
		case NINT8:			c=0xff00;	break;
		case NUINT8:		c=0xff00;	break;
		case QUOTE:			c=0xff0000;	break;
		case COLON:			c=0xff0000;	break;
		case OBJECT:		c=0xff7f7f;	break;
		case PAIR:			c=0xff0000;	break;
		case LINE:			c=0xff0000;	break;
		case RUNPARSED:	c=0xff0000;	break;
		case TREE:			c=0xff0000;	break;
		case TODO:			c=0xff0000;	break;
		case FILE_:			c=0xff0000;	break;
		case PROC_:
		case INFIXPROC:	c=0xff0000;	break;
		case COMPILED:		c=0xff7fff;	break;
		case ARRAY:			c=0xffff;	break;
		case BITMAP_:		c=0xffff;	break;
		case TEXTURE:		c=0xffff;	break;
		case TURTLE_:		c=0xffff;	break;
		case TABLE:			c=0xffff;	break;
		case STRUCT:		c=0xffff;	break;
		case GRAPHIC:		c=0xffff;	break;
		case DYNAMIC_LIB:	c=0xffff;	break;
		case INTARRAY:		c=0xffff;	break;
		case INT16ARRAY:	c=0xffff;	break;
		case FLOATARRAY:	c=0xffff;	break;
		case PRIM:
		case MACRO:
		case TAILFORM:
		case INFIX:			c=0xff00ff;	break;
		default:				c=0xffffff;
		}
	}
	for (i=0; i< palsize; i++)
		paletteBGR[i]=	((palette[i] & 0xff) << 16)
						|   (palette[i] & 0xff00)
						|	((palette[i] & 0xff0000) >> 16)
						|  0xff000000;
}

MemoryWindow::~MemoryWindow() 
{
	delete pen;
	delete brush;
	delete palette;
	delete paletteBGR;
	if (theApp && theApp->mainFrame)
		theApp->mainFrame->viewMenuMemory->Check(false);
	memoryWindow=0;
	theApp->memoryFrame=0;
}

void MemoryWindow::OnErase(wxEraseEvent &event)
{
//	event.Skip();
//	Refresh(false);
//	Update();
//	wxPaintEvent pev;
//	OnPaint(pev);
}

void MemoryWindow::OnSize(wxSizeEvent &event)
{
	wxSize size=event.GetSize();
	width=size.x-20;
	SetVirtualSize(width, height);
	SetScrollRate(0, blocky);
//	Refresh();
//	Update();
	wxPaintEvent pev;
	OnPaint(pev);
//	event.Skip();
}

void MemoryWindow::OnPaint(wxPaintEvent &event)
{
//	if (!dc)
//		return;
	if (inOnPaint)
	{	event.Skip();
		return;
	}
	inOnPaint=true;
	wxPaintDC pdc(this);
	pdc.DestroyClippingRegion();
//	wxPoint pt = dc->GetDeviceOrigin();
//	dc->SetDeviceOrigin(pt.x,
//                      pt.y-m_yScrollPosition*m_yScrollPixelsPerLine);
	GetClientSize(&winwidth, &winheight);
	SetVirtualSize(winwidth, height);
	double y=m_yScrollPosition*m_yScrollPixelsPerLine;
	int size=winwidth*(winheight/blocky+3)*blocky;
	ColorP pixels=new Color[size], 
		pix=pixels, 
		end=&pixels[(size/winwidth)*winwidth];
	if (!pixels)
	{	error(OUT_OF_MEM); inOnPaint=false; return;
	}
	int x=0, restx=winwidth % blockx, maxx=winwidth-restx-blockx, 
		n=0, nodeNr=int(y/blocky)*(winwidth/blockx);
	SegmentP seg;
	pix+=winwidth*blocky-int(fmod(y/blocky, 1)*winwidth)/winwidth;
	for (seg=Segment::start;  seg;  seg=seg->prev) 
	{	if (nodeNr-n >= 0  &&  nodeNr-n < GC_::segment_size)
		{	for (Object *nd=&seg->base[nodeNr-n]; 
					unsigned(nd) <= seg->maxu && pix < end; nd++)
			{	int c;
				if (nd->type < palsize)
					c=paletteBGR[nd->type];
				else
					c=0xffffffff;
				ColorP start=pix, first=pix;
				for (int j=blocky-border; j && pix < end; j--)
				{	for (int i=blockx-border; i && pix < end; i--)
						*pix++=c;	// don't mess with SetPixel, this way is much faster
					for (int i=border; i && pix < end; i--)
						*pix++=0;
					first+=winwidth;
					pix=first;
				}
				pix=start+blockx;
				x+=blockx;
				if (x > maxx)
				{	start=first=pix;
					for (int j=blocky-border; j && pix < end; j--)
					{	for (int i=restx; i && pix < end; i--)
							*pix++=0;
						first+=winwidth;
						pix=first;
					}
					pix=start+restx;
					x=0;
					pix+=(blocky-border-1)*winwidth;
					for (int i=winwidth*border; i && pix < end; i--)
						*pix++=0;
				}
			}
			if (seg)
				seg=seg->prev;
			break;
		}
		n+=GC_::segment_size;
	}
	for (;  seg;  seg=seg->prev) 
	{	for (Object *nd=seg->base; 
			unsigned(nd) <= seg->maxu && pix < end; nd++)
		{	int c;
			if (nd->type < palsize)
				c=paletteBGR[nd->type];
			else
				c=0xffffffff;
			ColorP start=pix, first=pix;
			for (int j=blocky-border; j && pix < end; j--)
			{	for (int i=blockx-border; i && pix < end; i--)
					*pix++=c;	// don't mess with SetPixel, this way is much faster
				for (int i=border; i && pix < end; i--)
					*pix++=0;
				first+=winwidth;
				pix=first;
			}
			pix=start+blockx;
			x+=blockx;
			if (x > maxx)
			{	start=first=pix;
				for (int j=blocky-border; j && pix < end; j--)
				{	for (int i=restx; i && pix < end; i--)
						*pix++=0;
					first+=winwidth;
					pix=first;
				}
				pix=start+restx;
				x=0;
				pix+=(blocky-border-1)*winwidth;
				for (int i=winwidth*border; i && pix < end; i--)
					*pix++=0;
			}
		}
		if (pix >= end)
			break;
	}
	if (pix < end)
	{	ColorP first=pix;
		for (int j=blocky-border; j>0 && pix < end; j--)
		{	for (int i=winwidth-x; i>0 && pix < end; i--)
				*pix++=0x80808080;
			first+=winwidth;
			pix=first;
		}
		pix-=x;
		for (int i=x; i>0 && pix < end; i--)
			*pix++=0;
		while (pix < end)
			*pix++=0x80808080;
	}
	pix=pixels;
	charP ch=charP(pixels);
	while (pix < end)
	{	*ch++=pix->b;
		*ch++=pix->g;
		*ch++=pix->r;
		pix++;
	}

/*	HBITMAP bmp=CreateBitmap(winwidth, winheight, 1, 32, pixels);
	HDC mdc=CreateCompatibleDC((HDC)dc->GetHDC());
	SelectObject(mdc, bmp);
	BitBlt((HDC)dc->GetHDC(), 0,y, winwidth, winheight, mdc, 0,0, SRCCOPY);
	DeleteObject(bmp);
	DeleteDC(mdc);
*/
//	wxBitmap bitmap((const char*)pixels, winwidth, winheight+2*blocky, 32);
//	wxMemoryDC mdc(&pdc);
//	mdc.SelectObject(bitmap);
//	pdc.Blit(0,-blocky-fmod(y,blocky), winwidth,winheight+2*blocky, &mdc, 0,0);
	wxImage image(winwidth, winheight+2*blocky, (unsigned char*)pixels, true);
//	wxImage image=bitmap.ConvertToImage();
	wxBitmap bitmap2(image);
	pdc.DrawBitmap(bitmap2,int(-blocky-fmod(y,blocky)), false);
	delete pixels;
	event.Skip();
//c_out << "OnPaint ";
	inOnPaint=false;
}

void MemoryWindow::updateNode(NodeP nd)
{
//	ThreadP th=Thread::This();
//	console->dispatchMessages();
//	if (!dc || !pen || !brush)
//		return;

//	th->GuiEnter();

	int nodeNr, i=0;
	for (SegmentP seg=Segment::start;  seg;  seg=seg->prev) 
	{	if (unsigned(nd) >= unsigned(seg->base)  &&  unsigned(nd) <= seg->maxu)
		{	nodeNr=(Object*)&*nd-(Object*)seg->base;
			nodeNr+=GC_::segment_size*i;
			break;
		}
		i++;
	}
	int x,y,c;
	blocksx=winwidth/blockx;
	x=blockx*(nodeNr % blocksx);
	y=blocky*(nodeNr / blocksx);
	y-=m_yScrollPosition*m_yScrollPixelsPerLine;
	if (nd->type < palsize)
		c=palette[nd->type];
	else
		c=0xffffff;
	pen->SetColour(c);
	brush->SetColour(c);
	wxClientDC cdc(this);
	cdc.DestroyClippingRegion();
	cdc.SetPen(*pen);
	cdc.SetBrush(*brush);

//wxPoint pt = cdc.GetDeviceOrigin();
//cdc.SetDeviceOrigin(pt.x, pt.y-m_yScrollPosition*m_yScrollPixelsPerLine*blocky);
//cdc.SetUserScale(1,1);
//cdc.SetLogicalScale(1,1);
#ifdef WIN32
//cdc.SetSystemScale(1,1);
#endif
	if (blockx-border > 1 || blocky-border > 1)
		cdc.DrawRectangle(x,y, blockx-border, blocky-border);
	else
		cdc.DrawPoint(x,y);
//cdc.SetDeviceOrigin(pt.x, pt.y);
cdc.SetPen(wxNullPen);
cdc.SetBrush(wxNullBrush);
//	th->GuiLeave();
}

void MemoryWindow::OnMouseMove(wxMouseEvent &event)
{
	if (!GC_::initialized)
		return;
	if (inMouseMove)
		return;
	inMouseMove=true;

	wxClientDC cdc(this);
	wxPoint pt = cdc.GetDeviceOrigin();
	cdc.SetDeviceOrigin(pt.x,
		pt.y-m_yScrollPosition*m_yScrollPixelsPerLine);
	wxPoint p=event.GetLogicalPosition(cdc);
	cdc.SetDeviceOrigin(pt.x, pt.y);
	int x,y, nodeNr;
	x=p.x/blockx;
	y=p.y/blocky;
	blocksx=winwidth/blockx;
	nodeNr=x+y*blocksx;
	NodeP nd=0;
	int n=0;
	for (SegmentP seg=Segment::start;  seg;  seg=seg->prev)
	{	if (nodeNr-n >= 0  &&  nodeNr-n < GC_::segment_size)
		{	nd=&seg->base[nodeNr-n];
			break;
		}
		n+=GC_::segment_size;
	}
	wxString text, ndtext;
	text << nodeNr << "                            \n";
	if (nd)
	{	StringP s=0;
		s=nd->_typeof();
		if (s)
		{	if (s->is.string)
				text << s->ptr();
			else
				s=s;
			if (nd)
			{	if (nd->is.string)
				{	
					if (StringP(nd)->ptr()[StringP(nd)->length()] != '\0')
						s=String::make(StringP(nd)->ptr(), 0, 
											StringP(nd)->length(), 
											nd->type, String::copy);
					else
						s=StringP(nd);
				}
				else
					s=nd->string();
				if (s && s->is.string)
				{	text << '\n';
					ndtext=s->ptr();
					ndtext.Replace("\t", " ");
					text << ndtext;
				}
			}
		}
	}
	else
		text << _("List") << "\n[]";
	if (tip)
		delete tip;
	tip = new wxTipWindow(this, text, 1024, &tip);
	int	x0=(event.GetX()/blockx)*blockx, 
			y0=(event.GetY()/blocky)*blocky;
	ClientToScreen(&x0, &y0);
	wxRect rect(x0, y0, blockx, blocky);
	if (tip)
		tip->SetBoundingRect(rect);
	inMouseMove=false;
}
