#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#ifdef MIDI
	#include <wxMidi.h>
// link with: porttime.lib portmidi.lib pm_dll.lib wxMidi.lib 
#endif

#include "Main.h"
#include "Node.h"
#include "Pair.h"
#include "Int.h"
#include "String.h"
#include "Object.h"
#include "Error.h"
#include "NodePStack.h"
#include "Midi.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#ifdef MIDI
wxMidiSystem *midiSystem;		//ptr to MIDI package
wxMidiOutDevice *midiOutDev;	//ptr to current MIDI output device
wxMidiInDevice  *midiInDev;	//ptr to current MIDI input device
#endif

void midiInit()
{
#ifdef MIDI
	midiOutDev = 0;
	midiInDev = 0;
	midiSystem = 0;
#endif
}

void midiCloseDevices()
{
#ifdef MIDI
	if (midiOutDev) 
	{	midiOutDev->Close();
		delete midiOutDev;
		midiOutDev=0;
	}
	if (midiInDev) 
	{	midiInDev->Close();
		delete midiInDev;
		midiInDev=0;
	}
#endif
}

NodeP lMidiCountDevices(PairPP)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	int numDevices = midiSystem->CountDevices();
	return new Int(numDevices);
#else
	return 0;
#endif
}

NodeP lMidiDeviceInfo(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	int devNr=arg[1]->a_int(1);
	if throwing return Unbound;
	wxMidiOutDevice *midiDev = new wxMidiOutDevice(devNr);
	if (!midiDev)
	{	error(OUT_OF_MEM); return Unbound;
	}
	PairP result=0;
	push(result, new String(midiDev->InterfaceUsed()));
	push(result, new String(midiDev->DeviceName()));
	if (midiDev->IsOutputPort()) 
		push(result, new String(_("output")));
	if (midiDev->IsInputPort()) 
		push(result, new String(_("input")));
	delete midiDev;
	return result;
#else
	return 0;
#endif
}

NodeP lMidiOpen(PairPP arg)
{
#ifdef MIDI
	ThreadP th=Thread::This();
	if (!th->security->filewrite)
	{	error(SECURITY_ERROR, new String(_("to use Midi"))); return Unbound;
	}
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	midiCloseDevices();

	int outdevNr=arg[1]->a_int(1);
	if throwing return Unbound;
	midiOutDev = new wxMidiOutDevice(outdevNr);
	if (!midiOutDev)
	{	error(OUT_OF_MEM); return Unbound;
	}
    // open output device
	wxString msg;
	long lantency = 0;
	wxMidiError err = midiOutDev->Open(lantency);
	if (err) 
	{	msg.Printf(_("Error %d in MidiOpen: %s \n"),
			err, midiSystem->GetErrorText(err));
		error(FILE_ERROR, new String(msg));
	}

	if (exist(2))
	{	int indevNr=arg[2]->a_int(2);
		if throwing return Unbound;
		//Open input device
		midiInDev = new wxMidiInDevice(indevNr);
		if (!midiInDev)
		{	error(OUT_OF_MEM); return Unbound;
		}
		err = midiInDev->Open();
		if (err)
		{	msg.Printf(_("Error %d in MidiOpen: %s \n"),
				err, midiSystem->GetErrorText(err));
			error(FILE_ERROR, new String(msg));
		}
	}
#endif
	return Unbound;
}

NodeP lMidiClose(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	midiCloseDevices();
#endif
	return Unbound;
}

NodeP lMidiMessage(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	if (!midiOutDev)
	{	error(FILE_ERROR, new String(_("Midi output device not open (use MidiOpen)")));
		return Unbound;
	}
	PairP list=arg[1]->a_list(1);
	if throwing return Unbound;
	if (!list)
	{	error(BAD_DATA, arg[1]);
		return Unbound;
	}
	wxMidiError err=wxMIDI_NO_ERROR;
	while (list && !err)
	{	int command=list->first()->a_int(1);
		if throwing return Unbound;
		next(list);
		if (!list)
		{	error(BAD_DATA, arg[1]);
			return Unbound;
		}
		int byte1=list->first()->a_int(2);
		if throwing return Unbound;
		next(list);
		if (!list)
		{	error(BAD_DATA, arg[1]);
			return Unbound;
		}
		int byte2=list->first()->a_int(3);
		if throwing return Unbound;
		next(list);
		wxMidiShortMessage msg(command, byte1, byte2);
		err = midiOutDev->Write(&msg);
		if (err) 
		{	wxString str;
			str.Printf(_("Error %d in MidiMessage: %s \n"),
				err, midiSystem->GetErrorText(err));
			error(FILE_ERROR, new String(str));
		}
	}
#endif
	return Unbound;
}

NodeP lMidiProgramChange(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	if (!midiOutDev)
	{	error(FILE_ERROR, new String(_("Midi output device not open (use MidiOpen)")));
		return Unbound;
	}
	int channel=arg[1]->a_int(1);
	if throwing return Unbound;
	int program=arg[2]->a_int(2);
	if throwing return Unbound;
	wxString str;
	wxMidiError err = midiOutDev->ProgramChange(channel, program);
	if (err) 
	{	str = wxString::Format(
			_("Error %d in MidiProgramChange\n"), err);
		error(FILE_ERROR, new String(str));
	}
#endif
	return Unbound;
}

NodeP lMidiNoteOn(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	if (!midiOutDev)
	{	error(FILE_ERROR, new String(_("Midi output device not open (use MidiOpen)")));
		return Unbound;
	}
	int channel=arg[1]->a_int(1);
	if throwing return Unbound;
	int pitch=arg[2]->a_int(2);
	if throwing return Unbound;
	int velocity=arg[3]->a_int(3);
	if throwing return Unbound;
	wxMidiError err = midiOutDev->NoteOn(channel, pitch, velocity);
	if (err) 
	{	wxString str;
		str.Printf(_("Error %d in MidiNoteOn: %s \n"),
					err, midiSystem->GetErrorText(err));
		error(FILE_ERROR, new String(str));
	}
#endif
	return Unbound;
}

NodeP lMidiNoteOff(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	if (!midiOutDev)
	{	error(FILE_ERROR, new String(_("Midi output device not open (use MidiOpen)")));
		return Unbound;
	}
	int channel=arg[1]->a_int(1);
	if throwing return Unbound;
	int pitch=arg[2]->a_int(2);
	if throwing return Unbound;
	int velocity=arg[3]->a_int(3);
	if throwing return Unbound;
	wxMidiError err = midiOutDev->NoteOff(channel, pitch, velocity);
	if (err) 
	{	wxString str;
		str.Printf(_("Error %d in MidiNoteOff: %s \n"),
					err, midiSystem->GetErrorText(err));
		error(FILE_ERROR, new String(str));
	}
#endif
	return Unbound;
}

NodeP lMidiAllSoundsOff(PairPP arg)
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	if (!midiOutDev)
	{	error(FILE_ERROR, new String(_("Midi output device not open (use MidiOpen)")));
		return Unbound;
	}
	wxMidiError err = midiOutDev->AllSoundsOff();
	if (err) 
	{	wxString str;
		str.Printf(_("Error %d in MidiAllSoundsOff: %s \n"),
					err, midiSystem->GetErrorText(err));
		error(FILE_ERROR, new String(str));
	}
#endif
	return Unbound;
}

#ifdef MIDI

#ifdef WIN32
union TimeUnion {_LARGE_INTEGER i; __int64 u;};
#else
union TimeUnion {long long  i; __int64 u;};
#endif

class MidiOutThread: public wxThread
{
	int channel, length;
	double *pitch, *duration, *volume;
public:
	static volatile int finished;
	static volatile bool stop, start;

	MidiOutThread(
		int channel_, int length_, 
		double *pitch_, double *duration_, double *volume_): 
		channel(channel_), length(length_), 
		pitch(pitch_), duration(duration_), volume(volume_)
	{}

	ExitCode Entry()
	{	//timeBeginPeriod(1);
		while (!start)
			Sleep(1);		
		double *p=pitch, *d=duration, *v=volume;
		int onote=0, ov=0;
		for (int i=length;  i>0;  i--)
		{	wxMidiError err = midiOutDev->NoteOff(channel, onote, ov);
			if (err)
				break;
			int note=int(*p), 
				frac=127+int(127*fmod(*p,1));
			if (note < 255)
			{	err = midiOutDev->NoteOn(channel, note, int(*v));
				if (err) 
					break;
				if (abs(frac)>0.01)
				{	wxMidiShortMessage msg(224+channel, 0, frac);
					err = midiOutDev->Write(&msg);
					if (err) 
						break;
				}
			}
			int n=int(*d);
			TimeUnion f,c;
#ifdef WIN32
			QueryPerformanceCounter(&c.i);
			QueryPerformanceFrequency(&f.i);
#else
//	TODO
			c.i=clock();
			f.i=CLOCKS_PER_SEC;
#endif
			if (!f.u)
				break;

			if (n-5 > 0)
				Sleep(n-5);
			double t,t0=double(c.u)/f.u;
			do
			{
#ifdef WIN32
				QueryPerformanceCounter(&c.i);
#else
//	TODO
				c.i=clock();
#endif
				t=double(c.u)/f.u;
			} while (1000*(t-t0) < n);
			
			onote=note;	ov=int(*v);
			p++; d++; v++;
			if (stop)
				break;
		}
		//timeEndPeriod();
		finished++;
		midiOutDev->NoteOff(channel, onote, ov);
		delete pitch;
		delete duration;
		delete volume;
		return 0;
	}
};

volatile int MidiOutThread::finished=0;
volatile bool MidiOutThread::stop=false;
volatile bool MidiOutThread::start=false;

#endif \\ MIDI

NodeP lMidiOutStream(PairPP arg) 
{
#ifdef MIDI
	if (!midiSystem)
		midiSystem = wxMidiSystem::GetInstance();
	if (!midiOutDev)
	{	error(FILE_ERROR, new String(_("Midi output device not open (use MidiOpen)")));
		return Unbound;
	}
	MidiOutThread::start=false;
	MidiOutThread::stop=false;
	int channel=arg[1]->a_int(1);
	if throwing return Unbound;

	int length=0;
	double *pitch=0, *duration=0, *volume;

	if (exist(4))
	{	PairP plist=arg[2]->a_list(2);
		if throwing return Unbound;
		NodeP	dnode=arg[3];
		if (!dnode)
		{	error(BAD_DATA, arg[3], 3);
			return Unbound;
		}
		NodeP	vnode=arg[4];
		if (!vnode)
		{	error(BAD_DATA, arg[4], 4); return Unbound;
		}
		length=plist->count();
		pitch   =new double[length]; double *p=pitch;
		duration=new double[length]; double *d=duration;
		volume  =new double[length]; double *v=volume;
		if (!p || !d || !v)
		{	error(OUT_OF_MEM); return Unbound;
		}
		if (dnode->islist())
		{	PairP dlist=PairP(dnode);
			if (vnode->islist())
			{	PairP vlist=PairP(vnode);
				while (plist && plist->first() 
					 && dlist && dlist->first()
					 && vlist && vlist->first())
				{	*p++=plist->first()->a_float(2);
					*d++=dlist->first()->a_float(2);
					*v++=vlist->first()->a_float(2);
					if throwing return Unbound;
					next(plist);
					next(dlist);
					next(vlist);
				}
			}
			else if (vnode->numberP())
			{	double vi=vnode->getnumber();
				while (plist && plist->first() 
					 && dlist && dlist->first())
				{	*p++=plist->first()->a_float(2);
					*d++=dlist->first()->a_float(2);
					*v++=vi;
					if throwing return Unbound;
					next(plist);
					next(dlist);
				}
			}
			else
			{	error(BAD_DATA, arg[4], 4); return Unbound;
			}
		}
		else if (dnode->numberP())
		{	double di=dnode->getnumber();
			if (vnode->islist())
			{	PairP vlist=PairP(vnode);
				while (plist && plist->first()
					 && vlist && vlist->first())
				{	*p++=plist->first()->a_float(2);
					*d++=di;
					*v++=vlist->first()->a_float(4);
					if throwing return Unbound;
					next(plist);
				}
			}
			else if (vnode->numberP())
			{	double vi=vnode->getnumber();
				while (plist && plist->first())
				{	*p++=plist->first()->a_float(2);
					*d++=di;
					*v++=vi;
					if throwing return Unbound;
					next(plist);
				}
			}
			else
			{	error(BAD_DATA, arg[4], 4); return Unbound;
			}
		}
		else
		{	error(BAD_DATA, arg[3], 3); return Unbound;
		}
	}else
	{	PairP list=arg[2]->a_list(2);
		if throwing return Unbound;
		if (!list)
			return Unbound;
		length=int(list->count()/3);
		pitch   =new double[length]; double *p=pitch;
		duration=new double[length]; double *d=duration;
		volume  =new double[length]; double *v=volume;
		if (!p || !d || !v)
		{	error(OUT_OF_MEM); return Unbound;
		}
		while (list && list->first() 
		&& list->tail() && list->second() 
		&& list->tail()->tail() && list->third())
		{	*p++=list->first()->a_float(2);
			*d++=list->second()->a_float(2);
			*v++=list->third()->a_float(2);
			if throwing return Unbound;
			next(list);
			next(list);
			next(list);
		}
	}
	if throwing return Unbound;
	MidiOutThread *t=new MidiOutThread(channel, length, pitch, duration, volume);
	if (!t)
	{	error(OUT_OF_MEM); return Unbound;
	}
	t->Create(0x100);
	t->Run();
#endif \\ MIDI
	return Unbound;
}

NodeP lMidiOutStreamsStart(PairPP)
{
#ifdef MIDI
/*	MidiOutThread::stop=true;
	Sleep(200);
	MidiOutThread::stop=false;
*/	MidiOutThread::start=true;
#endif \\ MIDI
	return Unbound;
}

NodeP lMidiOutStreamsStop(PairPP)
{
#ifdef MIDI
	MidiOutThread::stop=true;
#endif \\ MIDI
	return Unbound;
}

NodeP lMidiOutStreamsFinished(PairPP)
{
#ifdef MIDI
	return new Int(MidiOutThread::finished);
#else
	return new Int(0);
#endif \\ MIDI
}

