
#include "stdafx.h"

#include "Main.h"
#include "GC.h"
#include "Node.h"
#include "Int.h"
#include "Float.h"
#include "Complex.h"
#include "String.h"
#include "Pair.h"
#include "Object.h"
#include "Array.h"
#include "IntArray.h"
#include "FloatArray.h"
#include "Primitive.h"
#include "File.h"
#include "Color.h"
#include "Console.h"
#include "LogoApp.h"
#include "Thread.h"		// needs installed wxWidgets headers
#include "Init.h"
#include "Error.h"

Export int __cdecl logomain(int argc, char **argv, void (__cdecl *callback)());

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif
/*
NodeP lmandelc(PairPP arg)
{
	ComplexdP znd=arg[1]->a_complexOrReal(1),
		cnd=arg[2]->a_complexOrReal(2);
	int maxiter=arg[3]->a_int(3);
//	if throwing
//		return Unbound;
	complexd z=znd->complex_, c=cnd->complex_;
	int i;
	for (i=1;  i< maxiter;  i++)
	{	z=z*z+c;
		if (norm(z) > 4)
			break;
	}
	IntP result=new Int(i);
	return result;
}

complexd sine(complexd z)
{
	double x,y;
	x=real(z); y=imag(z); 
	return complexd(sin(x)*cosh(y), cos(x)*sinh(y));
}

NodeP literate_sinec(PairPP arg)
{
	ComplexdP znd=arg[1]->a_complexOrReal(1),
		cnd=arg[2]->a_complexOrReal(2);
	int maxiter=arg[3]->a_int(3);
//	if throwing
//		return Unbound;
	complexd z=znd->complex_, c=cnd->complex_;
	int i;
	for (i=1;  i< maxiter;  i++)
	{	z=sine(z)+z*z+c;
		if (abs(real(z)) > 10 
		||  abs(imag(z)) > 10)
			break;
	}
	IntP result=new Int(i);
	return result;
}
*/

class Molecules3d
{
	bool tooFast, tooSlow, gravity;
	int topteil, maxf, cmin, 
		msizehx, msizehy, msizehz;
	intP c, banz, ri[3];
	double
		deltaTFac, sqrDeltaTFac, maxb, dopt, dbind, 
		ffein, cfein;
	doubleP r[3],v[3],a[3], f, f2, gravV;
	NodePP b;
	NodePP m;

	void faster()
	{	int k,l;
		for (k=topteil-1; k>=0; k--)
			for (l=2; l>=0; l--)
				v[l][k]=v[l][k]*deltaTFac;
		
		for (k=maxf; k>=0; k--)
		{	f[k]=f[k]*sqrDeltaTFac;
			f2[k]=f2[k]*sqrDeltaTFac;
		}
		*gravV=(*gravV)*sqrDeltaTFac;
	}

	void slower()
	{	int k,l;
		for (k=topteil-1; k>=0; k--)
			for (l=2; l>=0; l--)
				v[l][k]=v[l][k]/deltaTFac;
		
		for (k=maxf; k>=0; k--)
		{	f[k]=f[k]/sqrDeltaTFac;
			f2[k]=f2[k]/sqrDeltaTFac;
		}
		*gravV=(*gravV)/sqrDeltaTFac;
	}

	void prepareVars()
	{	int i,k;
		if (tooSlow) faster();
		if (tooFast) slower();

		tooSlow=true;
		tooFast=false;

		for (i=topteil-1; i>=0; i--)
		{	for (k=2; k>=0; k--)
				a[k][i]=0;
			c[i]=0;
		}
	}

	void energyLoss(int i, int j)
	{	int k;
		double h[3];
		intP bi;
		for (k=2; k>=0; k--)
			h[k]=(v[k][i]+v[k][j])/2;
		
		for (k=2; k>=0; k--)
		{	v[k][i]=h[k];
			v[k][j]=h[k];
		}	
		banz[i]=banz[i]+1;
		banz[j]=banz[j]+1;
		if (b[i])
		{	bi=IntArrayP(b[i])->data();
			for (k=maxb-1; k>=0; k--, bi++)
				if (!*bi)
				{	*bi=j+1;
					break;
				}
		}
		if (b[j])
		{	bi=IntArrayP(b[j])->data();
			for (k=maxb-1; k>=0; k--, bi++)
				if (!*bi)
				{	*bi=i+1;
					break;
				}
		}
	}

	bool unboundf(int i, int j)
	{	int k;
		if (b[i])
		{	intP bi=IntArrayP(b[i])->data();
			for (k=maxb-1; k>=0; k--)
				if (*bi++==j+1)
					return false;
		}
		return true;
	}

	void ionize(int i, int j)
	{	int k;
		intP bi;
		if (b[i])
		{	bi=IntArrayP(b[i])->data();
			for (k=maxb-1; k>=0; k--, bi++)
				if (*bi==j+1)
				{	*bi=0;
					banz[i]--;

					if (b[j])
					{	bi=IntArrayP(b[j])->data();
						for (k=maxb-1; k>=0; k--, bi++)
							if (*bi==i+1)
							{	*bi=0;
								banz[j]--;
								break;
							}
						break;
					}
				}
		}
	}

	NodeP removeInt(PairP list, int i)
	{
		PairP last=0, result=0;
		while (list)
		{	if (list->first() && IntP(list->first())->int_==i)
			{	if (last)
					last->settail(list->tail());
				else
					result=list->tail();
				next(list);
			}else
			{	if (!result)
					result=list;
				last=list;
				next(list);
			}
		}
		return result;
	}

public:
	Molecules3d()
	{
		tooFast=tooSlow=false;
	}

	void moveThem(
		int topteil_, int maxf_, int maxb_, 
		doubleP r_[3], doubleP v_[3], doubleP a_[3], intP ri_[3], 
		double f_[], double f2_[], int c_[], int banz_[], 
		NodePP b_, NodePP m_, 
		doubleP gravV_, 
		double dbind_, double dopt_, double deltaTFac_, 
		double ffein_, double cfein_, int cmin_, bool gravity_,
		int msizehx_, int msizehy_, int msizehz_)
	{
		ThreadP th=Thread::This();
		topteil=topteil_; maxf=maxf_; maxb=maxb_; 
		f=f_; f2=f2_; c=c_; banz=banz_;
		b=b_;	m=m_;
		gravV=gravV_;
		dbind=dbind_; dopt=dopt_; deltaTFac=deltaTFac_;
		ffein=ffein_; cfein=cfein_; cmin=cmin_; gravity=gravity_;
		msizehx=msizehx_; msizehy=msizehy_; msizehz=msizehz_;

		bool unbound;
		int i,j,k, ix,iy,iz, di;
		double d,da[3], f0,force, ms[3];
		PairP l=0;

		for (k=2; k>=0; k--)
		{	r[k]=r_[k];
			ri[k]=ri_[k];
			v[k]=v_[k];
			a[k]=a_[k];
		}
		ms[0]=msizehx;
		ms[1]=msizehy;
		ms[2]=msizehz;

		sqrDeltaTFac=deltaTFac;
		unbound=true;
	nomml:
		prepareVars();
		for (i=topteil-1; i>=0; i--)
		{	for (ix=ri[0][i]-1; ix<=ri[0][i]+1; ix++)
			{	for (iy=ri[1][i]-1; iy<=ri[1][i]+1; iy++)
				{	for (iz=ri[2][i]-1; iz<=ri[2][i]+1; iz++)
					{	
						l=PairP(ArrayP(ArrayP(m[ix])->
							data()[iy])->
							data()[iz]);
						while (l)
						{	j=IntP(l->first())->int_;
							l=l->tail();
							da[0]=r[0][i]-r[0][j];
							if (abs(da[0]) > dbind)
								continue;
							da[1]=r[1][i]-r[1][j];
							if (abs(da[1]) > dbind)
								continue;
							da[2]=r[2][i]-r[2][j];
							if (abs(da[2]) > dbind)
								continue;
							d=sqrt( 
								Float::sqr(da[0])+
								Float::sqr(da[1])+
								Float::sqr(da[2]));
							
							if (d > dopt*1.45)
								ionize(i,j);
							if (d < dopt*1.4)
							{	unbound=unboundf(i,j);
								if (unbound
								&& (banz[i] < maxb) 
								&& (banz[j] < maxb))
								{	//if not yet bound & free
									if (abs(d-dopt)/dopt < 0.5)
									{	//and d around dopt
										energyLoss(i,j); //then "emitt a Photon"
										unbound=false;
									}
								}	
							}
							if (d > dbind)
								continue;
							
							d=d*ffein;
							di=int(d);
							if (di >= maxf-1) 
								di=maxf-1;
							if (unbound)
								f0=f2[di];
							else
								f0=f[di];
							force=f0; //+(d-Int d)*(f.(di+1)-f0)
							for (k=2; k>=0; k--)
								a[k][i]=a[k][i]+da[k]*force;
						}
					}
				}
			}
		}
		for (i=topteil-1; i>=0; i--)
		{	double cd=sqrt( 
				Float::sqr(a[0][i])+
				Float::sqr(a[1][i])+
				Float::sqr(a[2][i]))*cfein;
			if (cd > 250)
			{
				tooFast=true;
				tooSlow=false;
				if (th->checkBreak() || throwing)
					return;
				goto nomml;
			}
			if (cd > cmin)
				tooSlow=false;
			c[i]=int(cd);
		}
		for (i=topteil-1; i>=0; i--)
		{	for (k=2; k>=0; k--)
				v[k][i]=v[k][i]+a[k][i];
			
			if (gravity)
				v[1][i]=v[1][i]+(*gravV);
			int 
				rxi=ri[0][i],
				ryi=ri[1][i],
				rzi=ri[2][i];
			ArrayP my;
			my=ArrayP(ArrayP(m[rxi])->data()[ryi]);
			NodeP list=0;
			list=removeInt(PairP(my->data()[rzi]), i);
			my->data()[rzi]=list;
			my->check_oldyoung(list);

			for (k=2; k>=0; k--)
				r[k][i]=r[k][i]+v[k][i];
			
			if (r[0][i] < -400+dopt || r[0][i] > 400-dopt)
			{	v[0][i]=-v[0][i];
				r[0][i]=r[0][i]+v[0][i];
			}
			if (r[1][i] < -300+dopt || r[1][i] > 300-dopt)
			{	v[1][i]=-v[1][i];
				r[1][i]=r[1][i]+v[1][i];
			}
			if (r[2][i] < -300+dopt || r[2][i] > 300-dopt)
			{	v[2][i]=-v[2][i];
				r[2][i]=r[2][i]+v[2][i];
			}
			for (k=2; k>=0; k--)
				ri[k][i]=int(floor(r[k][i]/dbind+0.5))+ms[k];
			rxi=ri[0][i];
			ryi=ri[1][i];
			rzi=ri[2][i];
			my=ArrayP(ArrayP(m[rxi])->data()[ryi]);
			list=new Pair(new Int(i), my->data()[rzi]);
			my->data()[rzi]=list;
			my->check_oldyoung(list);
		}
	}
} molecules3d;

NodeP lMolecules3dMoveThem(PairPP arg)
//	Molecules3dMoveThem 
//		topteil maxf maxb  r v a ri  f f2 c banz b m
//		gravV  dbind dopt deltaTFac
//		ffein cfein cmin gravity
{
	enum	// Interface
	{	TOPTEIL=1, MAXF, MAXB,  R, V, A, RI,  F, F2, C, BANZ, B, M,
		GRAVV,  DBIND, DOPT, DELTATFAC,
		FFEIN, CFEIN, CMIN, GRAVITY
	};
	int topteil=arg[TOPTEIL]->a_int(TOPTEIL),
		maxf=arg[MAXF]->a_int(MAXF),
		maxb=arg[MAXB]->a_int(MAXB);
	if throwing return Unbound;

	ArrayP 
		ra=arg[R]->a_array(R),
		va=arg[V]->a_array(V),
		aa=arg[A]->a_array(A),
		ria=arg[RI]->a_array(RI);
	if throwing return Unbound;
	if (ra->length() < 3)
	{	error(BAD_DATA, ra, R); return Unbound;
	}
	if (va->length() < 3)
	{	error(BAD_DATA, va, V); return Unbound;
	}
	if (aa->length() < 3)
	{	error(BAD_DATA, aa, A); return Unbound;
	}
	if (ria->length() < 3)
	{	error(BAD_DATA, ria, RI); return Unbound;
	}
	FloatArrayP rfa, vfa, afa;
	IntArrayP riia;
	doubleP r[3], v[3], a[3];
	intP ri[3];
	for (int k=2; k>=0; k--)
	{	rfa=ra->data()[k]->a_floatarray(R);
		vfa=va->data()[k]->a_floatarray(V);
		afa=aa->data()[k]->a_floatarray(A);
		riia=ria->data()[k]->a_intarray(RI);
		if throwing return Unbound;
		if (rfa->length() < topteil)
		{	error(BAD_DATA, ra, R); return Unbound;
		}
		if (vfa->length() < topteil)
		{	error(BAD_DATA, va, V); return Unbound;
		}
		if (afa->length() < topteil)
		{	error(BAD_DATA, aa, A); return Unbound;
		}
		if (riia->length() < topteil)
		{	error(BAD_DATA, riia, RI); return Unbound;
		}
		if throwing return Unbound;
		r[k]=rfa->data();
		v[k]=vfa->data();
		a[k]=afa->data();
		ri[k]=riia->data();
	}

	FloatArrayP
		ffa=arg[F]->a_floatarray(F),
		f2fa=arg[F2]->a_floatarray(F2);
	IntArrayP cia=arg[C]->a_intarray(C),
		banzia=arg[BANZ]->a_intarray(BANZ);
	if throwing return Unbound;
	if (ffa->length() < maxf)
	{	error(BAD_DATA, ffa, F); return Unbound;
	}
	if (f2fa->length() < maxf)
	{	error(BAD_DATA, f2fa, F2); return Unbound;
	}
	if (cia->length() < topteil)
	{	error(BAD_DATA, cia, C); return Unbound;
	}
	if (banzia->length() < topteil)
	{	error(BAD_DATA, banzia, BANZ); return Unbound;
	}
	doubleP 
		f=ffa->data(),
		f2=f2fa->data();
	intP c=cia->data(),
		banz=banzia->data();

	ArrayP ba=arg[B]->a_array(B);
	if (ba->length() < topteil)
	{	error(BAD_DATA, ba, B); return Unbound;
	}
	for (int i=topteil-1; i>=0; i--)
	{	IntArrayP list=ba->data()[i]->a_intarray(B);
		if throwing return Unbound;
	}
	NodePP b=NodePP(ba->data());

	ArrayP mx=arg[M]->a_array(M);
	if throwing return Unbound;
	int msizehx=-mx->origin();
	NodePP m=mx->data();
	ArrayP my=m[0]->a_array(M);
	if throwing return Unbound;
	int msizehy=-my->origin();
	ArrayP mz=my->data()[0]->a_array(M);
	if throwing return Unbound;
	int msizehz=-mz->origin();

	FloatP gravVf=FloatP(arg[GRAVV]);
	if (gravVf->type!=FLOAT_)
	{	error(BAD_DATA, gravVf, GRAVV); return Unbound;
	}
	double *gravV=&gravVf->float_,
		dbind=arg[DBIND]->a_float(DBIND),
		dopt=arg[DOPT]->a_float(DOPT),
		deltaTFac=arg[DELTATFAC]->a_float(DELTATFAC),
		ffein=arg[FFEIN]->a_float(FFEIN),
		cfein=arg[CFEIN]->a_float(CFEIN);
	int cmin=arg[CMIN]->a_int(CMIN);
	bool gravity=(arg[GRAVITY]->a_bool(GRAVITY)==1);
	if throwing return Unbound;

	molecules3d.moveThem(
		topteil, maxf, maxb,  r, v, a, ri,  f, f2, c, banz,  b, m, 
		gravV,  dbind, dopt, deltaTFac, 
		ffein, cfein, cmin, gravity,
		msizehx, msizehy, msizehz);

	return Unbound;
}
