#include "stdafx.h"
#include "Main.h"
#include "Node.h"
#include "Pair.h"
#include "Float.h"
#include "FloatArray.h"
#include "Point.h"
#include "Error.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

double sqr(double x) {return x*x;}

double NMRCarterNu(double I, double m, 
						  double nu0, double nuQ, double nuref, 
						  double eta, double K1, double K2,
						  double theta, double phi)
{
	theta*=Float::deg2rad;
	phi*=Float::deg2rad;
	double mu=cos(theta), cos2phi=cos(2*phi);
	return nu0 + (0.5*nuQ*(m-0.5)+K1*nuref/2)*(3*sqr(mu)-1) 
		- 0.5*(K2*nuref+(m-0.5)*nuQ*eta)*sqr(sin(theta))*cos2phi
		+ (sqr(nuQ)/(32*nu0))*(1-sqr(mu))
			*((102*m*(m-1)-18*I*(I+1)+39)*sqr(mu)*(1+2.0/3*eta*cos2phi)
				-(6*m*(m-1)-2*I*(I+1)+3)*(1-2.0/3*eta*cos2phi))
		+ sqr(eta)/72*sqr(nuQ)/nu0*(24*m*(m-1)-4*I*(I+1)+9-(30*m*(m-1)-2*I*(I+1)+12)*sqr(mu)
			-(51.0/2*m*(m-1)-9.0/4*I*(I+1)+39.0/4)*sqr(cos2phi)*sqr(sqr(mu)-1));
}

NodeP lNMRCarterNu(PairPP arg)
{
	enum Params
	{	_I=1, _m, 
		_nu0, _nuQ, _nuref, 
		_eta, _K1, _K2,
		_theta, _phi,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		m  =arg[_m  ]->a_float(_m  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		theta=arg[_theta]->a_float(_theta),
		phi=arg[_phi]->a_float(_phi);
	if throwing return Unbound;

	return new Float(
		NMRCarterNu(I, m, nu0, nuQ, nuref, 
						 eta, K1, K2, theta, phi));
}

NodeP lNMRCarterIntensity_oneTheta(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_theta, _nphi, _Inu0, _nnu,
		_intensity, _expo,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		theta=arg[_theta]->a_float(_theta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		expo=arg[_expo]->a_float(_expo);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dphi=360/nphi;

	double ifactors[5][5]=
	{	{1,0,0,0,0},
		{4,3,0,0,0},
		{9,8,5,0,0},
		{16,15,12,7},
		{25,24,21,16,9},
	};

	for (double m=-I+1; m<=I; m++)
	{
		double ifactor=ifactors[int(I-0.5)][int(Float::abs(m-0.5))];
		double factor=sin(Float::deg2rad*theta);
//			y=(sin x)*f+(abs cos x)^(1000*(1-f))*10*(1-f)
		if (factor!=0)
			for (double phi=Float::rnd()*dphi/factor; phi<360; phi+=dphi/factor)
			{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
											eta, K1, K2, theta, phi);
				int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
				if (inu>=0 && inu<ilen)
					intensity[inu]++;
			}
		checkThrowing();
		if throwing return Unbound;
	}
	return lintensity;
}

NodeP lNMRCarterIntensity(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_ntheta, _nphi, _Inu0, _nnu,
		_intensity, _expo,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		ntheta=arg[_ntheta]->a_int(_ntheta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		expo=arg[_expo]->a_float(_expo);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dtheta=180/ntheta,
		dphi=360/nphi;

	double ifactors[5][5]=
	{	{1,0,0,0,0},
		{4,3,0,0,0},
		{9,8,5,0,0},
		{16,15,12,7},
		{25,24,21,16,9},
	};
	for (double m=-I+1; m<=I; m++)
	{	double ifactor=ifactors[int(I-0.5)][abs(int(m-0.5))];
		for (double theta=Float::rnd()*dtheta; theta<180; theta+=dtheta)
		{	double factor=sin(Float::deg2rad*theta)*expo
				+pow(Float::abs(cos(Float::deg2rad*theta)), 1000*(1-expo))*20*(1-expo);
//			y=(sin x)*f+(abs cos x)^(1000*(1-f))*10*(1-f)
			if (factor!=0)
				for (double phi=Float::rnd()*dphi/factor; phi<360; phi+=dphi/factor)
				{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
												eta, K1, K2, theta, phi);
					int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
					if (inu>=0 && inu<ilen)
						intensity[inu]++;
				}
			checkThrowing();
			if throwing return Unbound;
		}
	}
	return lintensity;
}

NodeP lNMRCarterIntensity_MAS(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_ntheta, _nphi, _Inu0, _nnu,
		_intensity, _expo,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		ntheta=arg[_ntheta]->a_int(_ntheta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		expo=arg[_expo]->a_float(_expo);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dtheta=180/ntheta,
		dphi=360/nphi,
		thetaMAS=54.7,
		r=Float::deg2rad;

	double ifactors[5][5]=
	{	{1,0,0,0,0},
		{4,3,0,0,0},
		{9,8,5,0,0},
		{16,15,12,7},
		{25,24,21,16,9},
	};
	for (double m=-I+1; m<=I; m++)
	{	double ifactor=ifactors[int(I-0.5)][abs(int(m-0.5))];
		for (double theta=Float::rnd()*dtheta; theta<180; theta+=dtheta)
		{	double factor=sin(r*theta)*expo
				+pow(Float::abs(cos(r*theta)), 1000*(1-expo))*20*(1-expo);
//			y=(sin x)*f+(abs cos x)^(1000*(1-f))*10*(1-f)
			DPoint 
				a(	cos(r*thetaMAS), 0, sin(r*thetaMAS));

			if (factor!=0)
				for (double phi=Float::rnd()*dphi/factor; phi<360; phi+=dphi/factor)
				{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
												eta, K1, K2, theta, phi);
					int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
					if (inu>=0 && inu<ilen)
					{	DPoint 
							b(	cos(r*phi)*cos(r*theta), 
								sin(r*phi)*cos(r*theta), 
								sin(r*theta)),
							c(a*b);
						double d=c.x+c.y+c.z;
						intensity[int(inu)]+=Float::abs(d);
					}
				}
			checkThrowing();
			if throwing return Unbound;
		}
	}
	return lintensity;
}

NodeP lNMRCarterIntensity_m(PairPP arg)
{
	enum Params
	{	_I=1, 
		_nu0, _nuQ, _nuref, 
		_eta,  
		_K1, _K2,
		_ntheta, _nphi, _Inu0, _nnu,
		_intensity,
		_m,
	};
	double 
		I  =arg[_I  ]->a_float(_I  ),
		nu0=arg[_nu0]->a_float(_nu0),
		nuQ=arg[_nuQ]->a_float(_nuQ),
		nuref=arg[_nuref]->a_float(_nuref),
		eta=arg[_eta]->a_float(_eta),
		K1 =arg[_K1 ]->a_float(_K1 ),
		K2 =arg[_K2 ]->a_float(_K2 ),
		ntheta=arg[_ntheta]->a_int(_ntheta),
		nphi  =arg[_nphi  ]->a_int(_nphi  ),
		Inu0  =arg[_Inu0  ]->a_float(_Inu0),
		nnu   =arg[_nnu   ]->a_float(_nnu ),
		m=arg[_m]->a_float(_m);
	FloatArrayP lintensity=arg[_intensity]->a_floatarray(_intensity);
	if throwing return Unbound;
	doubleP intensity=lintensity->data();
	int ilen=lintensity->length();
	double 
		dtheta=180/ntheta,
		dphi=360/nphi;

	for (double theta=Float::rnd()*dtheta; theta<180; theta+=dtheta)
	{	for (double phi=Float::rnd()*dphi; phi<360; phi+=dphi)
		{	double nu=NMRCarterNu(I, m, nu0, nuQ, nuref, 
										eta, K1, K2, theta, phi);
			int inu=int(((nu-Inu0)/nnu+0.5)*ilen);
			if (inu>=0 && inu<ilen)
				intensity[inu]++;
		}
		checkThrowing();
		if throwing return Unbound;
	}
	return lintensity;
}

NodeP lGaussianSmooth(PairPP arg)
{
	enum Params
	{	_width=1, _smoothWidth, _compWidth, _data
	};
	double 
		width=arg[_width]->a_float(_width),
		smoothWidth=arg[_smoothWidth]->a_float(_smoothWidth),
		compWidth=arg[_compWidth]->a_float(_compWidth);
	FloatArrayP ldata=arg[_data]->a_floatarray(_data);
	if throwing return Unbound;
	doubleP s=ldata->data();
	int len=ldata->length();
	if (smoothWidth==0)
	{	error(BAD_DATA, arg[2], 2); return Unbound;
	}
	FloatArrayP result=new FloatArray(len);
	if throwing return Unbound;
	doubleP r=result->data();
	double factor=1.0/len*width/smoothWidth/sqrt(2.0);
	int clen=int(len/width*compWidth);
	if (clen>len)
		clen=len;
	doubleP e=(doubleP)GC_::malloc(2*clen*sizeof(double));
	for (int j=-clen; j<clen; j++)
		e[j+clen]=exp(-sqr(j*factor));
/*	for (int i=0; i<len; i++)
	{	double y=s[i];
		for (int j=-clen; j<clen; j++)
		{	int k=i+j;
			if (k >= 0 && k < len)
			{	r[k]+=y*e[j+clen];
			}
		}
	}
*/	for (int i=0; i<len; i++)
	{	double y=0;
		for (int j=-clen; j<clen; j++)
		{	int k=i-j;
			if (k >= 0 && k < len)
			{	y+=s[k]*e[j+clen];
			}
		}
		r[i]=y;
	}
	GC_::free(e); e=0;
	return result;
}

NodeP lNMRBlochSim(PairPP arg)
{
	enum Params 
	{	_Mx=1, _My, _P,
		_signalOn, _frequency, _MSamples, _B1c, _dBmax,
		_tau, _tauab, _taubc,
		_spins,
	};
	FloatArrayP nMx=arg[_Mx]->a_floatarray(_Mx);
	FloatArrayP nMy=arg[_My]->a_floatarray(_My);
	FloatArrayP nP =arg[_P ]->a_floatarray(_P );
	bool signalOn=(arg[_signalOn]->a_bool(_signalOn)!=0);
	double frequency=arg[_frequency]->a_float(_frequency);
	double MSamples=arg[_MSamples]->a_float(_MSamples);
	double B1c=arg[_B1c]->a_float(_B1c);
	double dBmax=arg[_dBmax]->a_float(_dBmax);
	double tau=arg[_tau]->a_float(_tau);
	double tauab=arg[_tauab]->a_float(_tauab);
	double taubc=arg[_taubc]->a_float(_taubc);
	int spins=arg[_spins]->a_int(_spins);
	if throwing return Unbound;
	int IMAX=nMx->count();
	double *Mx=nMx->data();
	double *My=nMy->data();
	double *P =nP ->data();
	double 
		IM=IMAX/MSamples,
		f=frequency*IM,
		inull=Float::rnd()*100, 
		tau_=tau/IM, tauab_=tauab/IM, taubc_=taubc/IM,
		t0=0.35-tauab_,
		t1=t0+2*tau_, 
		t2=t1+tauab_, 
		t3=t2+tau_,
		t4=t3+taubc_,
		t5=t4+2*tau_,
		t6=t5+tauab_;
	const int n=1000;
	B1c=B1c/frequency;
	double 
		PI2=2*Float::pi,
		gamma=f*PI2,
		T2=0.1, T2e=T2, T1=10,
		B[3], B0=1, B1=0, 
		t=t0, dt=1.0/IMAX,
		M0=1, M[n][3], *Mj=0, Mnorm=0,
		dM_dt[n][3], d2M_dt2[3], *dM_dtj,
		Bj[3], dB[n][3];
	int i0=int(t*IMAX);
	
//	srand(0);
	for (int j=0; j<n; j++)
	{	M[j][0]=0;
		M[j][1]=M0;
		M[j][2]=0;
		dM_dt[j][0]=0;
		dM_dt[j][1]=0;
		dM_dt[j][2]=0;
		dB[j][0]=0;
		dB[j][1]=sqr(sqr(Float::rnd()))*dBmax;
		dB[j][2]=0;
	}
	for (int i=i0; i<IMAX; i++)
	{	if (signalOn)
			B1=B1c*((t<t1)|((t>t2)&(t<t3))|((t>t4)&(t<t5)));
		else
			B1=B1c*(       ((t>t2)&(t<t3))|((t>t4)&(t<t5)));
		B[0]=B1*sin(PI2*f*t);
		B[1]=B0;
		B[2]=0; //B1*cos -omega*t
		for (int j=0; j<spins; j++)
		{	Mj=M[j];
			Bj[0]=B[0]+dB[j][0];
			Bj[1]=B[1]+dB[j][1];
			Bj[2]=B[2]+dB[j][2];
			dM_dtj=dM_dt[j];
			dM_dtj[0]=(gamma*(Mj[1]*Bj[2]-Mj[2]*Bj[1])-    Mj[0] /T2e);
			dM_dtj[1]=(gamma*(Mj[2]*Bj[0]-Mj[0]*Bj[2])+(M0-Mj[1])/T1);
			dM_dtj[2]=(gamma*(Mj[0]*Bj[1]-Mj[1]*Bj[0])-    Mj[2] /T2);
			d2M_dt2[0]=dM_dtj[0]-dM_dt[j][0];
			d2M_dt2[1]=dM_dtj[1]-dM_dt[j][1];
			d2M_dt2[2]=dM_dtj[2]-dM_dt[j][2];
			M[j][0]=Mj[0]+dM_dtj[0]*dt+d2M_dt2[0]*dt/2;
			M[j][1]=Mj[1]+dM_dtj[1]*dt+d2M_dt2[1]*dt/2;
			M[j][2]=Mj[2]+dM_dtj[2]*dt+d2M_dt2[2]*dt/2;
			Mnorm=sqrt(sqr(M[j][0])+sqr(M[j][1])+sqr(M[j][2]));
			if (Mnorm > M0) 
			{	M[j][0]=M[j][0]/Mnorm;
				M[j][1]=M[j][1]/Mnorm;
				M[j][2]=M[j][2]/Mnorm;
			}
		}
		t+=dt;

		double Mges[3]={0,0,0};
		for (int j=0; j<spins; j++)
		{	Mges[0]+=M[j][0];
			Mges[1]+=M[j][1];
			Mges[2]+=M[j][2];
		}
		Mges[0]=Mges[0]/spins;
		Mges[1]=Mges[1]/spins;
		Mges[2]=Mges[2]/spins;
		if (i>=0)
		{	Mx[i]=Mges[0];// M[0] in coil direction, M[1] parallel B0
			My[i]=Mges[1];
			P [i]=B1/B1c;
//			Ur[i]=char(10*sin(PI2*f*t));
		}
	}
	return Unbound;
}