#ifndef GC_included
	#include "GC.h"
#endif

#define OPERATORS_DECL				\
	virtual NodeP negate();			\
	virtual NodeP reciprocal();	\
	virtual NodeP BitNot();			\
	virtual NodeP random();			\
	virtual NodeP _int();			\
	virtual NodeP toint16();		\
	virtual NodeP toint8();			\
	virtual NodeP touint8();		\
	virtual NodeP truncate();		\
	virtual NodeP round();			\
	virtual NodeP _float();			\
	virtual NodeP _bigfloat();		\
	virtual NodeP tocomplex();		\
	virtual NodeP Sin();				\
	virtual NodeP Cos();				\
	virtual NodeP Tan();				\
	virtual NodeP ArcSin();			\
	virtual NodeP ArcCos();			\
	virtual NodeP radSin();			\
	virtual NodeP radCos();			\
	virtual NodeP radTan();			\
	virtual NodeP radArcSin();		\
	virtual NodeP radArcCos();		\
	virtual NodeP abs();				\
	virtual NodeP Sqr();				\
	virtual NodeP Sqrt();			\
	virtual NodeP exp();				\
	virtual NodeP LN();				\
	virtual NodeP Log10();			\
	virtual NodeP real();			\
	virtual NodeP imag();			\
	virtual NodeP conjugate();		\
	virtual NodeP Signum();			\
	virtual NodeP add(NodeP arg);	\
	virtual NodeP sub(NodeP arg);	\
	virtual NodeP mul(NodeP arg);	\
	virtual NodeP div(NodeP arg);	\
	virtual NodeP mod(NodeP arg);	\
	virtual NodeP Remainder(NodeP arg);	\
	virtual NodeP aShift(NodeP arg);		\
	virtual NodeP lShift(NodeP arg);		\
	virtual NodeP BitOr(NodeP arg);		\
	virtual NodeP BitAnd(NodeP arg);		\
	virtual NodeP BitXOr(NodeP arg);		\
	virtual NodeP Power(NodeP arg);		\
	virtual NodeP ArcTan(NodeP arg);		\
	virtual NodeP radArcTan(NodeP arg);	\
	virtual void xCopy(NodeP arg);		\
	virtual void xAdd(NodeP arg);			\
	virtual void xSub(NodeP arg);			\
	virtual void xMul(NodeP arg);			\
	virtual void xDiv(NodeP arg);			\
	virtual void xMod(NodeP arg);			\

#define GRAPHIC_DECL												\
	virtual void setPixel(NodeP color);						\
	virtual void setPixelXY(NodeP yin, NodeP color);	\
	virtual void setPixelXYZ(NodeP yin, NodeP zin, NodeP color);	\
	virtual void line(NodeP color);							\
	virtual void setXYZ(NodeP ynode, NodeP znode);		\

extern "C++" Export NodeP Unbound;

enum NodeType
{
	nt_LIST     =1<<0,
	nt_WORD     =1<<1,
	nt_AGGREGATE=1<<2,
	nt_PRIM     =1<<3,
	nt_STRING   =1<<4,
	nt_NUMBER   =1<<5,
	nt_PUNCT    =1<<6,
	nt_BACKSL   =1<<7,
	nt_LINE     =1<<8,
	nt_WINDOW   =1<<9,
	nt_ENUM     =1<<10,

	    UNBOUND = 0,
	     NTFREE = nt_ENUM,
	 SPARE_NODE = nt_ENUM*2,

	     STRING = nt_WORD | nt_STRING,
	BACKSLASHED = nt_WORD | nt_STRING | nt_BACKSL,
	VBAR_STRING = nt_WORD | nt_STRING | nt_BACKSL | nt_ENUM,

	       INT_ = nt_WORD | nt_NUMBER,
	     FLOAT_ = nt_WORD | nt_NUMBER | nt_ENUM,
	    COMPLEX = nt_WORD | nt_NUMBER | nt_ENUM*2,
	     NINT16 = nt_WORD | nt_NUMBER | nt_ENUM*3,
	      NINT8 = nt_WORD | nt_NUMBER | nt_ENUM*4,
	     NUINT8 = nt_WORD | nt_NUMBER | nt_ENUM*5,
	   BIGFLOAT = nt_WORD | nt_NUMBER | nt_ENUM*6,

	      QUOTE = nt_WORD | nt_PUNCT | nt_LINE,
	      COLON = nt_WORD | nt_PUNCT | nt_LINE | nt_ENUM,
	   VARPARAM = nt_WORD | nt_PUNCT | nt_LINE | nt_ENUM*2,

	     OBJECT = nt_WORD,

	       PAIR = nt_AGGREGATE | nt_LIST,
	       LINE = nt_AGGREGATE | nt_LIST | nt_LINE,
	    RUNLINE = nt_AGGREGATE | nt_LIST | nt_LINE | nt_ENUM,
	  RUNPARSED = nt_AGGREGATE | nt_LIST | nt_LINE | nt_ENUM*2,
	       TREE = nt_AGGREGATE | nt_LIST | nt_LINE | nt_ENUM*3,
	       TODO = nt_AGGREGATE | nt_LIST | nt_ENUM*4,
			FILE_ = nt_AGGREGATE | nt_LIST | nt_ENUM*5,
	       PNIL = nt_AGGREGATE | nt_LIST | nt_ENUM*6,
	 WINDOWPAIR = nt_AGGREGATE | nt_LIST | nt_ENUM*7,
	   PROCCALL = nt_AGGREGATE | nt_LIST | nt_ENUM*8,

			PROC_ = nt_AGGREGATE | nt_ENUM, 
	      ARRAY = nt_AGGREGATE | nt_ENUM*2,
	    BITMAP_ = nt_AGGREGATE | nt_ENUM*3,
	    TEXTURE = nt_AGGREGATE | nt_ENUM*4,
	    TURTLE_ = nt_AGGREGATE | nt_ENUM*5,
	      TABLE = nt_AGGREGATE | nt_ENUM*6,
	     STRUCT = nt_AGGREGATE | nt_ENUM*7,
		 GRAPHIC = nt_AGGREGATE | nt_ENUM*8,
	DYNAMIC_LIB = nt_AGGREGATE | nt_ENUM*9,
		INTARRAY = nt_AGGREGATE | nt_ENUM*10,
	 INT16ARRAY = nt_AGGREGATE | nt_ENUM*11,
	 FLOATARRAY = nt_AGGREGATE | nt_ENUM*12,
	 LOGOOBJECT = nt_AGGREGATE | nt_ENUM*13,
		BOXSIZER = nt_AGGREGATE | nt_ENUM*14,
STATICBOXSIZER = nt_AGGREGATE | nt_ENUM*15,
	   PTRTABLE = nt_AGGREGATE | nt_ENUM*16,
	  INFIXPROC = nt_AGGREGATE | nt_ENUM*17, 
		COMPILED = nt_AGGREGATE | nt_ENUM*18, 

	     WINDOW = nt_AGGREGATE | nt_WINDOW,
	      FRAME = nt_AGGREGATE | nt_WINDOW | nt_ENUM*1,
	      GRAPH = nt_AGGREGATE | nt_WINDOW | nt_ENUM*2,
	 STATICTEXT = nt_AGGREGATE | nt_WINDOW | nt_ENUM*3,
	     BUTTON = nt_AGGREGATE | nt_WINDOW | nt_ENUM*4,
	   CHECKBOX = nt_AGGREGATE | nt_WINDOW | nt_ENUM*5,
	RADIOBUTTON = nt_AGGREGATE | nt_WINDOW | nt_ENUM*6,
	TOGGLEBUTTON= nt_AGGREGATE | nt_WINDOW | nt_ENUM*7,
	     SLIDER = nt_AGGREGATE | nt_WINDOW | nt_ENUM*8,
	      GAUGE = nt_AGGREGATE | nt_WINDOW | nt_ENUM*9,
	 INTCONTROL = nt_AGGREGATE | nt_WINDOW | nt_ENUM*10,
	FLOATCONTROL= nt_AGGREGATE | nt_WINDOW | nt_ENUM*11,
	TEXTCONTROL = nt_AGGREGATE | nt_WINDOW | nt_ENUM*12,
		 LISTBOX = nt_AGGREGATE | nt_WINDOW | nt_ENUM*13,
	  CHOICEBOX = nt_AGGREGATE | nt_WINDOW | nt_ENUM*14,
	   COMBOBOX = nt_AGGREGATE | nt_WINDOW | nt_ENUM*15,
	LISTCONTROL = nt_AGGREGATE | nt_WINDOW | nt_ENUM*16,

			 PRIM = nt_PRIM,
	      MACRO = nt_PRIM | nt_ENUM,
	   TAILFORM = nt_PRIM | nt_ENUM*2,
	      INFIX = nt_PRIM | nt_ENUM*3,
};

enum Priority
{	priority_STOP  =0,
	priority_OUTPUT=1,
	priority_MAYBE =2,  priority_TAIL=2,
	priority_MACRO =3,
	prefix = 10,
	priority_assignment  = prefix-3,
	priority_logical     = prefix-2,
	priority_bit         = prefix-1,
	priority_prefix      = prefix,
	priority_comparsion  = prefix+1,
	priority_addition    = prefix+2,
	priority_product     = prefix+3,
	priority_power       = prefix+4,
	priority_factorial   = prefix+5,
	priority_negate      = prefix+6,
	priority_dot         = prefix+7,
};

struct ObjectBits
{	bool
		proc        : 1,  val        : 1,  table        : 1,
		proc_buried : 1,  val_buried : 1,  table_buried : 1,
		proc_traced : 1,  val_traced : 1,  table_traced : 1,
		proc_stepped: 1,  val_stepped: 1,  table_stepped: 1,
		proc_macro  : 1,  proc_dyn   : 1,  permanent    : 1,
		permanentobj: 1,                   table_expand : 1;
};

enum ObjectType
{	// Object flags.  Ones settable by users via bury_helper must come in threes
	// for proc, val, plist even if meaningless for some of those.
	NORMAL_OBJECT= 0,
	PROC_OBJECT =1<<0,  VAL_OBJECT =1<<1,  TABLE_OBJECT =1<<2,
	PROC_BURIED =1<<3,  VAL_BURIED =1<<4,  TABLE_BURIED =1<<5,
	PROC_TRACED =1<<6,  VAL_TRACED =1<<7,  TABLE_TRACED =1<<8,
	PROC_STEPPED=1<<9,  VAL_STEPPED=1<<10, TABLE_STEPPED=1<<11,
	PROC_MACRO  =1<<12,
	PROC_DYN    =1<<13,
	PERMANENT   =1<<14,
	PERMANENTOBJ=1<<15,
	TABLE_EXPAND =1<<16,

	 BURIED = PROC_BURIED  | VAL_BURIED  | TABLE_BURIED,
	 TRACED = PROC_TRACED  | VAL_TRACED  | TABLE_TRACED,
	STEPPED = PROC_STEPPED | VAL_STEPPED | TABLE_STEPPED,

	 PROCS =  PROC_OBJECT |  PROC_BURIED |  PROC_TRACED |  PROC_STEPPED,
	  VALS =   VAL_OBJECT |   VAL_BURIED |   VAL_TRACED |   VAL_STEPPED,
	TABLES = TABLE_OBJECT | TABLE_BURIED | TABLE_TRACED | TABLE_STEPPED,
};

class Point;
class DPoint;
class Turtle;
class wxDynamicLibrary;
class Slider;
class Frame;
class DataInputStream;
class wxDataOutputStream;

class Node: public GC_	// is the basic data structure of Logo.
{
	public:
		struct Bits
		{	bool
				list       :1,
				word       :1,
				aggregate  :1,
				primitive  :1,
				string     :1,
				number     :1,
				punct      :1,
				backslashed:1,
				line       :1,
				window     :1;
		};
		union {NodeType type; Bits is;};

		static const Node aNode;
		static const int vmtNode;
		static const int vmtPair;
		static const int vmtRunparsed;
		static const int vmtLine;
		static const int vmtTree;
		static const int boolerror;

		operator_new;
		Node(): type(UNBOUND) {}
		Node(NodeType type_): type(type_) {}

		virtual void free() {}
		virtual void inc_gc() {}
		virtual int youngs() {return 0;}
		virtual int youngsPtr(PtrToInt_ &ptrToInt) {return 0;}
		virtual void check_oldyoungs() {}
		virtual void mark_all_but_objects();

		virtual DataInputStream &read(DataInputStream &istr);
		virtual wxDataOutputStream &write(wxDataOutputStream &ostr);

		virtual void setPtrs(IntToPtr_ &intToPtr);
		virtual void aftersetPtrs();
		virtual void recreate();
		virtual void aftersetPtrs2();
		virtual void aftersetPtrs3();

		virtual NodeP copy(PtrToInt_ &ptrToInt);

		virtual StringP string();

		virtual void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		virtual void printBin(FILEP strm);
		virtual NodeP contentsTree(wxArrayString &namearr, 
											wxArrayString &valarr, 
											wxArrayPtrVoid &objarr, 
											ObjectType want, 
											ObjectType dont, 
											ObjectType must, 
											int &count, int level);
		virtual bool expanded() {return false;}
		virtual bool expandable() {return false;}
		virtual void setExpanded(bool state) {}

		virtual    int   numberP() {return 0;}
		virtual  NodeP    number();
		virtual double getnumber(); 

		OPERATORS_DECL
		virtual int compare(NodeP nd);
		virtual ObjectP intern();
		virtual ObjectP internRoot();
		virtual int count();
		virtual int _sizeof();
		virtual StringP _typeof();
		virtual bool equalP(NodeP b);
		virtual NodeP deepCopy();
		virtual NodeP deepCopyLine();
		virtual NodeP first_();
		virtual NodeP butFirst();
		virtual NodeP firsts();
		virtual NodeP butFirsts();
		virtual NodeP last();
		virtual NodeP butLast();
		virtual NodeP fPut(NodeP nd);
		virtual NodeP lPut(NodeP nd);
		virtual double _min(double d);
		virtual double _max(double d);
		virtual NodeP Norm();
		virtual NodeP Norm3();
		virtual NodeP maxNorm();
		virtual NodeP resize(int newsize);
		virtual NodeP lowPassFilter(double weight);
		virtual NodeP lowPassFilter2(double weight);
		virtual NodeP saturateAbove(double limit);
		virtual NodeP saturateBelow(double limit);
		virtual NodeP    item(int index);
		virtual NodeP    item(NodeP index, NodeP result);
		virtual NodeP   items(NodeP start, NodeP end);
		virtual NodeP  mdItem(NodeP index);
		virtual NodeP bititem(NodeP index);
		virtual NodeP setitem(int index, NodeP newitem);
		virtual NodeP setitem(NodeP index, NodeP newitem, bool safe);
		virtual NodeP setitems(NodeP start, NodeP newitems);
		virtual NodeP memberP(NodeP subobj, bool notp=false, bool substr=false);
		virtual void removeitem(NodeP index);
		virtual NodeP reverse();
		virtual NodeP rotate(int places);
		virtual NodeP shuffle();
		virtual NodeP remove(NodeP thing);
		virtual NodeP destructiveRemove(NodeP thing);
		virtual NodeP remdup();
		virtual NodeP upper();
		virtual NodeP lower();

		virtual PairP list();
		virtual ArrayP array(int origin=1);
		virtual IntArrayP intarray(int origin=1);
		virtual Int16ArrayP int16array(int origin=1);
		virtual FloatArrayP floatarray(int origin=1);
		virtual BitmapP bitmap(int width);
		virtual TableP table();

		virtual void untreeify();
		virtual FileP findInFilelist(bool remove);
		virtual Priority priority();
		virtual Priority infixPriority();
		virtual LineP gather_args(LinePP args, bool inparen, int oldPriority, ObjectPP ifnode);
		virtual NodeP readbin(FileP file);
		virtual void setParent(WindowNodeP parent_) {}
		virtual void setflag(ObjectType want, ObjectType newflag);
		virtual void clearflag(ObjectType want, ObjectType clrflag);
		virtual PairP contents(ObjectType want, ObjectType dont);
		virtual void combSortInPlace();
		virtual NodeP combSort();
		virtual int circcounter() {return 0;}
		virtual void setcirccounter(int ccnt) {}
		virtual bool circular2() {return false;}
		virtual wxWindow *wxwindow();
		virtual void draw(Turtle *turtle) {}
		GRAPHIC_DECL

		void settype_NTFREE()   {type=NTFREE;    *intP(this) = vmtNode;     }
		void settype_PAIR()     {type=PAIR;      *intP(this) = vmtPair;     }
		void settype_RUNPARSED(){type=RUNPARSED; *intP(this) = vmtRunparsed;}
		void settype_TREE()     {type=TREE;      *intP(this) = vmtTree;     }
		void settype_LINE()     {type=LINE;      *intP(this) = vmtLine;     }

		bool islist()     {return this? is.list      : true ;}
		bool isnolist()   {return this? !is.list     : false;}
		bool isaggregate(){return this? is.aggregate : true ;}
		bool isstring()   {return this? is.string    : false;}
		bool isarray()    {return this? type==ARRAY  : false;}
		bool istree()     {return this? type==TREE   : false;}

		bool equal(NodeP b) {return this? !compare(b) : b? false : true;}

		PairP maybe_quote();
		void silent_load(const char *prefix);

		NodeP paren_infix(LinePP rest, int old_priority, 
			bool inparen, bool firstInParen);

		bool circular();
		bool circular(NodeP b);

		static NodeP boolnode(bool tf);
		static void checkboolerror(int tf, NodeP a, NodeP b);

		static void incContentsArrays(wxArrayString &namearr, 
												wxArrayString &valarr, 
												wxArrayPtrVoid &objarr, 
												int &count);

		      int a_bool      (int nr);	// these are for argument type checking
		     IntP a_integer   (int nr);
		      int a_int       (int nr);
		   Int16P a_integer16 (int nr);
		    short a_int16     (int nr);
		    Int8P a_integer8  (int nr);
		     char a_int8      (int nr);
		     IntP a_pos_int   (int nr);
		   double a_float     (int nr);
		    NodeP a_number    (int nr);
	   ComplexdP a_complex   (int nr);
	   ComplexdP a_complexOrReal(int nr);
		BigFloatP a_bigfloat  (int nr);
		  StringP a_char      (int nr);
		  StringP a_string    (int nr);
		    NodeP a_name      (int nr);
		    NodeP a_bfable    (int nr);
		    PairP a_list      (int nr);
		    PairP a_runnable  (int nr);
		    LineP a_line      (int nr);
		  BitmapP a_bitmap    (int nr);
	      TableP a_table     (int nr);
	      ArrayP a_array     (int nr);
	   IntArrayP a_intarray  (int nr);
	 Int16ArrayP a_int16array(int nr);
	 FloatArrayP a_floatarray(int nr);
		    Point a_3darray   (int nr);
		    PairP a_vector_helper(bool floatok, bool three, bool neg, int nr);
		    PairP a_vector    (int nr);
		    PairP a_3dvector  (int nr);
		    PairP a_rgb       (int nr);
		    PairP a_rgba      (int nr);
		    PairP a_int_vector(int nr);
		    PairP a_pos_int_vector(int nr);
		    Color a_color			(int nr);
	 TurtleNodeP a_turtleNode	(int nr);
	 WindowNodeP a_window		(int nr);
	  FrameNodeP a_frame			(int nr);
   GraphicNodeP a_graphicNode	(int nr);
   wxDynamicLibrary *a_dynamicLibrary(int nr);
	 SliderNodeP a_slider		(int nr);
	  GaugeNodeP a_gauge			(int nr);
	  IntControlNodeP a_intControl	(int nr);
	FloatControlNodeP a_floatControl	(int nr);
			ButtonNodeP a_button			(int nr);
	ToggleButtonNodeP a_toggleButton	(int nr);
		CheckBoxNodeP a_checkBox	(int nr);
	RadioButtonNodeP a_radioButton(int nr);
			GraphNodeP a_graph		(int nr);
	 StaticTextNodeP a_staticText	(int nr);
	TextControlNodeP a_textControl(int nr);
		 ListBoxNodeP a_listBox		(int nr);
	  ChoiceBoxNodeP a_choiceBox	(int nr);
	   ComboBoxNodeP a_comboBox	(int nr);
	ListControlNodeP a_listControl(int nr);
		BoxSizerNodeP a_boxSizer	(int nr);
	wxWindow *a_wxwindow(int nr);
};

inline NodeType type_of(NodeP nd) {return nd? nd->type : PNIL;}

extern wxDataOutputStream &operator<<(wxDataOutputStream &ostr, bool &b);

Class Var;
typedef Var* VarP;

#define Node_included
