
#ifdef _DEBUG
	#define SDEBUG
#endif

class NodePStack
{
	public:
#ifdef SDEBUG
		struct Item
		{	NodeP i;
			EvalStackMarker marker;
			charP name;
			int line;
		};

		typedef Item *ItemP;

		union
		{	ItemP ibase;
			ItemP (*ibase100)[100];
		};
		ItemP iindex, imax;
#endif
		int item, maxitem;
		int size;

		union 
		{	NodeP *base;
			NodeP (*base100)[100];
		};
		NodeP *index, *max;

		NodePStack();
		NodePStack(int size_);
		~NodePStack();

		void free();
		void copyTo(NodePStack &out);
		bool grow();
		bool shrink();
		void copy(NodePStack &out, IntToPtr_ &intToPtr, PtrToInt_ &ptrToInt);

		bool operator!=(const PairPP b) {return index != (NodePP)b;}
		bool operator>=(const PairPP b) {return index >= (NodePP)b;}
		bool greaterEqual(const PairPP b) {return index >= (NodePP)b;}
		operator NodePP() {return index;}
		operator int() {return index-base;}
		PairPP operator-(const int &i) {return PairPP(index-i);}
#ifdef SDEBUG
		void 
			push(NodeP val), 
			push(NodeP val, EvalStackMarker marker, charP name, int line), 
			pop(NodeP &val, EvalStackMarker marker), 
			pop(PairP &val, EvalStackMarker marker), 
			pop(LineP &val, EvalStackMarker marker), 
			pop(ObjectP &val, EvalStackMarker marker), 
			pop(LogoObjectP &val, EvalStackMarker marker), 
			pop(ProcedureP &val, EvalStackMarker marker), 
			pop(EvalStackMarker marker),
			pushList(PairP head), 
			pushList(PairP head, EvalStackMarker marker, charP name, int line), 
			push(ArrayP a), 
			push(ArrayP a, EvalStackMarker marker, charP name, int line), 
			pop_till(PairPP sp),
			clean();

/*		NodePStack 
			&operator<<(const NodeP  val),
			&operator>>(      NodeP &val),
			&operator>>(      PairP &val),
			&operator>>(      LineP &val),
			&operator>>(    ObjectP &val),
			&operator>>(LogoObjectP &val),
			&operator>>( ProcedureP &val);
*/
		void 
			push(
				const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
				const NodeP n5, const NodeP n6, const NodeP n7, 
				EvalStackMarker marker),
			push(
				const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
				const NodeP n5, const NodeP n6, const NodeP n7, const NodeP n8,
				EvalStackMarker marker),
			pop(	NodeP &n1,   NodeP &n2, 
					NodeP &n3, ObjectP &n4, LineP &n5,  ObjectP &n6, LineP &n7,
					EvalStackMarker marker),
			pop(	PairP &n1, NodeP &n2, 
					NodeP &n3, ObjectP &n4, LineP &n5,  ObjectP &n6, LineP &n7,
					EvalStackMarker marker);
#else
	void pop() {index--;}

	void push(const NodeP  val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=val;
	}
	NodePStack &operator<<(const NodeP  val) 
	{	if (index+1 > max)
			if (!grow())
				return *this;
		*++index=val;
		return *this;
	}
	void pop(      NodeP &val) {    val =            *index-- ;}
	void pop(      PairP &val) {    val =      PairP(*index--);}
	void pop(      LineP &val) {    val =      LineP(*index--);}
	void pop(    ObjectP &val) {    val =    ObjectP(*index--);}
	void pop(LogoObjectP &val) {    val =LogoObjectP(*index--);}
	void pop( ProcedureP &val) {    val = ProcedureP(*index--);}

	void push(const NodeP n1, const NodeP n2) 
	{	if (index+2 > max)
			if (!grow())
				return;
		index[1]=n1; index[2]=n2;
		index+=2;
	}
	void push(const NodeP n1, const NodeP n2, const NodeP n3) 
	{	if (index+3 > max)
			if (!grow())
				return;
		index[1]=n1; index[2]=n2; index[3]=n3;
		index+=3;
	}
	void push(
		const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
		const NodeP n5, const NodeP n6, const NodeP n7) 
	{	if (index+7 > max)
			if (!grow())
				return;
		index[1]=n1; index[2]=n2; index[3]=n3; index[4]=n4; 
		index[5]=n5; index[6]=n6; index[7]=n7;
		index+=7;
	}
	void push(
		const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
		const NodeP n5, const NodeP n6, const NodeP n7, const NodeP n8) 
	{	if (index+8 > max)
			if (!grow())
				return;
		index[1]=n1; index[2]=n2; index[3]=n3; index[4]=n4; 
		index[5]=n5; index[6]=n6; index[7]=n7; index[8]=n8;
		index+=8;
	}
	void pop(NodeP &n1, NodeP &n2)
	{
		register NodeP *i=index;
		n2 =            i[-0] ; 
		n1 =            i[-1] ; 
		index=i-2;
	}
	void pop(ObjectP &n1, PairP &n2)
	{
		register NodeP *i=index;
		n2 =      PairP(i[-0]); 
		n1 =    ObjectP(i[-1]); 
		index=i-2;
	}
	void pop(NodeP &n1, NodeP &n2, NodeP &n3)
	{
		register NodeP *i=index;
		n3 =            i[-0] ; 
		n2 =            i[-1] ; 
		n1 =            i[-2] ; 
		index=i-3;
	}
	void pop(ObjectP &n1, NodeP &n2, NodeP &n3)
	{
		register NodeP *i=index;
		n3 =            i[-0] ; 
		n2 =            i[-1] ; 
		n1 =    ObjectP(i[-2]); 
		index=i-3;
	}
	void pop(NodeP &n1, ObjectP &n2, NodeP &n3)
	{
		register NodeP *i=index;
		n3 =            i[-0] ; 
		n2 =    ObjectP(i[-1]); 
		n1 =            i[-2] ; 
		index=i-3;
	}
	void pop(ObjectP &n1, ObjectP &n2, PairP &n3)
	{
		register NodeP *i=index;
		n3 =      PairP(i[-0]); 
		n2 =    ObjectP(i[-1]); 
		n1 =    ObjectP(i[-2]); 
		index=i-3;
	}
	void pop(NodeP &n1, ObjectP &n2, LineP &n3)
	{
		register NodeP *i=index;
		n3 =      LineP(i[-0]); 
		n2 =    ObjectP(i[-1]); 
		n1 =            i[-2] ; 
		index=i-3;
	}
	void pop(
		NodeP &n1, NodeP &n2, 
		NodeP &n3, ObjectP &n4, LineP &n5, 
		ObjectP &n6, LineP &n7)
	{
		register NodeP *i=index;
		n7 =      LineP(*i); 
		n6 =    ObjectP(i[-1]); 
		n5 =      LineP(i[-2]); 
		n4 =    ObjectP(i[-3]); 
		n3 =            i[-4] ; 
		n2 =            i[-5] ; 
		n1 =            i[-6] ; 
		index=i-7;
	}
	void pop(
		PairP &n1, NodeP &n2, 
		NodeP &n3, ObjectP &n4, LineP &n5, 
		ObjectP &n6, LineP &n7)
	{
		register NodeP *i=index;
		n7 =      LineP(*i); 
		n6 =    ObjectP(i[-1]); 

		n5 =      LineP(i[-2]); 
		n4 =    ObjectP(i[-3]); 
		n3 =            i[-4] ; 
		n2 =            i[-5] ; 
		n1 =      PairP(i[-6]); 
		index=i-7;
	}
	void pushList(PairP head);
	void push(ArrayP a);
	void pop_till(PairPP sp) {index=NodePP(sp);}
	void clean() {pop_till(PairPP(base));}
#endif
		void pop_till(int i) 
		{	pop_till((PairPP)&base[i]);
		}
		void mark(), markPtr(PtrToInt_ &ptrToInt);

		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);

		PairP to_list(PairPP sp), to_list_reversed(PairPP sp);
		PairP to_list(int i), to_list_reversed(int i);
};
