#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include "Main.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "String.h"
#include "Int.h"
#include "Pair.h"
#include "Procedure.h"
#include "Primitive.h"
#include "Object.h"
#include "Table.h"
#include "LogoObject.h"
#include "Init.h"
#include "Error.h"
#include "Thread.h"
#include "Stack.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//#define _DEBUG_PRINT_OBJECTS 1

void Object::setname(StringP    nd) 
{	_name = nd;  check_oldyoung(nd);
}
void Object::setprim(PrimitiveP nd) 
{	_var = nd;  
	bit.proc=true;
	bit.val=false;
	check_oldyoung(nd);
}
void Object::setvar  (NodeP      nd) 
{	_var = nd;  
	bit.proc=false;
	bit.val=(nd != Unbound);
	check_oldyoung(nd);
}
void Object::setvarPermanent(NodeP nd) 
{	_var = nd;  
	bit.proc=false;
	bit.val=true;
	bit.permanent=true;
	check_oldyoung(nd);
}
void Object::setproc_or_prim(NodeP nd) 
{	_var = nd;  
	bit.proc=(nd!=Unbound);
	bit.val=false;
	check_oldyoung(nd);
}
void Object::setproc   (ProcedureP nd) 
{	wxASSERT(nd==Unbound || nd->type==PROC_ || nd->type==INFIXPROC);
	_var = nd;  
	bit.proc=(nd!=Unbound);
	bit.val=false;
	check_oldyoung(nd);
}
void Object::settable(TableP nd) 
{	_table= nd;
	check_oldyoung(nd);
	bit.table=(_table!=Unbound);
}
void Object::setparent(ObjectP nd) 
{	wxASSERT(!nd || nd->type==OBJECT);	
	_parent = nd;
	check_oldyoung(nd);
}
void Object::setowner(ObjectP nd) 
{	wxASSERT(!nd || nd->type==OBJECT);	
	_owner = nd;
	check_oldyoung(nd);
}

Object::Object(): 
Node(OBJECT), _tail(0), _name(0), _var(0), _table(0), _parent(0), _owner(0), 
creationLine(0), creationFile(0), dummy(0), dummy2(0)
{
	flag=ObjectType(0);
}

Object::Object(StringP name_, ObjectP tail_): 
Node(OBJECT), 
_var(Unbound), _table(TableP(Unbound)), _parent(0), _owner(0), 
creationLine(0), creationFile(0), dummy(0), dummy2(0), flag(NORMAL_OBJECT)
{
	ThreadP th=Thread::This();
	setname(name_); 
	settail(tail_);
	setowner(th->user);
}

void Object::free() 
{
}

void Object::inc_gc() 
{	fifo << name() << var() << table() << tail() << parent() << owner();
}
int Object::youngs() 
{	return gmark(name())+gmark(var())
			+gmark(table())+gmark(tail())
			+gmark(parent())+gmark(owner());
}
int Object::youngsPtr(PtrToInt_ &ptrToInt) 
{	return 
		 gmarkPtr(name(), ptrToInt)
		+gmarkPtr(var(), ptrToInt)
		+gmarkPtr(table(), ptrToInt)
		+gmarkPtr(tail(), ptrToInt)
		+gmarkPtr(parent(), ptrToInt),
		+gmarkPtr(owner(), ptrToInt);
}
void Object::check_oldyoungs() 
{	check_oldyoung(name()); 
	check_oldyoung(var()); 
	check_oldyoung(table()); 
	check_oldyoung(tail());
	check_oldyoung(parent());
	check_oldyoung(owner());
}

void Object::setPtrs(IntToPtr_ &intToPtr)
{	settail        (ObjectP((Object*)intToPtr.find(int(_tail ))));
	setname        (StringP((String*)intToPtr.find(int(_name ))));
	setvar         (          (Node*)intToPtr.find(int(_var  )));
	settable       (  TableP((Table*)intToPtr.find(int(_table))));
	setparent      (ObjectP((Object*)intToPtr.find(int(_parent))));
	setowner       (ObjectP((Object*)intToPtr.find(int(_owner))));
}
DataInputStream &Object::read(DataInputStream &istr) 
{	return istr >> (int&)_tail >> (int&)_name 
		>> (int&)_var >> (int&)_table  >> (int&)_parent >> (int&)_owner
		>> tint;
}
wxDataOutputStream &Object::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(_tail) 
		<< GC_::ptrToInt->find(_name)
		<< GC_::ptrToInt->find(_var )
		<< GC_::ptrToInt->find(_table)
		<< GC_::ptrToInt->find(_parent)
		<< GC_::ptrToInt->find(_owner)
		<< tint;
}

NodeP Object::copy(PtrToInt_ &ptrToInt)
{	ObjectP obj=new Object();
	if throwing return Unbound;
	obj->_tail=ObjectP(ptrToInt.find(tail()));
	obj->_name=StringP(ptrToInt.find(name()));
	obj->_var=NodeP(ptrToInt.find(var()));
	obj->_table=TableP(ptrToInt.find(table()));
	obj->_parent=ObjectP(ptrToInt.find(parent()));
	obj->_owner=ObjectP(ptrToInt.find(owner()));
	obj->bit=bit;
	return obj;
}

StringP Object::longName()
{
	if (!parent())
		return name();
	StringP dblcolon=new String("::"), str=name();
	if throwing return _Null_Word->name();
	ThreadP th=Thread::This();
	wxASSERT(!parent() || (parent() && parent()->type==OBJECT));
	ObjectP obj=parent();
	while (obj && obj!=th->namespceobj)
	{	str=new String(dblcolon, str);
		if throwing return _Null_Word->name();
		str=new String(obj->name(), str);
		if throwing return _Null_Word->name();
		wxASSERT(!obj->parent() || (obj->parent() && obj->parent()->type==OBJECT));
		obj=obj->parent();
	}
	return str;
}

StringP Object::fullName()
{
	if (!parent())
		return name();
	StringP dblcolon=new String("::"), str=name();
	if throwing return _Null_Word->name();
	ThreadP th=Thread::This();
	wxASSERT(!parent() || (parent() && parent()->type==OBJECT));
	ObjectP obj=parent();
	while (obj)
	{	str=new String(dblcolon, str);
		if throwing return _Null_Word->name();
		str=new String(obj->name(), str);
		if throwing return _Null_Word->name();
		wxASSERT(!obj->parent() || (obj->parent() && obj->parent()->type==OBJECT));
		obj=obj->parent();
	}
	return str;
}

bool Object::inNamespace(ObjectP namespce)
{
	ObjectP p=this;
	while (p)
	{	if (p==namespce)
			return true;
		p=p->parent();
	}
	return false;
}

void Object::print() 
{	c_out << name() << ' ';
	if (dribbler) dribbler << name() << ' ';
}

void Object::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
#ifdef _DEBUG_PRINT_OBJECTS
	fullName()->print(strm, width, depth, noBrackets-1, noSpacers-1);
#else
//	::print(strm, width, 'O');
	name()->print(strm, width, depth, noBrackets-1, noSpacers-1);
#endif
}

void Object::printBin(FILEP strm)
{
	int width=Int::max;
	name()->print(strm, width,-1);
}

int Object::numberP() {return name()->numberP();}
double Object::getnumber() {return name()->getnumber();}
NodeP Object::number() {return name()->number();}

int Object::compare(NodeP nd) 
{
	if (!nd) return Node::boolerror;
	if (this==nd) return 0;
	else return name()->compare(nd);
}

int Object::_sizeof() {return _name->_sizeof();}

NodeP Object::fPut(NodeP nd)
{	return name()->fPut(nd);
}

NodeP Object::lPut(NodeP nd)
{	return name()->lPut(nd);
}

NodeP Object::item(int i) {return name()->item(i);}

NodeP Object::item(NodeP index, NodeP result) 
{	return name()->item(index, result);
}

NodeP Object::items(NodeP start, NodeP end) 
{	return name()->items(start, end);
}

NodeP   Object::bititem(NodeP index) 
{	return name()->bititem(index);
}

NodeP Object::memberP(NodeP subobj, bool notp, bool substr)
{	return name()->memberP(subobj, notp, substr);
}

NodeP Object::setitem(NodeP index, NodeP newitem, bool safe)
{	return _name->setitem(index, newitem, safe);
}

void Object::removeitem(NodeP index)
{	_name->removeitem(index);
}

NodeP Object::reverse()
{
	return _name->reverse();
}

NodeP Object::remove(NodeP thing) {return name()->remove(thing);}
NodeP Object::remdup() {return name()->remdup();}
NodeP Object::upper() {return _name->upper();}
NodeP Object::lower() {return _name->lower();}
PairP Object::list() {return name()->list();}
ArrayP Object::array(int origin) {return string()->array(origin);}

PairP Object::contents(ObjectType want, ObjectType dont)
{
	ThreadP th=Thread::This();
	PairP list=0;
	if (!(dont & flag)  &&  table()!=Unbound  &&  (!th->user || th->user==owner()))
		list=table()->contents(want, dont);

	if (want & flag  &&  !(dont & flag)  &&  (!th->user || th->user==owner()))
		push(list, longName());

	return list;
}

PairP Object::contentsNameVar(ObjectType want)
{
	ThreadP th=Thread::This();
	PairP contentsList=0;
	NodeP v=0;
	if (want & flag  &&  (!th->user || th->user==owner()))
		v=var();
	if ((bit.val || bit.table) &&  (!th->user || th->user==owner()))
		push(contentsList, 
		new Pair(name(), 
		new Pair(v, 
		new Pair(table()->contentsNameVar(want), 0))));
	return contentsList;
}

NodeP Object::contentsTree(wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont,
									ObjectType must, 
									int &count, int level)
{
	objarr[count]=this;

	namearr[count]=wxString(' ',level-1);
	if (var() && var()->expandable())
		if (var()->expanded())
			namearr[count]+='-';
		else
			namearr[count]+='+';
	namearr[count]+=name()->ptr();

	NodeP v=Unbound;
	if (var()!=Unbound  &&  (want & flag)  &&  !(dont & flag))
		if (var())
		{	if (expanded())
			{	if (var()->type==OBJECT)
				{	valarr[count]=ObjectP(var())->name()->_typeof()->ptr();
					incContentsArrays(namearr, valarr, objarr, count);
					v=ObjectP(var())->name()->contentsTree(
						namearr, valarr, objarr, want, dont, must, count, level+1);
				}else if (var()->type==LOGOOBJECT)
				{	valarr[count]=LogoObjectP(var())->name()->ptr();
					incContentsArrays(namearr, valarr, objarr, count);
					v=var()->contentsTree(
						namearr, valarr, objarr, want, dont, must, count, level+1);
				}else
				{	valarr[count]=var()->_typeof()->ptr();
					incContentsArrays(namearr, valarr, objarr, count);
					v=var()->contentsTree(
						namearr, valarr, objarr, want, dont, must, count, level+1);
				}
			}else
			{	StringP str=0;
				if (var()->is.string)
					str=StringP(var())->expandFormat();
				else if (var()->type==OBJECT)
					str=ObjectP(var())->name()->expandFormat();
				else
					str=var()->string();
				if (str && str->is.string)
					valarr[count]=str->noFormat()->ptr();
				v=str;
				incContentsArrays(namearr, valarr, objarr, count);
			}
		}else
		{	valarr[count]="[]";
			v=0;
			incContentsArrays(namearr, valarr, objarr, count);
		}

	NodeP tab=Unbound;
	
	if (bit.table)
		if (bit.table_expand)
		{	tab=table()->contentsTree(
				namearr, valarr, objarr, want, dont, must, count, level+1);
		}
		
	if ((!tab || tab==Unbound) && v==Unbound)
		return 0;

	if ((want & flag  &&  !(dont & flag)) 
	||  (must & flag  &&  tab))
		return this;

	return 0;
}

bool Object::expanded() 
{	return bit.table_expand;
}

void Object::setExpanded(bool state) 
{
	bit.table_expand=state;
	if (var())
	{	if (var()->type==OBJECT)
			ObjectP(var())->name()->bit.table_expand=state;
		else
			var()->setExpanded(state);
	}
	if (bit.table)
	{	table()->setExpanded(state);
	}
}

void Object::setflag(ObjectType want, ObjectType newflag)
{
	if (var() && var()!=Unbound 
	&& (var()->type==TABLE || var()->type==LOGOOBJECT))
		var()->setflag(want, newflag);
	if (table()!=Unbound)
		table()->setflag(want, newflag);
	if (flag & want)
		setflag(newflag);
}

void Object::clearflag(ObjectType want, ObjectType newflag)
{
	if (var() && var()!=Unbound 
	&& (var()->type==TABLE || var()->type==LOGOOBJECT))
		var()->clearflag(want, newflag);
	if (table()!=Unbound)
		table()->clearflag(want, newflag);
	if (flag & want)
		clearflag(newflag);
}

void Object::mark_all_but_objects() 
{
	name()->mark(); proc_or_prim()->mark(); 
	var()->mark(); 
	table()->mark_all_but_objects();
}

void Object::tell_shadow()	// Warn the user if a local variable shadows a global one. 
{
	if (bit.val_stepped)
		error(SHADOW_WARN, this);
}

bool Object::not_local(int stackPtr)	// Check if a local variable is already in this frame 
{
	ThreadP th=Thread::This();
	for ( ;  stackPtr > th->var;  stackPtr--) 
//		if (equal(th->varStack.base[stackPtr].object)) 
		if (ObjectP(this)==th->varStack->base[stackPtr].i.object) 
			return false;
	return true;
}

int Object::find_limit() 
{
	if (!this) 
#ifdef _DEBUG
		return -1;	//256;
#else
		return -1;
#endif
	NodeP num = var()->number();
	
	if (type_of(num) == INT_) 
		return IntP(num)->int_;
	else
#ifdef _DEBUG
		return -1;
#else
		return -1;
#endif
}
