//___________________________________________________________________
//
//                           Class Point
//___________________________________________________________________

class TextInputStream;
class TextOutputStream;
class DataInputStream;
class wxDataOutputStream;

class Point
{	
	public:
		union {float x; float a[1];};
		float y,z;

		Point(){x=y=z=0;}
		Point(float r){x=r; y=r; z=r;}
		Point(const float &_x, const float &_y, const float &_z){x=_x; y=_y; z=_z;}

		Point operator+(){return *this;}
		Point operator-(){return Point(-x,-y,-z);}

		Point round(){return Point(int(x+0.5), int(y+0.5), int(z+0.5));}
		Point   abs();
		Point  sqrt(){return Point(::sqrt(x), ::sqrt(y), ::sqrt(z));}

		Point operator+(const float &b){return Point(x+b, y+b, z+b);}
		Point operator-(const float &b){return Point(x-b, y-b, z-b);}
		Point operator*(const float &b){return Point(x*b, y*b, z*b);}
		Point operator/(const float &b){return Point(x/b, y/b, z/b);}
		Point operator%(const float &b);

		Point operator+(const Point &b){return Point(x+b.x, y+b.y, z+b.z);}
		Point operator-(const Point &b){return Point(x-b.x, y-b.y, z-b.z);}
		Point operator*(const Point &b){return Point(x*b.x, y*b.y, z*b.z);}
		Point operator/(const Point &b){return Point(x/b.x, y/b.y, z/b.z);}
		Point operator%(const Point &b);

		Point &operator+=(const float &b){x+=b; y+=b; z+=b; return *this;}
		Point &operator-=(const float &b){x-=b; y-=b; z-=b; return *this;}
		Point &operator*=(const float &b){x*=b; y*=b; z*=b; return *this;}
		Point &operator/=(const float &b){x/=b; y/=b; z/=b; return *this;}

		Point &operator=(const Point &b)
		{	x=b.x; y=b.y; z=b.z; 
			return *this;
		}
		Point &operator+=(const Point &b){x+=b.x; y+=b.y; z+=b.z; return *this;}
		Point &operator-=(const Point &b){x-=b.x; y-=b.y; z-=b.z; return *this;}
		Point &operator*=(const Point &b){x*=b.x; y*=b.y; z*=b.z; return *this;}
		Point &operator/=(const Point &b){x/=b.x; y/=b.y; z/=b.z; return *this;}

		bool operator==(const float &b){return (x==b && y==b && z==b);}
		bool operator!=(const float &b){return (x!=b || y!=b ||z!=b);}
		bool operator< (const float &b){return (Norm() < b);}
		bool operator> (const float &b){return (Norm() > b);}

		bool operator==(const Point &b){return (x==b.x && y==b.y && z==b.z);}
		bool operator!=(const Point &b){return (x!=b.x || y!=b.y || z!=b.z);}
		bool operator< (const Point &b){return Norm() < b.Norm();}
		bool operator> (const Point &b){return Norm() > b.Norm();}

		float Norm() const {return ::sqrt(x*x+y*y+z*z);}
		float min_() const {return x<y? (x<z?x:z):(y<z?y:z);}
		bool in(const Point &a, const Point &b) 
		{return (a.x<x && x<b.x && a.y<y && y<b.y && a.z<z && z<b.z);}

		Point X(const Point &b)
		{	return Point(y*b.z-z*b.y, z*b.x-x*b.z, x*b.y-y*b.x);
		}

	friend TextInputStream &operator>>(TextInputStream &istr, Point &p);
	friend TextOutputStream &operator<<(TextOutputStream &ostr, Point &p);
	friend DataInputStream &operator>>(DataInputStream &istr, Point &p);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Point &p);
};

typedef Point* PointP;

class DPoint
{	
	public:
		union {double x; double a[1];};
		double y,z;

		DPoint(){x=y=z=0;}
		DPoint(double r){x=r; y=r; z=r;}
		DPoint(const double &_x, const double &_y, const double &_z){x=_x; y=_y; z=_z;}

		operator Point(){return Point(x,y,z);}

		DPoint operator+(){return *this;}
		DPoint operator-(){return DPoint(-x,-y,-z);}

		DPoint round(){return DPoint(int(x+0.5), int(y+0.5), int(z+0.5));}
		DPoint   abs();
		DPoint  sqrt(){return DPoint(::sqrt(x), ::sqrt(y), ::sqrt(z));}

		DPoint operator+(const double &b){return DPoint(x+b, y+b, z+b);}
		DPoint operator-(const double &b){return DPoint(x-b, y-b, z-b);}
		DPoint operator*(const double &b){return DPoint(x*b, y*b, z*b);}
		DPoint operator/(const double &b){return DPoint(x/b, y/b, z/b);}
		DPoint operator%(const double &b);

		DPoint operator+(const DPoint &b){return DPoint(x+b.x, y+b.y, z+b.z);}
		DPoint operator-(const DPoint &b){return DPoint(x-b.x, y-b.y, z-b.z);}
		DPoint operator*(const DPoint &b){return DPoint(x*b.x, y*b.y, z*b.z);}
		DPoint operator/(const DPoint &b){return DPoint(x/b.x, y/b.y, z/b.z);}
		DPoint operator%(const DPoint &b);

		DPoint &operator+=(const double &b){x+=b; y+=b; z+=b; return *this;}
		DPoint &operator-=(const double &b){x-=b; y-=b; z-=b; return *this;}
		DPoint &operator*=(const double &b){x*=b; y*=b; z*=b; return *this;}
		DPoint &operator/=(const double &b){x/=b; y/=b; z/=b; return *this;}

		DPoint &operator=(const DPoint &b)
		{	x=b.x; y=b.y; z=b.z; 
			return *this;
		}
		DPoint &operator+=(const DPoint &b){x+=b.x; y+=b.y; z+=b.z; return *this;}
		DPoint &operator-=(const DPoint &b){x-=b.x; y-=b.y; z-=b.z; return *this;}
		DPoint &operator*=(const DPoint &b){x*=b.x; y*=b.y; z*=b.z; return *this;}
		DPoint &operator/=(const DPoint &b){x/=b.x; y/=b.y; z/=b.z; return *this;}

		bool operator==(const double &b){return (x==b && y==b && z==b);}
		bool operator!=(const double &b){return (x!=b || y!=b ||z!=b);}
		bool operator< (const double &b){return (Norm() < b);}
		bool operator> (const double &b){return (Norm() > b);}

		bool operator==(const DPoint &b){return (x==b.x && y==b.y && z==b.z);}
		bool operator!=(const DPoint &b){return (x!=b.x || y!=b.y || z!=b.z);}
		bool operator< (const DPoint &b){return Norm() < b.Norm();}
		bool operator> (const DPoint &b){return Norm() > b.Norm();}

		double Norm() const {return ::sqrt(x*x+y*y+z*z);}
		double min_() const {return x<y? (x<z?x:z):(y<z?y:z);}
		bool in(const DPoint &a, const DPoint &b) 
		{return (a.x<x && x<b.x && a.y<y && y<b.y && a.z<z && z<b.z);}

		DPoint X(const DPoint &b)
		{	return DPoint(y*b.z-z*b.y, z*b.x-x*b.z, x*b.y-y*b.x);
		}

	friend TextInputStream &operator>>(TextInputStream &istr, DPoint &p);
	friend TextOutputStream &operator<<(TextOutputStream &ostr, DPoint &p);
	friend DataInputStream &operator>>(DataInputStream &istr, DPoint &p);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, DPoint &p);
};

typedef DPoint* DPointP;

class PointFiFo		// is a First-in-First-out buffer class for Points.
{
	public:
		bool allocated, _overflow;
		int size, expand_size;
		PointP base, max, in, out;

		PointFiFo(int expand_size_) 
		{	expand_size=expand_size_;
			allocated=_overflow=false; 
			size=0; 
			base=max=in=out=0; 
			expand();
		}
		~PointFiFo();

		PointFiFo &operator<<(const Point val);
		PointFiFo &operator>>(Point &val);

		operator bool() {return in!=out;}
		PointFiFo &operator=(PointFiFo &b);
		bool overflow() {return _overflow;}
		void reset_in() {in=base;}
		void reset_out() {out=base;}
		void reset() {in=out=base;}
		void expand();
};

class DPointFiFo		// is a First-in-First-out buffer class for Points.
{
	public:
		bool allocated, _overflow;
		int size, expand_size;
		DPointP base, max, in, out;

		DPointFiFo(int expand_size_) 
		{	expand_size=expand_size_;
			allocated=_overflow=false; 
			size=0; 
			base=max=in=out=0; 
			expand();
		}
		~DPointFiFo();

		DPointFiFo &operator<<(const Point val);
		DPointFiFo &operator>>(Point &val);
		DPointFiFo &operator<<(const DPoint val);
		DPointFiFo &operator>>(DPoint &val);

		operator bool() {return in!=out;}
		DPointFiFo &operator=(DPointFiFo &b);
		bool overflow() {return _overflow;}
		void reset_in() {in=base;}
		void reset_out() {out=base;}
		void reset() {in=out=base;}
		void expand();
};

class Matrix
{
	public:
		Point x,y,z;

		Matrix() {}
		Matrix(float r) {x=Point(r); y=Point(r); z=Point(r);}
		Matrix(const Point &_x, const Point &_y, const Point &_z)
		{	x=_x; y=_y; z=_z;
		}

		Matrix &operator=(const Matrix &b)
		{	x=b.x; y=b.y; z=b.z; 
			return *this;
		}

		Point operator*(const Point &b){return x*b.x+y*b.y+z*b.z;}

		Matrix invert()
		{	float 
				d1=y.y*z.z-z.y*y.z,
				d2=y.z*z.x-z.z*y.x,
				d3=y.x*z.y-z.x*y.y,
	
				det=x.x*d1+x.y*d2+x.z*d3;

			if (det==0)
				return *this;

			Matrix b;
			b.x.x=d1/det;
			b.x.y=(z.y*x.z-x.y*z.z)/det;
			b.x.z=(x.y*y.z-y.y*x.z)/det;				
			b.y.x=d2/det;
			b.y.y=(z.z*x.x-x.z*z.x)/det;
			b.y.z=(x.z*y.x-y.z*x.x)/det;				
			b.z.x=d3/det;
			b.z.y=(z.x*x.y-x.x*z.y)/det;
			b.z.z=(x.x*y.y-y.x*x.y)/det;
			return b;
		}

	friend TextInputStream &operator>>(TextInputStream &istr, Matrix &m);
	friend TextOutputStream &operator<<(TextOutputStream &ostr, Matrix &m);
	friend DataInputStream &operator>>(DataInputStream &istr, Matrix &m);
	friend wxDataOutputStream &operator<<(wxDataOutputStream &ostr, Matrix &m);
};
