// QuickFill.h
//
// Author : John R. Shaw (shawj2@earthlink.net)
// Date   : Jan. 26 2004
//
// Copyright (C) 2004 John R. Shaw
// All rights reserved.
//
// This code may be used in compiled form in any way you desire. This
// file may be redistributed unmodified by any means PROVIDING it is 
// not sold for profit without the authors written consent, and 
// providing that this notice and the authors name is included. If 
// the source code in this file is used in any commercial application 
// then a simple email would be nice.
//
// Warranties and Disclaimers:
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND
// INCLUDING, BUT NOT LIMITED TO, WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
// IN NO EVENT WILL JOHN R. SHAW BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES,
// INCLUDING DAMAGES FOR LOSS OF PROFITS, LOSS OR INACCURACY OF DATA,
// INCURRED BY ANY PERSON FROM SUCH PERSON'S USAGE OF THIS SOFTWARE
// EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
// Please email bug reports, bug fixes, enhancements, requests and
// comments to: shawj2@earthlink.net
//

#ifndef _QUICKFILL_H_
#define _QUICKFILL_H_

//#ifndef _GDIPLUSGRAPHICS_H
//#include <gdiplus.h>
//#endif

//#ifndef __CDIBDATA_H__
//#include "CDibData.h"
//#endif

class CQuickFill
{
	// Public methods
public:
	int QuickFill(int x,int y,	Color fill_color, Color border_color=CLR_INVALID);

	RECT GetInvalidRect() const { return m_rcInvalid; }

	void SetFillMask(BYTE* pMask=NULL); // Mask[8] or NULL
	BOOL SetPatternBitmap(Color *pColor=NULL, int w=0, int h=0, 
		Color clrClear=CLR_INVALID);
	COLORREF SetPatternClearColor(Color clrClear=CLR_INVALID);
	COLORREF GetPatternClearColor() { return m_clrPatClear; }
	BOOL HasPattern();
	BOOL HasMask();

#ifdef QUICKFILL_SLOW
	void SetSlowMode(BOOL bEnable) { m_bSlowMode = bEnable; }
	BOOL IsSlowMode() { return m_bSlowMode; }
#endif

	// Protected scanning and drawing methods
protected:
	int width, height, patternWidth, patternHeight;
	Color *pixels, *pattern;

	DWORD GetPixel(int x, int y);
	BOOL SetPixel(int x, int y, DWORD dwValue);
	void DrawHorizontalLine(int x1, int x2, int y, DWORD dwValue);
	int ScanLeft(int x,int y,int xmin,DWORD dwValue);
	int SearchLeft(int x,int y,int xmin,DWORD dwValue);
	int ScanRight(int x,int y,int xmax,DWORD dwValue);
	int SearchRight(int x,int y,int xmax,DWORD dwValue);

	// Contructors/Destuctors
public:
	CQuickFill(int w, int h);
	~CQuickFill();

	// Public debug information methods
#ifdef QUICKFILL_TEST
	unsigned GetStackSize() { return m_CurStackSize; }
	unsigned GetMaxStackSize() { return m_MaxStackSize; }
	unsigned GetVisitSize() { return m_VisitSize; }
	int GetCurrentLine() { return m_CurrentLine; }
	int GetDirection() { return m_Direction; }
	int GetRevisitCount() { return m_nRevisitCount; }
	void SetShowRevisit(BOOL bFlg=TRUE) { m_bShowRevisits = bFlg; }
	BOOL GetShowRevisitFlag() { return m_bShowRevisits; }
#endif

	// Private list members
private:
	typedef struct hlineNode_st {
		int x1,x2,y,dy;
		struct hlineNode_st *pNext,*pPrev;
	} HLINE_NODE;

	HLINE_NODE*	m_pVisitList;
	HLINE_NODE*	m_pLineList;
	HLINE_NODE*	m_pFreeList;
	BOOL	m_bXSortOn;
	BOOL	m_bMemError;
	int		m_LastY;

	int MakeList(void);
	void FreeList(void);
	void PushLine(int x1,int x2,int y,int dy);
	void PopLine(int *x1,int *x2,int *y,int *dy);
	HLINE_NODE* FindNextLine(int x1,int x2,int y);
	void PopThis(HLINE_NODE *pThis);
	void PushOpposite(int OldX1,int OldX2,int x1,int x2,int y,int dy);

	void PushVisitedLine(int x1,int x2,int y);
	BOOL ExpandVisitedBlock(int x1,int x2,int y);
	BOOL IsRevisit(int x1,int x2,int y);


	// Private bitmap data members

//	CDibData m_DibData;
//	CDibData m_DibPattern;
	int m_xPatMod;
	int m_yPatMod;
	COLORREF m_clrPatClear;
	DWORD m_PatClearValue;
	BYTE m_FillMask[8];

	// Private information members
	RECT m_rcInvalid;

	// Private debug information
private:
#ifdef QUICKFILL_TEST
	unsigned m_CurStackSize;
	unsigned m_MaxStackSize;
	unsigned m_VisitSize;
	int m_CurrentLine;
	int m_Direction;
	int m_nRevisitCount;
	BOOL m_bShowRevisits;
#endif
#ifdef QUICKFILL_SLOW
	BOOL m_bSlowMode;
#endif
};

#endif // _QUICKFILL_H_
