//___________________________________________________________________
//
//                              Stack
//
//                           Definitions
//___________________________________________________________________

#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/datstrm.h>

#include <stdio.h>
#include <memory.h>

#include "Main.h"

#include "File.h"
#include "Object.h"
#include "Pair.h"
#include "Line.h"
#include "Array.h"
#include "Table.h"
#include "LogoObject.h"
#include "Table.h"
#include "String.h"

#include "Error.h"
#include "Init.h"
#include "Eval.h"
#include "NodePStack.h"
#include "Stack.h"
#include "ToDoStack.h"

#include "IntToPtr_.h"
#include "PtrToInt_.h"

#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#define DEBUG_STACK
//#define DEBUG_STACK2
//#define SDEBUG

#ifndef max
	#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
	#define realloc(m, len) _realloc_dbg((m), (len), _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

const int PAGE_SIZE=0x100;
const charP emptyString="";

bool IntStack::noisy=false;

IntStack::IntStack() 
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
		return;
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

IntStack::IntStack(int size_) 
{
	size=0;
	base=ItemP(GC_::calloc(size_, sizeof(Item)));
	if (!base)
		return;
	size=size_;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

IntStack::~IntStack() 
{
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: IntStack has still " 
			<< int(index-base) << " elements on destruction" << endL;
#endif
	GC_::free(base); base=0;
}

void IntStack::free()
{
	if (base)
		GC_::free(base);
	base=0;
	size=0;
	index=max=0;
}

void IntStack::copyTo(IntStack &out)
{
	out.free();
	out.base=ItemP(GC_::malloc(size*sizeof(Item)));
	if (!out.base)
		return;
	out.size=size;
	if (base)
		for (int i=index-base; i>=0; i--)
			out.base[i]=base[i];
	out.max = out.base+(max-base);
	out.index=out.base+(index-base);
#ifdef DEBUG_STACK
	out.item=item;
#endif
	return;
}

bool IntStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (noisy)
		c_out << _("IntStack size is now grown to ") << size << endL;
#endif
	return true;
}

bool IntStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (noisy)
		c_out << _("IntStack size is now shrinked to ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG
	void IntStack::push(int val) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		index++;
		index->i=val;
		index->name=emptyString;
		index->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void IntStack::push(int val, EvalStackMarker marker, charP name, int line) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		++index;
		index->i=val;
		index->marker=marker;
		index->name=name;
		index->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void IntStack::pop(EvalStackMarker marker) 
	{	if (index>base) 
		{
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			index->i=0;
			index->marker=ESM_dummy;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: IntStack underflow on pop()") << endL;
	}
	void IntStack::pop(int &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = index->i;
			index->i=0;
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			index->marker=ESM_dummy;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: IntStack underflow on pop(") << val << ')' << endL;
	}
	void IntStack::pop(bool &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = (index->i != 0);
			index->i=0;
			wxASSERT(index->marker==marker);
			index->marker=ESM_dummy;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: IntStack underflow on pop(") << int(val) << ')' << endL;
	}
	void IntStack::pop(TailCall &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = TailCall(index->i);
			index->i=0;
			wxASSERT(index->marker==marker);
			index->marker=ESM_dummy;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: IntStack underflow on pop(") << val << ')' << endL;
	}
	void IntStack::pop(ValStatus &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = ValStatus(index->i);
			index->i=0;
			wxASSERT(index->marker==marker);
			index->marker=ESM_dummy;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: IntStack underflow on pop(") << val << ')' << endL;
	}
	void IntStack::pop_till(int sp)
	{
		int dummy;
		while (index > base+sp)
			pop(dummy, ESM_dummy);
	}
/*	IntStack &IntStack::operator<<(const int val)
	{	push(val);
		return *this;
	}

	IntStack &IntStack::operator>>(int &val)
	{	pop(val);
		return *this;
	}
	IntStack &IntStack::operator>>(bool &val)
	{	pop((int&)val);
		return *this;
	}
	IntStack &IntStack::operator>>(TailCall &val)
	{	pop((int&)val);
		return *this;
	}
	IntStack &IntStack::operator>>(ValStatus &val)
	{	pop((int&)val);
		return *this;
	}

	void IntStack::push(int i1, int i2)
	{	push(i1);
		push(i2);
	}
	void IntStack::push(int i1, int i2, int i3)
	{	push(i1);
		push(i2);
		push(i3);
	}
	void IntStack::pop(int &i1, int &i2)
	{	pop(i1);
		pop(i2);
	}
	void IntStack::pop(int &i1, int &i2, int &i3)
	{	*this >> i3 >> i2 >> i1;
	}
	void IntStack::pop(ValStatus &i1, TailCall &i2)
	{	*this >> i2 >> i1;
	}
	void IntStack::pop(ValStatus &i1, TailCall &i2, int &i3)
	{	*this >> i3 >> i2 >> i1;
	}
*/
#endif

DataInputStream &IntStack::read(DataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	int val;
		istr >> val;
		push(val);
	}
	return istr;
}
wxDataOutputStream &IntStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
		ostr << (++itemptr)->i;
	return ostr;
}


FloatStack::FloatStack() 
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
		return;
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

FloatStack::~FloatStack() 
{
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: FloatStack has still " 
			<< int(index-base) << " elements on destruction" << endL;
#endif
	GC_::free(base); base=0;
}

bool FloatStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("FloatStack size is now grown to ") << size << endL;
#endif
	return true;
}

bool FloatStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("FloatStack size is now shrinked to ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG
	void FloatStack::push(double val) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		index++;
		index->i=val;
		index->name=emptyString;
		index->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void FloatStack::push(double val, EvalStackMarker marker, charP name, int line) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		++index;
		index->i=val;
		index->marker=marker;
		index->name=name;
		index->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void FloatStack::pop(double &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			val = index->i;
			index->i=0;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: FloatStack underflow on pop(") << val << ')' << endL;
	}
/*	FloatStack &FloatStack::operator<<(const double val)
	{	push(val);
		return *this;
	}

	FloatStack &FloatStack::operator>>(double &val)
	{	pop(val);
		return *this;
	}
	void FloatStack::push(double d1, double d2, double d3)
	{	*this << d1 << d2 << d3;
	}
	void FloatStack::pop(double &d1, double &d2, double &d3)
	{	*this >> d3 >> d2 >> d1;
	}
*/
#endif

DataInputStream &FloatStack::read(DataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	double val;
		istr >> val;
		push(val);
	}
	return istr;
}
wxDataOutputStream &FloatStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
		ostr << (++itemptr)->i;
	return ostr;
}

NodePStack::NodePStack()
{
	size=0;
	base=(NodeP*)GC_::calloc(PAGE_SIZE, sizeof(NodeP));
	if (!base)
	{	size=0;
		return;
	}
#ifdef SDEBUG
	ibase=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!ibase)
	{	size=0;
		return;
	}
#endif
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef SDEBUG
	imax = &ibase[size-1];
	iindex=ibase;
#endif
#ifdef DEBUG_STACK
	item=0;
#endif
}

NodePStack::NodePStack(int size_)
{
	size=0;
	base=(NodeP*)GC_::calloc(size_, sizeof(NodeP));
	if (!base)
	{	size=0;
		return;
	}
#ifdef SDEBUG
	ibase=ItemP(GC_::calloc(size_, sizeof(Item)));
	if (!ibase)
	{	size=0;
		return;
	}
#endif
	size=size_;
	max = &base[size-1];
	index=base;
#ifdef SDEBUG
	imax = &ibase[size-1];
	iindex=ibase;
#endif
#ifdef DEBUG_STACK
	item=0;
#endif
}

NodePStack::~NodePStack() 
{
#ifdef DEBUG_STACK2
//	if (index != base)
//		c_out << "Warning: NodePStack has still " 
//			<< int(index-base) << " elements on destruction" << endL;
#endif
	clean();
	if (base)
		GC_::free(base);
	base=0;
#ifdef SDEBUG
	if (ibase)
		GC_::free(ibase);
	ibase=0;
#endif
}

void NodePStack::free()
{
	if (base)
		GC_::free(base);
	base=0;
	size=0;
	index=max=0;
#ifdef SDEBUG
	if (ibase)
		GC_::free(ibase);
	ibase=0;
	iindex=imax=0;
#endif
}

void NodePStack::copyTo(NodePStack &out)
{
	out.free();
	out.base=NodePP(GC_::malloc(size*sizeof(NodeP)));
	if (!out.base)
		return;
	out.size=size;
#ifdef SDEBUG
	out.ibase=ItemP(GC_::malloc(size*sizeof(Item)));
	if (!out.ibase)
		return;
	out.imax = out.ibase+(imax-ibase);
	out.iindex=out.ibase+(iindex-ibase);
	if (base)
		for (int i=index-base; i>=0; i--)
			out.ibase[i]=ibase[i];
#endif
	if (base)
		for (int i=index-base; i>=0; i--)
			out.base[i]=base[i];
	out.max = out.base+(max-base);
	out.index=out.base+(index-base);
#ifdef DEBUG_STACK
	out.item=item;
#endif
	return;
}
void NodePStack::copy(NodePStack &out, IntToPtr_ &intToPtr, PtrToInt_ &ptrToInt)
{
	out.free();
	out.size=0;
	int s=size;
	out.base=NodePP(GC_::malloc(s*sizeof(NodeP)));
	if (!out.base)
	{	out.size=0;
		return;
	}
	out.size=s;
#ifdef SDEBUG
	out.ibase=ItemP(GC_::malloc(s*sizeof(Item)));
	if (!out.ibase)
		return;
	out.imax = out.ibase+(imax-ibase);
	out.iindex=out.ibase+(iindex-ibase);
	for (int i=index-base; i>=0; i--)
	{	out.ibase[i].i=NodeP(intToPtr.find(ptrToInt.find(ibase[i].i)));
		out.ibase[i].line  =ibase[i].line;
		out.ibase[i].marker=ibase[i].marker;
		out.ibase[i].name  =ibase[i].name;
	}
#endif
	for (int i=index-base; i>=0; i--)
		out.base[i]=NodeP(intToPtr.find(ptrToInt.find(base[i])));
	out.max = out.base+(max-base);
	out.index=out.base+(index-base);
#ifdef DEBUG_STACK
	out.item=item;
#endif
	return;
}

bool NodePStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	NodeP *ob=base;
	base=(NodeP*)GC_::realloc(base, sizeof(NodeP)*size*2);
	if (!base)
	{	base=ob;
		return false;	
	}
#ifdef SDEBUG
	ItemP iob=ibase;
	ibase=ItemP(GC_::realloc(ibase, sizeof(Item)*size*2));
	if (!ibase)
	{	ibase=iob;
		return false;	
	}
#endif
	size*=2;
	index=index-ob+base;
	max = base+size-1;
#ifdef SDEBUG
	iindex=iindex-iob+ibase;
	imax = ibase+size-1;
#endif
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("NodePStack size is now grown to ") << size << endL;
#endif
	return true;
}

bool NodePStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	NodeP *ob=base;
	base=(NodeP*)GC_::realloc(base, sizeof(NodeP)*s);
	if (!base)
	{	base=ob;
		return false;	
	}
#ifdef SDEBUG
	ItemP iob=ibase;
	ibase=ItemP(GC_::realloc(ibase, sizeof(Item)*s));
	if (!ibase)
	{	ibase=iob;
		return false;	
	}
#endif
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef SDEBUG
	iindex=iindex-iob+ibase;
	imax = &ibase[size-1];
#endif
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("NodePStack size is now shrinked to ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG

	void NodePStack::clean()
	{
		if (index!=base && theApp 
		&& theApp->action!=EXIT_LOGO 
		&& theApp->action!=EXIT_DELAYED)
		{	c_out << "; Warning: NodePStack is still containing " 
				<< int(index-base) << " elements!" << endL;

			while (index>base)
			{	NodeP nd; pop(nd, ESM_dummy); 
//				int width=Int::max;
//				File::print(conout, width, nd);
			}
			c_out << endL;
		}		
	}

	void NodePStack::push(NodeP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=val;

		iindex++;
		iindex->i=val;
		iindex->marker=ESM_dummy;
		iindex->name=emptyString;
		iindex->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void NodePStack::push(NodeP val, EvalStackMarker marker, charP name, int line) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		*++index=val;

		iindex++;
		iindex->i=val;
		iindex->marker=marker;
		iindex->name=name;
		iindex->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void NodePStack::pop(NodeP &val, EvalStackMarker marker) 
	{
		if (index>base) 
		{	val = *index;  *index--=0;

			wxASSERT(iindex->marker==marker || marker==ESM_dummy);
			iindex->i=0;
			iindex->name=0;
			iindex->line=0;
			iindex--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: NodePStack underflow on pop(") << val << ')' << endL;
		if (type_of(val)==NTFREE)
			c_out << _("Error in NodePStack::pop(") << val
				<< _("): NodePStack was damaged during garbage collection") << endL;
	}
	void NodePStack::pop(PairP &val, EvalStackMarker marker) 
	{
		if (index>base) 
		{	val = PairP(*index);  *index--=0;

			wxASSERT(iindex->marker==marker || marker==ESM_dummy);
			iindex->i=0;
			iindex->name=0;
			iindex->line=0;
			iindex--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: NodePStack underflow on pop(") << val << ')' << endL;
		if (type_of(val)==NTFREE)
			c_out << _("Error in NodePStack::pop(") << val
				<< _("): NodePStack was damaged during garbage collection") << endL;
	}
	void NodePStack::pop(LineP &val, EvalStackMarker marker) 
	{
		if (index>base) 
		{	val = LineP(*index);  *index--=0;

			wxASSERT(iindex->marker==marker || marker==ESM_dummy);
			iindex->i=0;
			iindex->name=0;
			iindex->line=0;
			iindex--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: NodePStack underflow on pop(") << val << ')' << endL;
		if (type_of(val)==NTFREE)
			c_out << _("Error in NodePStack::pop(") << val
				<< _("): NodePStack was damaged during garbage collection") << endL;
	}
	void NodePStack::pop(ObjectP &val, EvalStackMarker marker) 
	{
		if (index>base) 
		{	val = ObjectP(*index);  *index--=0;

			wxASSERT(iindex->marker==marker || marker==ESM_dummy);
			iindex->i=0;
			iindex->name=0;
			iindex->line=0;
			iindex--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: NodePStack underflow on pop(") << val << ')' << endL;
		if (type_of(val)==NTFREE)
			c_out << _("Error in NodePStack::pop(") << val
				<< _("): NodePStack was damaged during garbage collection") << endL;
	}
	void NodePStack::pop(LogoObjectP &val, EvalStackMarker marker) 
	{
		if (index>base) 
		{	val = LogoObjectP(*index);  *index--=0;

			wxASSERT(iindex->marker==marker || marker==ESM_dummy);
			iindex->i=0;
			iindex->name=0;
			iindex->line=0;
			iindex--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: NodePStack underflow on pop(") << val << ')' << endL;
		if (type_of(val)==NTFREE)
			c_out << _("Error in NodePStack::pop(") << val
				<< _("): NodePStack was damaged during garbage collection") << endL;
	}
	void NodePStack::pop(EvalStackMarker marker) 
	{
		if (index>base) 
		{	*index--=0;

			wxASSERT(iindex->marker==marker || marker==ESM_dummy);
			iindex->i=0;
			iindex->name=0;
			iindex->line=0;
			iindex--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: NodePStack underflow on pop()") << endL;
	}
	void NodePStack::pushList(PairP head) 
	{
		while (head)
		{	push(NodeP(head->first()));
			next(head);
		}
	}
	void NodePStack::pushList(PairP head, EvalStackMarker marker, charP name, int line) 
	{
		while (head)
		{	push(NodeP(head->first()), marker, name, line);
			next(head);
		}
	}
	void NodePStack::push(ArrayP a) 
	{
		int l=a->length();
		NodePP d=a->data();
		while (l--)
			push(*d++);
	}
	void NodePStack::push(ArrayP a, EvalStackMarker marker, charP name, int line) 
	{
		int l=a->length();
		NodePP d=a->data();
		while (l--)
			push(*d++, marker, name, line);
	}
	void NodePStack::pop_till(PairPP sp)
	{
		while (index > (NodeP*)sp)
			pop(ESM_dummy);
	}
/*	NodePStack &NodePStack::operator<<(const NodeP val)
	{	push(val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(NodeP &val)
	{	pop(val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(PairP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(LineP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(ObjectP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(LogoObjectP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	NodePStack &NodePStack::operator>>(ProcedureP &val)
	{	pop((NodeP&)val);
		return *this;
	}
	void NodePStack::push(
		const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
		const NodeP n5, const NodeP n6, const NodeP n7) 
	{	
		*this << n1 << n2 << n3 << n4 << n5 << n6 << n7;
	}
	void NodePStack::push(
		const NodeP n1, const NodeP n2, const NodeP n3, const NodeP n4, 
		const NodeP n5, const NodeP n6, const NodeP n7, const NodeP n8) 
	{	
		*this << n1 << n2 << n3 << n4 << n5 << n6 << n7 << n8;
	}
	void NodePStack::pop(
		NodeP &n1, NodeP &n2, 
		NodeP &n3, ObjectP &n4, LineP &n5, 
		ObjectP &n6, LineP &n7)
	{
		*this >> n7 >> n6 >> n5 >> n4 >> n3 >> n2 >> n1;
	}
	void NodePStack::pop(
		PairP &n1, NodeP &n2, 
		NodeP &n3, ObjectP &n4, LineP &n5, 
		ObjectP &n6, LineP &n7)
	{
		*this >> n7 >> n6 >> n5 >> n4 >> n3 >> n2 >> n1;
	}
*/
#else

void NodePStack::pushList(PairP head) 
{	while (head) 
	{	NodeP val=NodeP(head->first());
		if (index+1 > max)
			if (!grow())
				return;
		*++index=val;
		next(head);
	}
}

void NodePStack::push(ArrayP a) 
{
	int l=a->length();
	NodePP d=a->data();
	while (l--)
	{	if (index+1 > max)
			if (!grow())
				return;
		*++index=*d++;
	}
}

#endif

void NodePStack::mark()
{
	for (NodeP *i=base; i<=index; i++)
		if (i && *i)
			(*i)->mark();
}

void NodePStack::markPtr(PtrToInt_ &ptrToInt)
{
	for (NodeP *i=base; i<=index; i++)
		if (i && *i)
			(*i)->markPtr(ptrToInt);
}

DataInputStream &NodePStack::read(DataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	NodeP val;
		istr >> (int&)val;
		push(val);
	}
	return istr;
}
wxDataOutputStream &NodePStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	NodePP itemptr=base;
	for (int i=siz;  i>=1;  i--)
		ostr << GC_::ptrToInt->find(*++itemptr);
	return ostr;
}

void NodePStack::setPtrs(IntToPtr_ &intToPtr)
{
	int siz=index-base;
	NodePP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		*itemptr = (Node*)intToPtr.find(int(*itemptr));
	}
}

PairP NodePStack::to_list(PairPP sp)
{
	PairP list=0;
	NodePP i=index;
	while (i != NodePP(sp))
		::push(list, *i--);
	return list;
}

PairP NodePStack::to_list(int i)
{
	return to_list((PairPP)&base[i]);
}

PairP NodePStack::to_list_reversed(PairPP sp)
{
	PairP list=0;
	while (index != NodePP(sp))
		::push(list, *++sp);
	return list;
}


PairPPStack::PairPPStack()
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

PairPPStack::~PairPPStack() 
{
#ifdef DEBUG_STACK2
	if (index!=base)
		c_out << "Warning: PairPPStack is still containing "
			<< int(index-base) << " elements on destruction!" << endL;
#endif
	GC_::free(base); base=0;
}

bool PairPPStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("NodePStack size is now ") << size << endL;
#endif
	return true;
}

bool PairPPStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("IntStack size is now ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG
	void PairPPStack::push(PairPP val) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		index++;
		index->i=val;
		index->name=emptyString;
		index->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void PairPPStack::push(PairPP val, EvalStackMarker marker, charP name, int line) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		++index;
		index->i=val;
		index->marker=marker;
		index->name=name;
		index->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void PairPPStack::pop(PairPP &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = index->i;
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			index->i=0;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: PairPPStack underflow on pop(") << val << ')' << endL;
	}
/*	PairPPStack &PairPPStack::operator<<(const PairPP val)
	{	push(val);
		return *this;
	}
	PairPPStack &PairPPStack::operator>>(PairPP &val)
	{	pop(val);
		return *this;
	}
*/
#endif

DataInputStream &PairPPStack::read(DataInputStream &istr) 
{
	ThreadP th=Thread::This();
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr;
		istr >> nr;
		push(nr+(PairP*)th->argStack->base);
	}
	return istr;
}
wxDataOutputStream &PairPPStack::write(wxDataOutputStream &ostr) 
{
	ThreadP th=Thread::This();
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr=(++itemptr)->i-(PairP*)th->argStack->base;
		ostr << nr;
	}
	return ostr;
}


VarStack::VarStack()
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

VarStack::VarStack(int size_)
{
	size=0;
	base=ItemP(GC_::calloc(size_, sizeof(Item)));
	if (!base)
	{	size=0;
		return;
	}
	size=size_;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

VarStack::~VarStack() 
{
#ifdef DEBUG_STACK2
	if (index!=base)
		c_out << "Warning: VarStack is still containing "
			<< int(index-base) << " elements on destruction!" << endL;
#endif
	GC_::free(base); base=0;
}

void VarStack::free()
{
	if (base)
		GC_::free(base);
	base=0;
	size=0;
	index=max=0;
}

void VarStack::copyTo(VarStack &out)
{
	out.free();
	out.base=ItemP(GC_::malloc(size*sizeof(Item)));
	if (!out.base)
		return;
	out.size=size;
	if (base)
		for (int i=index-base; i>=0; i--)
			out.base[i]=base[i];
	out.max = out.base+(max-base);
	out.index=out.base+(index-base);
#ifdef DEBUG_STACK
	out.item=item;
#endif
	return;
}

void VarStack::copy(VarStack &out, IntToPtr_ &intToPtr, PtrToInt_ &ptrToInt)
{
//	out.free();
	int s=size;
	out.base=ItemP(GC_::malloc(s*sizeof(Item)));
	if (!out.base)
	{	out.size=0;
		return;
	}
	out.size=s;
	for (int i=index-base; i>=0; i--)
	{	out.base[i].i.object=ObjectP(intToPtr.find(ptrToInt.find(base[i].i.object)));
		out.base[i].i.val=NodeP(intToPtr.find(ptrToInt.find(base[i].i.val)));
		out.base[i].i.alias=base[i].i.alias;
#ifdef SDEBUG
		out.base[i].marker=base[i].marker;
		out.base[i].line=base[i].line;
		out.base[i].name=base[i].name;
#endif
	}
	out.max = out.base+(max-base);
	out.index=out.base+(index-base);
#ifdef DEBUG_STACK
	out.item=item;
#endif
	return;
}

bool VarStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("NodePStack size is now grown to ") << size << endL;
#endif
	return true;
}

bool VarStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("VarStack size is now shrinked to ") << size << endL;
#endif
	return true;
}

void VarStack::push(ObjectP object, NodeP val, int alias) 
{	if (index+2 > max)
		if (!grow())
			return;
	index++;
	wxASSERT(!object || (object && object->type==OBJECT));
	index->i.object=object;
	index->i.val=val;
	index->i.alias=alias;
	if (!object->parent() 
	|| object->parent()==_Null_Word 
	||	!(object->parent()->flag & BURIED))
		index->i.flag=NORMAL_OBJECT;
	else
		index->i.flag=VAL_BURIED;
	index->i.bit.val = (val != Unbound);
#ifdef SDEBUG
	index->name=emptyString;
	index->line=0;
#endif
//		index->val=object->var();
#ifdef DEBUG_STACK
	item++;
	maxitem=max(item, maxitem);
#endif
}
void VarStack::push(ObjectP object, NodeP val) 
{	push(object, val, false);
}

void VarStack::push(ObjectP object, NodeP val, int alias, 
						  EvalStackMarker marker, charP name, int line)
{	if (index+2 > max)
		if (!grow())
			return;
	index++;
	wxASSERT(!object || (object && object->type==OBJECT));
	index->i.object=object;
	index->i.val=val;
	index->i.alias=alias;
	if (!object->parent() 
	|| object->parent()==_Null_Word 
	||	!(object->parent()->flag & BURIED))
		index->i.flag=NORMAL_OBJECT;
	else
		index->i.flag=VAL_BURIED;
	index->i.bit.val = (val != Unbound);
#ifdef SDEBUG
	index->marker=marker;
	index->name=name;
	index->line=line;
#endif
//		index->val=object->var();
#ifdef DEBUG_STACK
	item++;
	maxitem=max(item, maxitem);
#endif
}
void VarStack::push(ObjectP object, NodeP val, EvalStackMarker marker, charP name, int line)
{	push(object, val, false, marker, name, line);
}

#ifdef SDEBUG
	void VarStack::pop(EvalStackMarker marker) 
	{
		if (index>base) 
		{//	index->object->setvar( index->val );
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			index->i.object=0; index->i.val=0; index->i.flag=NORMAL_OBJECT;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: VarStack underflow on pop(") << index->i.object << ')' << endL;
		if (type_of(index->i.object)==NTFREE)
			c_out << _("Error in VarStack::pop(") << index->i.object
				<< _("): VarStack was damaged during garbage collection") << endL;
		if (type_of(index->i.val)==NTFREE)
			c_out << _("Error in VarStack::pop(") << index->i.val
				<< _("): VarStack was damaged during garbage collection") << endL;
	}

	void VarStack::pop_till(ItemP oldSP)	// pop saved vars till old stack pointer
	{	while (index > oldSP)
			pop(ESM_dummy);
	}
#endif

void VarStack::push_alias(ObjectP object, ObjectP alias)
{
	ItemP i=index;
	int aliasnr=-1;
	while (i>base)	// search the obj
	{	if (i->i.object==alias)
		{	if (i->i.alias>0)
			{	aliasnr=i->i.alias;
				alias=ObjectP(i->i.val);
			}else if (i->i.alias==-1)
				alias=ObjectP(i->i.val);
			else
				aliasnr=i-base;
			break;
		}
		i--;
	}
	push(object, alias, aliasnr);
}
/*	
NodeP VarStack::find(ObjectP obj)
{
	ThreadP th=Thread::This();
	int framenr=*th->frameStack;
	int fp=th->frame, lastfp=th->lastFrame;
	ItemP i=index, 
		frame=base+fp;
	if (framenr==0)
		framenr=2;
	while (i>base && framenr>=1)
	{	while (i>frame)
		{	if (i->i.object==obj)
			{	if (i->i.alias==-1)	// special value for global
					return ObjectP(i->i.val)->var();
				else if (i->i.alias)
				{	i=base+i->i.alias;
					return i->i.val;
				}else
					return i->i.val;
			}
			i--;
		}
		lastfp=th->frameStack->base[framenr-1].i;
		i=base+fp;
		fp=th->frameStack->base[lastfp].i;
		frame=base+fp;
		framenr-=2;
	}
	if (th->thisobj)
	{	TableP t=th->thisobj->table();
		ObjectP o=t->find(ObjectP(obj)->name());
		if (o)
			return o->var();
	}
	wxASSERT(obj);
	return obj->var();
}
*/
NodeP VarStack::find(ObjectP obj)
{
	ThreadP th=Thread::This();
	int framenr=*th->frameStack-2;
	int fp;
	ItemP i=index, frame;
	if (framenr==0)
		framenr=1;
	fp=th->frameStack->base[framenr+2].i;
	frame=base+fp;
	while (i>base)
	{	while (i>=frame)	// was: i>frame
		{	if (i->i.object==obj)
			{	if (!i->i.alias)
					return i->i.val;
				else if (i->i.alias==-1)	// special value for global
					return ObjectP(i->i.val)->var();
				else
				{	i=base+i->i.alias;
					return i->i.val;
				}
			}
			i--;
		}
		fp=th->frameStack->base[framenr+1].i;
		frame=base+fp;
		fp=th->frameStack->base[framenr+2].i;
		i=base+fp;
		framenr=th->frameStack->base[framenr].i;
	}
	if (th->thisobj)
	{	TableP t=th->thisobj->table();
		ObjectP o=t->find(ObjectP(obj)->name());
		if (o)
			return o->var();
	}
	wxASSERT(obj);
	return obj->var();
}

bool VarStack::isvar(ObjectP obj)
{
	ThreadP th=Thread::This();
	int framenr=*th->frameStack-2;
	int fp;
	ItemP i=index, frame;
	if (framenr==0)
		framenr=1;
	fp=th->frameStack->base[framenr+2].i;
	frame=base+fp;
	while (i>base)
	{	while (i>=frame)
		{	if (i->i.object==obj)
			{	if (!i->i.alias)
					return true;
				else if (i->i.alias==-1)	// special value for global
					return true;
				else
				{	i=base+i->i.alias;
					return true;
				}
			}
			i--;
		}
		fp=th->frameStack->base[framenr+1].i;
		frame=base+fp;
		fp=th->frameStack->base[framenr+2].i;
		i=base+fp;
		framenr=th->frameStack->base[framenr].i;
	}
	if (th->thisobj)
	{	TableP t=th->thisobj->table();
		ObjectP o=t->find(ObjectP(obj)->name());
		if (o)
			return true;
	}
	wxASSERT(obj);
	if (obj->var() != Unbound)
		return true;
	return false;
}

void VarStack::setVar(ObjectP obj, NodeP val)
{
	ThreadP th=Thread::This();
	int framenr=*th->frameStack-2;
	int fp;
	ItemP i=index, frame;
	if (framenr==0)
		framenr=1;
	fp=th->frameStack->base[framenr+2].i;
	frame=base+fp;
	while (i>base)
	{	while (i>=frame)
		{	if (i->i.object==obj)
			{	if (!i->i.alias)
				{	i->i.val=val;
					i->i.bit.val = (val != Unbound);
				}else if (i->i.alias==-1)
				{	ObjectP(i->i.val)->setvar(val);
				}else
				{	i=base+i->i.alias;
					i->i.val=val;
					i->i.bit.val = (val != Unbound);
				}
				return;
			}
			i--;
		}
		fp=th->frameStack->base[framenr+1].i;
		frame=base+fp;
		fp=th->frameStack->base[framenr+2].i;
		i=base+fp;
		framenr=th->frameStack->base[framenr].i;
	}
	if (th->thisobj)
	{	TableP t=th->thisobj->table();
		ObjectP o=t->find(ObjectP(obj)->name());
		if (o)
		{	o->setvar(val);
			return;
		}
	}
	wxASSERT(obj);
	if (obj->bit.permanent)
	{	error(IS_PRIM, obj);	return;
	}
	obj->setvar(val);
	return;
}
/*
void VarStack::setVar(ObjectP obj, NodeP val)
{
	ThreadP th=Thread::This();
	Var *i=index, *frame=base;	// TODO: add lexical frame pointers
	while (i>frame)
	{	if (i->object==obj)
		{	if (i->alias==-1)
			{	ObjectP(i->val)->setvar(val);
			}
			else if (i->alias)
			{	i=base+i->alias;
				i->val=val;
			}else
				i->val=val;
			return;
		}
		i--;
	}
	wxASSERT(obj);
	obj->setvar(val);
	return;
}
*/

PairP VarStack::contents(ObjectType want, ObjectType dont)
{
	ThreadP th=Thread::This();
	PairP names=0;
	int framenr=*th->frameStack-2;
	int fp;
	ItemP i=index, frame;
	if (framenr==0)
		framenr=1;
	fp=th->frameStack->base[framenr+2].i;
	frame=base+fp;
	while (i>base)
	{	while (i>=frame)
		{	ObjectType flag=i->i.flag;
			if (flag & want  &&  !(flag & dont))
			{	if (th->namespce==root
				|| i->i.object->inNamespace(th->namespceobj))
					names=new Pair(i->i.object->longName(), names);
				if throwing return PairP(Unbound);
			}
			i--;
		}
		fp=th->frameStack->base[framenr+1].i;
		frame=base+fp;
		fp=th->frameStack->base[framenr+2].i;
		i=base+fp;
		framenr=th->frameStack->base[framenr].i;
	}
	return names;
}

bool VarStack::setflag(ObjectP obj, ObjectType set)
{
	ThreadP th=Thread::This();
	bool done=false;
	int framenr=*th->frameStack-2;
	int fp;
	ItemP i=index, frame;
	if (framenr==0)
		framenr=1;
	fp=th->frameStack->base[framenr+2].i;
	frame=base+fp;
	while (i>base)
	{	while (i>=frame)
		{	if (i->i.object==obj)
			{	if (!i->i.alias)
				{	i->i.tint |= set;
					done=true;
				}else if (i->i.alias==-1)	// special value for global
					;
				else
				{	i=base+i->i.alias;
					i->i.tint |= set;
					done=true;
				}
			}
			i--;
		}
		fp=th->frameStack->base[framenr+1].i;
		frame=base+fp;
		fp=th->frameStack->base[framenr+2].i;
		i=base+fp;
		framenr=th->frameStack->base[framenr].i;
	}
	if (th->thisobj)
	{	TableP t=th->thisobj->table();
		ObjectP o=t->find(ObjectP(obj)->name());
		if (o)
		{	o->tint |= set;
			done=true;
		}
	}
	return done;
}

bool VarStack::clearflag(ObjectP obj, ObjectType clear)
{
	ThreadP th=Thread::This();
	bool done=false;
	int framenr=*th->frameStack-2;
	int fp;
	ItemP i=index, frame;
	if (framenr==0)
		framenr=1;
	fp=th->frameStack->base[framenr+2].i;
	frame=base+fp;
	while (i>base)
	{	while (i>=frame)
		{	if (i->i.object==obj)
			{	if (!i->i.alias)
				{	i->i.tint &= ~clear;
					done=true;
				}else if (i->i.alias==-1)	// special value for global
					done=false;
				else
				{	i=base+i->i.alias;
					i->i.tint &= ~clear;
					done=true;
				}
			}
			i--;
		}
		fp=th->frameStack->base[framenr+1].i;
		frame=base+fp;
		fp=th->frameStack->base[framenr+2].i;
		i=base+fp;
		framenr=th->frameStack->base[framenr].i;
	}
	if (th->thisobj)
	{	TableP t=th->thisobj->table();
		ObjectP o=t->find(ObjectP(obj)->name());
		if (o)
		{	o->tint &= ~clear;
			done=true;
		}
	}
	return done;
}

void VarStack::mark()
{
	for (ItemP i=base+1; i<=index; i++)
		if (i)
			i->i.mark();
}

void VarStack::markPtr(PtrToInt_ &ptrToInt)
{
	for (ItemP i=base+1; i<=index; i++)
		if (i)
			i->i.markPtr(ptrToInt);
}
void VarStack::setPtrs(IntToPtr_ &intToPtr)
{
	int size=index-base;
	ItemP itemptr=base;
	for (int i=size;  i>=1;  i--)
	{	itemptr++;
		itemptr->i.object = (Object*)intToPtr.find(int(itemptr->i.object));
		itemptr->i.val    =   (Node*)intToPtr.find(int(itemptr->i.val   ));
	}
}
DataInputStream &VarStack::read(DataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	ObjectP object;
		NodeP val;
		istr 
			>> (int&)object
			>> (int&)val;
		if (index+2 > max)
			if (!grow())
				return istr;
		++index;
		index->i.object=object;
		index->i.val=val;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	return istr;
}
wxDataOutputStream &VarStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		ostr 
			<< GC_::ptrToInt->find(itemptr->i.object)
			<< GC_::ptrToInt->find(itemptr->i.val);
	}
	return ostr;
}


VarPStack::VarPStack()
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=-1;
#endif
}

VarPStack::~VarPStack() 
{
#ifdef DEBUG_STACK2
	if (index!=base && theApp)
		c_out << "Warning: VarPStack is still containing " 
			<< int(index-base) << " elements on destruction!" << endL;
#endif
	GC_::free(base); base=0;
}

bool VarPStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("NodePStack size is now ") << size << endL;
#endif
	return true;
}

bool VarPStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("IntStack size is now ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG
	void VarPStack::push(VarP val) 
	{	if (index+1 > max)
			if (!grow())
				return;
		index++;
		index->i=val;
		index->name=emptyString;
		index->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void VarPStack::push(VarP val, EvalStackMarker marker, charP name, int line) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		++index;
		index->i=val;
		index->marker=marker;
		index->name=name;
		index->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void VarPStack::pop(VarP &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = index->i;
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			index->i=0;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: VarPStack underflow on pop(") << val << ')' << endL;
	}
/*	VarPStack &VarPStack::operator<<(const VarP val)
	{	push(val);
		return *this;
	}
	VarPStack &VarPStack::operator>>(VarP &val)
	{	pop(val);
		return *this;
	}
*/
#endif

DataInputStream &VarPStack::read(DataInputStream &istr) 
{
	ThreadP th=Thread::This();
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr;
		istr >> nr;
		push(VarP(nr+VarP(th->varStack->base)));
	}
	return istr;
}
wxDataOutputStream &VarPStack::write(wxDataOutputStream &ostr) 
{
	ThreadP th=Thread::This();
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	int nr=(++itemptr)->i-VarP(th->varStack->base); // special case here for argPStack!!!
		ostr << nr;
	}
	return ostr;
}


ToDoStack::ToDoStack() 
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

ToDoStack::~ToDoStack() 
{
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: ToDoStack has still "
			<< int(index-base) << " elements on destruction" << endL;
#endif
	GC_::free(base); base=0;
}

bool ToDoStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("ToDoStack size is now grown to ") << size << endL;
#endif
	return true;
}

bool ToDoStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("ToDoStack size is now shrinked to ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG
	void ToDoStack::push(ToDo val) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		index++;
		index->i=val;
		index->name=emptyString;
		index->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void ToDoStack::push(ToDo val, EvalStackMarker marker, charP name, int line) 
	{
		if (index+1 > max)
			if (!grow())
				return;
		++index;
		index->i=val;
		index->marker=marker;
		index->name=name;
		index->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void ToDoStack::pop(ToDo &val, EvalStackMarker marker) 
	{
		if (index>=base) 
		{	val = index->i;
			index->i=eval_dummy_label;
			index->name=0;
			index->line=0;
			index--;
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: ToDoStack underflow on pop(") 
				<< val.mnemonic << ')' << endL;
	}
/*	ToDoStack &ToDoStack::operator<<(const ToDo val)
	{	push(val);
		return *this;
	}
	ToDoStack &ToDoStack::operator>>(ToDo &val)
	{	pop(val);
		return *this;
	}
*/
#endif

DataInputStream &ToDoStack::read(DataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	ToDo td;
#ifdef PROC_EVAL
	#ifdef SDEBUG
		istr >> (int&)td.mnemonic;
	#endif
		istr >> td.i;
#else
		istr >> (int&)td.mnemonic;
#endif
		push(td);
	}
	return istr;
}
wxDataOutputStream &ToDoStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
#ifdef PROC_EVAL
	#ifdef SDEBUG
		ostr << (int&)itemptr->mnemonic;
	#endif
		ostr << itemptr->i;
#else
		ostr << (int&)itemptr->i.mnemonic;
#endif
	}
	return ostr;
}

//#define SDEBUG
ErrorStack::ErrorStack() 
{
	size=0;
	base=ItemP(GC_::calloc(PAGE_SIZE, sizeof(Item)));
	if (!base)
	{	size=0;
		return;
	}
	size=PAGE_SIZE;
	max = &base[size-1];
	index=base;
#ifdef DEBUG_STACK
	item=0;
#endif
}

ErrorStack::~ErrorStack() 
{
#ifdef DEBUG_STACK2
	if (index != base)
		c_out << "Warning: ErrorStack has still "
			<< int(index-base) << " elements on destruction" << endL;
#endif
	GC_::free(base); base=0;
}

bool ErrorStack::grow()
{
	ThreadP th=Thread::This();
	if (th->security->stacksize < unsigned(size)*2)
	{	error(SECURITY_ERROR, new String(_("stacksize too big")));
	}
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*size*2));
	if (!base)
	{	base=ob;
		return false;	
	}
	size*=2;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("ErrorStack size is now grown to ") << size << endL;
#endif
	return true;
}

bool ErrorStack::shrink()
{
	int s=int(index-base);
	if (s > size/2)
		return true;
	s=2^int(log(double(s))/log(2.0));
	if (s < PAGE_SIZE)
		s=PAGE_SIZE;
	if (s > size/2)
		return true;
	ItemP ob=base;
	base=ItemP(GC_::realloc(base, sizeof(Item)*s));
	if (!base)
	{	base=ob;
		return false;	
	}
	size=s;
	index=index-ob+base;
	max = &base[size-1];
#ifdef DEBUG_STACK
	if (IntStack::noisy)
		c_out << _("ErrorStack size is now shrinked to ") << size << endL;
#endif
	return true;
}

#ifdef SDEBUG
	void ErrorStack::push(ObjectP fun_, ObjectP ufun_, LineP this_line_) 
	{	if (index+2 > max)
			if (!grow())
				return;
		index++;
		index->i.fun = fun_;
		index->i.ufun = ufun_;	
		index->i.this_line = this_line_;
		index->name=emptyString;
		index->line=0;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void ErrorStack::push(ObjectP fun_, ObjectP ufun_, LineP this_line_, 
								EvalStackMarker marker, charP name, int line) 
	{	if (index+2 > max)
			if (!grow())
				return;
		index++;
		index->i.fun = fun_;
		index->i.ufun = ufun_;	
		index->i.this_line = this_line_;
		index->marker=marker;
		index->name=name;
		index->line=line;
#ifdef DEBUG_STACK
		item++;
		maxitem=max(item, maxitem);
#endif
	}
	void ErrorStack::pop(EvalStackMarker marker) 
	{
		if (index>=base) 
		{	
			wxASSERT(index->marker==marker || marker==ESM_dummy);
			index->i.fun=index->i.ufun=0; index->i.this_line=0;
			index->name=0;
			index->line=0;
			index--;  
#ifdef DEBUG_STACK
			item--;
#endif
		}else
			c_out << _("Error: ErrorStack underflow on pop(")
				<< index->i.fun << ", " 
				<< index->i.ufun << ", " 
				<< index->i.this_line << ')' << endL;
	}
#endif

void ErrorStack::mark()
{
	for (ItemP i=base; i<=index; i++)
		i->i.mark();
}
void ErrorStack::markPtr(PtrToInt_ &ptrToInt)
{
	for (ItemP i=base; i<=index; i++)
		i->i.markPtr(ptrToInt);
}
DataInputStream &ErrorStack::read(DataInputStream &istr) 
{
	int siz;
	istr >> siz;
#ifdef DEBUG_STACK
	maxitem=item=siz;
#endif
	index=base;
	for (int i=siz;  i>=1;  i--)
	{	ErrorStruct e;
		istr 
			>> (int&)e.fun
			>> (int&)e.ufun
			>> (int&)e.this_line;
		push(e.fun, e.ufun, e.this_line);
	}
	return istr;
}
wxDataOutputStream &ErrorStack::write(wxDataOutputStream &ostr) 
{
	int siz=index-base;
	ostr << siz;

	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		ostr 
			<< GC_::ptrToInt->find(itemptr->i.fun      )
			<< GC_::ptrToInt->find(itemptr->i.ufun     )
			<< GC_::ptrToInt->find(itemptr->i.this_line);
	}
	return ostr;
}
void ErrorStack::setPtrs(IntToPtr_ &intToPtr)
{
	int siz=index-base;
	ItemP itemptr=base;
	for (int i=siz;  i>=1;  i--)
	{	itemptr++;
		itemptr->i.fun   = (Object*)intToPtr.find(int(itemptr->i.fun ));
		itemptr->i.ufun  = (Object*)intToPtr.find(int(itemptr->i.ufun));
		itemptr->i.this_line=(Line*)intToPtr.find(int(itemptr->i.this_line));
	}
}

NodeP lshrinkStacks(PairPP)
{
	ThreadP th=Thread::This();
	th->intStack->shrink();
	th->varPStack->shrink();
	th->argPStack->shrink();
	th->floatStack->shrink();
	th->stack->shrink();
	th->argStack->shrink();
	th->varStack->shrink();
	th->toDoStack->shrink();
	th->errorStack->shrink();
	return Unbound;
}

NodeP lsetStackNoisy(PairPP arg)
{
	bool noisy=arg[1]->a_bool(1)!=0;
	if throwing return Unbound;
	IntStack::noisy=noisy;
	return Unbound;
}
