//___________________________________________________________________
//
//                       Class StaticStaticBoxSizer
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#include <wx/sizer.h>
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/tglbtn.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/slider.h>
#include <wx/textctrl.h>
#include <wx/listbox.h>
#include <wx/listctrl.h>

#include "spinctld.h"
#include "spinctli.h"

#include "Main.h"
#include "File.h"
#include "Pair.h"
#include "String.h"
#include "Object.h"
#include "Int.h"
#include "Error.h"

#include "Event.h"
#include "Window.h"
#include "BoxSizer.h"
#include "StaticBoxSizer.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Point.h"
#include "Graph.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

StaticBoxSizer::StaticBoxSizer(StaticBoxSizerNodeP boxSizerNode_, 
										 int orient_, wxWindow* parent_, wxString label_):
	wxStaticBoxSizer(orient_, parent_, label_),
	boxSizerNode(boxSizerNode_)
{}

StaticBoxSizer::~StaticBoxSizer()
{
	if (boxSizerNode)
		boxSizerNode->boxSizer=0;
}

wxString StaticBoxSizer::wxstring()
{
	wxString str;
	str << "StaticBoxSizer";
	return str;
}

DataInputStream &StaticBoxSizer::read(DataInputStream &istr) 
{	
	return istr;
}
wxDataOutputStream &StaticBoxSizer::write(wxDataOutputStream &ostr) 
{
	return ostr;
}

StaticBoxSizerNode::StaticBoxSizerNode(): 
BoxSizerNode() 
{}

void StaticBoxSizerNode::free() 
{	
	if (boxSizer) 
	{	boxSizer->boxSizerNode=0;
		boxSizer=0;
	}
}

void StaticBoxSizerNode::inc_gc() {fifo << children;}
int StaticBoxSizerNode::youngs() {return gmark(children);}
int StaticBoxSizerNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(children, ptrToInt);
}
void StaticBoxSizerNode::check_oldyoungs() {check_oldyoung(children);}

DataInputStream &StaticBoxSizerNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	istr >> orient >> (int&)children >> StaticBoxSizerP(boxSizer)->label;
		boxSizer=BoxSizerP(new StaticBoxSizer(StaticBoxSizerNodeP(this), orient, 
			StaticBoxSizerP(boxSizer)->GetStaticBox()->GetParent(), StaticBoxSizerP(boxSizer)->label));
		if (!boxSizer)
		{	error(OUT_OF_MEM); return istr;
		}
		boxSizer->read(istr);
	}
	return istr;
}
wxDataOutputStream &StaticBoxSizerNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (boxSizer)
	{	ostr << 1 << orient << GC_::ptrToInt->find(children);
		boxSizer->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void StaticBoxSizerNode::setPtrs(IntToPtr_ &intToPtr) 
{
	children = PairP((Pair*)intToPtr.find(int(children)));
}

void StaticBoxSizerNode::aftersetPtrs2() 
{
	PairP list=children;
	while (list)
	{	wxWindow *win=0;
		wxSizer *sizer=0;
		bigSwitch(list->first(), win, sizer);
		next(list);
		int proportion=IntP(list->first())->int_;
		next(list);
		int    flag   =IntP(list->first())->int_;
		next(list);
		int   border  =IntP(list->first())->int_;
		next(list);
		if (sizer)
			boxSizer->Add(sizer, proportion, flag, border);
		else if (win)
			boxSizer->Add(win, proportion, flag, border);
		else
			win=win;	// dummy line for debugging
	}
}

StringP StaticBoxSizerNode::string()
{
	if (boxSizer)
		return new String(boxSizer->wxstring());
	return _Null_Word->name();
}

wxString StaticBoxSizerNode::wxstring()
{
	if (boxSizer)
		return boxSizer->wxstring();
	return wxEmptyString;
}

void StaticBoxSizerNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

void StaticBoxSizerNode::bigSwitch(NodeP arg, wxWindow *&win, wxSizer *&sizer)
{
	if (arg->is.window)
		win=WindowNodeP(arg)->window;
	else if (arg->type==BOXSIZER)
		sizer=StaticBoxSizerNodeP(arg)->boxSizer;
	else
		error(BAD_DATA, arg, 2);
}

NodeP StaticBoxSizerNode::bigArgSwitch(NodeP arg, 
											wxWindow *&win, wxSizer *&sizer)
{
	if (arg->is.window)
	{	WindowNodeP windowNode=arg->a_window(2);
		if throwing return Unbound;
		win=windowNode->window;
	}else if (arg->type==BOXSIZER)
	{	BoxSizerNodeP b=arg->a_boxSizer(2);
		if throwing return Unbound;
		sizer=b->boxSizer;
	}else
		error(BAD_DATA, arg, 2);
	return Unbound;
}

NodeP lStaticBoxSizer(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	int orient = wxHORIZONTAL;
	{	int o=arg[1]->a_int(1);
		if throwing return Unbound;
		orient=o;	
	}
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[2])
	{	parentNode=arg[2]->a_window(2);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[3]->a_string(3);
		if throwing return Unbound;
		label=l->ptr();	
	}
//	GUI_ENTER
	StaticBoxSizerNodeP boxSizerNode=new StaticBoxSizerNode();
	if throwing return Unbound;
	StaticBoxSizerP boxSizer=new StaticBoxSizer(boxSizerNode, orient, parent, label);
	if (!boxSizer)
	{	error(OUT_OF_MEM); return Unbound;
	}
	boxSizerNode->boxSizer=BoxSizerP(boxSizer);
	boxSizerNode->orient=orient;
//	GUI_LEAVE
	return boxSizerNode;
}
