#include "stdafx.h"
#include <math.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/thread.h>
#include <wx/datstrm.h>

#include <m_apm.h>

#include "Main.h"
#include "Thread.h"
#include "Error.h"
#include "File.h"
#include "IntToPtr_.h"
#include "PtrToInt_.h"
#include "Init.h"
#include "String.h"
#include "Pair.h"
#include "Int.h"
#include "Float.h"
#include "BigFloat.h"
#include "Complex.h"
#include "Object.h"
#include "Array.h"
#include "NodePStack.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif


Backslash backslash;

Backslash::Backslash()
{
	chars="()[]{}'\b\t\n<>\r\t\n\r;|~\"\\ +-*/^=:?.";	// ' instead of \a
	size=strlen(chars);
	begin=1;	// first char used for quoteds

	int c;
	for (c=0;  c<0x100;  c++)
	{	  array[c]=char(c);
		unarray[c]=char(c);
		    _is[c]=false;
	}
	for (c=0;  c < size;  c++)
	{	  array[(unsigned char)(chars[c])] = c+begin;
		unarray[c+begin] = chars[c];
		int h=c+begin;
		if (h<7 || h>13 || (h>10 && h<13))	// exclude \a, \b, \t, \n and \r
			_is[h] = true;
	}
}


const int String::digits=200;	// max number of digits in a number
bool String::caseIgnored=true;
int String::tabSize=3;

String::String(int length): Node(STRING) 
{
	_length=0; 
	_head= HeadP( GC_::calloc(sizeof(char), length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	_ptr= _head->str;
	_head->refCount=1;
	flag=ObjectType(0);
}

String::String(charP str): Node(STRING) 
{
	_length=0;
	int length=strlen(str); 
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, str, _length);
	_ptr= _head->str;
	_head->refCount=1;
	flag=ObjectType(0);
}

String::String(wxString str): Node(STRING) 
{
	_length=0;
	_ptr=charP(str.c_str()); 
	int length=str.Length(); 
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, _ptr, _length);
	_ptr= _head->str;
	_head->refCount=1;
	flag=ObjectType(0);
}

String::String(StringP sp): Node(STRING) 
{
	int length = sp->_length;
	_length=0;
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head)
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, sp->_ptr, _length);
	_ptr= _head->str;
	_head->refCount=1;
	flag=ObjectType(0);
}

String::String(StringP sp, StringP sp2): Node(STRING) 
{
	int length = sp->_length + sp2->_length;
	_length=0;
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return;
	}
	_length=length;
	copy(_head->str, sp->_ptr, sp->_length);
	copy(_head->str + sp->_length, sp2->_ptr, sp2->_length);
	_ptr= _head->str;
	_head->refCount=1;
	flag=ObjectType(0);
}

String::String(charP ptr_, HeadP head_, int length_, NodeType typ, CopyRoutineP copy):
	Node(typ)
{
	if (!head_) 
	{	head_= HeadP( GC_::malloc(length_+sizeof(int)+1) );
		if (!head_) 
		{	_head=0;
			_length=0;
			_ptr=0;
			return;
		}
		copy(head_->str, ptr_, length_);
		ptr_= head_->str;
		head_->refCount=0;
	}
	_length = length_;
      _ptr = ptr_;
	  _head = head_;

	_head->refCount++;
	flag=ObjectType(0);
}

String::String(PairPP list, int length_, NodeType typ, PairCopyRoutineP copy):
	Node(typ)
{
	_head = HeadP( GC_::malloc(length_+sizeof(int)+1) );
	if (!_head) 
	{	_length=0;
		_ptr=0;
		return;
	}
	copy(_head->str,  list, length_);
	_ptr = _head->str;
	_length = length_;
	_head->refCount=1;
	flag=ObjectType(0);
}

StringP String::make(charP ptr, HeadP head, int length, NodeType type, CopyRoutineP copy)
{
	if (length==0  &&  _Null_Word) 
	{	return StringP(_Null_Word);
	}else if (length >= 0)
		return new String(ptr, head, length, type, copy);
	error(BAD_DATA, 0);
	return _Null_Word->name();
}

StringP String::fromList(PairPP arg, int length, NodeType type, PairCopyRoutineP copy)
{
	if (length==0  &&  _Null_Word) 
		return StringP(_Null_Word);
	else
		return new String(arg, length, type, copy);

}

void String::free()
{
	if (_head  &&  !--_head->refCount)
	{	GC_::free(_head); _head=0;
	}
}

DataInputStream &String::read(DataInputStream &istr) 
{	wxString str;
	istr >> str;

	_ptr=charP(str.c_str()); 
	int length=str.Length(); 
	_length=0;
	_head= HeadP( GC_::malloc(length+sizeof(int)+1) );
	if (!_head) 
	{	_ptr=0;
		return istr;
	}
	_length=length;
	copy(_head->str, _ptr, _length);
	_ptr= _head->str;
	_head->refCount=1;
	return istr;
}
wxDataOutputStream &String::write(wxDataOutputStream &ostr) 
{	wxString str(_ptr, _length);
	return ostr << type << str;
}

NodeP String::copy(PtrToInt_ &ptrToInt)
{	StringP str=_Null_Word->name();
	if (_length >= 0)
		str=new String(this);
	else
		error(BAD_DATA, this);
	return str;
}

int String::_sizeof() {return length();}

void String::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers) 
{
	if (!is.string)
		return;
	int wid=0, dots=0;
	charP cp = ptr();
	
	if (width > 0 && width < length())
		wid=width-3;
	else
		wid=length();
	int w=width;
	if (width > 0 && width < length())
		w=width-3;
	if (wid < length()) 
		dots++;
	if ( ! is.backslashed)
	{	for (int i=0; i < wid; i++) 
			::print(strm, w, *cp++);
	}
	else if (!Thread::This()->backslashes) 
		for (int i=0; i < wid; i++) 
			::print(strm, w, backslash.decode(*cp++));
	else 
	{	int i;
		for (i=0; i < wid; i++) 
		{	if (backslash.is(cp[i])) 
				break;
		}
		if (i < length())		// word was in vbars
		{	if (strchr("\":", *cp)) 
				::print(strm, w, *cp++);
			::print(strm, w, '|');
			for (i=0; i < wid; i++) 
				::print(strm, w, backslash.decode(*cp++));
			::print(strm, w, '|');
		}else
		{	for (i=0; i < wid; i++) 
			{	if (strchr(backslash.chars,(int)*cp)) 
					::print(strm, w, '\\');
				::print(strm, w, *cp++);
			}
		}
	}
	width=w;
	if (dots) 
		::print(strm, width, "...");
}

void String::printBin(FILEP strm) 
{
	int width=Int::max;
	print(strm, width, Int::max);
}

NodeP String::contentsTree(wxArrayString &namearr, 
									wxArrayString &valarr, 
									wxArrayPtrVoid &objarr, 
									ObjectType want, 
									ObjectType dont,
									ObjectType must, 
									int &count, int level)
{
	objarr[count]=this;

	if (!expanded())
	{	StringP str=expandFormat();
		if (str->ptr())
		{	valarr[count]=str->ptr();
			valarr[count].Truncate(str->length());
		}
		incContentsArrays(namearr, valarr, objarr, count);
		return this;
	}
	int l=length(), i=1;
	PairP list=0, val=0, lastnode=0, tnode=0;
	charP start=ptr();
	while (l > 0) 
	{	charP end=strchr(start, '\n');
		int len;
		if (end)
		{	len=end-start;
			if (len > l)
				len=l;
		}else
			len=l;
		StringP str = make(start, 0, len, type, copy_mend);
		str=str->expandFormat();
		namearr[count]=wxString(' ',level-1);
		namearr[count] << i;
		valarr[count]=str->ptr();
		i++;
		l-=len;
		start=end+1;
		incContentsArrays(namearr, valarr, objarr, count);
		if (checkThrowing()) break;
	}
	return this;
}

int String::numberP() 
{
	if (_length >= digits) 
		return 0;

	charP p=_ptr; 
	int l=0, left=0, right=0;	// left counts digits left of decimal point

	if (l <_length  &&  *p=='-')
		p++, l++;
	
	while (l <_length  &&  is_digit(*p))
		p++, l++, left++;
	
	if (l <_length  &&  *p=='.') 
	{	p++, l++;
		while (l <_length  &&  is_digit(*p))
			p++, l++, right++;
	}
	if (l <_length  &&  (left || right)  &&  (*p=='E' || *p=='e')) 
	{	p++, l++;
		if (l <_length  &&  (*p=='+' || *p=='-'))
			p++, l++;
		
		while (l <_length  &&  is_digit(*p))
			p++, l++, right++;
	}
	if (l <_length  &&  (left || right)  &&  (*p=='I' || *p=='i')) 
	{	p++, l++;
	}
	if ((left==0 && right==0)  ||  l != _length)
		return 0;
	else
		return right+1;
}

NodeP String::number()
{
	char s2[digits];
	charP s=s2;
	int right=numberP();

	if (_length < digits  &&  right) 
	{	if (is.backslashed)
			copy_noparity(s, _ptr, _length);
		else
			copy(s, _ptr, _length);

		if (*s=='+') ++s;
		if (s2[_length-1]=='.') 
			s2[_length-1]= 0;

		if (right-1  ||  _length > 9) 
		{	if (s2[_length-1]=='I' || s2[_length-1]=='i')
				return new Complexd(0,atof(s));
			else
				if (_length < Float::mantDigits)
					return new Float(atof(s));
				else
					return new BigFloat(s);
		}else 
		{	if (s2[_length-1]=='I' || s2[_length-1]=='i')
				return new Complexd(0,atol(s));
			else
				return new Int(atol(s));
		}
	}else 
		return Unbound;
}

double String::getnumber()
{
	char s2[digits];

	charP s=s2;
	int right=numberP();

	if (_length < digits  &&  right) 
	{	if (is.backslashed)
			copy_noparity(s, _ptr, _length);
		else
			copy(s, _ptr, _length);

		if (*s=='+') ++s;
		if (s2[_length-1]=='.') 
			s2[_length-1]= 0;

		if (right-1  ||  _length > 9) 
			return atof(s);
		else 
			return atol(s);
	}else 
		return Float::quietNaN;
}

int String::compare(NodeP nd)
{
	if (!nd || nd->islist() || nd->isarray()) 
		return Node::boolerror;

	String &b = *nd->string();

	if (type==STRING  &&  b.type==STRING) 
	{	if (length()==b.length()  &&  ptr()==b.ptr())
			return 0;
		else 
		{	int cmp_len = length() < b.length() ?
							  length() : b.length();
// case-dependent comparsion should be done by changing the strcomp function pointer
			int icmp = caseIgnored ? 
					comp_low (ptr(), b.ptr(), cmp_len) :
					comp_case(ptr(), b.ptr(), cmp_len);

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);

			if (length() != b.length()  &&  icmp==0)
				icmp = length() - b.length();

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);
			return icmp;
		}
	}else if (is.backslashed || b.is.backslashed) 
	{	if (length()==b.length()  &&  ptr()==b.ptr())
			return 0;
		else 
		{	int cmp_len = length() < b.length() ?
							  length() : b.length();
			int icmp = caseIgnored ?
					comp_noparitylow(ptr(), b.ptr(), cmp_len) :
					comp_noparity   (ptr(), b.ptr(), cmp_len);

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);

			if ((length() != b.length()) && icmp==0)
				icmp = length() - b.length();

			icmp=icmp<0? -1 : (icmp>0? 1 : 0);
			return icmp;
		}
	}else 
		return Node::boolerror;
}

int String::count() {return length();}

double String::_max(double d) {return number()->_max(d);}

double String::_min(double d) {return number()->_min(d);}

NodeP String::Norm()
{	return new Float(Float::abs(getnumber()));
}

NodeP String::maxNorm()
{	return new Float(Float::abs(getnumber()));
}

StringP String::first() {return make(ptr(), head(), 1, type, copy);}

NodeP String::first_() 
{	return first();
}

NodeP String::butFirst()
{	if (length() > 1)
		return make(ptr()+1, head(), length()-1, type, copy);
	else
		return _Null_Word;
}

NodeP String::last()	
{	if (length())
		return make(ptr()+length()-1, head(), 1, type, copy);
	return _Null_Word;
}

NodeP String::butLast()
{	if (length() > 1)
		return make(ptr(), head(), length()-1, type, copy);
	return _Null_Word;
}

NodeP String::fPut(NodeP nd)
{	if (!nd)
	{	error(BAD_DATA, nd); return Unbound;
	}
	StringP str=nd->string();
	return new String(str, StringP(this));
}

NodeP String::lPut(NodeP nd)
{	if (!nd)
	{	error(BAD_DATA, nd); return Unbound;
	}
	StringP str=nd->string();
	return new String(StringP(this), str);
}

NodeP String::item(int i)
{	if (i > 0  &&  i <= _length)
		return new String(_ptr+i-1, _head, 1, type, String::copy);
	error(BAD_DATA, new Int(i)); return Unbound;
}

NodeP String::memberP(NodeP subobj, bool notp, bool substr)
{
	StringP sub  = subobj->string();

	if ( sub != Unbound && this != Unbound 
	&& sub->_length <= _length 
	&& (substr || sub->_length==1) )
	{
		int len=_length - sub->_length;
		StringP tmp = String::make(_ptr, _head, sub->_length, type, copy);

		for (int i=0; i <= len; i++) 
		{	if ( sub->equalP(tmp) ) 
			{	if (notp) 
				{	tmp->_length = len+sub->_length-i;
					return tmp;
				}else 
					return _True;
			}
			tmp->_ptr++;
		}
	}
	return notp ?   _Null_Word  :  _False;
}

NodeP String::items(NodeP start, NodeP end) 
{	if (start->type==INT_)
	{	if (end->type==INT_)
		{	int s=IntP(start)->int_, e=IntP(end)->int_;
			if (s > 0  &&  s <= _length)
				if (e >= 0  &&  e <= _length) 
				{	if (e < s)
						return _Null_Word;
					StringP newstr=String::make(ptr()+s-1, head(), e-s+1, type, copy);
					return newstr;
				}else
				{	error(BAD_DATA, end, 2); return Unbound;
				}
			else
			{	error(BAD_DATA, start, 1); return Unbound;
			}
		}
		error(BAD_DATA, end, 2); return Unbound;
	}
	error(BAD_DATA, start, 1); return Unbound;
}

NodeP String::setitem(NodeP index, NodeP newitem, bool safe)
{	int i=index->a_int(1)-1;
	StringP ch=newitem->a_char(3);
	if (!throwing)
	{	if (i<0 || i>=_length)
		{	error(BAD_DATA, index);
			return Unbound;
		}
		_ptr[i]=ch->_ptr[0];
	}
	return Unbound;
}

void String::removeitem(NodeP index)
{	int i=index->a_int(1)-1;
	if throwing return;
	
	if (i<0 || i>=_length)
	{	error(BAD_DATA, index);
		return;
	}
	charP cp=&_ptr[i];
	while (i<_length-1)
	{	*cp=cp[1];
		cp++;
		i++;
	}
	_length--;
}

NodeP String::reverse()
{
	StringP result=make(_ptr, 0, _length, STRING, copy_reverse);
	return result;
}

NodeP String::remove(NodeP thing)
{
	if (!thing)
	{	error(BAD_DATA, thing);
		return Unbound;
	}
	StringP str=new String(this);
	if throwing return Unbound;
	wxString s=str->ptr();

	StringP thingstr=new String(thing->string());
	if throwing return Unbound;
	wxString thstr=thingstr->ptr();
	if (thstr.Length()==0) 
		return this;
	
	int pos=s.Find(thstr);
	unsigned int thl=thstr.Length();
	while (pos!=-1)
	{	s.Remove(pos, thl);
		pos=s.Find(thstr);
	}
	if (s.Length()==0) 
		return StringP(_Null_Word);
	return new String(s);
}

NodeP String::remdup()
{
	int i, table[256];
	for (i=255; i>=0; i--)
		table[i]=0;

	charP end=&_ptr[_length-1], s=end;
	for (i=_length; i>0; i--)
	{	table[*s--]=1;
	}
	StringP res=new String(StringP(this));
	if throwing return res;
	charP s2=res->_ptr;
	s=end;
	int l=0;
	for (i=_length; i>0; i--)
	{	if (table[*s]==1)
		{	table[*s]=-1;
			*s2++=*s;
			l++;
		}
		s--;
	}
	StringP result=make(res->_ptr, 0, l, STRING, copy_reverse);
	return result;
}

NodeP String::upper() {return make(ptr(), 0, length(), type, copy_cap);}
NodeP String::lower() {return make(ptr(), 0, length(), type, copy_low);}

PairP String::list()
{
	charP cp=ptr()+length()-1;
	PairP out=0;
	for (int i=length(); i; i--, cp--)
	{	out=new Pair(new String(cp, 0, 1, STRING, copy), out);
		if throwing return PairP(Unbound);
	}
	return out;
}

ArrayP String::array(int origin) 
{
	int len=length();
	charP cp=ptr();
	ArrayP out=new Array(len, origin);
	if throwing return ArrayP(Unbound);
	NodePP d=out->data();
	for (int i=len; i; i--)
		*d++=new String(cp++, 0, 1, STRING, copy);
	return out;
}

//void String::untreeify() {}

StringP String::noparity_lower() {return make(ptr(), 0, length(), STRING, copy_noparitylow);}

charP String::expand_slash() 
{
	charP result, cp, cp2;
	int i, j;
	
	for (cp=_ptr, i=0, j=_length;  --j >= 0; )
		if (backslash.is(*cp++)) 
			i++;

	result = charP( GC_::malloc(_length+i+1) );
	if (!result) return 0;

	for (cp=_ptr, cp2=result, j=_length;  --j >= 0; ) 
	{	if (backslash.is(*cp)) 
			*cp2++='\\';
		*cp2++= backslash.decode(*cp++);
	}
	*cp2='\0';
	return result;
}

StringP String::expandSlash()
{
	charP tmp=expand_slash();
	if throwing return StringP(Unbound);
	int len=strlen(tmp);
	StringP result=new String(len);
	if throwing return StringP(Unbound);
	copy(result->_ptr, tmp, len);
	GC_::free(tmp); tmp=0;
	result->type=BACKSLASHED;
	return result;
}

StringP String::expandFormat()
{
	StringP tmp=new String(tabSize*_length);
	if throwing return StringP(Unbound);
	copy_expand_format(tmp->_ptr, _ptr, _length);
	int len=strlen(tmp->_ptr);
	StringP result=new String(len);
	if throwing return StringP(Unbound);
	copy(result->_ptr, tmp->_ptr, len);
	return result;
}

StringP String::noFormat()
{
	StringP d=new String(_length);
	if throwing return StringP(Unbound);
	copy_noformat(d->_ptr, _ptr, _length);
	return d;
}

StringP String::noParity()
{
	StringP d=new String(_length);
	if throwing return StringP(Unbound);
	copy_noparity(d->_ptr, _ptr, _length);
	return d;
}

StringP String::mend()
{
	StringP d=new String(_length);
	if throwing return StringP(Unbound);
	copy_mend(d->_ptr, _ptr, _length);
	return d;
}

charP String::copy(charP d, charP s, int n) 
{
	strncpy(d, s, n);
	d[n]='\0';
	return d;
}

charP String::copy_quote(charP d, charP s, int n) 
{
	d[0]='"';
	strncpy(d+1, s, n-1);
	d[n]='\0';
	return d;
}

charP String::copy_colon(charP d, charP s, int n) 
{
	d[0]=':';
	strncpy(d+1, s, n-1);
	d[n]='\0';
	return d;
}

charP String::copy_noparity(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++)
		d[i]= backslash.decode(s[i]);

	d[n]='\0';
	return d;
}

charP String::copy_cap(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (islower((unsigned char)(*s)))
			*d++= wxToupper(*s++);
		else
			*d++=*s++;

	*d='\0';
	return temp;
}

charP String::copy_low(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (wxIsupper((unsigned char)(*s)))
			*d++= wxTolower(*s++);
		else
			*d++=*s++;
	*d='\0';
	return temp;
}

charP String::copy_unspace(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (*s==' ')
		{	*d++= '-'; s++;
		}else
			*d++=*s++;
	*d='\0';
	return temp;
}

charP String::copy_noformat(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
		if (*s=='\t' || *s=='\n')
		{	s++;
		}else
			*d++=*s++;
	*d='\0';
	return temp;
}

charP String::copy_noparitylow(charP d, charP s, int n) 
{
	charP temp=d;
	for (int i=0; i < n; i++) 
	{	char c = backslash.decode(*s++);
		if (wxIsupper((unsigned char)c))
			*d++= wxTolower(c);
		else
			*d++= c;
	}
	*d='\0';
	return temp;
}

charP String::copy_vbar(charP d, charP s, int n) 
{
	charP start=d;
	int i=n;
	while (i>=0) 
	{	while (*s=='|') 
		{	s++;
			i--;
		}
		if (i>=0)
		{	*d++=*s++;
			i--;
		}
	}
	*d='\0';
	return start;
}

charP String::copy_mend_nosemi(charP d, charP s, int n) 
{
	bool vbar=false;
	for (int i=0; i < n; ) 
	{	while (*s=='|') 
		{	vbar = !vbar;	s++;
		}
		if (vbar) 
			if (strchr(backslash.chars, int(*s)))
				d[i++]= backslash.encode(*s++);
			else
				d[i++]=*s++;
		else 
		{	while ((*s=='~' && s[1]=='\n')) 
				while (*s=='~' && s[1]=='\n') 
					s+=2;

			if (*s !='|') 
				d[i++]=*s++;
		}
	}
	d[n]='\0';
	return d;
}

charP String::copy_mend(charP d, charP s, int n) 
{
	bool vbar=false;
	for (int i=0; i < n; ) 
	{	while (*s=='|') 
		{	vbar = !vbar;	s++;
		}
		if (vbar) 
			if (strchr(backslash.chars, (int)*s))
				d[i++]= backslash.encode(*s++);
			else
				d[i++]=*s++;
		else 
		{	while (*s==';' || (*s=='~' && s[1]=='\n')) 
			{
				while (*s=='~' && s[1]=='\n') 
					s+=2;

				if (*s==';')
					do s++; 
					while (*s !='\0' && *s !='~' && s[1] !='\n');
			}
			if (*s !='|') 
				d[i++]=*s++;
		}
	}
	d[n]='\0';
	return d;
}

charP String::copy_word(charP d, PairPP arg, int n)   // KLUDGE!
{
	ThreadP th=Thread::This();
	while (NodePP(arg)!=th->argStack->index) 
	{	++arg;
		if (!*arg)
			continue;
		StringP str = (*arg)->string();
		strncpy(d, str->ptr(), str->length());
		d += str->length();
	}
	*d='\0';
	return d;
}

charP String::copy_reverse(charP d, charP s, int n) 
{
	charP temp=d;
	s+=n-1;
	for (int i=0; i < n; i++) 
		*d++=*s--;
	*d='\0';
	return temp;
}

charP String::copy_expand_format(charP d, charP s, int n) 
{
	charP result=d;
	for (int i=n;  i > 0;  i--) 
	{	if (*s=='\t')
			for (int j=0;  j < tabSize;  j++)
				*d++=' ';
		else if (*s=='\n')
			;
//		{	*d++='\\';
//			*d++='n';
//		}
		else
			*d++=*s;
		s++;
	}
	*d='\0';
	return result;
}

int String::comp_noparity(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++) 
	{	if (backslash.decode(*d) != backslash.decode(*s))
			return backslash.decode(*d) - backslash.decode(*s);
		d++, s++;
	}
	return 0;
}

int String::comp_case(charP d, charP s, int n) 
{
	return strncmp(d, s, n);
/*	for (int i=0; i < n; i++) 
	{	if (*d !=*s) 
			return *d - *s;
		d++, s++;
	}
	return 0;
*/
}

int String::comp_low(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++) 
	{	if (*d !=*s) 
		{	if (wxIsupper((unsigned char)*s)) 
			{	if (wxIsupper((unsigned char)*d)) 

				{	if (wxTolower(*d) != wxTolower(*s))
						return wxTolower(*d) - wxTolower(*s);
				}else 
				{	if (*d != wxTolower(*s))
						return *d - wxTolower(*s);
				}
			}else 
				if (wxIsupper((unsigned char)*d)) 
				{	if (wxTolower(*d) !=*s)
						return wxTolower(*d) - *s;
				}else 
					return *d - *s;
		}
		d++, s++;
	}
	return 0;
}

int String::comp_noparitylow(charP d, charP s, int n) 
{
	for (int i=0; i < n; i++) 
	{	char c1 = backslash.decode(*d),
			  c2 = backslash.decode(*s);
		if (c1 != c2) 
		{	if (wxIsupper((unsigned char)c2)) 
			{	if (wxIsupper((unsigned char)c1)) 
				{	if (wxTolower(c1) != wxTolower(c2))
						return wxTolower(c1) - wxTolower(c2);
				}else 
				{	if (c1 != wxTolower(c2))
						return c1 - wxTolower(c2);
				}
			}else 
				if (wxIsupper((unsigned char)c1)) 
				{	if (wxTolower(c1)!=c2)
						return wxTolower(c1) - c2;
				}else 
					return c1 - c2;
		}
		d++, s++;
	}
	return 0;
}
