#include "stdafx.h"
#include <malloc.h>

#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/datstrm.h>
#include <wx/thread.h>

#include "Main.h"
#include "File.h"
#include "PtrToInt_.h"
#include "IntToPtr_.h"
#include "String.h"
#include "Pair.h"
#include "Object.h"
#include "Struct.h"
#include "Int.h"
#include "Int16.h"
#include "Int8.h"
#include "UInt8.h"
#include "Float.h"
#include "Complex.h"
#include "IntArray.h"
#include "Int16Array.h"
#include "FloatArray.h"

#include "Error.h"
#include "NodePStack.h"
#include "Init.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  calloc(n, m)	 _calloc_dbg((n), (m),   _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

Struct::Struct(PairP data_): 
	Node(STRUCT), size(0), base(0), _first(0)
{
	base = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(data_->count()+1));
	if (!base)
		return;
	size=data_->count();
	ObjectP last=0;
	PairP np;
	for (np=data_;  np;  next(np))
	{	StringP f=np->first()->first()->string();
		if (String::caseIgnored)
			f=StringP(f->lower());
		ObjectP obj=find(f);
		if (obj)
		{	error(BAD_DATA, f);
			return;
		}
		obj=insert(f);

		NodeP t=np->first()->second()->intern(), n=t;
		if (n==_Int)
			n=new Int();
		else if (n==_Int16)
			n=new Int16();

		else if (n==_Int8)
			n=new Int8();
		else if (n==_UInt8)
			n=new UInt8();
		else if (n==_Float)
			n=new Float;
		else if (n==_Complex_)
			n=new Complexd();
		else if (n==_Word_)
			n=new String();
		else if (n==_IntArray_)
			n=new IntArray();
		else if (n==_Int16Array_)
			n=new Int16Array();
		else if (n==_FloatArray_)
			n=new FloatArray();
		else 
			error(BAD_DATA, t);
		
		if throwing return;
		
		if (np->first()->tail()->tail())
		{	NodeP third=np->first()->third(), n3=third;
		
			if (t==_Int)
				n3=n3->_int();

			else if (t==_Int16)
				n3=n3->toint16();
			else if (t==_Int8)
				n3=n3->toint8();
			else if (t==_UInt8)
				n3=n3->touint8();
			else if (t==_Float)
				n3=n3->_float();
			else if (t==_Word_)
			{	if (n3->numberP())
					n3=new String(int(third->getnumber()));
				else
					n3=n3->string();
			}
			else if (t==_IntArray_)
			{	if (n3->numberP())
					n3=new IntArray(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else if (t==_Int16Array_)
			{	if (n3->numberP())
					n3=new Int16Array(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else if (t==_FloatArray_)
			{	if (n3->numberP())
					n3=new FloatArray(int(third->getnumber()));
				else
				{	error(BAD_DATA, third);
					return;
				}
			}
			else
			{	error(BAD_DATA, third);
				return;
			}
			if (n->type!=n3->type && !(n->is.string && n3->is.string))
			{	error(BAD_DATA, third);
				return;
			}
			n=n3;
		}
		if throwing return;
		obj->setvar(n);
		if (last)
			last->settable(TableP(obj));
		else
			setfirst(obj);
		last=obj;
	}
	last->settable(TableP(0));
}

void Struct::free() 
{
	GC_::free((NodeP*)base); base=0;
}

void Struct::inc_gc()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			fifo << obj;
}

int Struct::youngs() 
{
	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmark(obj);
	return youngs;
}

int Struct::youngsPtr(PtrToInt_ &ptrToInt) 
{
	int youngs=0;
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			youngs+=gmarkPtr(obj, ptrToInt);
	return youngs;
}

void Struct::check_oldyoungs()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			check_oldyoung(obj);


}

void Struct::setPtrs(IntToPtr_ &intToPtr)
{
	setfirst(ObjectP((Object*)intToPtr.find(int(_first))));
	ObjectPP itemptr=base;


	for (int i=size;  i;  i--)
	{	*itemptr=(Object*)intToPtr.find(int(*itemptr));
		check_oldyoung(*itemptr);
		itemptr++;
	}
}
DataInputStream &Struct::read(DataInputStream &istr) 
{	
	int siz;
	istr >> siz >> (int&)_first;

	base = (ObjectP*) GC_::calloc(sizeof(ObjectP), siz+1);
	if (!base)
		return istr;
	size=siz;
	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
	{	istr >> (int&)*itemptr;
		itemptr++;
	}
	return istr;
}
wxDataOutputStream &Struct::write(wxDataOutputStream &ostr) 
{
	ostr << type << size 
		<< GC_::ptrToInt->find(first());

	ObjectPP itemptr=base;
	for (int i=size;  i;  i--)
		ostr << GC_::ptrToInt->find(*itemptr++);
	return ostr;
}

NodeP Struct::copy(PtrToInt_ &ptrToInt)
{
	StructP str=new Struct();
	if throwing return Unbound;
	str->base = (ObjectP*) GC_::calloc(1, sizeof(ObjectP)*(size+1));
	if (!str->base)
		return 0;
	str->size=size;
	str->_first=ObjectP(ptrToInt.find(first()));
	ObjectPP itemptr=base, itemptr2=str->base;
	for (int i=size;  i;  i--)
		*itemptr2++ = ObjectP(ptrToInt.find(*itemptr++));
	return str;
}

StringP Struct::string()
{
	const int maxlength=65536;
	char textbuf[maxlength];
	ThreadP th=Thread::This();
	th->stringptr=textbuf;
	th->stringlen=maxlength;
	stringbuffer << NodeP(this);
	*th->stringptr='\0';
	return new String(textbuf);
}

void Struct::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	PairP lst=list();
	if (lst)
		lst->print(strm, width, depth, noBrackets, noSpacers);
}

void Struct::printBin(FILEP strm)
{
	for (ObjectP obj=first();  obj;  obj=ObjectP(obj->table()))
		if (obj->var())
			obj->var()->printBin(strm);
}

NodeP Struct::readbin(FileP file)
{
	if (Thread::This()->interactive && file->file()==File::conin)
	{	error(FILE_ERROR, new String
			(_("Can't read struct binary from keyboard")));
		return Unbound;
	}
	for (ObjectP obj=first();  obj;  obj=ObjectP(obj->table()))
	{	NodeP nd=obj->var()->readbin(file);
		obj->setvar(nd);
	}
	if (file->eof()) return 0;
	return this;
}

int Struct::hash(StringP str)
{
	unsigned int *i= (unsigned int*) str->ptr(),   index=0;

	for (char len=char(str->length());  len>0;  len-=sizeof(int))
	{	switch (len)
		{	case 0:	break;
			case 1:	index^= *i & 0x000000ff;	break;
			case 2:	index^= *i & 0x0000ffff;	break;
			case 3:	index^= *i & 0x00ffffff;	break;
			case 4:	index^= *i;	break;
			default:	index^= *i++ << 5;
		}
	};
	index %= size;
	return index;

}

ObjectP Struct::find(StringP str) 
{
	ObjectP obj;
	for (obj=base[hash(str)];  obj;  next(obj))
		if (str->equal(obj->name()))
			break;

	return obj;
}

ObjectP Struct::operator[](StringP str) {return find(str);}

ObjectP Struct::insert(StringP str)
{
	ObjectP &obj=base[hash(str)];
	obj = new Object(str, obj);
	check_oldyoung(obj);
	return obj;
}

void Struct::mark()
{
	ObjectPP itemptr=base;

	for (int i=size;  i;  i--)
		for (ObjectP obj = *itemptr++;  obj;  next(obj))
			obj->mark();
}

int Struct::count()
{
	int c=0;
	for (ObjectP obj=_first;  obj;  obj=ObjectP(obj->table()))
	{	c++;
	}
	return c;
}

int Struct::_sizeof()
{
	int size=0;
	for (ObjectP obj=_first;  obj;  obj=ObjectP(obj->table()))
		size+=obj->var()->_sizeof();
	return size;
}

NodeP Struct::item(NodeP index, NodeP result)
{	ObjectP obj=find(index->string());
	if (!obj)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return obj->var();
}

NodeP String::bititem(NodeP index)
{	int i=index->a_int(1);
	if throwing return Unbound;
	if (i<0 || i >= _length*int(sizeof(char))<<3)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	return boolnode((_ptr[i>>3] >> (i & 7)) & 1);
}

NodeP Struct::setitem(NodeP index, NodeP newitem, bool safe)
{	StringP str=index->a_string(1);
	ObjectP obj=0;
	if (String::caseIgnored)
		obj=find(StringP(str->lower()));
	else
		obj=find(str);
	if (!obj)
	{	error(BAD_DATA, index);
		return Unbound;
	}
	NodeP v=obj->var();
	if (!newitem)
	{	error(BAD_DATA, newitem, 3);
		return Unbound;
	}
	if (newitem->type==v->type)
	{	obj->setvar(newitem);
		return Unbound;
	}
	NodeP ni=0;
	if (v->type==INT_)
		ni=newitem->_int();
	else if (v->type==NINT16)
		ni=newitem->toint16();
	else if (v->type==NINT8)
		ni=newitem->toint8();
	else if (v->type==NUINT8)
		ni=newitem->touint8();
	else if (v->type==FLOAT_)
		ni=newitem->_float();
	else if (v->type==STRING)
	{	ni=newitem->string();
		if (StringP(ni)->length() != StringP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}
	else if (v->type==INTARRAY)
	{	ni=newitem->intarray();
		if throwing return Unbound;
		if (IntArrayP(ni)->length() != IntArrayP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}
	else if (v->type==INT16ARRAY)
	{	ni=newitem->int16array();
		if throwing return Unbound;
		if (Int16ArrayP(ni)->length() != Int16ArrayP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}
	else if (v->type==FLOATARRAY)
	{	ni=newitem->floatarray();
		if throwing return Unbound;
		if (FloatArrayP(ni)->length() != FloatArrayP(v)->length())
		{	error(BAD_DATA, newitem, 3);
			return Unbound;
		}
	}

	else
	{	error(BAD_DATA, newitem, 3);
		return Unbound;
	}
	if (ni->type!=v->type)
	{	error(BAD_DATA, newitem, 3);
		return Unbound;
	}
	obj->setvar(ni);
	return Unbound;
}

PairP Struct::list()
{
	PairP li=0, last=0, tail=0;
	for (ObjectP obj=_first;  obj;  obj=ObjectP(obj->table()))
	{	tail=	new Pair(
				new Pair(obj->name(), 
				new Pair(obj->var(), 0)), 0);
		if throwing return PairP(Unbound);
		if (last)
			last->settail(tail);
		else
			li=tail;
		last=tail;
	}
	return li;
}
