//___________________________________________________________________
//
//                          Class TextControl
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/textctrl.h>
#include <wx/stattext.h>
#include <wx/frame.h>
#include <wx/datstrm.h>

#include <math.h>

#include "Main.h"
#include "File.h"
#include "Object.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Int.h"
#include "UInt8.h"
#include "Point.h"
#include "NodePStack.h"

#include "Event.h"
#include "TextControl.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Error.h"
#include "Color.h"
#include "Graph.h"
#include "Console.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

//___________________________________________________________________

class MyTextCtrl : public wxTextCtrl
{
public:
	MyTextCtrl(wxWindow *parent, wxWindowID id,
				  const wxString &value = wxEmptyString,
              const wxPoint &pos = wxDefaultPosition,
              const wxSize &size = wxDefaultSize,
              long style = 0,
              const wxValidator& validator = wxDefaultValidator,
              const wxString &name = wxTextCtrlNameStr);



	// MSW sends extra kill focus event
	virtual ~MyTextCtrl()
	{
		if (parent) 
			parent->text = 0;
		parent=0;
	}

	TextControl *parent;

	void OnChar(wxKeyEvent &event);         // pass chars to TextControl
	void OnKeyDown(wxKeyEvent &event);
	void OnKeyUp(wxKeyEvent &event);

private:
	DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(MyTextCtrl,wxTextCtrl)
	EVT_CHAR		(MyTextCtrl::OnChar		)
	EVT_KEY_DOWN(MyTextCtrl::OnKeyDown	)
	EVT_KEY_UP	(MyTextCtrl::OnKeyUp		)
END_EVENT_TABLE()

MyTextCtrl::MyTextCtrl(wxWindow *parent_, wxWindowID id,
							  const wxString &value,
							  const wxPoint &pos, const wxSize &size,
							  long style,
							  const wxValidator& validator,
							  const wxString &name):
wxTextCtrl(parent_, id, value, pos, size, style, validator, name)
{
	parent = (TextControl*)parent_;
}

void MyTextCtrl::OnChar(wxKeyEvent &event)
{
	if (parent) 
		parent->OnChar(event);
	event.Skip();
}

void MyTextCtrl::OnKeyDown(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyDown(event);
	event.Skip();
}

void MyTextCtrl::OnKeyUp(wxKeyEvent &event)
{
	if (parent) 
		parent->OnKeyUp(event);
	event.Skip();
}
//___________________________________________________________________

BEGIN_EVENT_TABLE(TextControl, wxControl)
	EVT_CHAR			(		TextControl::OnChar		)
	EVT_KEY_DOWN	(		TextControl::OnKeyDown	)
	EVT_KEY_UP		(		TextControl::OnKeyUp		)
	EVT_TEXT			(-1,	TextControl::OnChange	)
	EVT_TEXT_ENTER	(-1,	TextControl::OnEnter		)
END_EVENT_TABLE()

TextControl::TextControl(wxWindow *parent)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	label=0;
	text=0;
}

TextControl::TextControl(TextControlNodeP textControlNode_, int style):
label(0), text(0), bestSize(-999, -999),
insertMode(true), writeBufferSize(0x1000)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=textControlNode_;
	wxWindow *parent=console;
	if (textControlNode_->parent)
		parent=textControlNode_->parent->window;
	create(parent, -1, wxEmptyString, 
			wxDefaultPosition, wxDefaultSize, style,
			wxDefaultValidator, _("TextControl"));
}

TextControl::TextControl(TextControlNodeP textControlNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &value_, 
			PairP onChange_,
			PairP onEnter_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
label(0), text(0), bestSize(-999, -999), value(value_),
insertMode(true), writeBufferSize(0x1000)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*(WindowNodeP*)(d(windowNode))=textControlNode_;
	*(PairP*)(d(onChange))=onChange_;
	*(PairP*)(d(onEnter))=onEnter_;
	bestSize = wxSize(-999, -999);
	textControlNode_->textControl=this;
	textControlNode_->style=style;
	create(parent, id, value, pos, size, style, validator, name);
}

bool TextControl::create(wxWindow* parent, wxWindowID id, 
			const wxString &value_, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name)
{
	if (!wxControl::Create(parent, id, pos, size, style|wxNO_BORDER,
								wxDefaultValidator, name))
	  return false;

	wxControl::SetLabel(name);
	wxControl::SetBackgroundColour(parent->GetBackgroundColour());
	wxControl::SetForegroundColour(parent->GetForegroundColour());

	int width = size.GetWidth(), height = size.GetHeight();

	wxSize best_size(DoGetBestSize());
	if (width  == -1) width  = best_size.GetWidth();
	if (height == -1) height = best_size.GetHeight();

	if (name != wxEmptyString)
	{	label = new wxStaticText(this, id, name, 
			wxPoint(0, 0), wxDefaultSize, 0, _("staticText"));
		if (!label)
		{	error(OUT_OF_MEM); return false;
		}
	}else
		label=0;

	if (label)
		text = new MyTextCtrl(this, id, value_,
			wxPoint(0,0),
			wxSize(width, height-label->GetSize().GetHeight()),
			style, validator, name);
	else
		text = new MyTextCtrl(this, id, value_,
			wxPoint(0,0),
			wxSize(width, height), style, validator, name);
	if (!text)
	{	error(OUT_OF_MEM); return false;
	}
	value=wxEmptyString;

	DoSetSize(pos.x, pos.y, width, height);
	SetBestFittingSize(wxSize(width, height));
	return true;
}

void TextControl::DoSetSize(int x, int y, int width, int height, int sizeFlags)
{
	wxSize bestSize(DoGetBestSize());
	if (width < 0)  width  = bestSize.GetWidth();
	if (height < 0) height = bestSize.GetHeight();

	wxWindow::DoSetSize(x, y, width, height, sizeFlags);

	int labelwidth  = 0;
	int labelheight = 0;
	if (label)
		label->GetSize(&labelwidth, &labelheight);

	if (text)   
		text->SetSize(0, labelheight, 
			width, height-labelheight-2);
}

wxSize TextControl::DoGetBestSize()
{
	if (bestSize.x == -999)
	{	wxTextCtrl *t=new wxTextCtrl((wxWindow*)this, wxID_ANY, GetValue(),
			wxPoint(0,0), wxDefaultSize, GetWindowStyle());
		if (!t)
		{	error(OUT_OF_MEM); return wxDefaultSize;
		}
		bestSize=t->GetBestSize();
		delete t;
		if (GetLabel()!=wxEmptyString)
		{	wxStaticText label((wxWindow*)this, wxID_ANY, GetLabel());
			wxSize ls=label.GetBestSize();
			bestSize.y+=ls.y;
			if (ls.x > bestSize.x)
				bestSize.x=ls.x;
		}
	}
	return bestSize;
}

TextControl::~TextControl()
{
	ThreadP th=Thread::This();
	if (*d(windowNode))
		WindowNodeP(*d(windowNode))->window=0;
	if (label)

	{	delete label;
		label=0;
	}
	if (text)
	{	delete text;
		text=0;
	}

	EventPP out=th->eventFiFo->out;
	EventP ev;
	while (*th->eventFiFo)
	{	*th->eventFiFo >> ev;
		switch (ev->type)
		{
		case Event::TextControlChange:
		case Event::TextControlEnter:
		case Event::TextControlChar:
		case Event::TextControlKeyDown:
		case Event::TextControlKeyUp:
			if (((EventText*)ev)->textControl==this)
				((EventText*)ev)->textControl=0;
			break;
		default:
			;
		}
	}
	th->eventFiFo->out=out;
}

void TextControl::SetValue(wxString &value)
{
	text->SetValue(value);
}

wxString TextControl::GetValue() const
{
	if (text)
		return text->GetValue();
	else
		return value;
}

#define TC_ON(NAME)														\
void TextControl::On##NAME(wxKeyEvent &event)					\
{																				\
	ThreadP th=Thread::This();											\
	EventP evt=new EventText##NAME(this, event);\
	if throwing return;													\
	*th->eventFiFo	<< evt;												\
	event.Skip();															\
}

TC_ON(Char)
TC_ON(KeyDown)
TC_ON(KeyUp)

void TextControl::OnChange(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventTextChange(this);
	if throwing return;
	*th->eventFiFo << evt;
	event.Skip();
}

void TextControl::OnEnter(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventTextEnter(this);
	if throwing return;
	*th->eventFiFo << evt;
	event.Skip();
}

wxString TextControl::wxstring()
{
	wxString str;
	str << _("TextControl") << " " << GetValue() << ' ' 
		<< '[' << GetName() << "] ";
	if (*d(onChange))
		str << (*d(onChange))->string()->ptr() << ' ';
	else
		str << "[] ";
	if (*d(onEnter))
		str << (*d(onEnter))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " << GetWindowStyle();
	return str;
}

DataInputStream &TextControl::read(DataInputStream &istr) 
{	
	wxPoint pos;
	wxSize size;
	wxString value, name, fontName;
	int fontSize, fontStyle, fontWeight, insertionPoint;
	unsigned char fr,fg,fb, br,bg,bb;
	
	istr >> value >> name 
		>> (int&)*d(onChange)
		>> (int&)*d(onEnter)
		>> pos.x >> pos.y >> size.x >> size.y
		>> (int&)insertMode >> insertionPoint
		>> fr >> fg >> fb 
		>> br >> bg >> bb 
		>> fontSize >> fontName >> fontStyle >> fontWeight;

	SetValue(value);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	if (!insertMode)
		toggleInsertMode();
	if (text)
		text->SetInsertionPoint(insertionPoint);
	SetForegroundColour(wxColour(fr,fg,fb));
	SetBackgroundColour(wxColour(br,bg,bb));
	wxFont f=GetFont();
	f.SetPointSize(fontSize);
	f.SetFaceName(fontName);
	f.SetStyle(fontStyle);
	f.SetWeight(fontWeight);
	SetFont(f);
	return istr;
}
wxDataOutputStream &TextControl::write(wxDataOutputStream &ostr) 
{
//	flush();
	ostr
		<< GetValue() << GetName() 
		<< GC_::ptrToInt->find(*d(onChange))
		<< GC_::ptrToInt->find(*d(onEnter ))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(insertMode);
	if (text)
		ostr << int(text->GetInsertionPoint());
	else
		ostr << 0;

	{	wxColour 
			f=text->GetForegroundColour(), 
			b=text->GetBackgroundColour();
		ostr
			<< f.Red() << f.Green() << f.Blue() 
			<< b.Red() << b.Green() << b.Blue();
	}
	wxFont f=text->GetFont();
	ostr 
		<< f.GetPointSize() << f.GetFaceName() 
		<< f.GetStyle() << f.GetWeight();
	return ostr;
}

void TextControl::flush()
{
	long from, to;
	if (text)
		text->GetSelection(&from, &to);
#ifdef WIN32
	if (!insertMode)
		to=from+writeBuffer.length();
	if (text)
		text->Replace(from, to, writeBuffer);
#else
	long ip=text->GetInsertionPoint();
	if (from < 0)
		from=ip;
	if (to < 0)
		to=ip;
	if (!insertMode)
	{	to=from+writeBuffer.length();
		if (to > text->GetLastPosition())
			to=text->GetLastPosition();
//		Remove(from, to);
	}
	text->Replace(from, to, writeBuffer);
//	WriteText(writeBuffer);
//	long ipnew=ip+writeBuffer.length();
//	SetInsertionPoint(ipnew);
#endif
	writeBuffer="";
}

void TextControl::toggleInsertMode()
{
	ThreadP th=Thread::This();
	flush();
	insertMode=!insertMode;
#ifdef _WIN32
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY , 0 );
	keybd_event( VK_INSERT, 0x0, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0);
#else
	text->EmulateKeyPress(wxKeyEvent(WXK_INSERT));
//	error(NOT_YET_SUPPORTED, new String("Console::toggleInsertMode"));
#endif
	th->dispatchMessages();
}

void TextControl::setInsertMode()
{
	if (!insertMode)
		toggleInsertMode();
}

void TextControl::setOverwriteMode()
{
	if (insertMode)
		toggleInsertMode();
}

NodeP TextControl::Cursor() 
{
	if (!text)
		return new Pair(new Int(0), new Pair(new Int(0),0));

	long pos=text->GetInsertionPoint();
	long x,y;
	text->PositionToXY(pos, &x, &y);
	return new Pair
	(	new Int(x),
		new Pair(new Int(y),0) );
}

int TextControl::getInsertionPoint() 
{
	if (!text)
		return 0;

	return text->GetInsertionPoint();
}

NodeP TextControl::setCursor(PairP pos) 
{
	if (!throwing && text) 
	{	flush();
		wxPoint coord;
		coord.x=IntP(pos-> first())->int_;
		coord.y=IntP(pos->second())->int_;

		if (coord.y >= text->GetNumberOfLines())
		{	int end=text->GetLastPosition();
			text->SetInsertionPoint(end);
			long x,y;
			text->PositionToXY(end, &x, &y);
			int missingLines=coord.y-y;
			wxString CRLFs;
			while (missingLines)
			{
#ifdef WIN32
				CRLFs+="\r\n";
#else
				CRLFs+="\n";
#endif
				missingLines--;
			}
			*text << CRLFs;
		}
		if (coord.x > text->GetLineLength(coord.y))
		{	text->SetInsertionPoint(
				text->XYToPosition( text->GetLineLength(coord.y), coord.y ));
			text->WriteText(wxString(
				coord.x+1-text->GetLineLength(coord.y), ' ').c_str());
		}
		text->SetInsertionPoint(text->XYToPosition(coord.x, coord.y));
	}
	return Unbound;
}

bool TextControl::SetBackgroundColour(const wxColour &c)
{
	if (text)
		return text->SetBackgroundColour(c);
	return false;
}

bool TextControl::SetForegroundColour(const wxColour &c)
{
	if (text)
		return text->SetForegroundColour(c);
	return false;
}

bool TextControl::SetFont(const wxFont &f)
{
	if (text)
		return text->SetFont(f);
	return false;
}

//___________________________________________________________________

#define NODE_ON(x)								\
PairP TextControlNode::on##x()				\
{														\
	if (textControl)								\
	return PairP(*textControl->d(TextControl::on##x));				\
	return 0;										\
}														\
														\
void TextControlNode::seton##x(NodeP nd)	\
{														\
	if (textControl)								\
	{	*textControl->d(TextControl::on##x) = PairP(nd);		\
		check_oldyoung(nd);						\
	}													\
}

NODE_ON(Char)
NODE_ON(KeyDown)
NODE_ON(KeyUp)
NODE_ON(Change)
NODE_ON(Enter)

TextControlNode::TextControlNode(): 
	WindowNode(TEXTCONTROL), 
	textControl((TextControlP&)window) 
{}

TextControlNode::TextControlNode(WindowNodeP parent_): 
	WindowNode(TEXTCONTROL, parent_), 
	textControl((TextControlP&)window) 
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void TextControlNode::free() 
{	
	if (textControl) 
	{	ThreadP th=Thread::This();
		*textControl->d(TextControl::windowNode)=0;
		textControl->Destroy(); 
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::TextControlChange:
			case Event::TextControlEnter:
			case Event::TextControlChar:
			case Event::TextControlKeyDown:
			case Event::TextControlKeyUp:
				if (((EventText*)ev)->textControl==textControl)
					((EventText*)ev)->textControl=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		textControl=0;
	}
}
void TextControlNode::inc_gc() 
{	fifo << onChange() << onEnter()
		 << onChar() << onKeyDown() << onKeyUp();
}
int TextControlNode::youngs() 
{	return gmark(onChange())+gmark(onEnter())
		+gmark(onChar())+gmark(onKeyDown())+gmark(onKeyUp());
}
int TextControlNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return 
		gmarkPtr(onChange(), ptrToInt)
		+gmarkPtr(onEnter(), ptrToInt)
		+gmarkPtr(onChar(), ptrToInt)
		+gmarkPtr(onKeyDown(), ptrToInt)
		+gmarkPtr(onKeyUp(), ptrToInt);
}
void TextControlNode::check_oldyoungs() 
{	check_oldyoung(onChange());
	check_oldyoung(onEnter());
	check_oldyoung(onChar());
	check_oldyoung(onKeyDown());
	check_oldyoung(onKeyUp());
}

DataInputStream &TextControlNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode >> style;
		textControl=new TextControl(TextControlNodeP(this), style);
		if (!textControl)
		{	error(OUT_OF_MEM); return istr;
		}
		textControl->read(istr);
		parent=parentNode;
	}
	return istr;
}
wxDataOutputStream &TextControlNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (textControl)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent) 
			<< style;
		textControl->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void TextControlNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (textControl)
	{	TextControlP t=textControl;
		setonChange ((Pair*)intToPtr.find(int(t->onChange )));
		setonEnter  ((Pair*)intToPtr.find(int(t->onEnter  )));
		setonChar   ((Pair*)intToPtr.find(int(t->onChar   )));
		setonKeyDown((Pair*)intToPtr.find(int(t->onKeyDown)));
		setonKeyUp  ((Pair*)intToPtr.find(int(t->onKeyUp  )));
	}
}

void TextControlNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	TextControlP t=textControl, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new TextControl(TextControlNodeP(this),
			p, -1, t->GetValue(), 
			PairP(*t->d(TextControl::onChange)), 
			PairP(*t->d(TextControl::onEnter )), 
			pos, t->GetSize(), 
			style, wxDefaultValidator, t->GetName());
	if (!t2)
	{	error(OUT_OF_MEM); return;
	}
	textControl=t2;
	setonChange (*t->d(TextControl::onChange ));
	setonEnter  (*t->d(TextControl::onEnter  ));
	if (!t->insertMode)
		t2->toggleInsertMode();
	t2->text->SetInsertionPoint(t->text->GetInsertionPoint());
	t2->text->SetForegroundColour(t->text->GetForegroundColour());
	t2->text->SetBackgroundColour(t->text->GetBackgroundColour());
	wxFont f=t->text->GetFont();
	t2->text->SetFont(f);
	delete t;
	textControl=t2;
}

StringP TextControlNode::string()
{
	if (textControl)
		return new String(textControl->wxstring());
	return _Null_Word->name();
}

wxString TextControlNode::wxstring()
{
	if (textControl)
		return textControl->wxstring();
	return wxEmptyString;
}

void TextControlNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lTextControl(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=Thread::This()->graph;
	WindowNodeP parentNode=0;
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString name=_("TextControl");
	{	StringP n=arg[2]->a_string(2);
		if throwing return Unbound;
		name=n->ptr();
	}
	wxString value;
	if (exist(3))
	{	StringP v=arg[3]->a_string(3);
		if throwing return Unbound;
		value=v->ptr();	
	}
	PairP onChange=0;
	if (exist(4))
	{	PairP oc=arg[4]->a_runnable(4);
		if throwing return Unbound;
		onChange=oc;
	}
	PairP onEnter=0;
	if (exist(5))
	{	PairP oe=arg[5]->a_runnable(5);
		if throwing return Unbound;
		onEnter=oe;
	}
	long style = 0;
	if (exist(6))
	{	int s=arg[6]->a_int(6);
		if throwing return Unbound;
		style=s;	

	}
	wxPoint pos=wxDefaultPosition;
	if (exist(7))
	{	PairP p=arg[7]->a_int_vector(7);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(8))
	{	PairP s=arg[8]->a_int_vector(8);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
//	GUI_ENTER
	TextControlNodeP textControlNode=new TextControlNode(parentNode);
	if throwing return Unbound;
	TextControl *tc=new TextControl(textControlNode,
			parent, id, value, 
			onChange, onEnter, 
			pos, size, style, wxDefaultValidator, name);
	if (!tc)
	{	error(OUT_OF_MEM); return Unbound;
	}
//	GUI_LEAVE
	return textControlNode;
}

NodeP lTextControlValue(PairPP arg)
{
	ThreadP th=Thread::This();
	if (exist(1))
	{	TextControlNodeP textControl=arg[1]->a_textControl(1);
		if throwing return Unbound;
//		GUI_ENTER
		StringP val=new String(textControl->textControl->GetValue());
//		GUI_LEAVE
		return val;
	}
	else
		return new String(th->textControlValue);
}

NodeP lTextControlSetValue(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->text->SetValue(value->ptr());
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlWrite(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->text->WriteText(value->ptr());
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlAppend(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP value=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->text->AppendText(value->ptr());
//	GUI_LEAVE
	return Unbound;
}

#define _TC_ON(NAME)									\
NodeP lTextControlOn##NAME(PairPP arg)			\
{															\
	TextControlNodeP textControl=arg[1]->a_textControl(1);\
	if throwing return Unbound;					\
	PairP function=arg[2]->a_runnable(2);		\
	if throwing return Unbound;					\
	textControl->seton##NAME(function);			\
	return Unbound;									\
}

_TC_ON(Change)
_TC_ON(Enter)
_TC_ON(Char)
_TC_ON(KeyDown)
_TC_ON(KeyUp)

NodeP lTextControlSetInsertionPointEnd(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->text->SetInsertionPointEnd();
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlInsertMode(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->setInsertMode();
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlOverwriteMode(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->setOverwriteMode();
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlCursor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
//	GUI_ENTER
	NodeP xy=textControl->textControl->Cursor();
//	GUI_LEAVE
	return xy;
}

NodeP lTextControlSetCursor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	PairP pos=arg[2]->a_pos_int_vector(2);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->setCursor(pos);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlSetBackgroundColor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	textControl->textControl->SetBackgroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}


NodeP lTextControlSetColor(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	Color c=arg[2]->a_color(2);
	if throwing return Unbound;
	wxColour wc(c);
//	GUI_ENTER
	textControl->textControl->SetForegroundColour(wc);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlSetFontSize(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=textControl->textControl->GetFont();
	f.SetPointSize(s);
	textControl->textControl->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlSetFontName(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	StringP s=arg[2]->a_string(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=textControl->textControl->GetFont();
	f.SetFaceName(s->ptr());
	textControl->textControl->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlSetFontStyle(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=textControl->textControl->GetFont();
	f.SetStyle(s);
	textControl->textControl->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlSetFontWeight(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int s=arg[2]->a_int(2);
	if throwing return Unbound;
//	GUI_ENTER
	wxFont f=textControl->textControl->GetFont();
	f.SetWeight(s);
	textControl->textControl->SetFont(f);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlEnable(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	textControl->textControl->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lTextControlDestroy(PairPP arg)
{
	TextControlNodeP textControl=arg[1]->a_textControl(1);
	if throwing return Unbound;
//	GUI_ENTER
	{	WindowNodeP p=textControl->parent;
		if (p)
		{	p->children = WindowPairP(p->children->destructiveRemove(textControl));
			p->check_oldyoung(p->children);
		}
	}
	textControl->textControl->Destroy();
//	GUI_LEAVE
	return Unbound;
}
