//___________________________________________________________________
//
//                         TextControl Header
//___________________________________________________________________

#ifndef WINDOWNODE
	#include "Window.h"
#endif

extern Function 
	lTextControl, lTextControlDestroy,
	lTextControlValue, lTextControlSetValue, 
	lTextControlWrite, lTextControlAppend,
	lTextControlSetInsertionPointEnd,
	lTextControlCursor, lTextControlSetCursor,
	lTextControlInsertMode, lTextControlOverwriteMode,
	lTextControlSetBackgroundColor, lTextControlSetColor,
	lTextControlSetFontSize, lTextControlSetFontName,
	lTextControlSetFontStyle, lTextControlSetFontWeight,
	lTextControlOnChar, lTextControlOnKeyDown, lTextControlOnKeyUp,
	lTextControlOnChange, lTextControlOnEnter,
	lTextControlEnable;

class wxStaticText;
class MyTextCtrl;

class TextControl: public wxControl
{
	public:
		wxStaticText *label;
		MyTextCtrl *text;
		wxSize bestSize;
		wxString value;
		bool insertMode;
		unsigned int writeBufferSize;
		wxString writeBuffer;
		enum Data
		{	WINDOW_DATA,
			onChange, uonChange, tonChange, 
			onEnter, uonEnter, tonEnter, 
			dataCount
		};
		NodeP data[dataCount];

		static int vmt;

		virtual NodePP d(int i) {return &data[i];}

		TextControl(wxWindow *parent);
		TextControl(TextControlNodeP textControlNode_, int style);
		TextControl(TextControlNodeP textControlNode_,
			wxWindow* parent, wxWindowID id, 
			const wxString &value, 
			PairP onChange_=PairP(0),
			PairP onEnter_=PairP(0),
			const wxPoint& pos = wxDefaultPosition, 
			const wxSize& size = wxDefaultSize, 
			long style=0, 
			const wxValidator& validator = wxDefaultValidator, 
			const wxString& name = _("TextControl"));

		bool create(wxWindow* parent, wxWindowID id, 
			const wxString &value, 
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name);

		~TextControl();

		virtual void DoSetSize(int x, int y, int width, int height,
						   int sizeFlags = wxSIZE_AUTO);

		virtual wxSize DoGetBestSize();

		virtual void SetValue(wxString &value);
		virtual wxString GetValue() const;

		void 
			OnClose	(wxCloseEvent &event),
			OnChange	(wxCommandEvent &event),
			OnEnter	(wxCommandEvent &event),
			OnChar	(wxKeyEvent &event),
			OnKeyDown(wxKeyEvent &event),
			OnKeyUp	(wxKeyEvent &event);

		wxString wxstring();
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

		void flush(), toggleInsertMode(), 
			setInsertMode(), setOverwriteMode();
		int getInsertionPoint();
		NodeP Cursor(), setCursor(PairP args);

		virtual bool SetBackgroundColour(const wxColour& colour);
		virtual bool SetForegroundColour(const wxColour& colour);
		virtual bool SetFont(const wxFont& font);

		DECLARE_EVENT_TABLE()
};
typedef TextControl *TextControlP;

class TextControlNode: public WindowNode
{
	public:
		TextControlP &textControl;
		static const TextControlNode aTextControlNode;

		operator_new;
		PairP onChange();
		PairP onEnter();
		PairP onChar();
		PairP onKeyDown();
		PairP onKeyUp();
		void setonChange	(NodeP function);
		void setonEnter	(NodeP function);
		void setonChar		(NodeP function);
		void setonKeyDown	(NodeP function);
		void setonKeyUp	(NodeP function);

		TextControlNode();
		TextControlNode(WindowNodeP parent_);

		void free();

		void inc_gc();
		int youngs();
		int youngsPtr(PtrToInt_ &ptrToInt);
		void check_oldyoungs();
		
		DataInputStream &read(DataInputStream &istr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);
		void setPtrs(IntToPtr_ &intToPtr);
		void recreate();

		StringP string();
		wxString wxstring();
		void print(FILEP strm, int &width, int depth, int noBrackets=0, int noSpacers=0);
		StringP _typeof();
};
