#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <stdio.h>
#include <math.h>

#ifdef OPENGL
	#include <GL/glu.h>
#endif

#include "Main.h"
#include "File.h"
#include "String.h"
#include "Bitmap.h"
#include "Texture.h"
#include "Int.h"
#include "Object.h"
#include "Pair.h"
#include "WindowPair.h"
#include "Error.h"
#include "Point.h"
#include "Color.h"
#include "Pen.h"
#include "Font.h"
#include "Screen.h"
#include "Turtle.h"
#include "Recorder.h"
#include "Graph.h"
#include "PtrToInt_.h"
#include "IntToPtr_.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

Texture::Texture(BitmapP b):
	Node(TEXTURE)
{
	setbitmap(b);
	if (Thread::This()->graph)
	{	setgraph(*Thread::This()->graph->d(Graph::windowNode));
		WindowPairP wp=new WindowPair(this, graph()->textures);
		if throwing return;
		graph()->textures=wp;
		graph()->check_oldyoung(graph()->textures);
	}
	else
		setgraph(0);	// should never be reached
	create();
}

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

void Texture::create()
{
	ThreadP th=Thread::This();
	if (!th->graph)
		return;
#ifdef OPENGL
	glGenTextures(1, (GLuint*)&_texnr);
	glBindTexture(GL_TEXTURE_2D, _texnr);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);	// Linear Filtering
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	// Linear Filtering

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	gluBuild2DMipmaps(GL_TEXTURE_2D, 4, _bitmap->winx, _bitmap->winy,
							GL_RGBA, GL_UNSIGNED_BYTE, _bitmap->pixels);
	glAlphaFunc(GL_GREATER, 0.01f);
//	glEnable(GL_ALPHA_TEST);
	glEnable(GL_TEXTURE_2D);
#endif
	R_Texture::maxnr=max(R_Texture::maxnr, _texnr);
	th->graph->screen->textured=true;
}

void Texture::free()
{
	if (graph() 
	&& graph()->textures!=NodeP(0) 
	&& graph()->textures->type==WINDOWPAIR) // not really correct!
	{	graph()->textures = WindowPairP(graph()->textures->destructiveRemove(this));
		graph()->check_oldyoung(graph()->textures);
	}
}

TextInputStream &Texture::read(TextInputStream &istr) 
{	return istr >> (int&)_bitmap >> (int&)_graph;
}
DataInputStream &Texture::read(DataInputStream &istr) 
{	return istr >> (int&)_bitmap >> (int&)_graph;
}
TextOutputStream &Texture::write(TextOutputStream &ostr) 
{	return ostr << type << ' '
		<< GC_::ptrToInt->find(_bitmap) << ' '
		<< GC_::ptrToInt->find(_graph ) << ' ';
}
wxDataOutputStream &Texture::write(wxDataOutputStream &ostr) 
{	return ostr << type 
		<< GC_::ptrToInt->find(_bitmap)
		<< GC_::ptrToInt->find(_graph );
}

void Texture::setPtrs(IntToPtr_ &intToPtr)
{
	setbitmap((Node*)intToPtr.find(int(_bitmap)));
	setgraph ((Node*)intToPtr.find(int(_graph )));
}

StringP Texture::string()
{
	return new String(_("Texture")); 
}

void Texture::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, _("Texture"));
}

PairP Texture::clone_all()
{
	ThreadP th=Thread::This();
	if (!th->graph || !*th->graph->d(Graph::windowNode))
	{	error(NOT_YET_SUPPORTED, new String(_("graphics")));
		return 0;
	}
	PairP tlist=GraphNodeP(*th->graph->d(Graph::windowNode))->textures,
		oldt=0, tnode=0, last=0;
	while (tlist)
	{	TextureP t=TextureP(tlist->first());
		if (t)
		{	tnode=new Pair(new Int(t->_texnr), 0);
			if throwing return PairP(Unbound);
			if (last)
				last->settail(tnode);
			else
				oldt=tnode;
			last=tnode;
			t->create();
		}
		next(tlist);
	}
	return oldt;
}

void Texture::unclone_all(PairP oldt)
{
	ThreadP th=Thread::This();
	if (!th->graph || !*th->graph->d(Graph::windowNode))
	{	error(NOT_YET_SUPPORTED, new String(_("graphics")));
		return;
	}
	PairP tlist=GraphNodeP(*th->graph->d(Graph::windowNode))->textures;
	while (tlist && oldt)
	{	TextureP t=TextureP(tlist->first());
		if (t)	
		{	t->_texnr=IntP(oldt->first())->int_;
			next(oldt);
		}
		next(tlist);
	}
}

