
#include <setjmp.h>

class NodePStack;
class IntStack;
class FloatStack;
class ToDoStack;
class VarStack;
class ErrorStack;
class Point;
class DPoint;
class Graph;
class EventFiFo;
class ListBox_;
class ChoiceBox;
class ComboBox;
class ListControl;
class Security;
class Turtle;
class PtrTable;

WX_DECLARE_LIST(Thread, ThreadList);	

class DispatchMutex: public wxMutex
{
public:
	void Lock();
};

class DispatchMutexLocker
{
	DispatchMutex *dm;
public:
	DispatchMutexLocker(DispatchMutex *dm_);
	~DispatchMutexLocker();
};

enum FiberState
{
	FIBER_RUNNING,
	FIBER_PAUSED,
	FIBER_KILLED,
};

struct Fiber
{
	int unknown, bottomStack, topStack, unknownarray[29];
};

class Thread: public wxThread
{
	public:
		ThreadP nextFiber, prevFiber, threadFiber, startFiber;
		Fiber *fiber;
		FiberState fiberState;

		Security *security;
		int memsize;

		NodePP bottomStack,	// points to the bottom of the C program stack. See begin of main!
			topStack;	// for Unix POSIX Threads

		// static vars from Main.h
		ControlType stoppingFlag;
		volatile bool inProcessIdle;
		bool inpause, instep, stepArmed, 
			updateVarsOnStep, lineInfoInTrace;
		int guiOwned, stepNr, fiberNr, inDispatch, 
			processIdleInterleaving;

		// static vars from eval:
		PairP nodeList;
		ToDoP what;

		bool dont_fix_ift, redefinable;
		int ift_iff_flag,
			repcount, user_repcount,	// counters for repeat & while
			ifori, iforend, iforstep,	// ...and for for
			trace_level,	// indentation level when tracing
			proccallInc;	// add this number to asp before proccall
		double fori, forend, forstep;	// counters for for

		TailCall tailcall;
		ValStatus val_status;

		NodeP last_call;	// the last proc called
		NodeP output_node;	// the value returned by the user function
		NodeP	didnt_output_name;  // the name of the proc that didn't OP 
		PairP loopcond;	// the loop condition list
		NodeP			fun;  // current function name 
		ObjectP	  ufun;  // current user-defined function name
		ObjectP lexufun;  // current user-defined lexical function
		LineP this_line;  // the current instruction line 
		LineP autoThisLine;  // for autoload: the currently parsed instruction
		LogoObjectP thisobj;	// current LogoObject
		PairP parents;		// current parents list for LogoObject constructor
		ObjectP last_ufun,	// the function that called this one 
							forname;		// name of the for control variable
		LineP last_line;	// the line that called this one 
		int var;			// dynamic frame pointer into varStack
		int frame;		// lexical frame pointer into varStack
		int lastFrame;	// pointer to the last frame pointer
		int fvar;		// same as var but for frameStack
		int lexicalDepth;
		int qmList;		// pointer into argStack for the Questionmark list
		ErrorStructP didnt_get_output;	// the name of the proc that wanted the OP

		NodePStack *argStack;	// argument Stack
		NodePStack    *stack;	// eval register stack
		NodePStack *funStack;	// for the calls window
		IntStack   *intStack;	// integer stack
		FloatStack *floatStack;	// float stack
		ToDoStack  *toDoStack;	// instruction stack 
		VarStack   *varStack;	// stack for storing global objects while local vars are bound
		IntStack  *varPStack;	// stack for numbers into varStack
		IntStack  *frameStack;// stack for frame pointers into varStack
		IntStack  *framePStack;// stack for numbers into frameStack
		IntStack  *argPStack;	// stack for numbers into argStack
		IntStack *argNrStack; // nr of actual arguments given to fun
		ErrorStack *errorStack;	// error stack

		// static vars from file.h:
		bool interactive, backslashes, loadNoisily;
		char digits, precision;
		charP 
			stringptr,  stringstart,  stringend, 
			rstringptr, rstringstart, rstringend;

		int stringlen, rstringlen,  indent,  
			marginX, marginY, 
			toPending, inputBlocking, 
			printWidthLimit, printDepthLimit;

		jmp_buf inputBlockingBuffer;

		StringP line, currentLine, deependProcName;
		FileP Reader, Writer, 
				Loader, Dribbler, Stringbuffer;

		TextControlNodeP textControl;
		NodeP startup;
		ArrayP specialForms;
		ObjectP logoLanguage;

		int lineLen, lineEnd;

		// static vars from Line.h:
		int col, tcol, row, readnr;
		StringP _filename;
		PairP filenames;

		// static vars from Error.h:
		NodeP throw_node;
		LineP err_msg;
		ErrorType erract_errtype;

		// static vars from Turtle.h
		int tess_nPoints, tess_prim;
		DPoint *tessNormal;

		Graph *graph;
		TurtleNodeP turtleNode;
		PtrTableP gtTable;

		// static vars from Event.h
		bool inEventDispatch;
		EventFiFo *eventFiFo;
		int keyboardValue,
			sliderValue, intControlValue;
		wxKeyEvent *keyEvent;
		bool checkBoxValue, toggleButtonValue, radioButtonValue;
		double floatControlValue;
		wxString textControlValue, comboBoxValue;
		ListBox_ *listBoxValue;
		ChoiceBox *choiceBoxValue;
		ComboBox *staticComboBox;
		ListControl *listControlValue;
		int listControlRow, listControlCol;
		wxString listControlText;
		wxMoveEvent *moveEvent;
		wxSizeEvent *sizeEvent;

		TableP namespce;
		ObjectP namespceobj;
		ObjectP user;

		NodeP erract;

		static ThreadP main;
		static HANDLE mainHandle;

		static ThreadList list;
		static DispatchMutex listCS;

		static int nrPaused;
		static int waitingForGui;
		static wxCriticalSection gui;
		static wxCriticalSection csWaitingForGui;

		static int circcounter;
		
		StringP filename();
		void setfilename(NodeP nd);

		void setcirccounter(int ccnt);

		Thread(wxThreadKind mode);

		~Thread();

		wxThreadError Create(unsigned int stacksize, 
			PairP nodeList_, ToDo what_, Graph *graph_, bool wxcreate);

		void destroy();

		ExitCode Entry();

		wxThreadInternal *GetInternal() {return wxThread::GetInternal();}

#ifndef WIN32
		bool TestDestroy();
#endif
		bool checkBreak();
		bool dispatch();
		bool processIdle();
		void dispatchMessages();
		void yield();
		void eventDispatch();
		void eventEatAll();

		void GuiEnter();
		void GuiLeave();
		bool GuiTryEnter();

		static Thread *This();
		static Thread *ThisThread();
};
