//___________________________________________________________________
//
//                        Class  ToggleButton
//
//                            Definitions
//___________________________________________________________________

#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <math.h>

#ifdef WIN32
	#include <wx/tglbtn.h>
#else
	#include <wx/checkbox.h>
#endif
#include <wx/frame.h>
#include <wx/datstrm.h>
#include <wx/stattext.h>
#include <wx/textctrl.h>

#include "Main.h"
#include "File.h"
#include "Object.h"
#include "LogoObject.h"
#include "Pair.h"
#include "WindowPair.h"
#include "String.h"
#include "Int.h"
#include "NodePStack.h"
#include "Color.h"
#include "Point.h"

#include "Event.h"
#include "ToggleButton.h"
#include "Frame.h"

#include "PtrToInt_.h"
#include "IntToPtr_.h"

#include "Init.h"
#include "Error.h"
#include "Graph.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

#ifdef WIN32
BEGIN_EVENT_TABLE(ToggleButton, wxToggleButton)
	EVT_TOGGLEBUTTON	(-1, ToggleButton::OnClick	)
END_EVENT_TABLE()
#else
BEGIN_EVENT_TABLE(ToggleButton, wxCheckBox)
	EVT_CHECKBOX	(-1, ToggleButton::OnClick	)
END_EVENT_TABLE()
#endif

ToggleButton::ToggleButton(ToggleButtonNodeP toggleButtonNode_)
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=toggleButtonNode_;
	wxWindow *parent=(wxWindow*)console;
	if (toggleButtonNode_->parent)
		parent=toggleButtonNode_->parent->window;
	Create(parent, -1, 
			wxEmptyString, 
			wxDefaultPosition, 
			wxDefaultSize, 
			0, 
			wxDefaultValidator, 
			_("ToggleButton"));
}

ToggleButton::ToggleButton(ToggleButtonNodeP toggleButtonNode_, 
			wxWindow* parent, wxWindowID id, 
			const wxString &label,
			PairP onClick_,
			const wxPoint& pos, 
			const wxSize& size, 
			long style, 
			const wxValidator& validator, 
			const wxString& name):
#ifdef WIN32
wxToggleButton(parent, id, label,
			pos, size, style,
			validator, name)
#else
wxCheckBox(parent, id, label,
			pos, size, style,
			validator, name)
#endif
{
	for (int i=0; i<dataCount; i++)
		data[i]=0;
	*d(windowNode)=toggleButtonNode_;
	toggleButtonNode_->window=WindowP(this);
	toggleButtonNode_->setonClick(onClick_);
}

ToggleButton::~ToggleButton()
{
	if (*d(windowNode))
		ToggleButtonNodeP(*d(windowNode))->toggleButton=0;
}

void ToggleButton::OnClick(wxCommandEvent &event)
{
	ThreadP th=Thread::This();
	EventP evt=new EventToggleButton(this);
	if throwing return;
	*th->eventFiFo << evt;
}

wxString ToggleButton::wxstring()
{
	wxString str;
	str << _("ToggleButton") << " [" << GetLabel() << "] " 
		<< '[' << GetName() << "] ";
	if (onClick)
		str << (*d(onClick))->string()->ptr() << ' ';
	else
		str << "[] ";
	str
		<< " [" << GetPosition().x << ' ' << GetPosition().y << "] " 
		<< " [" << GetSize().x << ' ' << GetSize().y << "] " 
		<< GetValue();
	return str;
}

DataInputStream &ToggleButton::read(DataInputStream &istr) 
{	
	wxString label, name;
	wxPoint pos;
	wxSize size;
	int style, state;
	
	istr >> label >> name >> (int&)*d(onClick)
		>> pos.x >> pos.y >> size.x >> size.y >> style >> state;

	SetLabel(label);
	SetName(name); 
	SetPosition(pos);
	SetSize(size);
	SetWindowStyle(style);
	SetValue(state!=0);
	return istr;
}
wxDataOutputStream &ToggleButton::write(wxDataOutputStream &ostr) 
{
	ostr
		<< GetLabel() << GetName() 
		<< GC_::ptrToInt->find(*d(onClick))
		<< GetPosition().x << GetPosition().y 
		<< GetSize().x << GetSize().y
		<< int(GetWindowStyle()) << int(GetValue());
	return ostr;
}

PairP ToggleButtonNode::onClick()
{
	if (toggleButton)
		return PairP(*toggleButton->d(ToggleButton::onClick));
	return 0;
}

void ToggleButtonNode::setonClick(NodeP nd) 
{
	if (toggleButton)
	{	ThreadP th=Thread::This();
		*toggleButton->d(ToggleButton:: onClick) = nd;
		*toggleButton->d(ToggleButton::uonClick) = th->lexufun;
		*toggleButton->d(ToggleButton::tonClick) = th->thisobj;
		check_oldyoung(nd);
	}
}

ToggleButtonNode::ToggleButtonNode(): 
	WindowNode(TOGGLEBUTTON), 
	toggleButton((ToggleButtonP&)window) 
{}

ToggleButtonNode::ToggleButtonNode(WindowNodeP parent_): 
	WindowNode(TOGGLEBUTTON, parent_), 
	toggleButton((ToggleButtonP&)window) 
{
	WindowNodeP p=parent;
	if (p)
	{	p->children = p->children->wlPut(this);
		p->check_oldyoung(p->children);
	}
}

void ToggleButtonNode::free() 
{	
	if (toggleButton) 
	{	ThreadP th=Thread::This();
		*toggleButton->d(ToggleButton::windowNode)=0;
		toggleButton->Destroy(); 
		{	WindowNodeP p=parent;
			if (p && p->children && p->children->type!=NTFREE)
			{	p->children = WindowPairP(
					p->children->destructiveRemove(this));
				p->check_oldyoung(p->children);
			}
		}
		EventPP out=th->eventFiFo->out;
		EventP ev;
		while (*th->eventFiFo)
		{	*th->eventFiFo >> ev;
			switch (ev->type)
			{
			case Event::tToggleButton:
				if (((EventToggleButton*)ev)->toggleButton==toggleButton)
					((EventToggleButton*)ev)->toggleButton=0;
				break;
			default:
				;
			}
		}
		th->eventFiFo->out=out;
		toggleButton=0;
	}
}
void ToggleButtonNode::inc_gc() {fifo << onClick();}
int ToggleButtonNode::youngs() {return gmark(onClick());}
int ToggleButtonNode::youngsPtr(PtrToInt_ &ptrToInt) 
{	return gmarkPtr(onClick(), ptrToInt);
}
void ToggleButtonNode::check_oldyoungs() {check_oldyoung(onClick());}

DataInputStream &ToggleButtonNode::read(DataInputStream &istr) 
{	
	int ok;
	istr >> ok;
	if (ok)
	{	WindowNodeP parentNode;
		istr >> (int&)parentNode;
		toggleButton=new ToggleButton(ToggleButtonNodeP(this));
		if (!toggleButton)
		{	error(OUT_OF_MEM); return istr;
		}
		toggleButton->read(istr);
		parent=parentNode;
	}
	return istr;
}

wxDataOutputStream &ToggleButtonNode::write(wxDataOutputStream &ostr) 
{
	ostr << type;
	if (toggleButton)
	{	ostr << 1 
			<< GC_::ptrToInt->find(parent);
		toggleButton->write(ostr);
	}
	else
		ostr << 0;
	return ostr;
}

void ToggleButtonNode::setPtrs(IntToPtr_ &intToPtr)
{
	parent=(WindowNode*)intToPtr.find(int(parent));
	if (toggleButton)
	{	setonClick((Pair*)intToPtr.find(int(toggleButton->onClick)));
	}
}

void ToggleButtonNode::recreate()
{
	wxWindow *p=0;
	if (parent)
		p=parent->window;
	ToggleButtonP t=toggleButton, t2=0;
	wxPoint pos=t->GetPosition();
	t2=new ToggleButton(ToggleButtonNodeP(this),
			p, -1, t->GetLabel(), 
			PairP(*t->d(ToggleButton::onClick)), pos, t->GetSize(), t->GetWindowStyle(), 
			wxDefaultValidator, t->GetName());
	if (!t2)
	{	error(OUT_OF_MEM); return;
	}
	t2->SetValue(t->GetValue());
	toggleButton=t2;
	setonClick(*t->d(ToggleButton::onClick));
	delete t;
	toggleButton=t2;
}

StringP ToggleButtonNode::string()
{
	if (toggleButton)
		return new String(toggleButton->wxstring());
	return _Null_Word->name();
}

wxString ToggleButtonNode::wxstring()
{
	if (toggleButton)
		return toggleButton->wxstring();
	return wxEmptyString;
}

void ToggleButtonNode::print(FILEP strm, int &width, int depth, int noBrackets, int noSpacers)
{
	::print(strm, width, wxstring());
}

NodeP lToggleButton(PairPP arg)
{
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI"))); return Unbound;
	}
	wxWindowID id=-1;
	wxWindow *parent=th->graph;
	WindowNodeP parentNode=0;
	if (th->graph)
		parentNode=GraphNodeP(*th->graph->d(Graph::windowNode));
	if (arg[1])
	{	parentNode=arg[1]->a_window(1);
		if throwing return Unbound;
		parent=parentNode->window;
	}
	wxString label=wxEmptyString;
	{	StringP l=arg[2]->a_string(2);
		if throwing return Unbound;
		label=l->ptr();
	}
	PairP onClick=0;
	{	PairP oc=arg[3]->a_runnable(3);
		if throwing return Unbound;
		onClick=oc;
	}
	long style = 0;
	if (exist(4))
	{	int s=arg[4]->a_int(4);
		if throwing return Unbound;
		style=s;	
	}
	wxPoint pos=wxDefaultPosition;
	if (exist(5))
	{	PairP p=arg[5]->a_int_vector(5);
		if throwing return Unbound;
		pos.x=IntP(p->first ())->int_;
		pos.y=IntP(p->second())->int_;
	}
	wxSize size=wxDefaultSize;
	if (exist(6))
	{	PairP s=arg[6]->a_int_vector(6);
		if throwing return Unbound;
		size.x=IntP(s->first ())->int_;
		size.y=IntP(s->second())->int_;
	}
	wxString name=label;
	if (exist(7))
	{	StringP n=arg[7]->a_string(7);
		if throwing return Unbound;
		name=n->ptr();
	}
//	GUI_ENTER
	ToggleButtonNodeP toggleButtonNode=new ToggleButtonNode(parentNode);
	if throwing return Unbound;
	ToggleButtonP toggleButton=new ToggleButton(toggleButtonNode,
			parent, id, label, 
			onClick, pos, size, style, wxDefaultValidator, name);
	if (!toggleButton)
	{	error(OUT_OF_MEM); return Unbound;
	}
	toggleButtonNode->toggleButton=toggleButton;
//	GUI_LEAVE
	return toggleButtonNode;
}

NodeP lToggleButtonOnClick(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	PairP function=arg[2]->a_runnable(2);
	if throwing return Unbound;
//	GUI_ENTER
	toggleButton->setonClick(function);
//	GUI_LEAVE
	return Unbound;
}

NodeP lToggleButtonEnable(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	toggleButton->toggleButton->Enable(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lToggleButtonDestroy(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
//	GUI_ENTER
	toggleButton->toggleButton->Destroy();
//	GUI_LEAVE
	return Unbound;
}

NodeP lToggleButtonSetValue(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
	int state=arg[2]->a_bool(2);
	if throwing return Unbound;
//	GUI_ENTER
	toggleButton->toggleButton->SetValue(state!=0);
//	GUI_LEAVE
	return Unbound;
}

NodeP lToggleButtonValue(PairPP arg)
{
	ToggleButtonNodeP toggleButton=arg[1]->a_toggleButton(1);
	if throwing return Unbound;
//	GUI_ENTER
	bool state=toggleButton->toggleButton->GetValue();
//	GUI_LEAVE
	return Node::boolnode(state);
}

