//___________________________________________________________________
//
//                          Class Turtle
//___________________________________________________________________

class Pen;
class Font_;
class TextInputStream;
class DataInputStream;
class TextOutputStream;
class wxDataOutputStream;
class Recorder;

struct DPointDPointColor
{	DPoint p, t;
	Color c;
};

typedef DPointDPointColor *DPointDPointColorP;

class Screen_;
class GLUtesselator;

enum FillModes
{
	FILL_NORMAL=1, FILL_SURFACE,
	FILL_TOLERANT, FILL_SURFACE_TOLERANT,
	FILL_SMALL, FILL_SMALL_SURFACE,
	FILL_SMALL_TOLERANT, FILL_SMALL_SURFACE_TOLERANT,
};

class Turtle: public Point
{
	public:
		enum Mode {wrapmode, fencemode, windowmode};

		Pen pen;
		Font_ font;
		bool shown, outOfBounds, inEraseMode, 
			cylinderLines, 
			polyMode, polyShown, 
			tessMode, tessShown, surfaceMode, shadows;

		Mode mode;
		double heading, colorTolerance;
		Point polypt1,polypt2, normal, texpt1, texpt2, texpos, lightPos;
		Color polyc1, polyc2, shadowcolor;
		Matrix u, hbase;
		Screen_ *screen;
		Recorder *recorder;
		unsigned int tessWindingRule;

//		int refCount;
		int polyPointNr, tessPointNr;
		GLUtesselator *tess;
		Turtle *old;
		TurtleNodeP turtleNode;
		int shapeNr;

		Turtle(const Turtle &b);
		Turtle(bool makeOld=true);
		~Turtle();

		Turtle &operator=(const Turtle &b);
		Turtle &operator=(const Point &b);

		void 
			create(Recorder *recorder_, Screen_ *screen_), destroy(),

			setScreenColor(Color c), plainXOrPen(),

			setColor(Color c),
			setPenSize(Point size),
			setPenSize(double sizeX, double sizeY),
			setPenMode(Pen::WinMode newmode),

			penErase(), penDown(), penReverse(), restorePen(),

			reset3d(),
			moveTo(),
			moveTo(Point target), lineTo(Point target),
			moveTo(double x, double y), lineTo(double x, double y),
			polyStart(),
			polyEnd(),
			tessStart(),
			tessDraw(),
			tessContour(),
			tessEnd(),
			surfaceStart(),
			surfaceColumn(),
			surfaceColumnSh(),
			surfaceEnd(),

			pix(int w,int h, Color color),
			floodFill(int x, int y, Color fill, Color rim),
			floodFillRecursive(int x, int y, Color fill, Color rim),

			floodFillSurface(int x, int y, Color fill, Color cin),
			floodFillSurfaceRecursive(int x, int y, Color fill, Color cin),

			floodFillTolerant(int x, int y, Color fill, Color rim),
			floodFillTolerantRecursive(int x, int y, Color fill, Color rim),

			floodFillSurfaceTolerant(int x, int y, Color fill, Color cin),
			floodFillSurfaceTolerantRecursive(int x, int y, Color fill, Color cin),

			setPixel(int x, int y, Color c),
			floodFillSmall(int x, int y, Color fill, Color rim, int safe=0),
			floodFillSmallSurface(int x, int y, Color fill, Color cin, int safe=0),
			floodFillSmallTolerant(int x, int y, Color fill, Color rim, int safe=0),
			floodFillSmallSurfaceTolerant(int x, int y, Color fill, Color cin, int safe=0),

			set2dRasterPos(),
			fill(int mode), 

			fillRect(Point a, Point b),
			fillEllipse(Point radius), 
			cylinder(double length, double baseradius, double topradius),
			partialCylinder(double length, double baseradius, double topradius, 
								 double startangle, double endangle, double scrunch=1, int slices=30),
			sphere(double radius, int slices=0, int stacks=0),
			ellipsoid(Point radius, int slices=0, int stacks=0),
			partialEllipsoid(Point radius, 
				double slicestart, double sliceend, int slices,
				double stackstart, double stackend, int stacks),
			partialEllipsoidSh(Point radius, 
									 double slicestart, double sliceend, int slices,
									 double stackstart, double stackend, int stacks),
			fillPie(double angle, Point radius),

			getInWindow(),
			draw(), show(), hide(),

			setHeading(double angle), 
			     right(double angle), 
			   upPitch(double angle), 
			 rightRoll(double angle),
			     spinX(double angle),
			     spinY(double angle),
			     spinZ(double angle),
			   setRoll(double angle),
			  setPitch(double angle),
			setOrientation(Point o),

			forward(double d),

			setPos(Point target),
			setPos(double x, double y, double z),
			setPos(NodeP xnode, NodeP ynode, NodeP znode),
			_setPos(Point target),
			_setPos(double x, double y, double z),
			_setPos(NodeP xnode, NodeP ynode, NodeP znode),
			setSpherePos(double r, double phi, double theta),
			setCylinderPos(double r, double phi, double y),
			home(),

			setMatrix(PairP basex,PairP basey,PairP basez,PairP trans),
			setMatrix(Point basex,Point basey,Point basez,Point trans),
			setTurtleMatrix(),
			wrap(), arc(double angle, double radius),
			EllipseArc(double angle, double minorRadius, 
							double majorRadius, double startAngle),
			Label(wxString str), 
			setLabelSize(double x, double y),
			setLabelFont(wxString face),
			setLabelWeight(int weight),
			setLabelAlign(int alignx, int aligny),
			setPixel(Point p, Color c);
		
		Color getPixel(int x, int y);
		wxString getLabelFont();
		double towards_helper(Point end), distance(Point target),
			getHeading(), pitch(), roll();
		NodeP matrix(), getNodePenMode();

		Point getWindowPos();
		Point orientation();

		int Pixel();
		Point LabelSize(wxString str);

		void
			bitPaste(BitmapP b),
			bitPasteFast(BitmapP b),
			scroll(int width, int height, double sx, double sy);

		BitmapP bitCopy(double w, double h);
		Point	scrollCalibrate(double sx, double sy);

		TextInputStream &readEssentials(TextInputStream &istr);
		DataInputStream &readEssentials(DataInputStream &istr);
		TextOutputStream &writeEssentials(TextOutputStream &ostr);
		wxDataOutputStream &writeEssentials(wxDataOutputStream &ostr);

		TextInputStream &read(TextInputStream &istr);
		DataInputStream &read(DataInputStream &istr);
		TextOutputStream &write(TextOutputStream &ostr);
		wxDataOutputStream &write(wxDataOutputStream &ostr);

	private:
		bool surfaceShown;
		int surfaceColumnNr;
		double size, half_bottom, _sin, _cos, wrapDelta;
		Point shoulder, head, polyv1;
		ColorFiFo 
			*colorFiFo, *colorFiFo2, 
			*colorFiFo00, *colorFiFo01, *colorFiFo02;
		PointFiFo 
			*surfaceFiFo, *surfaceFiFo2, *surfaceFiFo3, 
			*surfaceFiFo00, *surfaceFiFo01, *surfaceFiFo02;
		DPointFiFo *tessFiFo;

		bool 
			wrapLeft (double d, Point end),
			wrapUp   (double d, Point end),
			wrapRight(double d, Point end),
			wrapDown (double d, Point end);

		void 
			sinCos(), drawRaw(), drawp(),
			checkMinX(), checkMaxX(), checkMinY(), checkMaxY();

		static void
			__stdcall tessglBegin(unsigned int prim),
			__stdcall tessglVertex(void *data),
			__stdcall tessglEnd(),
			__stdcall tessglCombine(DPointP p, DPointDPointColorP d[4], 
							float w[4], void **out),
			__stdcall tessglError(unsigned int errorCode);

	friend Function lTurtleMatrix, lsetTurtleMatrix;
};
