#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/app.h>
#include <wx/listctrl.h>
#include <wx/aui/aui.h>
#include <wx/config.h>
#include <wxflatmenu/flat_menu.h>
#include <wxflatmenu/flat_menu_bar.h>

#include "Main.h"
#include "File.h"
#include "Object.h"
#include "Pair.h"
#include "Table.h"
#include "LogoObject.h"
#include "Table.h"
#include "Array.h"
#include "FloatArray.h"
#include "Float.h"
#include "String.h"

#include "Vars.h"
#include "MainFrame.h"
#include "LogoApp.h"
#include "Stack.h"
#include "Init.h"
#include "Thread.h"
#include "Aui.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
#endif

BEGIN_EVENT_TABLE(StackVars, wxListCtrl)
	EVT_LIST_ITEM_FOCUSED(-1, StackVars::OnClick)
	EVT_LIST_ITEM_ACTIVATED(-1, StackVars::OnClick)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(Vars, wxListCtrl)
	EVT_LIST_ITEM_FOCUSED(-1, Vars::OnClick)
	EVT_LIST_ITEM_ACTIVATED(-1, Vars::OnClick)
END_EVENT_TABLE()

bool updateVarsOnStep=true;

StackVars::StackVars(wxWindow* parent, wxWindowID id, 
			  const wxPoint& pos, 
			  const wxSize& size, 
			  long style, 
			  const wxValidator& validator, 
			  const wxString& name):
wxListCtrl(parent, id, pos, size, style, validator, name)
{
	inupdate=false;
	wxString value=_("Value    "), v;
	for (int i=1; i<100; i++)
	{	v.Printf("%-3i...... ", i*10);
		value+=v;
	}

	InsertColumn(0, _("Name"), wxLIST_FORMAT_RIGHT, 100);
	InsertColumn(1,  value, wxLIST_FORMAT_LEFT , 500);

	SetFont(wxFont(8,
           wxMODERN,
           wxFONTSTYLE_NORMAL,
           wxNORMAL,
           false,
           "Courier New"));

	theApp->stackvars=this;
	show();
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuStackVars->Check(true);
	return;
}

StackVars::~StackVars()
{
	theApp->stackvars=0;
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuStackVars->Check(false);
}

void StackVars::show(bool shown)
{
	update();
	wxAuiPaneInfo &pane=theApp->aui->GetPane(this);
	if (pane.IsOk())
	{	pane.Show();
		
	}else
	{	wxString pi;
		if (theApp->config->Read("VarsPaneInfo", &pi))
			theApp->aui->LoadPaneInfo(pi, pane);
		else
			pane=wxAuiPaneInfo()
				.Name(wxT("StackVars")).Caption(_("StackVars"))
				.Bottom()
				.MinimizeButton().PinButton()
				.DestroyOnClose();
		theApp->aui->AddPane(this, pane);
	}
	theApp->aui->Update();
}

wxString StackVars::OnGetItemText(long item, long column) const
{
	if (inupdate)
		return wxEmptyString;
	if (!column)
		return namearr[item];
	return valarr[item];
}

void StackVars::OnClick(wxListEvent &ev)
{
	NodeP nd=NodeP(objarr[ev.GetIndex()]);
	if (!nd)
		return;
	nd->setExpanded(!nd->expanded());
	update();
}

void StackVars::update()
{
	if (inupdate)
		return;

	inupdate=true;

	ThreadP th=Thread::This();

//	charP reserve=new char[100000];
//	if (!reserve)
//		return;
//	delete reserve;

	int oldMaxSL=Pair::maxStringLength;
	Pair::maxStringLength=256;
	count=0;
	int stacksize=*th->varStack;
	count+=stacksize;
	int y=0;

	th->GuiEnter();

	namearr.Clear(); namearr.Alloc(count); namearr.Add(wxEmptyString, count);
	 valarr.Clear();  valarr.Alloc(count);  valarr.Add(wxEmptyString, count);
	 objarr.Clear();  objarr.Alloc(count);  objarr.Add(0, count);

	count=0;
	VarStack::ItemP base=th->varStack->base;
	for (VarStack::ItemP i=th->varStack->index;  i>base;  i--)
	{	StringP name=i->i.object->fullName();
		namearr[count]="";
		NodeP nd=i->i.val;
		int lev=2;
		if (nd && nd->type==OBJECT)
		{	nd=ObjectP(nd)->name();
		}else if (nd && nd->type==LOGOOBJECT)
		{	nd=LogoObjectP(nd)->name();
		}
		if (nd && nd->expandable())
			if (nd->expanded())
				namearr[count]+='-';
			else
				namearr[count]+='+';
		namearr[count]+=name->ptr();
		objarr[count]=nd;
		StringP val=0;
		if (nd)
		{	StringP valstr=_Null_Word->name();
			if (i->i.alias==-1)
			{	ObjectP obj=ObjectP(i->i.val);
				if (obj->var())
					valstr=obj->var()->string();
				StringP shortval=new String(
						valstr->ptr(), 0, valstr->length(), STRING, 
						String::copy_noformat);
				if throwing break;
				valstr=new String(wxString('&')+shortval->ptr());
				if throwing break;
				if (valstr  &&  valstr->is.string)
				{	StringP shortval=new String(
						valstr->ptr(), 0, valstr->length(), STRING, 
						String::copy_noformat);
					if throwing break;
					if (shortval->ptr())
						valarr[y]=shortval->ptr();
				}
			}
			else if (i->i.alias>0)	
			{	VarStack::ItemP j=th->varStack->base+i->i.alias;
				if (j->i.val)
					valstr=j->i.val->string();
				StringP shortval=new String(
						valstr->ptr(), 0, valstr->length(), STRING, 
						String::copy_noformat);
				if throwing break;
				valstr=new String(wxString('&')+shortval->ptr());
				if throwing break;
				if (valstr  &&  valstr->is.string)
				{	StringP shortval=new String(
						valstr->ptr(), 0, valstr->length(), STRING, 
						String::copy_noformat);
					if throwing break;
					if (shortval->ptr())
						valarr[y]=shortval->ptr();
				}
			}
			else
			{	if (nd->expanded())
				{	if (nd->type==LOGOOBJECT)
					{	valarr[count]=LogoObjectP(nd)->name()->ptr();
						Node::incContentsArrays(namearr, valarr, objarr, count);
					}else
					{	valarr[count]=nd->_typeof()->ptr();
						Node::incContentsArrays(namearr, valarr, objarr, count);
					}
				}
				nd->contentsTree(namearr, valarr, objarr,
					ObjectType(VALS | PROCS), ObjectType(PERMANENT | PERMANENTOBJ), TABLES,
					count, lev);
				if throwing break;
			}
		}else
			valarr[y]=wxEmptyString;
		y++;
	}
	SetItemCount(count);
	if (count >= 1)
		RefreshItems(0, count-1);

	th->GuiLeave();

	Pair::maxStringLength=oldMaxSL;

	inupdate=false;
}
//___________________________________________________________________

Vars::Vars(wxWindow* parent, wxWindowID id, 
			  const wxPoint& pos, 
			  const wxSize& size, 
			  long style, 
			  const wxValidator& validator, 
			  const wxString& name):
wxListCtrl(parent, id, pos, size, style, validator, name)
{
	inupdate=false;
	wxString value=_("Value    "), v;
	for (int i=1; i<100; i++)
	{	v.Printf("%-3i...... ", i*10);
		value+=v;
	}

	InsertColumn(0, _("Name"), wxLIST_FORMAT_RIGHT, 100);
	InsertColumn(1,  value, wxLIST_FORMAT_LEFT , 500);

	theApp->vars=this;
	SetFont(wxFont(8,
           wxMODERN,
           wxFONTSTYLE_NORMAL,
           wxNORMAL,
           false,
           "Courier New"));
	show();
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuVars->Check(true);
	return;
}

Vars::~Vars()
{
	theApp->vars=0;
	if (theApp->mainFrame)
		theApp->mainFrame->viewMenuVars->Check(false);
}

void Vars::show(bool shown)
{
	update();
	wxAuiPaneInfo &pane=theApp->aui->GetPane(this);
	if (pane.IsOk())
	{	pane.Show();
		
	}else
	{	wxString pi;
		if (theApp->config->Read("VarsPaneInfo", &pi))
			theApp->aui->LoadPaneInfo(pi, pane);
		else
			pane=wxAuiPaneInfo()
				.Name("Vars").Caption(_("Vars"))
				.Bottom()
				.MinimizeButton().PinButton()
				.DestroyOnClose();
		theApp->aui->AddPane(this, pane);
	}
	theApp->aui->Update();
}

wxString Vars::OnGetItemText(long item, long column) const
{
	if (inupdate)
		return wxEmptyString;
	if (!column)
		return namearr[item];
	return valarr[item];
}

void Vars::OnClick(wxListEvent &ev)
{
	NodeP nd=NodeP(objarr[ev.GetIndex()]);
	if (!nd)
		return;
	nd->setExpanded(!nd->expanded());
/*
	if (nd->type==OBJECT)
	{	ObjectP obj=ObjectP(nd);
		obj->bit.table_expand=!obj->bit.table_expand;
		NodeP var=obj->var();
		if (var && var->type==LOGOOBJECT)
		{	LogoObjectP(var)->bit.table_expand=
				obj->bit.table_expand;
			LogoObjectP(var)->table()->bit.table_expand=
				obj->bit.table_expand;
		}
		else if (var && var->is.list)
		{	PairP list=PairP(var);
			bool expand=obj->bit.table_expand;
			list->bit.table_expand=expand;
		}
		else if (var && var->type==ARRAY)
			ArrayP(var)->bit.table_expand=
				obj->bit.table_expand;
		else if (var && var->type==FLOATARRAY)
			FloatArrayP(var)->bit.table_expand=
				obj->bit.table_expand;
		else if (var && var->type==TABLE)
			TableP(var)->bit.table_expand=
				obj->bit.table_expand;
		else if (var && var->type==OBJECT)
			ObjectP(var)->name()->bit.table_expand=
				obj->bit.table_expand;
		else if (var && var->is.string)
			StringP(var)->bit.table_expand=
				obj->bit.table_expand;
		if (obj->table()!=Unbound)
		{	obj->table()->bit.table_expand=obj->bit.table_expand;
		}
	}else if (nd->type==LOGOOBJECT)
	{	LogoObjectP lobj=LogoObjectP(nd);
		lobj->bit.table_expand=!lobj->bit.table_expand;
		lobj->table()->bit.table_expand=
			lobj->bit.table_expand;
	}
	else if (nd->type==TABLE)
	{	TableP table=TableP(nd);
		table->bit.table_expand=!table->bit.table_expand;
	}
	else if (nd->type==ARRAY)
	{	ArrayP arr=ArrayP(nd);
		arr->bit.table_expand=!arr->bit.table_expand;
	}
	else if (nd->type==FLOATARRAY)
	{	FloatArrayP arr=FloatArrayP(nd);
		arr->bit.table_expand=!arr->bit.table_expand;
	}
	else if (nd->is.list)
	{	PairP list=PairP(nd);
		bool expand=!list->bit.table_expand;
		list->bit.table_expand=expand;
	}
	else if (nd->type==OBJECT)
	{	ObjectP obj=ObjectP(nd);
		obj->name()->bit.table_expand=!obj->name()->bit.table_expand;
	}
	else if (nd->is.string)
	{	StringP str=StringP(nd);
		str->bit.table_expand=!str->bit.table_expand;
	}
//	else
//		c_out << "updateVars error!" << endL;
*/
	update();
}

void Vars::branch(PairP tree, int &y, int level)
{
	while (tree && tree->is.list)
	{
		NodeP type=tree->first();
		if (type==type_Object)
		{	next(tree);
			StringP name=StringP(tree->first());
			next(tree);
			ObjectP obj=ObjectP(tree->first());
			objarr[y]=obj;
			next(tree);
			PairP var=tree->first();
			next(tree);
			ObjectP type=ObjectP(tree->first()->first());
			if ((var  &&  var->is.list  
			 && (ObjectP(var->first())==type_Table 
			  || ObjectP(var->first())==type_Array
			  || ObjectP(var->first())==type_FloatArray
			  || ObjectP(var->first())==type_Pair
			  || ObjectP(var->first())==type_String
			  || ObjectP(var->first())==type_LogoObject))
			|| (tree && tree->is.list
			 && (type==type_Table
			  || type==type_Array
			  || type==type_FloatArray
			  || type==type_Pair
			  || type==type_String
			  || type==type_LogoObject)))
				namearr[y]=wxString(' ',level-1)+wxString('-')+name->ptr();
			else if ((obj->var() 
			 && (obj->var()->type==TABLE
			  || obj->var()->type==ARRAY
			  || obj->var()->type==FLOATARRAY
			  || obj->var()->is.list
			  || obj->var()->type==OBJECT
			  || obj->var()->type==LOGOOBJECT))
			|| obj->table()!=Unbound)
				namearr[y]=wxString(' ',level-1)+wxString('+')+name->ptr();
			else
				namearr[y]=wxString(' ',level-1)+name->ptr();

			if (var && var->is.list 
			&& (ObjectP(var->first())==type_Table
			 || ObjectP(var->first())==type_Array
			 || ObjectP(var->first())==type_FloatArray
			 || ObjectP(var->first())==type_Pair
			 || ObjectP(var->first())==type_String))
			{	valarr[y]=var->first()->string()->ptr();
				y++;
				branch(var, y, level+1);
			}else if (var && var->is.list 
			&& ObjectP(var->first())==type_LogoObject)
			{	valarr[y]=var->tail()->first()->string()->ptr();
				y++;
				branch(var->fourth(), y, level+1);
			}else if (var && var->is.list 
			&& ObjectP(var->first())==type_Object)
			{	valarr[y]=var->second()->string()->ptr();
				y++;
				branch(var->second(), y, level+1);
			}else
			{	if (var)
				{	StringP varstr=var->string();
					if (varstr  &&  varstr->is.string)
					{	StringP shortvar=new String(
							varstr->ptr(), 0, varstr->length(), STRING, 
							String::copy_noformat);
						if throwing break;
						if (shortvar->ptr())
						{	valarr[y]=shortvar->ptr();
							y++;
						}
					}
				}else
				{	valarr[y]="[]";
					y++;
				}
			}
			if (tree)
				branch(tree->first(), y, level+1);
		}else if (type==type_LogoObject)
		{	next(tree);
			StringP name=StringP(tree->first());
			next(tree);
			LogoObjectP obj=LogoObjectP(tree->first());
			objarr[y]=obj;
			next(tree);
			PairP tab=tree->first();
			if (tab  &&  tab->is.list  &&  ObjectP(tab->first())==type_Table)
				namearr[y]=wxString(' ',level-1)+wxString('-')+name->ptr();
			else if (obj->table()!=Unbound)
				namearr[y]=wxString(' ',level-1)+wxString('+')+name->ptr();
			else
				namearr[y]=wxString(' ',level-1)+name->ptr();

			if (tab && tab->is.list && ObjectP(tab->first())==type_Table)
			{	valarr[y]=tab->first()->string()->ptr();
				y++;
				branch(tab, y, level-1);
			}else
			{	if (tab)
				{	StringP varstr=tab->string();
					if (varstr  &&  varstr->is.string)
					{	StringP shortvar=new String(
							varstr->ptr(), 0, varstr->length(), STRING, 
							String::copy_noformat);
						if throwing break;
						if (shortvar->ptr())
						{	valarr[y]=shortvar->ptr();
							y++;
						}
					}
				}else
				{	valarr[y]="[]";
					y++;
				}
			}
		}else if (type==type_Table)
		{	next(tree);
			ObjectP obj=ObjectP(tree->first());
			if (obj && obj->type==OBJECT)
				objarr[y]=ObjectP(obj)->name();
			else
				objarr[y]=obj;
			next(tree);
			if (tree && tree->is.list)
			{	while (tree)
				{	branch(tree->first(), y, level);
					next(tree);
				}
			}else
			{	if (tree)
				{	StringP varstr=tree->string();
					if (varstr  &&  varstr->is.string)
					{	StringP shortvar=new String(
							varstr->ptr(), 0, varstr->length(), STRING, 
							String::copy_noformat);
						if throwing break;
						if (shortvar->ptr())
						{	valarr[y]=shortvar->ptr();
							y++;
						}
					}
				}else
				{	valarr[y]="[]";
					y++;
				}
			}
			break;
		}else if (type==type_Array)
		{	next(tree);
			ArrayP obj=ArrayP(tree->first());
			objarr[y]=obj;
			next(tree);
			if (tree && tree->is.list)
			{	int i=obj->origin();
				while (tree)
				{	namearr[y]=wxString(' ',level-1);
					namearr[y] << i;
					NodeP nd=tree->first();
					if (nd && nd->type==OBJECT)
						objarr[y]=ObjectP(nd)->name();
					else
						objarr[y]=nd;
					next(tree);

					ObjectP type=ObjectP(tree->first()->first());
					PairP var=tree->first();
					if ((var  &&  var->is.list  
					 && (ObjectP(var->first())==type_Table 
					  || ObjectP(var->first())==type_Array
					  || ObjectP(var->first())==type_FloatArray
					  || ObjectP(var->first())==type_Pair
					  || ObjectP(var->first())==type_String
					  || ObjectP(var->first())==type_LogoObject))
					|| (tree && tree->is.list
					 && (type==type_Table
					  || type==type_Array
					  || type==type_FloatArray
					  || type==type_Pair
					  || type==type_String
					  || type==type_LogoObject)))
					{	namearr[y]=wxString(' ',level-1)+wxString('-');
						namearr[y] << i;
					}else if (nd 
					 && (nd->type==TABLE
					  || nd->type==ARRAY
					  || nd->type==FLOATARRAY
					  || nd->is.list
					  || nd->type==OBJECT
					  || nd->type==LOGOOBJECT))
					{	namearr[y]=wxString(' ',level-1)+wxString('+');
						namearr[y] << i;
					}else
					{	namearr[y]=wxString(' ',level-1);
						namearr[y] << i;
					}

					if (var && var->is.list 
					&& (ObjectP(var->first())==type_Table
					 || ObjectP(var->first())==type_Array
					 || ObjectP(var->first())==type_FloatArray
					 || ObjectP(var->first())==type_Pair
					 || ObjectP(var->first())==type_String))
					{	valarr[y]=var->first()->string()->ptr();
						y++;
						branch(var, y, level+1);
					}else if (var && var->is.list 
					&& ObjectP(var->first())==type_LogoObject)
					{	valarr[y]=var->tail()->first()->string()->ptr();
						y++;
						branch(var->fourth(), y, level+1);
					}else if (var && var->is.list 
					&& ObjectP(var->first())==type_Object)
					{	valarr[y]=var->second()->string()->ptr();
						y++;
						branch(var->second(), y, level+1);
					}else
					{	if (var)
						{	StringP varstr=var->string();
							if (varstr  &&  varstr->is.string)
							{	StringP shortvar=new String(
									varstr->ptr(), 0, varstr->length(), STRING, 
									String::copy_noformat);
								if throwing break;
								if (shortvar->ptr())
								{	valarr[y]=shortvar->ptr();
									y++;
								}
							}
						}else
						{	valarr[y]="[]";
							y++;
						}
					}
					i++;
					next(tree);
				}
			}
			break;
		}else if (type==type_FloatArray)
		{	next(tree);
			FloatArrayP obj=FloatArrayP(tree->first());
			objarr[y]=obj;
			next(tree);
			if (tree && tree->is.list)
			{	int i=obj->origin();
				while (tree)
				{	namearr[y]=wxString(' ',level-1);
					namearr[y] << i;
					FloatP f=FloatP(tree->first());
					StringP varstr=f->string();
					if (varstr  &&  varstr->is.string)
					{	StringP shortvar=new String(
							varstr->ptr(), 0, varstr->length(), STRING, 
							String::copy_noformat);
						if throwing break;
						if (shortvar->ptr())
						{	valarr[y]=shortvar->ptr();
							y++;
						}
					}
					i++;
					next(tree);
				}
			}else
			{	if (tree)
				{	StringP varstr=tree->string();
					if (varstr  &&  varstr->is.string)
					{	StringP shortvar=new String(
							varstr->ptr(), 0, varstr->length(), STRING, 
							String::copy_noformat);
						if throwing break;
						if (shortvar->ptr())
						{	valarr[y]=shortvar->ptr();
							y++;
						}
					}
				}else
				{	valarr[y]="[]";
					y++;
				}
			}
			break;
		}else if (type==type_Pair)
		{	next(tree);
			PairP obj=tree->first();
			objarr[y]=obj;
			next(tree);
			if (tree && tree->is.list)
			{	int i=1;
				PairP list=tree->first();
				while (list)
				{	namearr[y]=wxString(' ',level-1);
					namearr[y] << i;
					NodeP nd=list->first()->first();
					if (nd && nd->type==OBJECT)
						objarr[y]=ObjectP(nd)->name();
					else
						objarr[y]=nd;
					next(list);

					ObjectP type=ObjectP(list->first()->first());
					PairP var=list->first();
					if ((var  &&  var->is.list  
					 && (ObjectP(var->first())==type_Table 
					  || ObjectP(var->first())==type_Array
					  || ObjectP(var->first())==type_FloatArray
					  || ObjectP(var->first())==type_Pair
					  || ObjectP(var->first())==type_String
					  || ObjectP(var->first())==type_LogoObject))
					|| (tree && tree->is.list
					 && (type==type_Table
					  || type==type_Array
					  || type==type_Pair
					  || type==type_String
					  || type==type_LogoObject)))
					{	namearr[y]=wxString(' ',level-1)+wxString('-');
						namearr[y] << i;
					}else if (nd 
					 && (nd->type==TABLE
					  || nd->type==ARRAY
					  || nd->type==FLOATARRAY
					  || nd->is.list
					  || nd->type==OBJECT
					  || nd->type==LOGOOBJECT))
					{	namearr[y]=wxString(' ',level-1)+wxString('+');
						namearr[y] << i;
					}else
					{	namearr[y]=wxString(' ',level-1);
						namearr[y] << i;
					}

					if (var && var->is.list 
					&& (ObjectP(var->first())==type_Table
					 || ObjectP(var->first())==type_Array
					 || ObjectP(var->first())==type_FloatArray
					 || ObjectP(var->first())==type_Pair
					 || ObjectP(var->first())==type_String))
					{	valarr[y]=var->first()->string()->ptr();
						y++;
						branch(var, y, level+1);
					}else if (var && var->is.list 
					&& ObjectP(var->first())==type_LogoObject)
					{	valarr[y]=var->tail()->first()->string()->ptr();
						y++;
						branch(var->fourth(), y, level+1);
					}else if (var && var->is.list 
					&& ObjectP(var->first())==type_Object)
					{	valarr[y]=var->second()->string()->ptr();
						y++;
						branch(var->second(), y, level+1);
					}else
					{	if (var)
						{	StringP varstr=var->string();
							if (varstr  &&  varstr->is.string)
							{	StringP shortvar=new String(
									varstr->ptr(), 0, varstr->length(), STRING, 
									String::copy_noformat);
								if throwing break;
								if (shortvar->ptr())
								{	valarr[y]=shortvar->ptr();
									y++;
								}
							}
						}else
						{	valarr[y]="[]";
							y++;
						}
					}
					i++;
					next(list);
				}
			}else
			{	if (tree)
				{	StringP varstr=tree->string();
					if (varstr  &&  varstr->is.string)
					{	StringP shortvar=new String(
							varstr->ptr(), 0, varstr->length(), STRING, 
							String::copy_noformat);
						if throwing break;
						if (shortvar->ptr())
						{	valarr[y]=shortvar->ptr();
							y++;
						}
					}
				}else
				{	valarr[y]="[]";
					y++;
				}
			}
			break;
		}else if (type==type_String)
		{	next(tree);
			PairP obj=tree->first();
			objarr[y]=obj;
			next(tree);
			PairP list=tree->first();
			if (list  &&  list->is.list)
				namearr[y]=wxString(' ',level-1)+wxString('-');
			else if (obj->table()!=Unbound)
				namearr[y]=wxString(' ',level-1)+wxString('+');
			else
				namearr[y]=wxString(' ',level-1);

			if (tree && tree->first()->is.list)
			{	int i=1;
				while (list)
				{	namearr[y]=wxString(' ',level-1);
					namearr[y] << i;
					branch(list, y, level+1);
					i++;
					next(list);
				}
			}else
			{	if (tree)
				{	StringP varstr=tree->string();
					if (varstr  &&  varstr->is.string)
					{	valarr[y]=varstr->ptr();
						y++;
					}
				}else
				{	valarr[y]="[]";
					y++;
				}
			}
			break;
		}else
		{	NodeP nd=tree->first();
			objarr[y]=nd;
			if (nd)
			{	StringP varstr=nd->string();
				if (varstr  &&  varstr->is.string)
				{	valarr[y]=varstr->ptr();
					y++;
				}
			}else
			{	valarr[y]="[]";
				y++;
			}
			break;
		}
		next(tree);
	}
}

void Vars::update()
{
	if (inupdate)
		return;

	inupdate=true;
	ThreadP th=Thread::This();

//	charP reserve=new char[100000];
//	if (!reserve)
//		return;
//	delete reserve;

	int PairMaxSL=Pair::maxStringLength,
		TableMaxSL=Table::maxStringLength,
		LogoObjectMaxSL=LogoObject::maxStringLength;

	int maxSL=256;
	Pair::maxStringLength=maxSL;
	Table::maxStringLength=maxSL;
	LogoObject::maxStringLength=maxSL;
	Array::maxStringLength=maxSL;

	count=1;
	namearr.Clear(); namearr.Alloc(count); namearr.Add(wxEmptyString, count);
	 valarr.Clear();  valarr.Alloc(count);  valarr.Add(wxEmptyString, count);
	 objarr.Clear();  objarr.Alloc(count);  objarr.Add(0, count);

	count=0;
	NodeP thisnames=0;
	if (th->thisobj)
	{	namearr[count]=_("This");
		objarr[count]=th->thisobj;
		if (th->thisobj->expanded())
		{	valarr[count]="";
			_True->incContentsArrays(namearr, valarr, objarr, count);
		}
		th->thisobj->contentsTree(namearr, valarr, objarr,
			ObjectType(VALS | PROCS), ObjectType(PERMANENT | PERMANENTOBJ), TABLES,
			count, 2);
	}
	if (root)
	{	root->bit.table_expand=true;
		root->contentsTree(namearr, valarr, objarr,
			ObjectType(VALS | PROCS), ObjectType(PERMANENT | PERMANENTOBJ), TABLES,
			count, 1);
	}
	th->GuiEnter();

	SetItemCount(count);
	if (count >= 1)
		RefreshItems(0, count-1);

	th->GuiLeave();

	Pair::maxStringLength=PairMaxSL;
	Table::maxStringLength=TableMaxSL;
	LogoObject::maxStringLength=LogoObjectMaxSL;

	inupdate=false;
}

NodeP lupdateVars(PairPP)
{
	if (theApp->stackvars) 
		theApp->stackvars->update();
	if (theApp->vars) 
		theApp->vars->update();
	return Unbound;
}

NodeP lupdateVarsOnStep(PairPP arg)
{
	int update=arg[1]->a_bool(1);
	if (!throwing)
		updateVarsOnStep=(update==1);
	return Unbound;
}
