#include "stdafx.h"
#ifdef WIN32
	#include <crtdbg.h>
#endif

#include <wx/string.h>
#include <wx/utils.h>
#include <wx/textctrl.h>
#include <wx/html/helpctrl.h>

#include <stdio.h>
#include <malloc.h>

#ifdef WIN32
	#include <process.h>
#else
	#include <unistd.h>
#endif

#include "Main.h"

#include "String.h"
#include "File.h"
#include "Object.h"
#include "LogoObject.h"
#include "Pair.h"
#include "Procedure.h"
#include "Primitive.h"
#include "Table.h"
#include "Line.h"
#include "List.h"
#include "Quote.h"
#include "Colon.h"
#include "Tree.h"
#include "Int.h"
#include "UInt8.h"
#include "Array.h"

#include "Error.h"
#include "NodePStack.h"
#include "Stack.h"
#include "Init.h"

#include "Color.h"
#include "Security.h"
#include "Thread.h"

#ifdef _DEBUG
	#undef THIS_FILE
	static char THIS_FILE[]= __FILE__;
	#define new new(_NORMAL_BLOCK, THIS_FILE, __LINE__)
	#define  malloc(m)       _malloc_dbg((m),        _NORMAL_BLOCK, __FILE__, __LINE__)
#endif

NodeP lPrimitives   (PairPP) 
{
	return root->contents(ObjectType(PROCS | PERMANENT),  ObjectType(VALS | PERMANENTOBJ));
}

NodeP lLogoConstants(PairPP) 
{
	return root->contents(ObjectType(VALS | PERMANENT), ObjectType(PROCS | PERMANENTOBJ));
}

NodeP lProcedures   (PairPP) 
{
	ThreadP th=Thread::This();
	return th->namespce->contents(PROCS, ObjectType(PROC_BURIED | PERMANENT));
}
NodeP lNames(PairPP) 
{
	ThreadP th=Thread::This();
	PairP names=0, thisnames=0, stacknames=0, staticnames=0;

	if (th->thisobj)
		thisnames=th->thisobj->contents(VALS, ObjectType(BURIED | PERMANENT));
	stacknames=th->varStack->contents(VALS, ObjectType(BURIED | PERMANENT));
	staticnames=th->namespce->contents(VALS, ObjectType(BURIED | PERMANENT | PERMANENTOBJ));

	names=thisnames;
	if (names)
		names=PairP(names->sentence(stacknames));
	else
		names=stacknames;
	if (names)
		names=PairP(names->sentence(staticnames));
	else
		names=staticnames;
	if (names)
		names=PairP(names->combSort());
	return 
		new Pair(0, 
		new Pair(names,0));
}
NodeP lTables(PairPP) 
{
	ThreadP th=Thread::This();
	return new List(
		0, 
		0, 
		th->namespce->contents(TABLES, ObjectType(TABLE_BURIED | PERMANENT)), 
		_ListEnd);
}

NodeP lContents (PairPP) 
{
	ThreadP th=Thread::This();
	PairP names=0, thisnames=0, stacknames=0, staticnames=0;

	if (th->thisobj)
		thisnames=th->thisobj->contents(VALS, ObjectType(BURIED | PERMANENT));
	stacknames=th->varStack->contents(VALS, ObjectType(BURIED | PERMANENT | PERMANENTOBJ));
	staticnames=th->namespce->contents(VALS, ObjectType(BURIED | PERMANENT | PERMANENTOBJ));

	names=thisnames;
	if (names)
		names=PairP(names->sentence(stacknames));
	else
		names=stacknames;
	if (names)
		names=PairP(names->sentence(staticnames));
	else
		names=staticnames;
	if (names)
		names=PairP(names->combSort());

	PairP ret=0;
	push(ret, th->namespce->contents(TABLES, ObjectType(BURIED | PERMANENT)));
	push(ret, names);
	push(ret, th->namespce->contents(PROCS, ObjectType(BURIED | PERMANENT)));
	return ret;
}

NodeP lburied (PairPP) 
{
	ThreadP th=Thread::This();
	if (!th->security->buried)
	{	error(SECURITY_ERROR, new String(_("to readout buried")));
		return Unbound;
	}
	PairP names=0, thisnames=0, stacknames=0, staticnames=0;

	if (th->thisobj)
		thisnames=th->thisobj->contents(ObjectType(VAL_BURIED), PERMANENT);
	stacknames=th->varStack->contents(ObjectType(VAL_BURIED), ObjectType(PERMANENT | PERMANENTOBJ));
	staticnames=th->namespce->contents(ObjectType(VAL_BURIED), ObjectType(PERMANENT | PERMANENTOBJ));

	names=thisnames;
	if (names)
		names=PairP(names->sentence(stacknames));
	else
		names=stacknames;
	if (names)
		names=PairP(names->sentence(staticnames));
	else
		names=staticnames;
	if (names)
		names=PairP(names->combSort());

	PairP ret=0;
	push(ret, th->namespce->contents(ObjectType(TABLE_BURIED), PERMANENT));
	push(ret, names);
	push(ret, th->namespce->contents(ObjectType(PROC_BURIED), PERMANENT));
	return ret;
}

NodeP ltraced (PairPP) 
{
	ThreadP th=Thread::This();
	PairP names=0, thisnames=0, stacknames=0, staticnames=0;

	if (th->thisobj)
		thisnames=th->thisobj->contents(ObjectType(VAL_TRACED), PERMANENT);
	stacknames=th->varStack->contents(ObjectType(VAL_TRACED), PERMANENT);
	staticnames=th->namespce->contents(ObjectType(  VAL_TRACED), PERMANENT);

	names=thisnames;
	if (names)
		names=PairP(names->sentence(stacknames));
	else
		names=stacknames;
	if (names)
		names=PairP(names->sentence(staticnames));
	else
		names=staticnames;
	if (names)
		names=PairP(names->combSort());

	PairP ret=0;
	push(ret, th->namespce->contents(ObjectType(TABLE_TRACED), PERMANENT));
	push(ret, names);
	push(ret, th->namespce->contents(ObjectType( PROC_TRACED), PERMANENT));
	return ret;
}

NodeP lstepped(PairPP) 
{
	ThreadP th=Thread::This();
	PairP names=0, thisnames=0, stacknames=0, staticnames=0;

	if (th->thisobj)
		thisnames=th->thisobj->contents(ObjectType(  VAL_STEPPED), PERMANENT);
	stacknames=th->varStack->contents(ObjectType(  VAL_STEPPED), PERMANENT);
	staticnames=th->namespce->contents(ObjectType(  VAL_STEPPED), PERMANENT);

	names=thisnames;
	if (names)
		names=PairP(names->sentence(stacknames));
	else
		names=stacknames;
	if (names)
		names=PairP(names->sentence(staticnames));
	else
		names=staticnames;
	if (names)
		names=PairP(names->combSort());

	PairP ret=0;
	push(ret, th->namespce->contents(ObjectType(TABLE_STEPPED), PERMANENT));
	push(ret, names);
	push(ret, th->namespce->contents(ObjectType( PROC_STEPPED), PERMANENT));
	return ret;
}

NodeP lbury (PairPP arg) {return arg[1]->  bury( BURIED);}
NodeP ltrace(PairPP arg) 
{	
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI")));
		return Unbound;
	}
	return arg[1]->  bury( TRACED);
}

NodeP lstep (PairPP arg) 
{	
	ThreadP th=Thread::This();
	if (!th->security->gui)
	{	error(SECURITY_ERROR, new String(_("to use GUI")));
		return Unbound;
	}
	return arg[1]->  bury(STEPPED);
}

NodeP lunBury (PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!th->security->buried)
	{	error(SECURITY_ERROR, new String(_("to unbury something")));
		return Unbound;
	}
	return arg[1]->unBury( BURIED);
}
NodeP lunTrace(PairPP arg) {return arg[1]->unBury( TRACED);}
NodeP lunStep (PairPP arg) {return arg[1]->unBury(STEPPED);}


NodeP lisdefined  (PairPP arg) {return arg[1]->procType(0);}
NodeP lisPrimitive(PairPP arg) {return arg[1]->procType(1);}
NodeP lisMacro    (PairPP arg) {return arg[1]->procType(2);}

NodeP lisProcedure(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);
	if (!throwing) 
	{	ObjectP obj=name->intern();
		if (obj != Unbound && obj->bit.proc && !obj->bit.permanent)
			return _True;
		return _False;
	}
	return Unbound;
}

NodeP lisName(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP name = arg[1]->a_name(1);
	if (!throwing) 
	{	ObjectP obj=name->intern();
		NodeP val=th->varStack->find(obj);
		return Node::boolnode( val != Unbound );
	}
	return Unbound;
}

NodeP lArity(PairPP arg)
{
   NodeP name = arg[1]->a_name(1);

   if (!throwing) 
	{	ObjectP obj = name->intern();
		
		if (obj->prim()->is.primitive)
		{	int min = obj->prim()->min;
			if (min == Primitive::ok_if_no_arg) 
				min = 0;
			return new List(
				new Int(min),
			   new Int(obj->prim()->std),
			   new Int(obj->prim()->max), 
				new Int(obj->prim()->_priority), _ListEnd);

		}
		else if (obj == Unbound) 
		{	error(DK_HOW_UNREC, name);
			return Unbound;
		}else 
		{	return new List(
				new Int(obj->proc()->min),
				new Int(obj->proc()->std),
				new Int(obj->proc()->max),
				new Int(priority_prefix), _ListEnd);
		}
	}
	return Unbound;
}

NodeP lThing(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP	name = arg[1]->a_name(1),
		 val=Unbound;

	if (!throwing) 
		val = th->varStack->find(name->intern());

	if (val==Unbound)
		error(NO_VALUE, arg[1]);
	return val;
}


NodeP lmake(PairPP arg) 
{
	ThreadP th=Thread::This();
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	ObjectP what=0;
		if (name->type==OBJECT)	// slow if many locals exist!
		{	what=ObjectP(name);
			th->varStack->setVar(what, arg[2]);
		}else
		{	what = name->intern();
			th->varStack->setVar(what, arg[2]);
		}
		if (what->bit.val_traced) 
		{
			NodeP tvar=0;	tvar = arg[2]->maybe_quote();
			writer << _("Make ") << QuoteP(new Quote(what)) << ' ' << tvar;

			ThreadP th=Thread::This();
			if (th->ufun) 
				writer << _(" in ") << th->ufun << endL << th->this_line;
			writer << endL;
		}
	}
	return Unbound;
}

NodeP llocal(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (th->tailcall==tailcall_tail) 
		return Unbound;

	PairP li = arg[1];
	while (li->islist()  &&  exist(2)  &&  !throwing)
		li = error(BAD_DATA, li);

	int vsp = *th->varStack;

	if (li->islist())		// local [a b]
		while (li  &&  !throwing) 
		{
			while (! li->first()->is.word  &&  !throwing)
				li->setfirst( error(BAD_DATA, li->first()) ); // prevent crash in _apply
			
			if (!throwing) 
			{	ObjectP a = li->first()->intern();
				li->setfirst(a);	// local [a b] faster next time
				if (a->not_local(vsp)) 
					PushVar(th->varStack, a, Unbound, local);
				else
					th->varStack->setVar(a, Unbound);
				a->tell_shadow();
			}
			if (checkThrowing())
				break;
			next(li);
		}
	else		// (local "a "b)
	{	int nr=1;
		while (arg!=PairPP(th->argStack->index)  &&  !throwing) 
		{
			if (arg!=PairPP(th->argStack->index)  
			&&  *++arg  &&  !(*arg)->is.word)
				error(BAD_DATA, *arg, nr);

			ObjectP a = (*arg)->intern();
			if (!throwing) 
			{
				if (a->not_local(vsp)) 
					PushVar(th->varStack, a, Unbound, local);
				else
					th->varStack->setVar(a, Unbound);
				a->tell_shadow();
			}
			if (checkThrowing()) break;
			nr++;
		}
	}
	th->var = *th->varStack;    // so eval won't undo our work
	return Unbound;
}


NodeP Pair::to_helper(bool macro_flag) 
{
	ThreadP th=Thread::This();
	if (th->lexufun  &&  loader.file()==File::conin)
	{	error(NOT_INSIDE, 0);
		return Unbound;
	}
	if (!this) 									// verifiy this
	{	error(NOT_ENOUGH, 0);
		return Unbound;
	}
	ObjectP obj = ObjectP(first());				// verify object
	if (type_of(obj) != OBJECT)
		error(BAD_DATA_UNREC, obj);
	else
		if ((obj->proc_or_prim() != Unbound  &&  loader.file()==File::conin)
		 ||  obj->proc_or_prim()->is.primitive )
			error(ALREADY_DEFINED, obj);
	
	th->deependProcName=obj->name();

	PairP args=tail(),  formals=0;
	int min=0, std=0, max=0, old_std=-12345;
		
	if (!throwing)
	{	NodeP old_proc = obj->proc_or_prim();		// remember std of old proc
		if (old_proc != Unbound) 
			old_std =  old_proc->is.primitive ? 
				PrimitiveP(old_proc)->std  :  ProcedureP(old_proc)->std;

		PairP lastnode=0, tnode=0;

		while (args)										// scan formals 
		{	NodeP arg=args->first();
			next(args);
	
			if (arg->islist()  &&  max !=-1)	// scan default arguments
			{	PairP &pair= (PairP&) arg;
				pair->make_runparse();
				if throwing return Unbound;
				RunparsedP &run = (RunparsedP&) pair;
				arg = run->parsed();
				max++;

				if (type_of(pair->first()) != COLON) 
					pair->setfirst(ObjectP(pair->first()));
				else
					pair->setfirst( ColonP(pair->first())->colon() );

				if (!pair->tail())		// formal is rest
					max=-1;
			}
			else if (type_of(arg)==COLON  &&  min==max)	// usual formals
			{	arg = ColonP(LineP(arg))->colon();
				min++, std++, max++;
			}
			else if (type_of(arg)==INT_)				// set std
			{	Int &irg = *IntP(arg);
			
				if ((unsigned)irg.int_ <= (unsigned) max
				&&            irg.int_ >=            min) 
					std = irg.int_;
			}else if (th->logoLanguage==_ucblogo)
			{	error(DK_HOW, obj); return Unbound;
			}else
				min++, std++, max++;	// colonless formal

			tnode = new Pair(arg, 0);
			if throwing return Unbound;

			if (formals) lastnode->settail( tnode );
			else formals=tnode;
			lastnode=tnode;
		}
	}
	if (!throwing) 
	{	PairP bodylist=0, bodywords=0;
		th->toPending++;    // for int or quit signal

		if (!throwing  &&  th->toPending  &&  !loader.eof())	// read procedure`s body
		{	//Line::nr=1;
			th->row=th->readnr;
			StringP body = loader.read("");
			if throwing return Unbound;
//			c_out << "to_helper " << body << endL;
			bodylist = new Line( body->parser(true), 0);
			if throwing return Unbound;
//			c_out << "to_helper " << bodylist << endL;
			if (NodeP(body)==_Null_Word)
				bodywords = 
					new Pair(th->currentLine, 
					new Pair(_End, 0));
			else
				bodywords = 
					new Pair(th->currentLine, 
					new Pair(body, 
					new Pair(_End, 0)));
		}
		if (th->toPending  &&  !throwing) 
		{	obj->setproc( ProcedureP( new Procedure(
				formals, bodylist, bodywords, obj, 
				priority_prefix, min, std, max, PROC_) ) );

			obj->bit.proc_macro = macro_flag;

			if (std != old_std  &&  old_std >= 0) 
				Line::the_generation++;
			
			if (loader.file()==File::conin 
			||  th->loadNoisily)
				c_out << _(";") << obj << _(" defined") << endL;
		}
		th->toPending=0;
	}
	th->deependProcName=0;
	return Unbound;
}

NodeP lto      (PairPP arg) {return error(BAD_DATA);}//arg[1]->to_helper(false, false);}
NodeP lbe      (PairPP arg) {return error(BAD_DATA);}//arg[1]->to_helper(false, true );}
NodeP l_Macro  (PairPP arg) {return error(BAD_DATA);}//arg[1]->to_helper(true , false);}
NodeP l_beMacro(PairPP arg) {return error(BAD_DATA);}//arg[1]->to_helper(true , true );}
NodeP lend     (PairPP arg) {return error(BAD_DATA);}

ProcedureP Line::define_helper(bool macroFlag, bool dynFlag, PairP name, StringP left) 
{
	ThreadP th=Thread::This();
	PairP val=0;
	ObjectP obj=0;
	int old_std=-1234;
	
	if (name) 								// verify name
	{	name = PairP(name->a_name(1));
		if (!throwing) 
		{	obj = name->intern();
			NodeP p = obj->proc_or_prim();

			if (!th->security->buried)
			{	ObjectP o=obj;
				if (p->type==PROC_)
					while (o)
					{	if (o->flag & BURIED)
						{	error(SECURITY_ERROR, new String(_("to redefine buried procedures")));
							return ProcedureP(Unbound);
						}
						o=o->parent();
					}
			}
			bool redef = th->redefinable;

			if (p->is.primitive && !redef)		// grumble if primitive and not redefinable
			{	error(IS_PRIM, obj);
				return ProcedureP(Unbound);
			}else if (p->is.primitive && !th->security->erase)
			{	error(SECURITY_ERROR, new String(_("to redefine primitives")));
				return ProcedureP(Unbound);
			}else									// remember old std arg nr
				if (p != Unbound) 
					old_std=  p->is.primitive ?
						PrimitiveP(p)->std  :  ProcedureP(p)->std;
		}
		if (!throwing)		// verify this (should be a listlist)
		{	val=this;
			while ( (val->isnolist()  ||  !val->all_lists() )  &&  !throwing) 
				val = error(BAD_DATA, val);
		}
	}else			// lambda (proc with no name)
	{	val=this;
		obj=this->intern();
	}
	if throwing return ProcedureP(Unbound);

	if (val)
		val=PairP(val->deepCopyLine());

	int min=0, std=0, max=0;
	PairP formals=0;
	if (val)
		formals = val->first();
	
	if (!throwing) 
	{	ObjectP olexufun=th->lexufun;
		if (dynFlag)
			th->lexufun=0;
		else
			th->lexufun=obj;
		if (formals) 
		{	formals->make_runparse();
			formals = RunparsedP(formals)->parsed();
			val->setfirst( formals );
		}
		PairP formal=formals;
		PairP lastnode=0, tnode=0;
		formals=0;

		while (formal) 
		{	NodeP arg = formal->first();
			next(formal);

			if (arg->islist()  &&  max !=-1) 
			{
				LineP(arg)->make_runparse();
				arg = RunparsedP(LineP(arg))->parsed();
				max++;

				LineP list = LineP(arg);
				switch (type_of(list->first()))
				{
				case COLON: 
					list->setfirst( ColonP(list->first())->colon() );
				default:
					list->setfirst(ObjectP(list->first()));
				}
				if (! list->tail())		// formal is rest
					max=-1;
			}else
			{	if (type_of(arg)==COLON  &&  min==max)	// usual formals
				{	arg = ColonP(LineP(arg))->colon();
					min++, std++, max++;
				}
				else if (type_of(arg)==INT_)
				{	IntP &i = (IntP&) arg;
					if (unsigned(i->int_) <= unsigned(max) 
					&&           i->int_  >=          min ) 
					std = i->int_;
				}else 
					if (min==max) 
						min++, max++, std++;
					else 
					{	error(BAD_DATA_UNREC, arg);
						break;
					}
			}
			tnode = new Pair(arg, 0);

			if (formals) lastnode->settail( tnode );
			else formals=tnode;
			lastnode=tnode;
			if (checkThrowing()) break;
		}
		th->lexufun=olexufun;
	}
	if (!throwing)
	{	PairP body=0;
		if (val)
			body = val->tail();

//c_out << formals << "DFormals" << endL;
//c_out << body << "DBody" << endL;
		Priority pri=priority_prefix;
		NodeType typ=PROC_;
		if (left)
		{	pri=priority_product;
			min=std=max=1;
			typ=INFIXPROC;
		}
		ProcedureP proc=new Procedure(
			formals, body, NIL, obj, pri, min, std, max, typ);
		if throwing return ProcedureP(Unbound);
		if (!name)
			return proc;
		else 
		{	obj->setproc(proc);
			obj->bit.proc_macro=macroFlag;
			obj->bit.proc_dyn=dynFlag;
			if (th->stepArmed)
				obj->bit.proc_stepped=true;
			
			if (std != old_std && old_std >=-1) 
				Line::the_generation++;

			if (th->loadNoisily)
				c_out << _(";") << obj << _(" defined") << endL;
			return proc;
		}
	}
	return ProcedureP(Unbound);
}

NodeP ldefine   (PairPP arg) {LineP(arg[2])->define_helper(false, true, arg[1]); return Unbound;}
NodeP ldefineBe (PairPP arg) {LineP(arg[2])->define_helper(false, false, arg[1]); return Unbound;}
NodeP l_defMacro(PairPP arg) {LineP(arg[2])->define_helper( true, true, arg[1]); return Unbound;}


NodeP lcopyDef(PairPP arg) 
{
	ThreadP th=Thread::This();
	int redef = th->redefinable;
	int new_std=0, old_std=0;
	ObjectP
		arg1 = ObjectP( arg[1]->a_name(1) ),
		arg2 = ObjectP( arg[2]->a_name(2) );

	if (arg2->numberP()) error(BAD_DATA_UNREC, arg2, 2);
	if (arg1->numberP()) error(BAD_DATA_UNREC, arg1);

	if (!throwing) 
	{	arg1= arg1->intern();
		arg2= arg2->intern();
	}
	if (!throwing  &&  arg2->proc_or_prim()==Unbound)
		error(DK_HOW, arg2, 2);

	if (!throwing  &&  arg1->proc_or_prim()->is.primitive)
	{	if (!redef)
			error(IS_PRIM, arg1);
		if (!th->security->erase)
			error(SECURITY_ERROR, new String(_("to overwrite primitives")));
	}
	if (!throwing) 
	{	NodeP
			old_proc=arg1->proc_or_prim(),
			new_proc=arg2->proc_or_prim();

		if (old_proc != Unbound) 
		{	if (old_proc->is.primitive)
				old_std=PrimitiveP(old_proc)->std;
			else if (old_proc->type==PROC_)
				old_std=ProcedureP(old_proc)->std;
		}
		if (new_proc->is.primitive)
			new_std=PrimitiveP(new_proc)->std;
		else if (new_proc->type==PROC_)
			new_std=ProcedureP(new_proc)->std;
		else
			new_std=-99;
		if (old_std != new_std) 
			Line::the_generation++;
		
		if (new_proc->is.primitive)
		{	PrimitiveP p=PrimitiveP(new_proc);
			arg1->setprim(p);
			arg1->setflag( PERMANENT );
			if (p->min==-1 && p->std==-1 && p->max==-1)
			{	th->specialForms=ArrayP(th->specialForms->lPut(arg1));
			}
		}else if (new_proc->type==PROC_)
		{	arg1->setproc(ProcedureP(new_proc));
			arg1->setflag( PROC_BURIED );
		}else
		{	arg1->setvar( new_proc );
			arg1->setflag( VAL_OBJECT|VAL_BURIED|PERMANENT );
		}
		if (arg2->flag & PERMANENT)
			arg1->setflag( PERMANENT );

		if (arg2->bit.proc_macro) 
			arg1->setflag( PROC_MACRO );
		else
			arg1->clearflag( PROC_MACRO );

		if (arg2->bit.proc_dyn) 
			arg1->setflag( PROC_DYN );
		else
			arg1->clearflag( PROC_DYN );

		arg1->setowner(th->user);
	}
	return Unbound;
}


NodeP lText(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	NodeP proc=0;	proc = name->intern()->proc_or_prim();

		if (proc==Unbound) 
		{	error(DK_HOW_UNREC, name); return Unbound;
		}
		else if (proc->is.primitive) 
		{	error(IS_PRIM, name); return Unbound;
		}
		else return new Pair(
			ProcedureP(proc)->formals(),
			ProcedureP(proc)->bodylist());
	}
	return Unbound;
}

NodeP lfullText(PairPP arg) 
{
	NodeP name = arg[1]->a_name(1);

	if (!throwing) 
	{	NodeP proc_or_prim = name->intern()->proc_or_prim();

		if (proc_or_prim==Unbound) 
		{	error(DK_HOW_UNREC, name); return Unbound;
		}
		else if (proc_or_prim->is.primitive) 
		{	error(IS_PRIM, name); return Unbound;
		}
		else return ProcedureP(proc_or_prim)->get_bodywords(name);
	}
	return Unbound;
}



NodeP Pair::printOut(int just_titles)	// >0 for _printOutTitles, <0 for _edit
{
	ThreadP th=Thread::This();
	PairP proclist=0, varlist=0, tablelist=0;
	three_lists(proclist, varlist, tablelist);
	
	th->backslashes=true;
	int width=Int::max;
	
	while (proclist) 
	{	NodeP proc_or_prim;
		if (!proclist->first()  ||  proclist->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, proclist->first());
			break;
		}else
			proc_or_prim = proclist->first()->intern()->proc_or_prim();
		
		if (proc_or_prim==Unbound)
		{	if (just_titles < 0) 
				writer << "to " << proclist->first() << endL << "end" << endL << endL;
			else 
			{	error(DK_HOW_UNREC, proclist->first());
				break;
			}
		}else 
			if (proc_or_prim->is.primitive) 
			{	error(IS_PRIM, proclist->first());
				break;
			}else 
			{	PairP bodywords = ProcedureP(proc_or_prim)->
					get_bodywords(proclist->first());

				if (just_titles > 0) 
				{	if (bodywords->first()->is.list)
						File::print(writer.file(), bodywords->first(), 1);
					else 
					{	charP str = bodywords->first()->string()->expand_slash();
						writer << str;
						GC_::free(str); str=0;
					}
				}else 
					while (bodywords) 
					{	if (bodywords->first()->is.list)
							File::print(writer.file(), bodywords->first(), 1);
						else 
						{	charP str = bodywords->first()->string()->expand_slash();
							writer << str;
							GC_::free(str); str=0;
						}
						next(bodywords);
					}
				writer << endL << endL;
			}
		next(proclist);
		if (checkThrowing()) break;
	}
	while (varlist && !throwing) 
	{	NodeP var;
		if (varlist->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, varlist->first());
			break;
		}else
			var = th->varStack->find(varlist->first()->intern());	// variable
		
		if (var==Unbound) 
		{	if (just_titles >= 0) 
			{	error(NO_VALUE, varlist->first());
				break;
			}
		}else 
			writer << varlist->first() << '=';
			File::print(writer.file(), var);
			writer << endL;

		next(varlist);
		if (checkThrowing()) break;
	}
/*	while (tablelist && !throwing) 
	{	if (tablelist->first()->is.aggregate) 
		{	error(BAD_DATA_UNREC, tablelist->first());
			break;
		}else 
		{	TableP table = tablelist->first()->intern()->table();

			writer << tablelist->first()
				<< "=Table " << table << endL;
		}
		next(tablelist);
		if (checkThrowing()) break;
	}
*/
	th->backslashes=false;
	return Unbound;
}

NodeP lprintOut      (PairPP arg) {return arg[1]->printOut(0);}
NodeP lprintOutTitles(PairPP arg) {return arg[1]->printOut(1);}


NodeP lerase(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!th->security->erase)
	{	error(SECURITY_ERROR, new String(_("to erase definitions"))); return Unbound;
	}
	PairP proclist=0, varlist=0, tablelist=0;
	arg[1]->three_lists(proclist, varlist, tablelist);
	
	if (proclist)
		Line::the_generation++;
	
	int redef = th->redefinable;

	while (proclist)
	{	if (proclist->first())
		{	if (proclist->first()->is.aggregate) 
			{	error(BAD_DATA_UNREC, proclist->first());
				break;
			}
			ObjectP obj = proclist->first()->intern();

			if ( !redef  &&  obj->proc_or_prim()->is.primitive )
			{	error(IS_PRIM, obj);
				break;
			}
			obj->setproc( ProcedureP(Unbound) );
			obj->flag=NORMAL_OBJECT;
			obj->setowner(0);
			obj->tint &= ~PROCS;
		}
		next(proclist);
	}
	while (varlist && !throwing)
	{	if (varlist->first())
		{	if (varlist->first()->is.aggregate) 
			{	error(BAD_DATA_UNREC, varlist->first());
				break;
			}
			ObjectP obj=varlist->first()->intern();
			obj->setvar( Unbound );
			obj->flag=NORMAL_OBJECT;
			obj->setowner(0);
			obj->tint &= ~VALS;
		}
		next(varlist);
	}
	while (tablelist && !throwing)
	{	if (tablelist->first())
		{	if (tablelist->first()->is.aggregate) 
			{	error(BAD_DATA_UNREC, tablelist->first());
				break;
			}
			ObjectP obj=tablelist->first()->intern();
			obj->settable(TableP(Unbound));
			obj->flag=NORMAL_OBJECT;
			obj->setowner(0);
			obj->tint &= ~TABLES;
		}
		next(tablelist);
	}
	return Unbound;
}

NodeP leraseAll       (PairPP) 
{
	ThreadP th=Thread::This();
	return th->namespce->erase_all( true,  true,  true);
}
NodeP leraseProcedures(PairPP) 
{
	ThreadP th=Thread::This();
	return th->namespce->erase_all( true, false, false);
}
NodeP leraseNames     (PairPP) 
{
	ThreadP th=Thread::This();
	return th->namespce->erase_all(false,  true, false);
}
NodeP leraseTables    (PairPP) 
{
	ThreadP th=Thread::This();
	return th->namespce->erase_all(false, false,  true);
}


NodeP lCalls(PairPP) // try to reconstruct the stackframes
{
	ThreadP th=Thread::This();
	LineP callslist=0;
	IntStack::ItemP 
	  argi=th->argNrStack->base+1,
	  argm=th->argNrStack->index;
	NodePP fun=th->funStack->base+1;

	while (argi <= argm)
	{	
		VarStack::ItemP var=&th->varStack->base[argi->i+1];
		LineP varlist=0, tnode=0, lastnode=0;
		argi++;
		int i=argi->i;
		while (i > 0)
		{	
			if (var->i.object)
			{	tnode=new Line(
					new Line(var->i.object, 
					new Line(var->i.val, 0)),0);
				if throwing return Unbound;
				if (varlist)
					lastnode->settail(tnode);
				else
					varlist=tnode;
				lastnode=tnode;
			}
			var++;
			i--;
		}
		if (*fun && (*fun)->type==OBJECT && ObjectP(*fun)->proc())
		{	LineP l=LineP(ObjectP(*fun)->proc()->bodylist()->first());
			th->col=l->col;
			th->tcol=l->tcol;
			th->row=l->row;
			th->setfilename(l->filename());
		}
		varlist=new Line(*fun, varlist);
		callslist=new Line(varlist, callslist);
		if throwing return Unbound;
		argi++;
		fun++;
	}
	return callslist;
}

wxString String::convertQMtoP(charP ptr, int len) 
{
	wxString result;
	char c;
	charP p=wxStringBuffer(result, 256);
	if (len > 255)
		len=255;

	while (len--)
	{	c = *ptr++;
		if (c=='?')
			c='P';
		*p++ = c;
	}
	*p=0;
	return result;
}

StringP String::unspace()
{
	return make(_ptr, 0, _length, type, copy_unspace);
}


NodeP lsetRedefinable(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!th->security->erase)
	{	error(SECURITY_ERROR, new String(_("to redefine Primitives"))); 
		return Unbound;
	}
	int redefinable=arg[1]->a_bool(1);
	if throwing return Unbound;
	th->redefinable=(redefinable!=0);
	return Unbound;
}

NodeP lisredefinable(PairPP) 
{
	ThreadP th=Thread::This();
	return Node::boolnode(th->redefinable);
}

NodeP lsetUser(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!th->security->user)
	{	error(SECURITY_ERROR, new String(_("to set the user"))); 
		return Unbound;
	}
	NodeP user=arg[1];
	if (user)
		th->user=user->intern();
	else
		th->user=0;
	return Unbound;
}

NodeP lgetUser(PairPP) 
{
	ThreadP th=Thread::This();
	if (!th->security->user)
	{	error(SECURITY_ERROR, new String(_("to read out the user"))); 
		return Unbound;
	}
	return th->user;
}

NodeP lgetOwner(PairPP arg) 
{
	ThreadP th=Thread::This();
	if (!th->security->user)
	{	error(SECURITY_ERROR, new String(_("to read out the owner"))); 
		return Unbound;
	}
	NodeP nd=arg[1];
	if (nd)
		return nd->intern()->owner();
	return 0;
}

NodeP lloadNoisily(PairPP arg)
{
	ThreadP th=Thread::This();
	return Node::boolnode(th->loadNoisily);
}

NodeP lsetLoadNoisily(PairPP arg)
{
	ThreadP th=Thread::This();
	int loadNoisily=arg[1]->a_bool(1);
	if throwing return Unbound;
	th->loadNoisily=(loadNoisily!=0);
	return Unbound;
}

NodeP lStartup(PairPP arg)
{
	ThreadP th=Thread::This();
	return th->startup;
}

NodeP lsetStartup(PairPP arg)
{
	ThreadP th=Thread::This();
	NodeP startup=arg[1]->a_runnable(1);
	if throwing return Unbound;
	th->startup=startup;
	return Unbound;
}

NodeP lisProcedureDynamic(PairPP arg)
{
	StringP name=arg[1]->a_string(1);
	if throwing return Unbound;
	ObjectP obj=name->intern();
	if throwing return Unbound;
	if (obj->bit.proc && !obj->bit.permanent)
	{	return Node::boolnode(obj->bit.proc_dyn);
	}else
		error(BAD_DATA, name);
	return Unbound;
}

